/*
 * qMaZda - Image Analysis and Pattern Recognition
 * 
 * Copyright 2013  Piotr M. Szczypiński <piotr.szczypinski@p.lodz.pl>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "../MzShared/helpers.h"



/*===========================================================================*/
/* http://www.developer.nokia.com/Community/Wiki/Using_OpenCV_with_Qt        */
QImage IplImage2QImage(const IplImage *iplImage)
{
    int height = iplImage->height;
    int width = iplImage->width;

    if  (iplImage->depth == IPL_DEPTH_8U && iplImage->nChannels == 3)
    {
        const uchar *qImageBuffer = (const uchar*)iplImage->imageData;
        QImage img(qImageBuffer, width, height, QImage::Format_RGB888);
        return img.rgbSwapped();
    }
    else if  (iplImage->depth == IPL_DEPTH_8U && iplImage->nChannels == 1)
    {
        const uchar *qImageBuffer = (const uchar*)iplImage->imageData;
        QImage img(qImageBuffer, width, height, QImage::Format_Indexed8);

        QVector<QRgb> colorTable;
        for (int i = 0; i < 256; i++)
        {
            colorTable.push_back(qRgb(i, i, i));
        }
        img.setColorTable(colorTable);
        return img;
    }
    else
    {
        //qWarning() << "Image cannot be converted.";
        return QImage();
    }
}

/*===========================================================================*/
/* http://www.developer.nokia.com/Community/Wiki/Using_OpenCV_with_Qt        */
//?????????????????????? Zoptymalizować
IplImage* QImage2IplImage(const QImage& qImage)
{
  int width = qImage.width();
  int height = qImage.height();

  // Creates a iplImage with 3 channels
  IplImage *img = cvCreateImage(cvSize(width, height), IPL_DEPTH_8U, 3);
  char * imgBuffer = img->imageData;

  //Remove alpha channel
  int jump = (qImage.hasAlphaChannel()) ? 4 : 3;

  for (int y = 0;y < height;y++)
  {
    uchar* a = (uchar*)qImage.scanLine(y);
    for (int i = 0; i < width * jump; i+=jump)
    {
        imgBuffer[2] = a[i];
        imgBuffer[1] = a[i+1];
        imgBuffer[0] = a[i+2];
        imgBuffer+=3;
    }
  }
  return img;
}


//====================================================================================================
// Konwersja obrazu RGB 8/16 bitow na monochromatyczny 8/16 bitow.
// colortransform - rodzaj konwersji
int mzImageConvert(char colortransform, IplImage* in, IplImage* out)
{
    int height = in->height;
    if(height != out->height) return -1;
    int width = in->width;
    if(width != out->width) return -2;
    if(out->nChannels != 1) return -3;
    if(in->nChannels != 3) return -4;
    if(in->depth != out->depth) return -5;


    for(int y = 0; y < height; y++)
    {
        /* Rownania zamiany RGB na Greyscale wg CCIR Recomendation 601-1 */
        /*       Y = 299/1000*R + 587/1000*G+ 114/1000*B                 */
        for(int x = 0; x < width; x++)
        {
            CvScalar pin = cvGet2D(in, y, x);
            CvScalar pout = cvScalar(0);

            switch(colortransform)
            {
                //Red
            case 'R': pout.val[0] = (pin.val[2]); break;
                //Green
            case 'G': pout.val[0] = (pin.val[1]); break;
                //Blue
            case 'B': pout.val[0] = (pin.val[0]); break;
                //U channel  B-Y
            case 'U':
                pout.val[0] = (((int)886*pin.val[0]-(int)587*pin.val[1]-(int)299*pin.val[2]+886*255)/1772);
                break;
                //V channel  R-Y
            case 'V':
                pout.val[0] = ((-(int)114*pin.val[0]-(int)587*pin.val[1]+(int)701*pin.val[2]+701*255)/1402);
                break;
                //Phase
            case 'H':
                {
                    double u = ((886.0*(double)pin.val[0]-587.0*(double)pin.val[1]-299.0*(double)pin.val[2])/886.0);
                    double v = ((-114.0*(double)pin.val[0]-587.0*(double)pin.val[1]+701.0*(double)pin.val[2])/701.0);
                    if(u == 0 && v == 0) pout.val[0] = 0;
                    else
                    {
                        double h = atan2(v, u);
                        if (h<0) h+=(2*M_PI);
                        pout.val[0] = (uint8_t)(h*40.58451049);
                    }
                }
                break;
                //Saturation
            case 'S':
                {
                    double u = ((886.0*(double)pin.val[0]-587.0*(double)pin.val[1]-299.0*(double)pin.val[2])/886.0);
                    double v = ((-114.0*(double)pin.val[0]-587.0*(double)pin.val[1]+701.0*(double)pin.val[2])/701.0);
                    pout.val[0] = (uint8_t)(0.937*sqrt(u*u+v*v));
                }
                break;
                //I
            case 'I':
                pout.val[0] = ((-(int)3213*pin.val[0]-(int)2744*pin.val[1]+(int)5957*pin.val[2]+1519290)/11916);
                break;
                //Q
            case 'Q':
                pout.val[0] = (((int)3111*pin.val[0]-(int)5226*pin.val[1]+(int)2115*pin.val[2]+1332630)/10452);
                break;
                //u znormal
            case 'u':
                {
                    double u = (886.0*(double)pin.val[0]-587.0*(double)pin.val[1]-299.0*(double)pin.val[2]);
                    double bb = (114.0*(double)pin.val[0]+587.0*(double)pin.val[1]+299.0*(double)pin.val[2]+1.0);
                    pout.val[0] = (uint8_t)((u/bb+1.0)*29.07);
                }
                break;
                //u znormal
            case 'v':
                {
                    double v = (-114.0*(double)pin.val[0]-587.0*(double)pin.val[1]+701.0*(double)pin.val[2]);
                    double bb = (114.0*(double)pin.val[0]+587.0*(double)pin.val[1]+299.0*(double)pin.val[2]+1.0);
                    pout.val[0] = (uint8_t)((v/bb+1.0)*76.245);
                }
                break;
                //i znorm
            case 'i':
                {
                    double i = (-321.3*(double)pin.val[0]-274.4*(double)pin.val[1]+595.7*(double)pin.val[2]);
                    double bb = (114.0*(double)pin.val[0]+587.0*(double)pin.val[1]+299.0*(double)pin.val[2]+1.0);
                    pout.val[0] = (uint8_t)((i/bb+2.8185)*53.006);
                }
                break;
                //q znorm
            case 'q':
                {
                    double q = (311.1*(double)pin.val[0]-522.6*(double)pin.val[1]+211.5*(double)pin.val[2]);
                    double bb = (114.0*(double)pin.val[0]+587.0*(double)pin.val[1]+299.0*(double)pin.val[2]+1.0);
                    pout.val[0] = (uint8_t)((q/bb+0.8903)*70.457);
                }
                break;

            case 'h':
                {
                    double i = (-321.3*(double)pin.val[0]-274.4*(double)pin.val[1]+595.7*(double)pin.val[2]);
                    double q = (311.1*(double)pin.val[0]-522.6*(double)pin.val[1]+211.5*(double)pin.val[2]);

                    if(i == 0 && q == 0) pout.val[0] = 0;
                    else
                    {
                        double h = atan2(q, i);
                        if (h<0) h+=(2*M_PI);
                        pout.val[0] = (uint8_t)(h*40.58451049);
                    }
                } break;
                //X (CIE XYZ)
                /*
                265 x {{0.4124, 0.3576,  0.1805},
                255 x {0.2126, 0.7152,  0.0722},
                234 x {0.0193, 0.1192,  0.9505}};
                */
            case 'X':
                {
                    double r = ((double)pin.val[2])/255.0;
                    if (r <= 0.04045) r = r / 12.92;
                    else r = pow(((r + 0.055) / 1.055), 2.4);
                    double g = ((double)pin.val[1])/255.0;
                    if (g <= 0.04045) g = g / 12.92;
                    else g = pow(((g + 0.055) / 1.055), 2.4);
                    double b = ((double)pin.val[0])/255.0;
                    if (b <= 0.04045) b = b / 12.92;
                    else b = pow(((b + 0.055) / 1.055), 2.4);
                    pout.val[0] = (uint8_t)(105.162*r+91.188*g+46.0275*b);
                } break;
                //Y (CIE XYZ)
            case 'Y':
                {
                    double r = ((double)pin.val[2])/255.0;
                    if (r <= 0.04045) r = r / 12.92;
                    else r = pow(((r + 0.055) / 1.055), 2.4);
                    double g = ((double)pin.val[1])/255.0;
                    if (g <= 0.04045) g = g / 12.92;
                    else g = pow(((g + 0.055) / 1.055), 2.4);
                    double b = ((double)pin.val[0])/255.0;
                    if (b <= 0.04045) b = b / 12.92;
                    else b = pow(((b + 0.055) / 1.055), 2.4);
                    pout.val[0] = (uint8_t)(54.213*r+182.376*g+18.411*b);
                } break;
                //Z (CIE XYZ)
            case 'Z':
                {
                    double r = ((double)pin.val[2])/255.0;
                    if (r <= 0.04045) r = r / 12.92;
                    else r = pow(((r + 0.055) / 1.055), 2.4);
                    double g = ((double)pin.val[1])/255.0;
                    if (g <= 0.04045) g = g / 12.92;
                    else g = pow(((g + 0.055) / 1.055), 2.4);
                    double b = ((double)pin.val[0])/255.0;
                    if (b <= 0.04045) b = b / 12.92;
                    else b = pow(((b + 0.055) / 1.055), 2.4);
                    pout.val[0] = (uint8_t)(4.5162*r+27.8928*g+222.417*b);
                } break;

                //L (CIE Lab)
            case 'L':
                {
                    double r = ((double)pin.val[2])/255.0;
                    if (r <= 0.04045) r = r / 12.92;
                    else r = pow(((r + 0.055) / 1.055), 2.4);
                    double g = ((double)pin.val[1])/255.0;
                    if (g <= 0.04045) g = g / 12.92;
                    else g = pow(((g + 0.055) / 1.055), 2.4);
                    double b = ((double)pin.val[0])/255.0;
                    if (b <= 0.04045) b = b / 12.92;
                    else b = pow(((b + 0.055) / 1.055), 2.4);

                    double yy = 0.2126*r+0.7152*g+0.0722*b;
                    yy = pow(yy, 1.0 / 3.0);
                    pout.val[0] = (uint8_t)(255*yy);
                }break;

                //a (CIE Lab)
            case 'a':
                {
                    double r = ((double)pin.val[2])/255.0;
                    if (r <= 0.04045) r = r / 12.92;
                    else r = pow(((r + 0.055) / 1.055), 2.4);
                    double g = ((double)pin.val[1])/255.0;
                    if (g <= 0.04045) g = g / 12.92;
                    else g = pow(((g + 0.055) / 1.055), 2.4);
                    double b = ((double)pin.val[0])/255.0;
                    if (b <= 0.04045) b = b / 12.92;
                    else b = pow(((b + 0.055) / 1.055), 2.4);

                    double xx = 0.43387*r+0.37622*g+0.18990*b;
                    double yy = 0.2126*r+0.7152*g+0.0722*b;
                    xx = pow(xx, 1.0 / 3.0);
                    yy = pow(yy, 1.0 / 3.0);
                    pout.val[0] = (uint8_t)(376.128385*(xx-yy+0.33898));
                }break;
                //b (CIE Lab)
            case 'b':
                {
                    double r = ((double)pin.val[2])/255.0;
                    if (r <= 0.04045) r = r / 12.92;
                    else r = pow(((r + 0.055) / 1.055), 2.4);
                    double g = ((double)pin.val[1])/255.0;
                    if (g <= 0.04045) g = g / 12.92;
                    else g = pow(((g + 0.055) / 1.055), 2.4);
                    double b = ((double)pin.val[0])/255.0;
                    if (b <= 0.04045) b = b / 12.92;
                    else b = pow(((b + 0.055) / 1.055), 2.4);

                    double yy = 0.2126*r+0.7152*g+0.0722*b;
                    double zz = 0.01772*r+0.109458*g+0.872819*b;
                    yy = pow(yy, 1.0 / 3.0);
                    zz = pow(zz, 1.0 / 3.0);
                    pout.val[0] = (uint8_t)(159.235668*(yy-zz+0.8007));
                }break;
                //Brightness
            default:
                pout.val[0] = (uint8_t)(((int)114*pin.val[0]+(int)587*pin.val[1]+(int)299*pin.val[2])/1000);
            }

            cvSet2D(out, y, x, pout);
        }
    }
    return 0;
}

/*===========================================================================*/

#ifdef _USE_DCMTK
#define GET_DICOM_INFO_LINE(tag) \
if (fileformat.getDataset()->findAndGetOFString(tag, infostring).good())\
{info->append(&((#tag ": ")[4])); info->append(infostring.c_str()); info->append("<br>");}

void GenerateDicomInfo(QString *info, DcmFileFormat fileformat)
{
    OFString infostring;

    info->append("<div style=\"color:#00007F\">");
    GET_DICOM_INFO_LINE(DCM_CommandGroupLength                   ) //DcmTagKey(0x0000, 0x0000)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_CommandGroupLengthToEnd     ) //DcmTagKey(0x0000, 0x0001)
    GET_DICOM_INFO_LINE(DCM_AffectedSOPClassUID                  ) //DcmTagKey(0x0000, 0x0002)
    GET_DICOM_INFO_LINE(DCM_RequestedSOPClassUID                 ) //DcmTagKey(0x0000, 0x0003)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_CommandRecognitionCode      ) //DcmTagKey(0x0000, 0x0010)
    GET_DICOM_INFO_LINE(DCM_CommandField                         ) //DcmTagKey(0x0000, 0x0100)
    GET_DICOM_INFO_LINE(DCM_MessageID                            ) //DcmTagKey(0x0000, 0x0110)
    GET_DICOM_INFO_LINE(DCM_MessageIDBeingRespondedTo            ) //DcmTagKey(0x0000, 0x0120)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_Initiator                   ) //DcmTagKey(0x0000, 0x0200)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_Receiver                    ) //DcmTagKey(0x0000, 0x0300)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_FindLocation                ) //DcmTagKey(0x0000, 0x0400)
    GET_DICOM_INFO_LINE(DCM_MoveDestination                      ) //DcmTagKey(0x0000, 0x0600)
    GET_DICOM_INFO_LINE(DCM_Priority                             ) //DcmTagKey(0x0000, 0x0700)
//    GET_DICOM_INFO_LINE(DCM_DataSetType                          ) //DcmTagKey(0x0000, 0x0800)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_NumberOfMatches             ) //DcmTagKey(0x0000, 0x0850)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_ResponseSequenceNumber      ) //DcmTagKey(0x0000, 0x0860)
    GET_DICOM_INFO_LINE(DCM_Status                               ) //DcmTagKey(0x0000, 0x0900)
    GET_DICOM_INFO_LINE(DCM_OffendingElement                     ) //DcmTagKey(0x0000, 0x0901)
    GET_DICOM_INFO_LINE(DCM_ErrorComment                         ) //DcmTagKey(0x0000, 0x0902)
    GET_DICOM_INFO_LINE(DCM_ErrorID                              ) //DcmTagKey(0x0000, 0x0903)
    GET_DICOM_INFO_LINE(DCM_AffectedSOPInstanceUID               ) //DcmTagKey(0x0000, 0x1000)
    GET_DICOM_INFO_LINE(DCM_RequestedSOPInstanceUID              ) //DcmTagKey(0x0000, 0x1001)
    GET_DICOM_INFO_LINE(DCM_EventTypeID                          ) //DcmTagKey(0x0000, 0x1002)
    GET_DICOM_INFO_LINE(DCM_AttributeIdentifierList              ) //DcmTagKey(0x0000, 0x1005)
    GET_DICOM_INFO_LINE(DCM_ActionTypeID                         ) //DcmTagKey(0x0000, 0x1008)
//    GET_DICOM_INFO_LINE(DCM_NumberOfRemainingSubOperations       ) //DcmTagKey(0x0000, 0x1020)
//    GET_DICOM_INFO_LINE(DCM_NumberOfCompletedSubOperations       ) //DcmTagKey(0x0000, 0x1021)
//    GET_DICOM_INFO_LINE(DCM_NumberOfFailedSubOperations          ) //DcmTagKey(0x0000, 0x1022)
//    GET_DICOM_INFO_LINE(DCM_NumberOfWarningSubOperations         ) //DcmTagKey(0x0000, 0x1023)
    GET_DICOM_INFO_LINE(DCM_MoveOriginatorApplicationEntityTitle ) //DcmTagKey(0x0000, 0x1030)
    GET_DICOM_INFO_LINE(DCM_MoveOriginatorMessageID              ) //DcmTagKey(0x0000, 0x1031)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_DialogReceiver              ) //DcmTagKey(0x0000, 0x4000)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_TerminalType                ) //DcmTagKey(0x0000, 0x4010)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_MessageSetID                ) //DcmTagKey(0x0000, 0x5010)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_EndMessageSet               ) //DcmTagKey(0x0000, 0x5020)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_DisplayFormat               ) //DcmTagKey(0x0000, 0x5110)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_PagePositionID              ) //DcmTagKey(0x0000, 0x5120)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_TextFormatID                ) //DcmTagKey(0x0000, 0x5130)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_NormalReverse               ) //DcmTagKey(0x0000, 0x5140)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_AddGrayScale                ) //DcmTagKey(0x0000, 0x5150)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_Borders                     ) //DcmTagKey(0x0000, 0x5160)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_Copies                      ) //DcmTagKey(0x0000, 0x5170)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_MagnificationType           ) //DcmTagKey(0x0000, 0x5180)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_Erase                       ) //DcmTagKey(0x0000, 0x5190)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_Print                       ) //DcmTagKey(0x0000, 0x51a0)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_Overlays                    ) //DcmTagKey(0x0000, 0x51b0)
            info->append("</div>");
            info->append("<div style=\"color:#00007F\">");
    GET_DICOM_INFO_LINE(DCM_FileMetaInformationGroupLength       ) //DcmTagKey(0x0002, 0x0000)
    GET_DICOM_INFO_LINE(DCM_FileMetaInformationVersion           ) //DcmTagKey(0x0002, 0x0001)
    GET_DICOM_INFO_LINE(DCM_MediaStorageSOPClassUID              ) //DcmTagKey(0x0002, 0x0002)
    GET_DICOM_INFO_LINE(DCM_MediaStorageSOPInstanceUID           ) //DcmTagKey(0x0002, 0x0003)
    GET_DICOM_INFO_LINE(DCM_TransferSyntaxUID                    ) //DcmTagKey(0x0002, 0x0010)
    GET_DICOM_INFO_LINE(DCM_ImplementationClassUID               ) //DcmTagKey(0x0002, 0x0012)
    GET_DICOM_INFO_LINE(DCM_ImplementationVersionName            ) //DcmTagKey(0x0002, 0x0013)
    GET_DICOM_INFO_LINE(DCM_SourceApplicationEntityTitle         ) //DcmTagKey(0x0002, 0x0016)
    GET_DICOM_INFO_LINE(DCM_PrivateInformationCreatorUID         ) //DcmTagKey(0x0002, 0x0100)
    GET_DICOM_INFO_LINE(DCM_PrivateInformation                   ) //DcmTagKey(0x0002, 0x0102)
            info->append("</div>");
            info->append("<div style=\"color:#00007F\">");
    GET_DICOM_INFO_LINE(DCM_FileSetID                            ) //DcmTagKey(0x0004, 0x1130)
    GET_DICOM_INFO_LINE(DCM_FileSetDescriptorFileID              ) //DcmTagKey(0x0004, 0x1141)
    GET_DICOM_INFO_LINE(DCM_SpecificCharacterSetOfFileSetDescriptorFile ) //DcmTagKey(0x0004, 0x1142)
    GET_DICOM_INFO_LINE(DCM_OffsetOfTheFirstDirectoryRecordOfTheRootDirectoryEntity ) //DcmTagKey(0x0004, 0x1200)
    GET_DICOM_INFO_LINE(DCM_OffsetOfTheLastDirectoryRecordOfTheRootDirectoryEntity ) //DcmTagKey(0x0004, 0x1202)
    GET_DICOM_INFO_LINE(DCM_FileSetConsistencyFlag               ) //DcmTagKey(0x0004, 0x1212)
    GET_DICOM_INFO_LINE(DCM_DirectoryRecordSequence              ) //DcmTagKey(0x0004, 0x1220)
    GET_DICOM_INFO_LINE(DCM_OffsetOfTheNextDirectoryRecord       ) //DcmTagKey(0x0004, 0x1400)
    GET_DICOM_INFO_LINE(DCM_RecordInUseFlag                      ) //DcmTagKey(0x0004, 0x1410)
    GET_DICOM_INFO_LINE(DCM_OffsetOfReferencedLowerLevelDirectoryEntity ) //DcmTagKey(0x0004, 0x1420)
    GET_DICOM_INFO_LINE(DCM_DirectoryRecordType                  ) //DcmTagKey(0x0004, 0x1430)
    GET_DICOM_INFO_LINE(DCM_PrivateRecordUID                     ) //DcmTagKey(0x0004, 0x1432)
    GET_DICOM_INFO_LINE(DCM_ReferencedFileID                     ) //DcmTagKey(0x0004, 0x1500)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_MRDRDirectoryRecordOffset    ) //DcmTagKey(0x0004, 0x1504)
    GET_DICOM_INFO_LINE(DCM_ReferencedSOPClassUIDInFile          ) //DcmTagKey(0x0004, 0x1510)
    GET_DICOM_INFO_LINE(DCM_ReferencedSOPInstanceUIDInFile       ) //DcmTagKey(0x0004, 0x1511)
    GET_DICOM_INFO_LINE(DCM_ReferencedTransferSyntaxUIDInFile    ) //DcmTagKey(0x0004, 0x1512)
    GET_DICOM_INFO_LINE(DCM_ReferencedRelatedGeneralSOPClassUIDInFile ) //DcmTagKey(0x0004, 0x151a)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_NumberOfReferences           ) //DcmTagKey(0x0004, 0x1600)
            info->append("</div>");
            info->append("<div style=\"color:#00007F\">");
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_IdentifyingGroupLengthToEnd ) //DcmTagKey(0x0008, 0x0001)
    GET_DICOM_INFO_LINE(DCM_SpecificCharacterSet                 ) //DcmTagKey(0x0008, 0x0005)
    GET_DICOM_INFO_LINE(DCM_LanguageCodeSequence                 ) //DcmTagKey(0x0008, 0x0006)
    GET_DICOM_INFO_LINE(DCM_ImageType                            ) //DcmTagKey(0x0008, 0x0008)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_RecognitionCode             ) //DcmTagKey(0x0008, 0x0010)
    GET_DICOM_INFO_LINE(DCM_InstanceCreationDate                 ) //DcmTagKey(0x0008, 0x0012)
    GET_DICOM_INFO_LINE(DCM_InstanceCreationTime                 ) //DcmTagKey(0x0008, 0x0013)
    GET_DICOM_INFO_LINE(DCM_InstanceCreatorUID                   ) //DcmTagKey(0x0008, 0x0014)
    GET_DICOM_INFO_LINE(DCM_SOPClassUID                          ) //DcmTagKey(0x0008, 0x0016)
    GET_DICOM_INFO_LINE(DCM_SOPInstanceUID                       ) //DcmTagKey(0x0008, 0x0018)
    GET_DICOM_INFO_LINE(DCM_RelatedGeneralSOPClassUID            ) //DcmTagKey(0x0008, 0x001a)
    GET_DICOM_INFO_LINE(DCM_OriginalSpecializedSOPClassUID       ) //DcmTagKey(0x0008, 0x001b)
    GET_DICOM_INFO_LINE(DCM_StudyDate                            ) //DcmTagKey(0x0008, 0x0020)
    GET_DICOM_INFO_LINE(DCM_SeriesDate                           ) //DcmTagKey(0x0008, 0x0021)
    GET_DICOM_INFO_LINE(DCM_AcquisitionDate                      ) //DcmTagKey(0x0008, 0x0022)
    GET_DICOM_INFO_LINE(DCM_ContentDate                          ) //DcmTagKey(0x0008, 0x0023)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_OverlayDate                  ) //DcmTagKey(0x0008, 0x0024)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_CurveDate                    ) //DcmTagKey(0x0008, 0x0025)
    GET_DICOM_INFO_LINE(DCM_AcquisitionDateTime                  ) //DcmTagKey(0x0008, 0x002a)
    GET_DICOM_INFO_LINE(DCM_StudyTime                            ) //DcmTagKey(0x0008, 0x0030)
    GET_DICOM_INFO_LINE(DCM_SeriesTime                           ) //DcmTagKey(0x0008, 0x0031)
    GET_DICOM_INFO_LINE(DCM_AcquisitionTime                      ) //DcmTagKey(0x0008, 0x0032)
    GET_DICOM_INFO_LINE(DCM_ContentTime                          ) //DcmTagKey(0x0008, 0x0033)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_OverlayTime                  ) //DcmTagKey(0x0008, 0x0034)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_CurveTime                    ) //DcmTagKey(0x0008, 0x0035)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_OldDataSetType              ) //DcmTagKey(0x0008, 0x0040)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_DataSetSubtype              ) //DcmTagKey(0x0008, 0x0041)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_NuclearMedicineSeriesType    ) //DcmTagKey(0x0008, 0x0042)
    GET_DICOM_INFO_LINE(DCM_AccessionNumber                      ) //DcmTagKey(0x0008, 0x0050)
    GET_DICOM_INFO_LINE(DCM_IssuerOfAccessionNumberSequence      ) //DcmTagKey(0x0008, 0x0051)
    GET_DICOM_INFO_LINE(DCM_QueryRetrieveLevel                   ) //DcmTagKey(0x0008, 0x0052)
    GET_DICOM_INFO_LINE(DCM_RetrieveAETitle                      ) //DcmTagKey(0x0008, 0x0054)
    GET_DICOM_INFO_LINE(DCM_InstanceAvailability                 ) //DcmTagKey(0x0008, 0x0056)
    GET_DICOM_INFO_LINE(DCM_FailedSOPInstanceUIDList             ) //DcmTagKey(0x0008, 0x0058)
    GET_DICOM_INFO_LINE(DCM_Modality                             ) //DcmTagKey(0x0008, 0x0060)
    GET_DICOM_INFO_LINE(DCM_ModalitiesInStudy                    ) //DcmTagKey(0x0008, 0x0061)
    GET_DICOM_INFO_LINE(DCM_SOPClassesInStudy                    ) //DcmTagKey(0x0008, 0x0062)
    GET_DICOM_INFO_LINE(DCM_ConversionType                       ) //DcmTagKey(0x0008, 0x0064)
    GET_DICOM_INFO_LINE(DCM_PresentationIntentType               ) //DcmTagKey(0x0008, 0x0068)
    GET_DICOM_INFO_LINE(DCM_Manufacturer                         ) //DcmTagKey(0x0008, 0x0070)
    GET_DICOM_INFO_LINE(DCM_InstitutionName                      ) //DcmTagKey(0x0008, 0x0080)
    GET_DICOM_INFO_LINE(DCM_InstitutionAddress                   ) //DcmTagKey(0x0008, 0x0081)
    GET_DICOM_INFO_LINE(DCM_InstitutionCodeSequence              ) //DcmTagKey(0x0008, 0x0082)
    GET_DICOM_INFO_LINE(DCM_ReferringPhysicianName               ) //DcmTagKey(0x0008, 0x0090)
    GET_DICOM_INFO_LINE(DCM_ReferringPhysicianAddress            ) //DcmTagKey(0x0008, 0x0092)
    GET_DICOM_INFO_LINE(DCM_ReferringPhysicianTelephoneNumbers   ) //DcmTagKey(0x0008, 0x0094)
    GET_DICOM_INFO_LINE(DCM_ReferringPhysicianIdentificationSequence ) //DcmTagKey(0x0008, 0x0096)
    GET_DICOM_INFO_LINE(DCM_CodeValue                            ) //DcmTagKey(0x0008, 0x0100)
    GET_DICOM_INFO_LINE(DCM_CodingSchemeDesignator               ) //DcmTagKey(0x0008, 0x0102)
    GET_DICOM_INFO_LINE(DCM_CodingSchemeVersion                  ) //DcmTagKey(0x0008, 0x0103)
    GET_DICOM_INFO_LINE(DCM_CodeMeaning                          ) //DcmTagKey(0x0008, 0x0104)
    GET_DICOM_INFO_LINE(DCM_MappingResource                      ) //DcmTagKey(0x0008, 0x0105)
    GET_DICOM_INFO_LINE(DCM_ContextGroupVersion                  ) //DcmTagKey(0x0008, 0x0106)
    GET_DICOM_INFO_LINE(DCM_ContextGroupLocalVersion             ) //DcmTagKey(0x0008, 0x0107)
    GET_DICOM_INFO_LINE(DCM_ContextGroupExtensionFlag            ) //DcmTagKey(0x0008, 0x010b)
    GET_DICOM_INFO_LINE(DCM_CodingSchemeUID                      ) //DcmTagKey(0x0008, 0x010c)
    GET_DICOM_INFO_LINE(DCM_ContextGroupExtensionCreatorUID      ) //DcmTagKey(0x0008, 0x010d)
    GET_DICOM_INFO_LINE(DCM_ContextIdentifier                    ) //DcmTagKey(0x0008, 0x010f)
    GET_DICOM_INFO_LINE(DCM_CodingSchemeIdentificationSequence   ) //DcmTagKey(0x0008, 0x0110)
    GET_DICOM_INFO_LINE(DCM_CodingSchemeRegistry                 ) //DcmTagKey(0x0008, 0x0112)
    GET_DICOM_INFO_LINE(DCM_CodingSchemeExternalID               ) //DcmTagKey(0x0008, 0x0114)
    GET_DICOM_INFO_LINE(DCM_CodingSchemeName                     ) //DcmTagKey(0x0008, 0x0115)
    GET_DICOM_INFO_LINE(DCM_CodingSchemeResponsibleOrganization  ) //DcmTagKey(0x0008, 0x0116)
    GET_DICOM_INFO_LINE(DCM_ContextUID                           ) //DcmTagKey(0x0008, 0x0117)
    GET_DICOM_INFO_LINE(DCM_TimezoneOffsetFromUTC                ) //DcmTagKey(0x0008, 0x0201)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_NetworkID                   ) //DcmTagKey(0x0008, 0x1000)
    GET_DICOM_INFO_LINE(DCM_StationName                          ) //DcmTagKey(0x0008, 0x1010)
    GET_DICOM_INFO_LINE(DCM_StudyDescription                     ) //DcmTagKey(0x0008, 0x1030)
    GET_DICOM_INFO_LINE(DCM_ProcedureCodeSequence                ) //DcmTagKey(0x0008, 0x1032)
    GET_DICOM_INFO_LINE(DCM_SeriesDescription                    ) //DcmTagKey(0x0008, 0x103e)
    GET_DICOM_INFO_LINE(DCM_SeriesDescriptionCodeSequence        ) //DcmTagKey(0x0008, 0x103f)
    GET_DICOM_INFO_LINE(DCM_InstitutionalDepartmentName          ) //DcmTagKey(0x0008, 0x1040)
    GET_DICOM_INFO_LINE(DCM_PhysiciansOfRecord                   ) //DcmTagKey(0x0008, 0x1048)
    GET_DICOM_INFO_LINE(DCM_PhysiciansOfRecordIdentificationSequence ) //DcmTagKey(0x0008, 0x1049)
    GET_DICOM_INFO_LINE(DCM_PerformingPhysicianName              ) //DcmTagKey(0x0008, 0x1050)
    GET_DICOM_INFO_LINE(DCM_PerformingPhysicianIdentificationSequence ) //DcmTagKey(0x0008, 0x1052)
    GET_DICOM_INFO_LINE(DCM_NameOfPhysiciansReadingStudy         ) //DcmTagKey(0x0008, 0x1060)
    GET_DICOM_INFO_LINE(DCM_PhysiciansReadingStudyIdentificationSequence ) //DcmTagKey(0x0008, 0x1062)
    GET_DICOM_INFO_LINE(DCM_OperatorsName                        ) //DcmTagKey(0x0008, 0x1070)
    GET_DICOM_INFO_LINE(DCM_OperatorIdentificationSequence       ) //DcmTagKey(0x0008, 0x1072)
    GET_DICOM_INFO_LINE(DCM_AdmittingDiagnosesDescription        ) //DcmTagKey(0x0008, 0x1080)
    GET_DICOM_INFO_LINE(DCM_AdmittingDiagnosesCodeSequence       ) //DcmTagKey(0x0008, 0x1084)
    GET_DICOM_INFO_LINE(DCM_ManufacturerModelName                ) //DcmTagKey(0x0008, 0x1090)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ReferencedResultsSequence    ) //DcmTagKey(0x0008, 0x1100)
    GET_DICOM_INFO_LINE(DCM_ReferencedStudySequence              ) //DcmTagKey(0x0008, 0x1110)
    GET_DICOM_INFO_LINE(DCM_ReferencedPerformedProcedureStepSequence ) //DcmTagKey(0x0008, 0x1111)
    GET_DICOM_INFO_LINE(DCM_ReferencedSeriesSequence             ) //DcmTagKey(0x0008, 0x1115)
    GET_DICOM_INFO_LINE(DCM_ReferencedPatientSequence            ) //DcmTagKey(0x0008, 0x1120)
    GET_DICOM_INFO_LINE(DCM_ReferencedVisitSequence              ) //DcmTagKey(0x0008, 0x1125)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ReferencedOverlaySequence    ) //DcmTagKey(0x0008, 0x1130)
    GET_DICOM_INFO_LINE(DCM_ReferencedStereometricInstanceSequence ) //DcmTagKey(0x0008, 0x1134)
    GET_DICOM_INFO_LINE(DCM_ReferencedWaveformSequence           ) //DcmTagKey(0x0008, 0x113a)
    GET_DICOM_INFO_LINE(DCM_ReferencedImageSequence              ) //DcmTagKey(0x0008, 0x1140)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ReferencedCurveSequence      ) //DcmTagKey(0x0008, 0x1145)
    GET_DICOM_INFO_LINE(DCM_ReferencedInstanceSequence           ) //DcmTagKey(0x0008, 0x114a)
    GET_DICOM_INFO_LINE(DCM_ReferencedRealWorldValueMappingInstanceSequence ) //DcmTagKey(0x0008, 0x114b)
    GET_DICOM_INFO_LINE(DCM_ReferencedSOPClassUID                ) //DcmTagKey(0x0008, 0x1150)
    GET_DICOM_INFO_LINE(DCM_ReferencedSOPInstanceUID             ) //DcmTagKey(0x0008, 0x1155)
    GET_DICOM_INFO_LINE(DCM_SOPClassesSupported                  ) //DcmTagKey(0x0008, 0x115a)
    GET_DICOM_INFO_LINE(DCM_ReferencedFrameNumber                ) //DcmTagKey(0x0008, 0x1160)
    GET_DICOM_INFO_LINE(DCM_SimpleFrameList                      ) //DcmTagKey(0x0008, 0x1161)
    GET_DICOM_INFO_LINE(DCM_CalculatedFrameList                  ) //DcmTagKey(0x0008, 0x1162)
    GET_DICOM_INFO_LINE(DCM_TimeRange                            ) //DcmTagKey(0x0008, 0x1163)
    GET_DICOM_INFO_LINE(DCM_FrameExtractionSequence              ) //DcmTagKey(0x0008, 0x1164)
    GET_DICOM_INFO_LINE(DCM_MultiFrameSourceSOPInstanceUID       ) //DcmTagKey(0x0008, 0x1167)
    GET_DICOM_INFO_LINE(DCM_TransactionUID                       ) //DcmTagKey(0x0008, 0x1195)
    GET_DICOM_INFO_LINE(DCM_FailureReason                        ) //DcmTagKey(0x0008, 0x1197)
    GET_DICOM_INFO_LINE(DCM_FailedSOPSequence                    ) //DcmTagKey(0x0008, 0x1198)
    GET_DICOM_INFO_LINE(DCM_ReferencedSOPSequence                ) //DcmTagKey(0x0008, 0x1199)
    GET_DICOM_INFO_LINE(DCM_StudiesContainingOtherReferencedInstancesSequence ) //DcmTagKey(0x0008, 0x1200)
    GET_DICOM_INFO_LINE(DCM_RelatedSeriesSequence                ) //DcmTagKey(0x0008, 0x1250)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_LossyImageCompressionRetired ) //DcmTagKey(0x0008, 0x2110)
    GET_DICOM_INFO_LINE(DCM_DerivationDescription                ) //DcmTagKey(0x0008, 0x2111)
    GET_DICOM_INFO_LINE(DCM_SourceImageSequence                  ) //DcmTagKey(0x0008, 0x2112)
    GET_DICOM_INFO_LINE(DCM_StageName                            ) //DcmTagKey(0x0008, 0x2120)
    GET_DICOM_INFO_LINE(DCM_StageNumber                          ) //DcmTagKey(0x0008, 0x2122)
    GET_DICOM_INFO_LINE(DCM_NumberOfStages                       ) //DcmTagKey(0x0008, 0x2124)
    GET_DICOM_INFO_LINE(DCM_ViewName                             ) //DcmTagKey(0x0008, 0x2127)
    GET_DICOM_INFO_LINE(DCM_ViewNumber                           ) //DcmTagKey(0x0008, 0x2128)
    GET_DICOM_INFO_LINE(DCM_NumberOfEventTimers                  ) //DcmTagKey(0x0008, 0x2129)
    GET_DICOM_INFO_LINE(DCM_NumberOfViewsInStage                 ) //DcmTagKey(0x0008, 0x212a)
    GET_DICOM_INFO_LINE(DCM_EventElapsedTimes                    ) //DcmTagKey(0x0008, 0x2130)
    GET_DICOM_INFO_LINE(DCM_EventTimerNames                      ) //DcmTagKey(0x0008, 0x2132)
    GET_DICOM_INFO_LINE(DCM_EventTimerSequence                   ) //DcmTagKey(0x0008, 0x2133)
    GET_DICOM_INFO_LINE(DCM_EventTimeOffset                      ) //DcmTagKey(0x0008, 0x2134)
    GET_DICOM_INFO_LINE(DCM_EventCodeSequence                    ) //DcmTagKey(0x0008, 0x2135)
    GET_DICOM_INFO_LINE(DCM_StartTrim                            ) //DcmTagKey(0x0008, 0x2142)
    GET_DICOM_INFO_LINE(DCM_StopTrim                             ) //DcmTagKey(0x0008, 0x2143)
    GET_DICOM_INFO_LINE(DCM_RecommendedDisplayFrameRate          ) //DcmTagKey(0x0008, 0x2144)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_TransducerPosition           ) //DcmTagKey(0x0008, 0x2200)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_TransducerOrientation        ) //DcmTagKey(0x0008, 0x2204)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_AnatomicStructure            ) //DcmTagKey(0x0008, 0x2208)
    GET_DICOM_INFO_LINE(DCM_AnatomicRegionSequence               ) //DcmTagKey(0x0008, 0x2218)
    GET_DICOM_INFO_LINE(DCM_AnatomicRegionModifierSequence       ) //DcmTagKey(0x0008, 0x2220)
    GET_DICOM_INFO_LINE(DCM_PrimaryAnatomicStructureSequence     ) //DcmTagKey(0x0008, 0x2228)
    GET_DICOM_INFO_LINE(DCM_AnatomicStructureSpaceOrRegionSequence ) //DcmTagKey(0x0008, 0x2229)
    GET_DICOM_INFO_LINE(DCM_PrimaryAnatomicStructureModifierSequence ) //DcmTagKey(0x0008, 0x2230)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_TransducerPositionSequence   ) //DcmTagKey(0x0008, 0x2240)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_TransducerPositionModifierSequence ) //DcmTagKey(0x0008, 0x2242)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_TransducerOrientationSequence ) //DcmTagKey(0x0008, 0x2244)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_TransducerOrientationModifierSequence ) //DcmTagKey(0x0008, 0x2246)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_AnatomicStructureSpaceOrRegionCodeSequenceTrial ) //DcmTagKey(0x0008, 0x2251)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_AnatomicPortalOfEntranceCodeSequenceTrial ) //DcmTagKey(0x0008, 0x2253)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_AnatomicApproachDirectionCodeSequenceTrial ) //DcmTagKey(0x0008, 0x2255)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_AnatomicPerspectiveDescriptionTrial ) //DcmTagKey(0x0008, 0x2256)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_AnatomicPerspectiveCodeSequenceTrial ) //DcmTagKey(0x0008, 0x2257)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_AnatomicLocationOfExaminingInstrumentDescriptionTrial ) //DcmTagKey(0x0008, 0x2258)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_AnatomicLocationOfExaminingInstrumentCodeSequenceTrial ) //DcmTagKey(0x0008, 0x2259)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_AnatomicStructureSpaceOrRegionModifierCodeSequenceTrial ) //DcmTagKey(0x0008, 0x225a)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_OnAxisBackgroundAnatomicStructureCodeSequenceTrial ) //DcmTagKey(0x0008, 0x225c)
    GET_DICOM_INFO_LINE(DCM_AlternateRepresentationSequence      ) //DcmTagKey(0x0008, 0x3001)
    GET_DICOM_INFO_LINE(DCM_IrradiationEventUID                  ) //DcmTagKey(0x0008, 0x3010)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_IdentifyingComments         ) //DcmTagKey(0x0008, 0x4000)
    GET_DICOM_INFO_LINE(DCM_FrameType                            ) //DcmTagKey(0x0008, 0x9007)
    GET_DICOM_INFO_LINE(DCM_ReferencedImageEvidenceSequence      ) //DcmTagKey(0x0008, 0x9092)
    GET_DICOM_INFO_LINE(DCM_ReferencedRawDataSequence            ) //DcmTagKey(0x0008, 0x9121)
    GET_DICOM_INFO_LINE(DCM_CreatorVersionUID                    ) //DcmTagKey(0x0008, 0x9123)
    GET_DICOM_INFO_LINE(DCM_DerivationImageSequence              ) //DcmTagKey(0x0008, 0x9124)
    GET_DICOM_INFO_LINE(DCM_SourceImageEvidenceSequence          ) //DcmTagKey(0x0008, 0x9154)
    GET_DICOM_INFO_LINE(DCM_PixelPresentation                    ) //DcmTagKey(0x0008, 0x9205)
    GET_DICOM_INFO_LINE(DCM_VolumetricProperties                 ) //DcmTagKey(0x0008, 0x9206)
    GET_DICOM_INFO_LINE(DCM_VolumeBasedCalculationTechnique      ) //DcmTagKey(0x0008, 0x9207)
    GET_DICOM_INFO_LINE(DCM_ComplexImageComponent                ) //DcmTagKey(0x0008, 0x9208)
    GET_DICOM_INFO_LINE(DCM_AcquisitionContrast                  ) //DcmTagKey(0x0008, 0x9209)
    GET_DICOM_INFO_LINE(DCM_DerivationCodeSequence               ) //DcmTagKey(0x0008, 0x9215)
    GET_DICOM_INFO_LINE(DCM_ReferencedPresentationStateSequence  ) //DcmTagKey(0x0008, 0x9237)
    GET_DICOM_INFO_LINE(DCM_ReferencedOtherPlaneSequence         ) //DcmTagKey(0x0008, 0x9410)
    GET_DICOM_INFO_LINE(DCM_FrameDisplaySequence                 ) //DcmTagKey(0x0008, 0x9458)
    GET_DICOM_INFO_LINE(DCM_RecommendedDisplayFrameRateInFloat   ) //DcmTagKey(0x0008, 0x9459)
    GET_DICOM_INFO_LINE(DCM_SkipFrameRangeFlag                   ) //DcmTagKey(0x0008, 0x9460)
            info->append("</div>");
            info->append("<div style=\"color:#0000FF\">");
    GET_DICOM_INFO_LINE(DCM_PatientName                          ) //DcmTagKey(0x0010, 0x0010)
    GET_DICOM_INFO_LINE(DCM_PatientID                            ) //DcmTagKey(0x0010, 0x0020)
    GET_DICOM_INFO_LINE(DCM_IssuerOfPatientID                    ) //DcmTagKey(0x0010, 0x0021)
    GET_DICOM_INFO_LINE(DCM_TypeOfPatientID                      ) //DcmTagKey(0x0010, 0x0022)
    GET_DICOM_INFO_LINE(DCM_IssuerOfPatientIDQualifiersSequence  ) //DcmTagKey(0x0010, 0x0024)
    GET_DICOM_INFO_LINE(DCM_PatientBirthDate                     ) //DcmTagKey(0x0010, 0x0030)
    GET_DICOM_INFO_LINE(DCM_PatientBirthTime                     ) //DcmTagKey(0x0010, 0x0032)
    GET_DICOM_INFO_LINE(DCM_PatientSex                           ) //DcmTagKey(0x0010, 0x0040)
    GET_DICOM_INFO_LINE(DCM_PatientInsurancePlanCodeSequence     ) //DcmTagKey(0x0010, 0x0050)
    GET_DICOM_INFO_LINE(DCM_PatientPrimaryLanguageCodeSequence   ) //DcmTagKey(0x0010, 0x0101)
    GET_DICOM_INFO_LINE(DCM_PatientPrimaryLanguageModifierCodeSequence ) //DcmTagKey(0x0010, 0x0102)
    GET_DICOM_INFO_LINE(DCM_OtherPatientIDs                      ) //DcmTagKey(0x0010, 0x1000)
    GET_DICOM_INFO_LINE(DCM_OtherPatientNames                    ) //DcmTagKey(0x0010, 0x1001)
    GET_DICOM_INFO_LINE(DCM_OtherPatientIDsSequence              ) //DcmTagKey(0x0010, 0x1002)
    GET_DICOM_INFO_LINE(DCM_PatientBirthName                     ) //DcmTagKey(0x0010, 0x1005)
    GET_DICOM_INFO_LINE(DCM_PatientAge                           ) //DcmTagKey(0x0010, 0x1010)
    GET_DICOM_INFO_LINE(DCM_PatientSize                          ) //DcmTagKey(0x0010, 0x1020)
    GET_DICOM_INFO_LINE(DCM_PatientSizeCodeSequence              ) //DcmTagKey(0x0010, 0x1021)
    GET_DICOM_INFO_LINE(DCM_PatientWeight                        ) //DcmTagKey(0x0010, 0x1030)
    GET_DICOM_INFO_LINE(DCM_PatientAddress                       ) //DcmTagKey(0x0010, 0x1040)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_InsurancePlanIdentification ) //DcmTagKey(0x0010, 0x1050)
    GET_DICOM_INFO_LINE(DCM_PatientMotherBirthName               ) //DcmTagKey(0x0010, 0x1060)
    GET_DICOM_INFO_LINE(DCM_MilitaryRank                         ) //DcmTagKey(0x0010, 0x1080)
    GET_DICOM_INFO_LINE(DCM_BranchOfService                      ) //DcmTagKey(0x0010, 0x1081)
    GET_DICOM_INFO_LINE(DCM_MedicalRecordLocator                 ) //DcmTagKey(0x0010, 0x1090)
    GET_DICOM_INFO_LINE(DCM_MedicalAlerts                        ) //DcmTagKey(0x0010, 0x2000)
    GET_DICOM_INFO_LINE(DCM_Allergies                            ) //DcmTagKey(0x0010, 0x2110)
    GET_DICOM_INFO_LINE(DCM_CountryOfResidence                   ) //DcmTagKey(0x0010, 0x2150)
    GET_DICOM_INFO_LINE(DCM_RegionOfResidence                    ) //DcmTagKey(0x0010, 0x2152)
    GET_DICOM_INFO_LINE(DCM_PatientTelephoneNumbers              ) //DcmTagKey(0x0010, 0x2154)
    GET_DICOM_INFO_LINE(DCM_EthnicGroup                          ) //DcmTagKey(0x0010, 0x2160)
    GET_DICOM_INFO_LINE(DCM_Occupation                           ) //DcmTagKey(0x0010, 0x2180)
    GET_DICOM_INFO_LINE(DCM_SmokingStatus                        ) //DcmTagKey(0x0010, 0x21a0)
    GET_DICOM_INFO_LINE(DCM_AdditionalPatientHistory             ) //DcmTagKey(0x0010, 0x21b0)
    GET_DICOM_INFO_LINE(DCM_PregnancyStatus                      ) //DcmTagKey(0x0010, 0x21c0)
    GET_DICOM_INFO_LINE(DCM_LastMenstrualDate                    ) //DcmTagKey(0x0010, 0x21d0)
    GET_DICOM_INFO_LINE(DCM_PatientReligiousPreference           ) //DcmTagKey(0x0010, 0x21f0)
    GET_DICOM_INFO_LINE(DCM_PatientSpeciesDescription            ) //DcmTagKey(0x0010, 0x2201)
    GET_DICOM_INFO_LINE(DCM_PatientSpeciesCodeSequence           ) //DcmTagKey(0x0010, 0x2202)
    GET_DICOM_INFO_LINE(DCM_PatientSexNeutered                   ) //DcmTagKey(0x0010, 0x2203)
    GET_DICOM_INFO_LINE(DCM_AnatomicalOrientationType            ) //DcmTagKey(0x0010, 0x2210)
    GET_DICOM_INFO_LINE(DCM_PatientBreedDescription              ) //DcmTagKey(0x0010, 0x2292)
    GET_DICOM_INFO_LINE(DCM_PatientBreedCodeSequence             ) //DcmTagKey(0x0010, 0x2293)
    GET_DICOM_INFO_LINE(DCM_BreedRegistrationSequence            ) //DcmTagKey(0x0010, 0x2294)
    GET_DICOM_INFO_LINE(DCM_BreedRegistrationNumber              ) //DcmTagKey(0x0010, 0x2295)
    GET_DICOM_INFO_LINE(DCM_BreedRegistryCodeSequence            ) //DcmTagKey(0x0010, 0x2296)
    GET_DICOM_INFO_LINE(DCM_ResponsiblePerson                    ) //DcmTagKey(0x0010, 0x2297)
    GET_DICOM_INFO_LINE(DCM_ResponsiblePersonRole                ) //DcmTagKey(0x0010, 0x2298)
    GET_DICOM_INFO_LINE(DCM_ResponsibleOrganization              ) //DcmTagKey(0x0010, 0x2299)
    GET_DICOM_INFO_LINE(DCM_PatientComments                      ) //DcmTagKey(0x0010, 0x4000)
    GET_DICOM_INFO_LINE(DCM_ExaminedBodyThickness                ) //DcmTagKey(0x0010, 0x9431)
            info->append("</div>");
            info->append("<div style=\"color:#00007F\">");
    GET_DICOM_INFO_LINE(DCM_ClinicalTrialSponsorName             ) //DcmTagKey(0x0012, 0x0010)
    GET_DICOM_INFO_LINE(DCM_ClinicalTrialProtocolID              ) //DcmTagKey(0x0012, 0x0020)
    GET_DICOM_INFO_LINE(DCM_ClinicalTrialProtocolName            ) //DcmTagKey(0x0012, 0x0021)
    GET_DICOM_INFO_LINE(DCM_ClinicalTrialSiteID                  ) //DcmTagKey(0x0012, 0x0030)
    GET_DICOM_INFO_LINE(DCM_ClinicalTrialSiteName                ) //DcmTagKey(0x0012, 0x0031)
    GET_DICOM_INFO_LINE(DCM_ClinicalTrialSubjectID               ) //DcmTagKey(0x0012, 0x0040)
    GET_DICOM_INFO_LINE(DCM_ClinicalTrialSubjectReadingID        ) //DcmTagKey(0x0012, 0x0042)
    GET_DICOM_INFO_LINE(DCM_ClinicalTrialTimePointID             ) //DcmTagKey(0x0012, 0x0050)
    GET_DICOM_INFO_LINE(DCM_ClinicalTrialTimePointDescription    ) //DcmTagKey(0x0012, 0x0051)
    GET_DICOM_INFO_LINE(DCM_ClinicalTrialCoordinatingCenterName  ) //DcmTagKey(0x0012, 0x0060)
    GET_DICOM_INFO_LINE(DCM_PatientIdentityRemoved               ) //DcmTagKey(0x0012, 0x0062)
    GET_DICOM_INFO_LINE(DCM_DeidentificationMethod               ) //DcmTagKey(0x0012, 0x0063)
    GET_DICOM_INFO_LINE(DCM_DeidentificationMethodCodeSequence   ) //DcmTagKey(0x0012, 0x0064)
    GET_DICOM_INFO_LINE(DCM_ClinicalTrialSeriesID                ) //DcmTagKey(0x0012, 0x0071)
    GET_DICOM_INFO_LINE(DCM_ClinicalTrialSeriesDescription       ) //DcmTagKey(0x0012, 0x0072)
    GET_DICOM_INFO_LINE(DCM_ClinicalTrialProtocolEthicsCommitteeName ) //DcmTagKey(0x0012, 0x0081)
    GET_DICOM_INFO_LINE(DCM_ClinicalTrialProtocolEthicsCommitteeApprovalNumber ) //DcmTagKey(0x0012, 0x0082)
    GET_DICOM_INFO_LINE(DCM_ConsentForClinicalTrialUseSequence   ) //DcmTagKey(0x0012, 0x0083)
    GET_DICOM_INFO_LINE(DCM_DistributionType                     ) //DcmTagKey(0x0012, 0x0084)
    GET_DICOM_INFO_LINE(DCM_ConsentForDistributionFlag           ) //DcmTagKey(0x0012, 0x0085)
            info->append("</div>");
            info->append("<div style=\"color:#00007F\">");
    GET_DICOM_INFO_LINE(DCM_ContrastBolusAgent                   ) //DcmTagKey(0x0018, 0x0010)
    GET_DICOM_INFO_LINE(DCM_ContrastBolusAgentSequence           ) //DcmTagKey(0x0018, 0x0012)
    GET_DICOM_INFO_LINE(DCM_ContrastBolusAdministrationRouteSequence ) //DcmTagKey(0x0018, 0x0014)
    GET_DICOM_INFO_LINE(DCM_BodyPartExamined                     ) //DcmTagKey(0x0018, 0x0015)
    GET_DICOM_INFO_LINE(DCM_ScanningSequence                     ) //DcmTagKey(0x0018, 0x0020)
    GET_DICOM_INFO_LINE(DCM_SequenceVariant                      ) //DcmTagKey(0x0018, 0x0021)
    GET_DICOM_INFO_LINE(DCM_ScanOptions                          ) //DcmTagKey(0x0018, 0x0022)
    GET_DICOM_INFO_LINE(DCM_MRAcquisitionType                    ) //DcmTagKey(0x0018, 0x0023)
    GET_DICOM_INFO_LINE(DCM_SequenceName                         ) //DcmTagKey(0x0018, 0x0024)
    GET_DICOM_INFO_LINE(DCM_AngioFlag                            ) //DcmTagKey(0x0018, 0x0025)
    GET_DICOM_INFO_LINE(DCM_InterventionDrugInformationSequence  ) //DcmTagKey(0x0018, 0x0026)
    GET_DICOM_INFO_LINE(DCM_InterventionDrugStopTime             ) //DcmTagKey(0x0018, 0x0027)
    GET_DICOM_INFO_LINE(DCM_InterventionDrugDose                 ) //DcmTagKey(0x0018, 0x0028)
    GET_DICOM_INFO_LINE(DCM_InterventionDrugCodeSequence         ) //DcmTagKey(0x0018, 0x0029)
    GET_DICOM_INFO_LINE(DCM_AdditionalDrugSequence               ) //DcmTagKey(0x0018, 0x002a)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_Radionuclide                 ) //DcmTagKey(0x0018, 0x0030)
    GET_DICOM_INFO_LINE(DCM_Radiopharmaceutical                  ) //DcmTagKey(0x0018, 0x0031)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_EnergyWindowCenterline       ) //DcmTagKey(0x0018, 0x0032)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_EnergyWindowTotalWidth       ) //DcmTagKey(0x0018, 0x0033)
    GET_DICOM_INFO_LINE(DCM_InterventionDrugName                 ) //DcmTagKey(0x0018, 0x0034)
    GET_DICOM_INFO_LINE(DCM_InterventionDrugStartTime            ) //DcmTagKey(0x0018, 0x0035)
    GET_DICOM_INFO_LINE(DCM_InterventionSequence                 ) //DcmTagKey(0x0018, 0x0036)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_TherapyType                  ) //DcmTagKey(0x0018, 0x0037)
    GET_DICOM_INFO_LINE(DCM_InterventionStatus                   ) //DcmTagKey(0x0018, 0x0038)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_TherapyDescription           ) //DcmTagKey(0x0018, 0x0039)
    GET_DICOM_INFO_LINE(DCM_InterventionDescription              ) //DcmTagKey(0x0018, 0x003a)
    GET_DICOM_INFO_LINE(DCM_CineRate                             ) //DcmTagKey(0x0018, 0x0040)
    GET_DICOM_INFO_LINE(DCM_InitialCineRunState                  ) //DcmTagKey(0x0018, 0x0042)
    GET_DICOM_INFO_LINE(DCM_SliceThickness                       ) //DcmTagKey(0x0018, 0x0050)
    GET_DICOM_INFO_LINE(DCM_KVP                                  ) //DcmTagKey(0x0018, 0x0060)
    GET_DICOM_INFO_LINE(DCM_CountsAccumulated                    ) //DcmTagKey(0x0018, 0x0070)
    GET_DICOM_INFO_LINE(DCM_AcquisitionTerminationCondition      ) //DcmTagKey(0x0018, 0x0071)
    GET_DICOM_INFO_LINE(DCM_EffectiveDuration                    ) //DcmTagKey(0x0018, 0x0072)
    GET_DICOM_INFO_LINE(DCM_AcquisitionStartCondition            ) //DcmTagKey(0x0018, 0x0073)
    GET_DICOM_INFO_LINE(DCM_AcquisitionStartConditionData        ) //DcmTagKey(0x0018, 0x0074)
    GET_DICOM_INFO_LINE(DCM_AcquisitionTerminationConditionData  ) //DcmTagKey(0x0018, 0x0075)
    GET_DICOM_INFO_LINE(DCM_RepetitionTime                       ) //DcmTagKey(0x0018, 0x0080)
    GET_DICOM_INFO_LINE(DCM_EchoTime                             ) //DcmTagKey(0x0018, 0x0081)
    GET_DICOM_INFO_LINE(DCM_InversionTime                        ) //DcmTagKey(0x0018, 0x0082)
    GET_DICOM_INFO_LINE(DCM_NumberOfAverages                     ) //DcmTagKey(0x0018, 0x0083)
    GET_DICOM_INFO_LINE(DCM_ImagingFrequency                     ) //DcmTagKey(0x0018, 0x0084)
    GET_DICOM_INFO_LINE(DCM_ImagedNucleus                        ) //DcmTagKey(0x0018, 0x0085)
    GET_DICOM_INFO_LINE(DCM_EchoNumbers                          ) //DcmTagKey(0x0018, 0x0086)
    GET_DICOM_INFO_LINE(DCM_MagneticFieldStrength                ) //DcmTagKey(0x0018, 0x0087)
    GET_DICOM_INFO_LINE(DCM_SpacingBetweenSlices                 ) //DcmTagKey(0x0018, 0x0088)
    GET_DICOM_INFO_LINE(DCM_NumberOfPhaseEncodingSteps           ) //DcmTagKey(0x0018, 0x0089)
    GET_DICOM_INFO_LINE(DCM_DataCollectionDiameter               ) //DcmTagKey(0x0018, 0x0090)
    GET_DICOM_INFO_LINE(DCM_EchoTrainLength                      ) //DcmTagKey(0x0018, 0x0091)
    GET_DICOM_INFO_LINE(DCM_PercentSampling                      ) //DcmTagKey(0x0018, 0x0093)
    GET_DICOM_INFO_LINE(DCM_PercentPhaseFieldOfView              ) //DcmTagKey(0x0018, 0x0094)
    GET_DICOM_INFO_LINE(DCM_PixelBandwidth                       ) //DcmTagKey(0x0018, 0x0095)
    GET_DICOM_INFO_LINE(DCM_DeviceSerialNumber                   ) //DcmTagKey(0x0018, 0x1000)
    GET_DICOM_INFO_LINE(DCM_DeviceUID                            ) //DcmTagKey(0x0018, 0x1002)
    GET_DICOM_INFO_LINE(DCM_DeviceID                             ) //DcmTagKey(0x0018, 0x1003)
    GET_DICOM_INFO_LINE(DCM_PlateID                              ) //DcmTagKey(0x0018, 0x1004)
    GET_DICOM_INFO_LINE(DCM_GeneratorID                          ) //DcmTagKey(0x0018, 0x1005)
    GET_DICOM_INFO_LINE(DCM_GridID                               ) //DcmTagKey(0x0018, 0x1006)
    GET_DICOM_INFO_LINE(DCM_CassetteID                           ) //DcmTagKey(0x0018, 0x1007)
    GET_DICOM_INFO_LINE(DCM_GantryID                             ) //DcmTagKey(0x0018, 0x1008)
    GET_DICOM_INFO_LINE(DCM_SecondaryCaptureDeviceID             ) //DcmTagKey(0x0018, 0x1010)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_HardcopyCreationDeviceID     ) //DcmTagKey(0x0018, 0x1011)
    GET_DICOM_INFO_LINE(DCM_DateOfSecondaryCapture               ) //DcmTagKey(0x0018, 0x1012)
    GET_DICOM_INFO_LINE(DCM_TimeOfSecondaryCapture               ) //DcmTagKey(0x0018, 0x1014)
    GET_DICOM_INFO_LINE(DCM_SecondaryCaptureDeviceManufacturer   ) //DcmTagKey(0x0018, 0x1016)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_HardcopyDeviceManufacturer   ) //DcmTagKey(0x0018, 0x1017)
    GET_DICOM_INFO_LINE(DCM_SecondaryCaptureDeviceManufacturerModelName ) //DcmTagKey(0x0018, 0x1018)
    GET_DICOM_INFO_LINE(DCM_SecondaryCaptureDeviceSoftwareVersions ) //DcmTagKey(0x0018, 0x1019)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_HardcopyDeviceSoftwareVersion ) //DcmTagKey(0x0018, 0x101a)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_HardcopyDeviceManufacturerModelName ) //DcmTagKey(0x0018, 0x101b)
    GET_DICOM_INFO_LINE(DCM_SoftwareVersions                     ) //DcmTagKey(0x0018, 0x1020)
    GET_DICOM_INFO_LINE(DCM_VideoImageFormatAcquired             ) //DcmTagKey(0x0018, 0x1022)
    GET_DICOM_INFO_LINE(DCM_DigitalImageFormatAcquired           ) //DcmTagKey(0x0018, 0x1023)
    GET_DICOM_INFO_LINE(DCM_ProtocolName                         ) //DcmTagKey(0x0018, 0x1030)
    GET_DICOM_INFO_LINE(DCM_ContrastBolusRoute                   ) //DcmTagKey(0x0018, 0x1040)
    GET_DICOM_INFO_LINE(DCM_ContrastBolusVolume                  ) //DcmTagKey(0x0018, 0x1041)
    GET_DICOM_INFO_LINE(DCM_ContrastBolusStartTime               ) //DcmTagKey(0x0018, 0x1042)
    GET_DICOM_INFO_LINE(DCM_ContrastBolusStopTime                ) //DcmTagKey(0x0018, 0x1043)
    GET_DICOM_INFO_LINE(DCM_ContrastBolusTotalDose               ) //DcmTagKey(0x0018, 0x1044)
    GET_DICOM_INFO_LINE(DCM_SyringeCounts                        ) //DcmTagKey(0x0018, 0x1045)
    GET_DICOM_INFO_LINE(DCM_ContrastFlowRate                     ) //DcmTagKey(0x0018, 0x1046)
    GET_DICOM_INFO_LINE(DCM_ContrastFlowDuration                 ) //DcmTagKey(0x0018, 0x1047)
    GET_DICOM_INFO_LINE(DCM_ContrastBolusIngredient              ) //DcmTagKey(0x0018, 0x1048)
    GET_DICOM_INFO_LINE(DCM_ContrastBolusIngredientConcentration ) //DcmTagKey(0x0018, 0x1049)
    GET_DICOM_INFO_LINE(DCM_SpatialResolution                    ) //DcmTagKey(0x0018, 0x1050)
    GET_DICOM_INFO_LINE(DCM_TriggerTime                          ) //DcmTagKey(0x0018, 0x1060)
    GET_DICOM_INFO_LINE(DCM_TriggerSourceOrType                  ) //DcmTagKey(0x0018, 0x1061)
    GET_DICOM_INFO_LINE(DCM_NominalInterval                      ) //DcmTagKey(0x0018, 0x1062)
    GET_DICOM_INFO_LINE(DCM_FrameTime                            ) //DcmTagKey(0x0018, 0x1063)
    GET_DICOM_INFO_LINE(DCM_CardiacFramingType                   ) //DcmTagKey(0x0018, 0x1064)
    GET_DICOM_INFO_LINE(DCM_FrameTimeVector                      ) //DcmTagKey(0x0018, 0x1065)
    GET_DICOM_INFO_LINE(DCM_FrameDelay                           ) //DcmTagKey(0x0018, 0x1066)
    GET_DICOM_INFO_LINE(DCM_ImageTriggerDelay                    ) //DcmTagKey(0x0018, 0x1067)
    GET_DICOM_INFO_LINE(DCM_MultiplexGroupTimeOffset             ) //DcmTagKey(0x0018, 0x1068)
    GET_DICOM_INFO_LINE(DCM_TriggerTimeOffset                    ) //DcmTagKey(0x0018, 0x1069)
    GET_DICOM_INFO_LINE(DCM_SynchronizationTrigger               ) //DcmTagKey(0x0018, 0x106a)
    GET_DICOM_INFO_LINE(DCM_SynchronizationChannel               ) //DcmTagKey(0x0018, 0x106c)
    GET_DICOM_INFO_LINE(DCM_TriggerSamplePosition                ) //DcmTagKey(0x0018, 0x106e)
    GET_DICOM_INFO_LINE(DCM_RadiopharmaceuticalRoute             ) //DcmTagKey(0x0018, 0x1070)
    GET_DICOM_INFO_LINE(DCM_RadiopharmaceuticalVolume            ) //DcmTagKey(0x0018, 0x1071)
    GET_DICOM_INFO_LINE(DCM_RadiopharmaceuticalStartTime         ) //DcmTagKey(0x0018, 0x1072)
    GET_DICOM_INFO_LINE(DCM_RadiopharmaceuticalStopTime          ) //DcmTagKey(0x0018, 0x1073)
    GET_DICOM_INFO_LINE(DCM_RadionuclideTotalDose                ) //DcmTagKey(0x0018, 0x1074)
    GET_DICOM_INFO_LINE(DCM_RadionuclideHalfLife                 ) //DcmTagKey(0x0018, 0x1075)
    GET_DICOM_INFO_LINE(DCM_RadionuclidePositronFraction         ) //DcmTagKey(0x0018, 0x1076)
    GET_DICOM_INFO_LINE(DCM_RadiopharmaceuticalSpecificActivity  ) //DcmTagKey(0x0018, 0x1077)
    GET_DICOM_INFO_LINE(DCM_RadiopharmaceuticalStartDateTime     ) //DcmTagKey(0x0018, 0x1078)
    GET_DICOM_INFO_LINE(DCM_RadiopharmaceuticalStopDateTime      ) //DcmTagKey(0x0018, 0x1079)
    GET_DICOM_INFO_LINE(DCM_BeatRejectionFlag                    ) //DcmTagKey(0x0018, 0x1080)
    GET_DICOM_INFO_LINE(DCM_LowRRValue                           ) //DcmTagKey(0x0018, 0x1081)
    GET_DICOM_INFO_LINE(DCM_HighRRValue                          ) //DcmTagKey(0x0018, 0x1082)
    GET_DICOM_INFO_LINE(DCM_IntervalsAcquired                    ) //DcmTagKey(0x0018, 0x1083)
    GET_DICOM_INFO_LINE(DCM_IntervalsRejected                    ) //DcmTagKey(0x0018, 0x1084)
    GET_DICOM_INFO_LINE(DCM_PVCRejection                         ) //DcmTagKey(0x0018, 0x1085)
    GET_DICOM_INFO_LINE(DCM_SkipBeats                            ) //DcmTagKey(0x0018, 0x1086)
    GET_DICOM_INFO_LINE(DCM_HeartRate                            ) //DcmTagKey(0x0018, 0x1088)
    GET_DICOM_INFO_LINE(DCM_CardiacNumberOfImages                ) //DcmTagKey(0x0018, 0x1090)
    GET_DICOM_INFO_LINE(DCM_TriggerWindow                        ) //DcmTagKey(0x0018, 0x1094)
    GET_DICOM_INFO_LINE(DCM_ReconstructionDiameter               ) //DcmTagKey(0x0018, 0x1100)
    GET_DICOM_INFO_LINE(DCM_DistanceSourceToDetector             ) //DcmTagKey(0x0018, 0x1110)
    GET_DICOM_INFO_LINE(DCM_DistanceSourceToPatient              ) //DcmTagKey(0x0018, 0x1111)
    GET_DICOM_INFO_LINE(DCM_EstimatedRadiographicMagnificationFactor ) //DcmTagKey(0x0018, 0x1114)
    GET_DICOM_INFO_LINE(DCM_GantryDetectorTilt                   ) //DcmTagKey(0x0018, 0x1120)
    GET_DICOM_INFO_LINE(DCM_GantryDetectorSlew                   ) //DcmTagKey(0x0018, 0x1121)
    GET_DICOM_INFO_LINE(DCM_TableHeight                          ) //DcmTagKey(0x0018, 0x1130)
    GET_DICOM_INFO_LINE(DCM_TableTraverse                        ) //DcmTagKey(0x0018, 0x1131)
    GET_DICOM_INFO_LINE(DCM_TableMotion                          ) //DcmTagKey(0x0018, 0x1134)
    GET_DICOM_INFO_LINE(DCM_TableVerticalIncrement               ) //DcmTagKey(0x0018, 0x1135)
    GET_DICOM_INFO_LINE(DCM_TableLateralIncrement                ) //DcmTagKey(0x0018, 0x1136)
    GET_DICOM_INFO_LINE(DCM_TableLongitudinalIncrement           ) //DcmTagKey(0x0018, 0x1137)
    GET_DICOM_INFO_LINE(DCM_TableAngle                           ) //DcmTagKey(0x0018, 0x1138)
    GET_DICOM_INFO_LINE(DCM_TableType                            ) //DcmTagKey(0x0018, 0x113a)
    GET_DICOM_INFO_LINE(DCM_RotationDirection                    ) //DcmTagKey(0x0018, 0x1140)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_AngularPosition              ) //DcmTagKey(0x0018, 0x1141)
    GET_DICOM_INFO_LINE(DCM_RadialPosition                       ) //DcmTagKey(0x0018, 0x1142)
    GET_DICOM_INFO_LINE(DCM_ScanArc                              ) //DcmTagKey(0x0018, 0x1143)
    GET_DICOM_INFO_LINE(DCM_AngularStep                          ) //DcmTagKey(0x0018, 0x1144)
    GET_DICOM_INFO_LINE(DCM_CenterOfRotationOffset               ) //DcmTagKey(0x0018, 0x1145)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_RotationOffset               ) //DcmTagKey(0x0018, 0x1146)
    GET_DICOM_INFO_LINE(DCM_FieldOfViewShape                     ) //DcmTagKey(0x0018, 0x1147)
    GET_DICOM_INFO_LINE(DCM_FieldOfViewDimensions                ) //DcmTagKey(0x0018, 0x1149)
    GET_DICOM_INFO_LINE(DCM_ExposureTime                         ) //DcmTagKey(0x0018, 0x1150)
    GET_DICOM_INFO_LINE(DCM_XRayTubeCurrent                      ) //DcmTagKey(0x0018, 0x1151)
    GET_DICOM_INFO_LINE(DCM_Exposure                             ) //DcmTagKey(0x0018, 0x1152)
    GET_DICOM_INFO_LINE(DCM_ExposureInuAs                        ) //DcmTagKey(0x0018, 0x1153)
    GET_DICOM_INFO_LINE(DCM_AveragePulseWidth                    ) //DcmTagKey(0x0018, 0x1154)
    GET_DICOM_INFO_LINE(DCM_RadiationSetting                     ) //DcmTagKey(0x0018, 0x1155)
    GET_DICOM_INFO_LINE(DCM_RectificationType                    ) //DcmTagKey(0x0018, 0x1156)
    GET_DICOM_INFO_LINE(DCM_RadiationMode                        ) //DcmTagKey(0x0018, 0x115a)
    GET_DICOM_INFO_LINE(DCM_ImageAndFluoroscopyAreaDoseProduct   ) //DcmTagKey(0x0018, 0x115e)
    GET_DICOM_INFO_LINE(DCM_FilterType                           ) //DcmTagKey(0x0018, 0x1160)
    GET_DICOM_INFO_LINE(DCM_TypeOfFilters                        ) //DcmTagKey(0x0018, 0x1161)
    GET_DICOM_INFO_LINE(DCM_IntensifierSize                      ) //DcmTagKey(0x0018, 0x1162)
    GET_DICOM_INFO_LINE(DCM_ImagerPixelSpacing                   ) //DcmTagKey(0x0018, 0x1164)
    GET_DICOM_INFO_LINE(DCM_Grid                                 ) //DcmTagKey(0x0018, 0x1166)
    GET_DICOM_INFO_LINE(DCM_GeneratorPower                       ) //DcmTagKey(0x0018, 0x1170)
    GET_DICOM_INFO_LINE(DCM_CollimatorGridName                   ) //DcmTagKey(0x0018, 0x1180)
    GET_DICOM_INFO_LINE(DCM_CollimatorType                       ) //DcmTagKey(0x0018, 0x1181)
    GET_DICOM_INFO_LINE(DCM_FocalDistance                        ) //DcmTagKey(0x0018, 0x1182)
    GET_DICOM_INFO_LINE(DCM_XFocusCenter                         ) //DcmTagKey(0x0018, 0x1183)
    GET_DICOM_INFO_LINE(DCM_YFocusCenter                         ) //DcmTagKey(0x0018, 0x1184)
    GET_DICOM_INFO_LINE(DCM_FocalSpots                           ) //DcmTagKey(0x0018, 0x1190)
    GET_DICOM_INFO_LINE(DCM_AnodeTargetMaterial                  ) //DcmTagKey(0x0018, 0x1191)
    GET_DICOM_INFO_LINE(DCM_BodyPartThickness                    ) //DcmTagKey(0x0018, 0x11a0)
    GET_DICOM_INFO_LINE(DCM_CompressionForce                     ) //DcmTagKey(0x0018, 0x11a2)
    GET_DICOM_INFO_LINE(DCM_DateOfLastCalibration                ) //DcmTagKey(0x0018, 0x1200)
    GET_DICOM_INFO_LINE(DCM_TimeOfLastCalibration                ) //DcmTagKey(0x0018, 0x1201)
    GET_DICOM_INFO_LINE(DCM_ConvolutionKernel                    ) //DcmTagKey(0x0018, 0x1210)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_UpperLowerPixelValues       ) //DcmTagKey(0x0018, 0x1240)
    GET_DICOM_INFO_LINE(DCM_ActualFrameDuration                  ) //DcmTagKey(0x0018, 0x1242)
    GET_DICOM_INFO_LINE(DCM_CountRate                            ) //DcmTagKey(0x0018, 0x1243)
    GET_DICOM_INFO_LINE(DCM_PreferredPlaybackSequencing          ) //DcmTagKey(0x0018, 0x1244)
    GET_DICOM_INFO_LINE(DCM_ReceiveCoilName                      ) //DcmTagKey(0x0018, 0x1250)
    GET_DICOM_INFO_LINE(DCM_TransmitCoilName                     ) //DcmTagKey(0x0018, 0x1251)
    GET_DICOM_INFO_LINE(DCM_PlateType                            ) //DcmTagKey(0x0018, 0x1260)
    GET_DICOM_INFO_LINE(DCM_PhosphorType                         ) //DcmTagKey(0x0018, 0x1261)
    GET_DICOM_INFO_LINE(DCM_ScanVelocity                         ) //DcmTagKey(0x0018, 0x1300)
    GET_DICOM_INFO_LINE(DCM_WholeBodyTechnique                   ) //DcmTagKey(0x0018, 0x1301)
    GET_DICOM_INFO_LINE(DCM_ScanLength                           ) //DcmTagKey(0x0018, 0x1302)
    GET_DICOM_INFO_LINE(DCM_AcquisitionMatrix                    ) //DcmTagKey(0x0018, 0x1310)
    GET_DICOM_INFO_LINE(DCM_InPlanePhaseEncodingDirection        ) //DcmTagKey(0x0018, 0x1312)
    GET_DICOM_INFO_LINE(DCM_FlipAngle                            ) //DcmTagKey(0x0018, 0x1314)
    GET_DICOM_INFO_LINE(DCM_VariableFlipAngleFlag                ) //DcmTagKey(0x0018, 0x1315)
    GET_DICOM_INFO_LINE(DCM_SAR                                  ) //DcmTagKey(0x0018, 0x1316)
    GET_DICOM_INFO_LINE(DCM_dBdt                                 ) //DcmTagKey(0x0018, 0x1318)
    GET_DICOM_INFO_LINE(DCM_AcquisitionDeviceProcessingDescription ) //DcmTagKey(0x0018, 0x1400)
    GET_DICOM_INFO_LINE(DCM_AcquisitionDeviceProcessingCode      ) //DcmTagKey(0x0018, 0x1401)
    GET_DICOM_INFO_LINE(DCM_CassetteOrientation                  ) //DcmTagKey(0x0018, 0x1402)
    GET_DICOM_INFO_LINE(DCM_CassetteSize                         ) //DcmTagKey(0x0018, 0x1403)
    GET_DICOM_INFO_LINE(DCM_ExposuresOnPlate                     ) //DcmTagKey(0x0018, 0x1404)
    GET_DICOM_INFO_LINE(DCM_RelativeXRayExposure                 ) //DcmTagKey(0x0018, 0x1405)
    GET_DICOM_INFO_LINE(DCM_ExposureIndex                        ) //DcmTagKey(0x0018, 0x1411)
    GET_DICOM_INFO_LINE(DCM_TargetExposureIndex                  ) //DcmTagKey(0x0018, 0x1412)
    GET_DICOM_INFO_LINE(DCM_DeviationIndex                       ) //DcmTagKey(0x0018, 0x1413)
    GET_DICOM_INFO_LINE(DCM_ColumnAngulation                     ) //DcmTagKey(0x0018, 0x1450)
    GET_DICOM_INFO_LINE(DCM_TomoLayerHeight                      ) //DcmTagKey(0x0018, 0x1460)
    GET_DICOM_INFO_LINE(DCM_TomoAngle                            ) //DcmTagKey(0x0018, 0x1470)
    GET_DICOM_INFO_LINE(DCM_TomoTime                             ) //DcmTagKey(0x0018, 0x1480)
    GET_DICOM_INFO_LINE(DCM_TomoType                             ) //DcmTagKey(0x0018, 0x1490)
    GET_DICOM_INFO_LINE(DCM_TomoClass                            ) //DcmTagKey(0x0018, 0x1491)
    GET_DICOM_INFO_LINE(DCM_NumberOfTomosynthesisSourceImages    ) //DcmTagKey(0x0018, 0x1495)
    GET_DICOM_INFO_LINE(DCM_PositionerMotion                     ) //DcmTagKey(0x0018, 0x1500)
    GET_DICOM_INFO_LINE(DCM_PositionerType                       ) //DcmTagKey(0x0018, 0x1508)
    GET_DICOM_INFO_LINE(DCM_PositionerPrimaryAngle               ) //DcmTagKey(0x0018, 0x1510)
    GET_DICOM_INFO_LINE(DCM_PositionerSecondaryAngle             ) //DcmTagKey(0x0018, 0x1511)
    GET_DICOM_INFO_LINE(DCM_PositionerPrimaryAngleIncrement      ) //DcmTagKey(0x0018, 0x1520)
    GET_DICOM_INFO_LINE(DCM_PositionerSecondaryAngleIncrement    ) //DcmTagKey(0x0018, 0x1521)
    GET_DICOM_INFO_LINE(DCM_DetectorPrimaryAngle                 ) //DcmTagKey(0x0018, 0x1530)
    GET_DICOM_INFO_LINE(DCM_DetectorSecondaryAngle               ) //DcmTagKey(0x0018, 0x1531)
    GET_DICOM_INFO_LINE(DCM_ShutterShape                         ) //DcmTagKey(0x0018, 0x1600)
    GET_DICOM_INFO_LINE(DCM_ShutterLeftVerticalEdge              ) //DcmTagKey(0x0018, 0x1602)
    GET_DICOM_INFO_LINE(DCM_ShutterRightVerticalEdge             ) //DcmTagKey(0x0018, 0x1604)
    GET_DICOM_INFO_LINE(DCM_ShutterUpperHorizontalEdge           ) //DcmTagKey(0x0018, 0x1606)
    GET_DICOM_INFO_LINE(DCM_ShutterLowerHorizontalEdge           ) //DcmTagKey(0x0018, 0x1608)
    GET_DICOM_INFO_LINE(DCM_CenterOfCircularShutter              ) //DcmTagKey(0x0018, 0x1610)
    GET_DICOM_INFO_LINE(DCM_RadiusOfCircularShutter              ) //DcmTagKey(0x0018, 0x1612)
    GET_DICOM_INFO_LINE(DCM_VerticesOfThePolygonalShutter        ) //DcmTagKey(0x0018, 0x1620)
    GET_DICOM_INFO_LINE(DCM_ShutterPresentationValue             ) //DcmTagKey(0x0018, 0x1622)
    GET_DICOM_INFO_LINE(DCM_ShutterOverlayGroup                  ) //DcmTagKey(0x0018, 0x1623)
    GET_DICOM_INFO_LINE(DCM_ShutterPresentationColorCIELabValue  ) //DcmTagKey(0x0018, 0x1624)
    GET_DICOM_INFO_LINE(DCM_CollimatorShape                      ) //DcmTagKey(0x0018, 0x1700)
    GET_DICOM_INFO_LINE(DCM_CollimatorLeftVerticalEdge           ) //DcmTagKey(0x0018, 0x1702)
    GET_DICOM_INFO_LINE(DCM_CollimatorRightVerticalEdge          ) //DcmTagKey(0x0018, 0x1704)
    GET_DICOM_INFO_LINE(DCM_CollimatorUpperHorizontalEdge        ) //DcmTagKey(0x0018, 0x1706)
    GET_DICOM_INFO_LINE(DCM_CollimatorLowerHorizontalEdge        ) //DcmTagKey(0x0018, 0x1708)
    GET_DICOM_INFO_LINE(DCM_CenterOfCircularCollimator           ) //DcmTagKey(0x0018, 0x1710)
    GET_DICOM_INFO_LINE(DCM_RadiusOfCircularCollimator           ) //DcmTagKey(0x0018, 0x1712)
    GET_DICOM_INFO_LINE(DCM_VerticesOfThePolygonalCollimator     ) //DcmTagKey(0x0018, 0x1720)
    GET_DICOM_INFO_LINE(DCM_AcquisitionTimeSynchronized          ) //DcmTagKey(0x0018, 0x1800)
    GET_DICOM_INFO_LINE(DCM_TimeSource                           ) //DcmTagKey(0x0018, 0x1801)
    GET_DICOM_INFO_LINE(DCM_TimeDistributionProtocol             ) //DcmTagKey(0x0018, 0x1802)
    GET_DICOM_INFO_LINE(DCM_NTPSourceAddress                     ) //DcmTagKey(0x0018, 0x1803)
    GET_DICOM_INFO_LINE(DCM_PageNumberVector                     ) //DcmTagKey(0x0018, 0x2001)
    GET_DICOM_INFO_LINE(DCM_FrameLabelVector                     ) //DcmTagKey(0x0018, 0x2002)
    GET_DICOM_INFO_LINE(DCM_FramePrimaryAngleVector              ) //DcmTagKey(0x0018, 0x2003)
    GET_DICOM_INFO_LINE(DCM_FrameSecondaryAngleVector            ) //DcmTagKey(0x0018, 0x2004)
    GET_DICOM_INFO_LINE(DCM_SliceLocationVector                  ) //DcmTagKey(0x0018, 0x2005)
    GET_DICOM_INFO_LINE(DCM_DisplayWindowLabelVector             ) //DcmTagKey(0x0018, 0x2006)
    GET_DICOM_INFO_LINE(DCM_NominalScannedPixelSpacing           ) //DcmTagKey(0x0018, 0x2010)
    GET_DICOM_INFO_LINE(DCM_DigitizingDeviceTransportDirection   ) //DcmTagKey(0x0018, 0x2020)
    GET_DICOM_INFO_LINE(DCM_RotationOfScannedFilm                ) //DcmTagKey(0x0018, 0x2030)
    GET_DICOM_INFO_LINE(DCM_IVUSAcquisition                      ) //DcmTagKey(0x0018, 0x3100)
    GET_DICOM_INFO_LINE(DCM_IVUSPullbackRate                     ) //DcmTagKey(0x0018, 0x3101)
    GET_DICOM_INFO_LINE(DCM_IVUSGatedRate                        ) //DcmTagKey(0x0018, 0x3102)
    GET_DICOM_INFO_LINE(DCM_IVUSPullbackStartFrameNumber         ) //DcmTagKey(0x0018, 0x3103)
    GET_DICOM_INFO_LINE(DCM_IVUSPullbackStopFrameNumber          ) //DcmTagKey(0x0018, 0x3104)
    GET_DICOM_INFO_LINE(DCM_LesionNumber                         ) //DcmTagKey(0x0018, 0x3105)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_AcquisitionComments         ) //DcmTagKey(0x0018, 0x4000)
    GET_DICOM_INFO_LINE(DCM_OutputPower                          ) //DcmTagKey(0x0018, 0x5000)
    GET_DICOM_INFO_LINE(DCM_TransducerData                       ) //DcmTagKey(0x0018, 0x5010)
    GET_DICOM_INFO_LINE(DCM_FocusDepth                           ) //DcmTagKey(0x0018, 0x5012)
    GET_DICOM_INFO_LINE(DCM_ProcessingFunction                   ) //DcmTagKey(0x0018, 0x5020)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_PostprocessingFunction       ) //DcmTagKey(0x0018, 0x5021)
    GET_DICOM_INFO_LINE(DCM_MechanicalIndex                      ) //DcmTagKey(0x0018, 0x5022)
    GET_DICOM_INFO_LINE(DCM_BoneThermalIndex                     ) //DcmTagKey(0x0018, 0x5024)
    GET_DICOM_INFO_LINE(DCM_CranialThermalIndex                  ) //DcmTagKey(0x0018, 0x5026)
    GET_DICOM_INFO_LINE(DCM_SoftTissueThermalIndex               ) //DcmTagKey(0x0018, 0x5027)
    GET_DICOM_INFO_LINE(DCM_SoftTissueFocusThermalIndex          ) //DcmTagKey(0x0018, 0x5028)
    GET_DICOM_INFO_LINE(DCM_SoftTissueSurfaceThermalIndex        ) //DcmTagKey(0x0018, 0x5029)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_DynamicRange                ) //DcmTagKey(0x0018, 0x5030)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_TotalGain                   ) //DcmTagKey(0x0018, 0x5040)
    GET_DICOM_INFO_LINE(DCM_DepthOfScanField                     ) //DcmTagKey(0x0018, 0x5050)
    GET_DICOM_INFO_LINE(DCM_PatientPosition                      ) //DcmTagKey(0x0018, 0x5100)
    GET_DICOM_INFO_LINE(DCM_ViewPosition                         ) //DcmTagKey(0x0018, 0x5101)
    GET_DICOM_INFO_LINE(DCM_ProjectionEponymousNameCodeSequence  ) //DcmTagKey(0x0018, 0x5104)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ImageTransformationMatrix    ) //DcmTagKey(0x0018, 0x5210)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ImageTranslationVector       ) //DcmTagKey(0x0018, 0x5212)
    GET_DICOM_INFO_LINE(DCM_Sensitivity                          ) //DcmTagKey(0x0018, 0x6000)
    GET_DICOM_INFO_LINE(DCM_SequenceOfUltrasoundRegions          ) //DcmTagKey(0x0018, 0x6011)
    GET_DICOM_INFO_LINE(DCM_RegionSpatialFormat                  ) //DcmTagKey(0x0018, 0x6012)
    GET_DICOM_INFO_LINE(DCM_RegionDataType                       ) //DcmTagKey(0x0018, 0x6014)
    GET_DICOM_INFO_LINE(DCM_RegionFlags                          ) //DcmTagKey(0x0018, 0x6016)
    GET_DICOM_INFO_LINE(DCM_RegionLocationMinX0                  ) //DcmTagKey(0x0018, 0x6018)
    GET_DICOM_INFO_LINE(DCM_RegionLocationMinY0                  ) //DcmTagKey(0x0018, 0x601a)
    GET_DICOM_INFO_LINE(DCM_RegionLocationMaxX1                  ) //DcmTagKey(0x0018, 0x601c)
    GET_DICOM_INFO_LINE(DCM_RegionLocationMaxY1                  ) //DcmTagKey(0x0018, 0x601e)
    GET_DICOM_INFO_LINE(DCM_ReferencePixelX0                     ) //DcmTagKey(0x0018, 0x6020)
    GET_DICOM_INFO_LINE(DCM_ReferencePixelY0                     ) //DcmTagKey(0x0018, 0x6022)
    GET_DICOM_INFO_LINE(DCM_PhysicalUnitsXDirection              ) //DcmTagKey(0x0018, 0x6024)
    GET_DICOM_INFO_LINE(DCM_PhysicalUnitsYDirection              ) //DcmTagKey(0x0018, 0x6026)
    GET_DICOM_INFO_LINE(DCM_ReferencePixelPhysicalValueX         ) //DcmTagKey(0x0018, 0x6028)
    GET_DICOM_INFO_LINE(DCM_ReferencePixelPhysicalValueY         ) //DcmTagKey(0x0018, 0x602a)
    GET_DICOM_INFO_LINE(DCM_PhysicalDeltaX                       ) //DcmTagKey(0x0018, 0x602c)
    GET_DICOM_INFO_LINE(DCM_PhysicalDeltaY                       ) //DcmTagKey(0x0018, 0x602e)
    GET_DICOM_INFO_LINE(DCM_TransducerFrequency                  ) //DcmTagKey(0x0018, 0x6030)
    GET_DICOM_INFO_LINE(DCM_TransducerType                       ) //DcmTagKey(0x0018, 0x6031)
    GET_DICOM_INFO_LINE(DCM_PulseRepetitionFrequency             ) //DcmTagKey(0x0018, 0x6032)
    GET_DICOM_INFO_LINE(DCM_DopplerCorrectionAngle               ) //DcmTagKey(0x0018, 0x6034)
    GET_DICOM_INFO_LINE(DCM_SteeringAngle                        ) //DcmTagKey(0x0018, 0x6036)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_DopplerSampleVolumeXPositionRetired ) //DcmTagKey(0x0018, 0x6038)
    GET_DICOM_INFO_LINE(DCM_DopplerSampleVolumeXPosition         ) //DcmTagKey(0x0018, 0x6039)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_DopplerSampleVolumeYPositionRetired ) //DcmTagKey(0x0018, 0x603a)
    GET_DICOM_INFO_LINE(DCM_DopplerSampleVolumeYPosition         ) //DcmTagKey(0x0018, 0x603b)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_TMLinePositionX0Retired      ) //DcmTagKey(0x0018, 0x603c)
    GET_DICOM_INFO_LINE(DCM_TMLinePositionX0                     ) //DcmTagKey(0x0018, 0x603d)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_TMLinePositionY0Retired      ) //DcmTagKey(0x0018, 0x603e)
    GET_DICOM_INFO_LINE(DCM_TMLinePositionY0                     ) //DcmTagKey(0x0018, 0x603f)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_TMLinePositionX1Retired      ) //DcmTagKey(0x0018, 0x6040)
    GET_DICOM_INFO_LINE(DCM_TMLinePositionX1                     ) //DcmTagKey(0x0018, 0x6041)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_TMLinePositionY1Retired      ) //DcmTagKey(0x0018, 0x6042)
    GET_DICOM_INFO_LINE(DCM_TMLinePositionY1                     ) //DcmTagKey(0x0018, 0x6043)
    GET_DICOM_INFO_LINE(DCM_PixelComponentOrganization           ) //DcmTagKey(0x0018, 0x6044)
    GET_DICOM_INFO_LINE(DCM_PixelComponentMask                   ) //DcmTagKey(0x0018, 0x6046)
    GET_DICOM_INFO_LINE(DCM_PixelComponentRangeStart             ) //DcmTagKey(0x0018, 0x6048)
    GET_DICOM_INFO_LINE(DCM_PixelComponentRangeStop              ) //DcmTagKey(0x0018, 0x604a)
    GET_DICOM_INFO_LINE(DCM_PixelComponentPhysicalUnits          ) //DcmTagKey(0x0018, 0x604c)
    GET_DICOM_INFO_LINE(DCM_PixelComponentDataType               ) //DcmTagKey(0x0018, 0x604e)
    GET_DICOM_INFO_LINE(DCM_NumberOfTableBreakPoints             ) //DcmTagKey(0x0018, 0x6050)
    GET_DICOM_INFO_LINE(DCM_TableOfXBreakPoints                  ) //DcmTagKey(0x0018, 0x6052)
    GET_DICOM_INFO_LINE(DCM_TableOfYBreakPoints                  ) //DcmTagKey(0x0018, 0x6054)
    GET_DICOM_INFO_LINE(DCM_NumberOfTableEntries                 ) //DcmTagKey(0x0018, 0x6056)
    GET_DICOM_INFO_LINE(DCM_TableOfPixelValues                   ) //DcmTagKey(0x0018, 0x6058)
    GET_DICOM_INFO_LINE(DCM_TableOfParameterValues               ) //DcmTagKey(0x0018, 0x605a)
    GET_DICOM_INFO_LINE(DCM_RWaveTimeVector                      ) //DcmTagKey(0x0018, 0x6060)
    GET_DICOM_INFO_LINE(DCM_DetectorConditionsNominalFlag        ) //DcmTagKey(0x0018, 0x7000)
    GET_DICOM_INFO_LINE(DCM_DetectorTemperature                  ) //DcmTagKey(0x0018, 0x7001)
    GET_DICOM_INFO_LINE(DCM_DetectorType                         ) //DcmTagKey(0x0018, 0x7004)
    GET_DICOM_INFO_LINE(DCM_DetectorConfiguration                ) //DcmTagKey(0x0018, 0x7005)
    GET_DICOM_INFO_LINE(DCM_DetectorDescription                  ) //DcmTagKey(0x0018, 0x7006)
    GET_DICOM_INFO_LINE(DCM_DetectorMode                         ) //DcmTagKey(0x0018, 0x7008)
    GET_DICOM_INFO_LINE(DCM_DetectorID                           ) //DcmTagKey(0x0018, 0x700a)
    GET_DICOM_INFO_LINE(DCM_DateOfLastDetectorCalibration        ) //DcmTagKey(0x0018, 0x700c)
    GET_DICOM_INFO_LINE(DCM_TimeOfLastDetectorCalibration        ) //DcmTagKey(0x0018, 0x700e)
    GET_DICOM_INFO_LINE(DCM_ExposuresOnDetectorSinceLastCalibration ) //DcmTagKey(0x0018, 0x7010)
    GET_DICOM_INFO_LINE(DCM_ExposuresOnDetectorSinceManufactured ) //DcmTagKey(0x0018, 0x7011)
    GET_DICOM_INFO_LINE(DCM_DetectorTimeSinceLastExposure        ) //DcmTagKey(0x0018, 0x7012)
    GET_DICOM_INFO_LINE(DCM_DetectorActiveTime                   ) //DcmTagKey(0x0018, 0x7014)
    GET_DICOM_INFO_LINE(DCM_DetectorActivationOffsetFromExposure ) //DcmTagKey(0x0018, 0x7016)
    GET_DICOM_INFO_LINE(DCM_DetectorBinning                      ) //DcmTagKey(0x0018, 0x701a)
    GET_DICOM_INFO_LINE(DCM_DetectorElementPhysicalSize          ) //DcmTagKey(0x0018, 0x7020)
    GET_DICOM_INFO_LINE(DCM_DetectorElementSpacing               ) //DcmTagKey(0x0018, 0x7022)
    GET_DICOM_INFO_LINE(DCM_DetectorActiveShape                  ) //DcmTagKey(0x0018, 0x7024)
    GET_DICOM_INFO_LINE(DCM_DetectorActiveDimensions             ) //DcmTagKey(0x0018, 0x7026)
    GET_DICOM_INFO_LINE(DCM_DetectorActiveOrigin                 ) //DcmTagKey(0x0018, 0x7028)
    GET_DICOM_INFO_LINE(DCM_DetectorManufacturerName             ) //DcmTagKey(0x0018, 0x702a)
    GET_DICOM_INFO_LINE(DCM_DetectorManufacturerModelName        ) //DcmTagKey(0x0018, 0x702b)
    GET_DICOM_INFO_LINE(DCM_FieldOfViewOrigin                    ) //DcmTagKey(0x0018, 0x7030)
    GET_DICOM_INFO_LINE(DCM_FieldOfViewRotation                  ) //DcmTagKey(0x0018, 0x7032)
    GET_DICOM_INFO_LINE(DCM_FieldOfViewHorizontalFlip            ) //DcmTagKey(0x0018, 0x7034)
    GET_DICOM_INFO_LINE(DCM_GridAbsorbingMaterial                ) //DcmTagKey(0x0018, 0x7040)
    GET_DICOM_INFO_LINE(DCM_GridSpacingMaterial                  ) //DcmTagKey(0x0018, 0x7041)
    GET_DICOM_INFO_LINE(DCM_GridThickness                        ) //DcmTagKey(0x0018, 0x7042)
    GET_DICOM_INFO_LINE(DCM_GridPitch                            ) //DcmTagKey(0x0018, 0x7044)
    GET_DICOM_INFO_LINE(DCM_GridAspectRatio                      ) //DcmTagKey(0x0018, 0x7046)
    GET_DICOM_INFO_LINE(DCM_GridPeriod                           ) //DcmTagKey(0x0018, 0x7048)
    GET_DICOM_INFO_LINE(DCM_GridFocalDistance                    ) //DcmTagKey(0x0018, 0x704c)
    GET_DICOM_INFO_LINE(DCM_FilterMaterial                       ) //DcmTagKey(0x0018, 0x7050)
    GET_DICOM_INFO_LINE(DCM_FilterThicknessMinimum               ) //DcmTagKey(0x0018, 0x7052)
    GET_DICOM_INFO_LINE(DCM_FilterThicknessMaximum               ) //DcmTagKey(0x0018, 0x7054)
    GET_DICOM_INFO_LINE(DCM_FilterBeamPathLengthMinimum          ) //DcmTagKey(0x0018, 0x7056)
    GET_DICOM_INFO_LINE(DCM_FilterBeamPathLengthMaximum          ) //DcmTagKey(0x0018, 0x7058)
    GET_DICOM_INFO_LINE(DCM_ExposureControlMode                  ) //DcmTagKey(0x0018, 0x7060)
    GET_DICOM_INFO_LINE(DCM_ExposureControlModeDescription       ) //DcmTagKey(0x0018, 0x7062)
    GET_DICOM_INFO_LINE(DCM_ExposureStatus                       ) //DcmTagKey(0x0018, 0x7064)
    GET_DICOM_INFO_LINE(DCM_PhototimerSetting                    ) //DcmTagKey(0x0018, 0x7065)
    GET_DICOM_INFO_LINE(DCM_ExposureTimeInuS                     ) //DcmTagKey(0x0018, 0x8150)
    GET_DICOM_INFO_LINE(DCM_XRayTubeCurrentInuA                  ) //DcmTagKey(0x0018, 0x8151)
    GET_DICOM_INFO_LINE(DCM_ContentQualification                 ) //DcmTagKey(0x0018, 0x9004)
    GET_DICOM_INFO_LINE(DCM_PulseSequenceName                    ) //DcmTagKey(0x0018, 0x9005)
    GET_DICOM_INFO_LINE(DCM_MRImagingModifierSequence            ) //DcmTagKey(0x0018, 0x9006)
    GET_DICOM_INFO_LINE(DCM_EchoPulseSequence                    ) //DcmTagKey(0x0018, 0x9008)
    GET_DICOM_INFO_LINE(DCM_InversionRecovery                    ) //DcmTagKey(0x0018, 0x9009)
    GET_DICOM_INFO_LINE(DCM_FlowCompensation                     ) //DcmTagKey(0x0018, 0x9010)
    GET_DICOM_INFO_LINE(DCM_MultipleSpinEcho                     ) //DcmTagKey(0x0018, 0x9011)
    GET_DICOM_INFO_LINE(DCM_MultiPlanarExcitation                ) //DcmTagKey(0x0018, 0x9012)
    GET_DICOM_INFO_LINE(DCM_PhaseContrast                        ) //DcmTagKey(0x0018, 0x9014)
    GET_DICOM_INFO_LINE(DCM_TimeOfFlightContrast                 ) //DcmTagKey(0x0018, 0x9015)
    GET_DICOM_INFO_LINE(DCM_Spoiling                             ) //DcmTagKey(0x0018, 0x9016)
    GET_DICOM_INFO_LINE(DCM_SteadyStatePulseSequence             ) //DcmTagKey(0x0018, 0x9017)
    GET_DICOM_INFO_LINE(DCM_EchoPlanarPulseSequence              ) //DcmTagKey(0x0018, 0x9018)
    GET_DICOM_INFO_LINE(DCM_TagAngleFirstAxis                    ) //DcmTagKey(0x0018, 0x9019)
    GET_DICOM_INFO_LINE(DCM_MagnetizationTransfer                ) //DcmTagKey(0x0018, 0x9020)
    GET_DICOM_INFO_LINE(DCM_T2Preparation                        ) //DcmTagKey(0x0018, 0x9021)
    GET_DICOM_INFO_LINE(DCM_BloodSignalNulling                   ) //DcmTagKey(0x0018, 0x9022)
    GET_DICOM_INFO_LINE(DCM_SaturationRecovery                   ) //DcmTagKey(0x0018, 0x9024)
    GET_DICOM_INFO_LINE(DCM_SpectrallySelectedSuppression        ) //DcmTagKey(0x0018, 0x9025)
    GET_DICOM_INFO_LINE(DCM_SpectrallySelectedExcitation         ) //DcmTagKey(0x0018, 0x9026)
    GET_DICOM_INFO_LINE(DCM_SpatialPresaturation                 ) //DcmTagKey(0x0018, 0x9027)
    GET_DICOM_INFO_LINE(DCM_Tagging                              ) //DcmTagKey(0x0018, 0x9028)
    GET_DICOM_INFO_LINE(DCM_OversamplingPhase                    ) //DcmTagKey(0x0018, 0x9029)
    GET_DICOM_INFO_LINE(DCM_TagSpacingFirstDimension             ) //DcmTagKey(0x0018, 0x9030)
    GET_DICOM_INFO_LINE(DCM_GeometryOfKSpaceTraversal            ) //DcmTagKey(0x0018, 0x9032)
    GET_DICOM_INFO_LINE(DCM_SegmentedKSpaceTraversal             ) //DcmTagKey(0x0018, 0x9033)
    GET_DICOM_INFO_LINE(DCM_RectilinearPhaseEncodeReordering     ) //DcmTagKey(0x0018, 0x9034)
    GET_DICOM_INFO_LINE(DCM_TagThickness                         ) //DcmTagKey(0x0018, 0x9035)
    GET_DICOM_INFO_LINE(DCM_PartialFourierDirection              ) //DcmTagKey(0x0018, 0x9036)
    GET_DICOM_INFO_LINE(DCM_CardiacSynchronizationTechnique      ) //DcmTagKey(0x0018, 0x9037)
    GET_DICOM_INFO_LINE(DCM_ReceiveCoilManufacturerName          ) //DcmTagKey(0x0018, 0x9041)
    GET_DICOM_INFO_LINE(DCM_MRReceiveCoilSequence                ) //DcmTagKey(0x0018, 0x9042)
    GET_DICOM_INFO_LINE(DCM_ReceiveCoilType                      ) //DcmTagKey(0x0018, 0x9043)
    GET_DICOM_INFO_LINE(DCM_QuadratureReceiveCoil                ) //DcmTagKey(0x0018, 0x9044)
    GET_DICOM_INFO_LINE(DCM_MultiCoilDefinitionSequence          ) //DcmTagKey(0x0018, 0x9045)
    GET_DICOM_INFO_LINE(DCM_MultiCoilConfiguration               ) //DcmTagKey(0x0018, 0x9046)
    GET_DICOM_INFO_LINE(DCM_MultiCoilElementName                 ) //DcmTagKey(0x0018, 0x9047)
    GET_DICOM_INFO_LINE(DCM_MultiCoilElementUsed                 ) //DcmTagKey(0x0018, 0x9048)
    GET_DICOM_INFO_LINE(DCM_MRTransmitCoilSequence               ) //DcmTagKey(0x0018, 0x9049)
    GET_DICOM_INFO_LINE(DCM_TransmitCoilManufacturerName         ) //DcmTagKey(0x0018, 0x9050)
    GET_DICOM_INFO_LINE(DCM_TransmitCoilType                     ) //DcmTagKey(0x0018, 0x9051)
    GET_DICOM_INFO_LINE(DCM_SpectralWidth                        ) //DcmTagKey(0x0018, 0x9052)
    GET_DICOM_INFO_LINE(DCM_ChemicalShiftReference               ) //DcmTagKey(0x0018, 0x9053)
    GET_DICOM_INFO_LINE(DCM_VolumeLocalizationTechnique          ) //DcmTagKey(0x0018, 0x9054)
    GET_DICOM_INFO_LINE(DCM_MRAcquisitionFrequencyEncodingSteps  ) //DcmTagKey(0x0018, 0x9058)
    GET_DICOM_INFO_LINE(DCM_Decoupling                           ) //DcmTagKey(0x0018, 0x9059)
    GET_DICOM_INFO_LINE(DCM_DecoupledNucleus                     ) //DcmTagKey(0x0018, 0x9060)
    GET_DICOM_INFO_LINE(DCM_DecouplingFrequency                  ) //DcmTagKey(0x0018, 0x9061)
    GET_DICOM_INFO_LINE(DCM_DecouplingMethod                     ) //DcmTagKey(0x0018, 0x9062)
    GET_DICOM_INFO_LINE(DCM_DecouplingChemicalShiftReference     ) //DcmTagKey(0x0018, 0x9063)
    GET_DICOM_INFO_LINE(DCM_KSpaceFiltering                      ) //DcmTagKey(0x0018, 0x9064)
    GET_DICOM_INFO_LINE(DCM_TimeDomainFiltering                  ) //DcmTagKey(0x0018, 0x9065)
    GET_DICOM_INFO_LINE(DCM_NumberOfZeroFills                    ) //DcmTagKey(0x0018, 0x9066)
    GET_DICOM_INFO_LINE(DCM_BaselineCorrection                   ) //DcmTagKey(0x0018, 0x9067)
    GET_DICOM_INFO_LINE(DCM_ParallelReductionFactorInPlane       ) //DcmTagKey(0x0018, 0x9069)
    GET_DICOM_INFO_LINE(DCM_CardiacRRIntervalSpecified           ) //DcmTagKey(0x0018, 0x9070)
    GET_DICOM_INFO_LINE(DCM_AcquisitionDuration                  ) //DcmTagKey(0x0018, 0x9073)
    GET_DICOM_INFO_LINE(DCM_FrameAcquisitionDateTime             ) //DcmTagKey(0x0018, 0x9074)
    GET_DICOM_INFO_LINE(DCM_DiffusionDirectionality              ) //DcmTagKey(0x0018, 0x9075)
    GET_DICOM_INFO_LINE(DCM_DiffusionGradientDirectionSequence   ) //DcmTagKey(0x0018, 0x9076)
    GET_DICOM_INFO_LINE(DCM_ParallelAcquisition                  ) //DcmTagKey(0x0018, 0x9077)
    GET_DICOM_INFO_LINE(DCM_ParallelAcquisitionTechnique         ) //DcmTagKey(0x0018, 0x9078)
    GET_DICOM_INFO_LINE(DCM_InversionTimes                       ) //DcmTagKey(0x0018, 0x9079)
    GET_DICOM_INFO_LINE(DCM_MetaboliteMapDescription             ) //DcmTagKey(0x0018, 0x9080)
    GET_DICOM_INFO_LINE(DCM_PartialFourier                       ) //DcmTagKey(0x0018, 0x9081)
    GET_DICOM_INFO_LINE(DCM_EffectiveEchoTime                    ) //DcmTagKey(0x0018, 0x9082)
    GET_DICOM_INFO_LINE(DCM_MetaboliteMapCodeSequence            ) //DcmTagKey(0x0018, 0x9083)
    GET_DICOM_INFO_LINE(DCM_ChemicalShiftSequence                ) //DcmTagKey(0x0018, 0x9084)
    GET_DICOM_INFO_LINE(DCM_CardiacSignalSource                  ) //DcmTagKey(0x0018, 0x9085)
    GET_DICOM_INFO_LINE(DCM_DiffusionBValue                      ) //DcmTagKey(0x0018, 0x9087)
    GET_DICOM_INFO_LINE(DCM_DiffusionGradientOrientation         ) //DcmTagKey(0x0018, 0x9089)
    GET_DICOM_INFO_LINE(DCM_VelocityEncodingDirection            ) //DcmTagKey(0x0018, 0x9090)
    GET_DICOM_INFO_LINE(DCM_VelocityEncodingMinimumValue         ) //DcmTagKey(0x0018, 0x9091)
    GET_DICOM_INFO_LINE(DCM_VelocityEncodingAcquisitionSequence  ) //DcmTagKey(0x0018, 0x9092)
    GET_DICOM_INFO_LINE(DCM_NumberOfKSpaceTrajectories           ) //DcmTagKey(0x0018, 0x9093)
    GET_DICOM_INFO_LINE(DCM_CoverageOfKSpace                     ) //DcmTagKey(0x0018, 0x9094)
    GET_DICOM_INFO_LINE(DCM_SpectroscopyAcquisitionPhaseRows     ) //DcmTagKey(0x0018, 0x9095)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ParallelReductionFactorInPlaneRetired ) //DcmTagKey(0x0018, 0x9096)
    GET_DICOM_INFO_LINE(DCM_TransmitterFrequency                 ) //DcmTagKey(0x0018, 0x9098)
    GET_DICOM_INFO_LINE(DCM_ResonantNucleus                      ) //DcmTagKey(0x0018, 0x9100)
    GET_DICOM_INFO_LINE(DCM_FrequencyCorrection                  ) //DcmTagKey(0x0018, 0x9101)
    GET_DICOM_INFO_LINE(DCM_MRSpectroscopyFOVGeometrySequence    ) //DcmTagKey(0x0018, 0x9103)
    GET_DICOM_INFO_LINE(DCM_SlabThickness                        ) //DcmTagKey(0x0018, 0x9104)
    GET_DICOM_INFO_LINE(DCM_SlabOrientation                      ) //DcmTagKey(0x0018, 0x9105)
    GET_DICOM_INFO_LINE(DCM_MidSlabPosition                      ) //DcmTagKey(0x0018, 0x9106)
    GET_DICOM_INFO_LINE(DCM_MRSpatialSaturationSequence          ) //DcmTagKey(0x0018, 0x9107)
    GET_DICOM_INFO_LINE(DCM_MRTimingAndRelatedParametersSequence ) //DcmTagKey(0x0018, 0x9112)
    GET_DICOM_INFO_LINE(DCM_MREchoSequence                       ) //DcmTagKey(0x0018, 0x9114)
    GET_DICOM_INFO_LINE(DCM_MRModifierSequence                   ) //DcmTagKey(0x0018, 0x9115)
    GET_DICOM_INFO_LINE(DCM_MRDiffusionSequence                  ) //DcmTagKey(0x0018, 0x9117)
    GET_DICOM_INFO_LINE(DCM_CardiacSynchronizationSequence       ) //DcmTagKey(0x0018, 0x9118)
    GET_DICOM_INFO_LINE(DCM_MRAveragesSequence                   ) //DcmTagKey(0x0018, 0x9119)
    GET_DICOM_INFO_LINE(DCM_MRFOVGeometrySequence                ) //DcmTagKey(0x0018, 0x9125)
    GET_DICOM_INFO_LINE(DCM_VolumeLocalizationSequence           ) //DcmTagKey(0x0018, 0x9126)
    GET_DICOM_INFO_LINE(DCM_SpectroscopyAcquisitionDataColumns   ) //DcmTagKey(0x0018, 0x9127)
    GET_DICOM_INFO_LINE(DCM_DiffusionAnisotropyType              ) //DcmTagKey(0x0018, 0x9147)
    GET_DICOM_INFO_LINE(DCM_FrameReferenceDateTime               ) //DcmTagKey(0x0018, 0x9151)
    GET_DICOM_INFO_LINE(DCM_MRMetaboliteMapSequence              ) //DcmTagKey(0x0018, 0x9152)
    GET_DICOM_INFO_LINE(DCM_ParallelReductionFactorOutOfPlane    ) //DcmTagKey(0x0018, 0x9155)
    GET_DICOM_INFO_LINE(DCM_SpectroscopyAcquisitionOutOfPlanePhaseSteps ) //DcmTagKey(0x0018, 0x9159)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_BulkMotionStatus             ) //DcmTagKey(0x0018, 0x9166)
    GET_DICOM_INFO_LINE(DCM_ParallelReductionFactorSecondInPlane ) //DcmTagKey(0x0018, 0x9168)
    GET_DICOM_INFO_LINE(DCM_CardiacBeatRejectionTechnique        ) //DcmTagKey(0x0018, 0x9169)
    GET_DICOM_INFO_LINE(DCM_RespiratoryMotionCompensationTechnique ) //DcmTagKey(0x0018, 0x9170)
    GET_DICOM_INFO_LINE(DCM_RespiratorySignalSource              ) //DcmTagKey(0x0018, 0x9171)
    GET_DICOM_INFO_LINE(DCM_BulkMotionCompensationTechnique      ) //DcmTagKey(0x0018, 0x9172)
    GET_DICOM_INFO_LINE(DCM_BulkMotionSignalSource               ) //DcmTagKey(0x0018, 0x9173)
    GET_DICOM_INFO_LINE(DCM_ApplicableSafetyStandardAgency       ) //DcmTagKey(0x0018, 0x9174)
    GET_DICOM_INFO_LINE(DCM_ApplicableSafetyStandardDescription  ) //DcmTagKey(0x0018, 0x9175)
    GET_DICOM_INFO_LINE(DCM_OperatingModeSequence                ) //DcmTagKey(0x0018, 0x9176)
    GET_DICOM_INFO_LINE(DCM_OperatingModeType                    ) //DcmTagKey(0x0018, 0x9177)
    GET_DICOM_INFO_LINE(DCM_OperatingMode                        ) //DcmTagKey(0x0018, 0x9178)
    GET_DICOM_INFO_LINE(DCM_SpecificAbsorptionRateDefinition     ) //DcmTagKey(0x0018, 0x9179)
    GET_DICOM_INFO_LINE(DCM_GradientOutputType                   ) //DcmTagKey(0x0018, 0x9180)
    GET_DICOM_INFO_LINE(DCM_SpecificAbsorptionRateValue          ) //DcmTagKey(0x0018, 0x9181)
    GET_DICOM_INFO_LINE(DCM_GradientOutput                       ) //DcmTagKey(0x0018, 0x9182)
    GET_DICOM_INFO_LINE(DCM_FlowCompensationDirection            ) //DcmTagKey(0x0018, 0x9183)
    GET_DICOM_INFO_LINE(DCM_TaggingDelay                         ) //DcmTagKey(0x0018, 0x9184)
    GET_DICOM_INFO_LINE(DCM_RespiratoryMotionCompensationTechniqueDescription ) //DcmTagKey(0x0018, 0x9185)
    GET_DICOM_INFO_LINE(DCM_RespiratorySignalSourceID            ) //DcmTagKey(0x0018, 0x9186)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ChemicalShiftMinimumIntegrationLimitInHz ) //DcmTagKey(0x0018, 0x9195)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ChemicalShiftMaximumIntegrationLimitInHz ) //DcmTagKey(0x0018, 0x9196)
    GET_DICOM_INFO_LINE(DCM_MRVelocityEncodingSequence           ) //DcmTagKey(0x0018, 0x9197)
    GET_DICOM_INFO_LINE(DCM_FirstOrderPhaseCorrection            ) //DcmTagKey(0x0018, 0x9198)
    GET_DICOM_INFO_LINE(DCM_WaterReferencedPhaseCorrection       ) //DcmTagKey(0x0018, 0x9199)
    GET_DICOM_INFO_LINE(DCM_MRSpectroscopyAcquisitionType        ) //DcmTagKey(0x0018, 0x9200)
    GET_DICOM_INFO_LINE(DCM_RespiratoryCyclePosition             ) //DcmTagKey(0x0018, 0x9214)
    GET_DICOM_INFO_LINE(DCM_VelocityEncodingMaximumValue         ) //DcmTagKey(0x0018, 0x9217)
    GET_DICOM_INFO_LINE(DCM_TagSpacingSecondDimension            ) //DcmTagKey(0x0018, 0x9218)
    GET_DICOM_INFO_LINE(DCM_TagAngleSecondAxis                   ) //DcmTagKey(0x0018, 0x9219)
    GET_DICOM_INFO_LINE(DCM_FrameAcquisitionDuration             ) //DcmTagKey(0x0018, 0x9220)
    GET_DICOM_INFO_LINE(DCM_MRImageFrameTypeSequence             ) //DcmTagKey(0x0018, 0x9226)
    GET_DICOM_INFO_LINE(DCM_MRSpectroscopyFrameTypeSequence      ) //DcmTagKey(0x0018, 0x9227)
    GET_DICOM_INFO_LINE(DCM_MRAcquisitionPhaseEncodingStepsInPlane ) //DcmTagKey(0x0018, 0x9231)
    GET_DICOM_INFO_LINE(DCM_MRAcquisitionPhaseEncodingStepsOutOfPlane ) //DcmTagKey(0x0018, 0x9232)
    GET_DICOM_INFO_LINE(DCM_SpectroscopyAcquisitionPhaseColumns  ) //DcmTagKey(0x0018, 0x9234)
    GET_DICOM_INFO_LINE(DCM_CardiacCyclePosition                 ) //DcmTagKey(0x0018, 0x9236)
    GET_DICOM_INFO_LINE(DCM_SpecificAbsorptionRateSequence       ) //DcmTagKey(0x0018, 0x9239)
    GET_DICOM_INFO_LINE(DCM_RFEchoTrainLength                    ) //DcmTagKey(0x0018, 0x9240)
    GET_DICOM_INFO_LINE(DCM_GradientEchoTrainLength              ) //DcmTagKey(0x0018, 0x9241)
    GET_DICOM_INFO_LINE(DCM_ChemicalShiftMinimumIntegrationLimitInppm ) //DcmTagKey(0x0018, 0x9295)
    GET_DICOM_INFO_LINE(DCM_ChemicalShiftMaximumIntegrationLimitInppm ) //DcmTagKey(0x0018, 0x9296)
    GET_DICOM_INFO_LINE(DCM_CTAcquisitionTypeSequence            ) //DcmTagKey(0x0018, 0x9301)
    GET_DICOM_INFO_LINE(DCM_AcquisitionType                      ) //DcmTagKey(0x0018, 0x9302)
    GET_DICOM_INFO_LINE(DCM_TubeAngle                            ) //DcmTagKey(0x0018, 0x9303)
    GET_DICOM_INFO_LINE(DCM_CTAcquisitionDetailsSequence         ) //DcmTagKey(0x0018, 0x9304)
    GET_DICOM_INFO_LINE(DCM_RevolutionTime                       ) //DcmTagKey(0x0018, 0x9305)
    GET_DICOM_INFO_LINE(DCM_SingleCollimationWidth               ) //DcmTagKey(0x0018, 0x9306)
    GET_DICOM_INFO_LINE(DCM_TotalCollimationWidth                ) //DcmTagKey(0x0018, 0x9307)
    GET_DICOM_INFO_LINE(DCM_CTTableDynamicsSequence              ) //DcmTagKey(0x0018, 0x9308)
    GET_DICOM_INFO_LINE(DCM_TableSpeed                           ) //DcmTagKey(0x0018, 0x9309)
    GET_DICOM_INFO_LINE(DCM_TableFeedPerRotation                 ) //DcmTagKey(0x0018, 0x9310)
    GET_DICOM_INFO_LINE(DCM_SpiralPitchFactor                    ) //DcmTagKey(0x0018, 0x9311)
    GET_DICOM_INFO_LINE(DCM_CTGeometrySequence                   ) //DcmTagKey(0x0018, 0x9312)
    GET_DICOM_INFO_LINE(DCM_DataCollectionCenterPatient          ) //DcmTagKey(0x0018, 0x9313)
    GET_DICOM_INFO_LINE(DCM_CTReconstructionSequence             ) //DcmTagKey(0x0018, 0x9314)
    GET_DICOM_INFO_LINE(DCM_ReconstructionAlgorithm              ) //DcmTagKey(0x0018, 0x9315)
    GET_DICOM_INFO_LINE(DCM_ConvolutionKernelGroup               ) //DcmTagKey(0x0018, 0x9316)
    GET_DICOM_INFO_LINE(DCM_ReconstructionFieldOfView            ) //DcmTagKey(0x0018, 0x9317)
    GET_DICOM_INFO_LINE(DCM_ReconstructionTargetCenterPatient    ) //DcmTagKey(0x0018, 0x9318)
    GET_DICOM_INFO_LINE(DCM_ReconstructionAngle                  ) //DcmTagKey(0x0018, 0x9319)
    GET_DICOM_INFO_LINE(DCM_ImageFilter                          ) //DcmTagKey(0x0018, 0x9320)
    GET_DICOM_INFO_LINE(DCM_CTExposureSequence                   ) //DcmTagKey(0x0018, 0x9321)
    GET_DICOM_INFO_LINE(DCM_ReconstructionPixelSpacing           ) //DcmTagKey(0x0018, 0x9322)
    GET_DICOM_INFO_LINE(DCM_ExposureModulationType               ) //DcmTagKey(0x0018, 0x9323)
    GET_DICOM_INFO_LINE(DCM_EstimatedDoseSaving                  ) //DcmTagKey(0x0018, 0x9324)
    GET_DICOM_INFO_LINE(DCM_CTXRayDetailsSequence                ) //DcmTagKey(0x0018, 0x9325)
    GET_DICOM_INFO_LINE(DCM_CTPositionSequence                   ) //DcmTagKey(0x0018, 0x9326)
    GET_DICOM_INFO_LINE(DCM_TablePosition                        ) //DcmTagKey(0x0018, 0x9327)
    GET_DICOM_INFO_LINE(DCM_ExposureTimeInms                     ) //DcmTagKey(0x0018, 0x9328)
    GET_DICOM_INFO_LINE(DCM_CTImageFrameTypeSequence             ) //DcmTagKey(0x0018, 0x9329)
    GET_DICOM_INFO_LINE(DCM_XRayTubeCurrentInmA                  ) //DcmTagKey(0x0018, 0x9330)
    GET_DICOM_INFO_LINE(DCM_ExposureInmAs                        ) //DcmTagKey(0x0018, 0x9332)
    GET_DICOM_INFO_LINE(DCM_ConstantVolumeFlag                   ) //DcmTagKey(0x0018, 0x9333)
    GET_DICOM_INFO_LINE(DCM_FluoroscopyFlag                      ) //DcmTagKey(0x0018, 0x9334)
    GET_DICOM_INFO_LINE(DCM_DistanceSourceToDataCollectionCenter ) //DcmTagKey(0x0018, 0x9335)
    GET_DICOM_INFO_LINE(DCM_ContrastBolusAgentNumber             ) //DcmTagKey(0x0018, 0x9337)
    GET_DICOM_INFO_LINE(DCM_ContrastBolusIngredientCodeSequence  ) //DcmTagKey(0x0018, 0x9338)
    GET_DICOM_INFO_LINE(DCM_ContrastAdministrationProfileSequence ) //DcmTagKey(0x0018, 0x9340)
    GET_DICOM_INFO_LINE(DCM_ContrastBolusUsageSequence           ) //DcmTagKey(0x0018, 0x9341)
    GET_DICOM_INFO_LINE(DCM_ContrastBolusAgentAdministered       ) //DcmTagKey(0x0018, 0x9342)
    GET_DICOM_INFO_LINE(DCM_ContrastBolusAgentDetected           ) //DcmTagKey(0x0018, 0x9343)
    GET_DICOM_INFO_LINE(DCM_ContrastBolusAgentPhase              ) //DcmTagKey(0x0018, 0x9344)
    GET_DICOM_INFO_LINE(DCM_CTDIvol                              ) //DcmTagKey(0x0018, 0x9345)
    GET_DICOM_INFO_LINE(DCM_CTDIPhantomTypeCodeSequence          ) //DcmTagKey(0x0018, 0x9346)
    GET_DICOM_INFO_LINE(DCM_CalciumScoringMassFactorPatient      ) //DcmTagKey(0x0018, 0x9351)
    GET_DICOM_INFO_LINE(DCM_CalciumScoringMassFactorDevice       ) //DcmTagKey(0x0018, 0x9352)
    GET_DICOM_INFO_LINE(DCM_EnergyWeightingFactor                ) //DcmTagKey(0x0018, 0x9353)
    GET_DICOM_INFO_LINE(DCM_CTAdditionalXRaySourceSequence       ) //DcmTagKey(0x0018, 0x9360)
    GET_DICOM_INFO_LINE(DCM_ProjectionPixelCalibrationSequence   ) //DcmTagKey(0x0018, 0x9401)
    GET_DICOM_INFO_LINE(DCM_DistanceSourceToIsocenter            ) //DcmTagKey(0x0018, 0x9402)
    GET_DICOM_INFO_LINE(DCM_DistanceObjectToTableTop             ) //DcmTagKey(0x0018, 0x9403)
    GET_DICOM_INFO_LINE(DCM_ObjectPixelSpacingInCenterOfBeam     ) //DcmTagKey(0x0018, 0x9404)
    GET_DICOM_INFO_LINE(DCM_PositionerPositionSequence           ) //DcmTagKey(0x0018, 0x9405)
    GET_DICOM_INFO_LINE(DCM_TablePositionSequence                ) //DcmTagKey(0x0018, 0x9406)
    GET_DICOM_INFO_LINE(DCM_CollimatorShapeSequence              ) //DcmTagKey(0x0018, 0x9407)
    GET_DICOM_INFO_LINE(DCM_XAXRFFrameCharacteristicsSequence    ) //DcmTagKey(0x0018, 0x9412)
    GET_DICOM_INFO_LINE(DCM_FrameAcquisitionSequence             ) //DcmTagKey(0x0018, 0x9417)
    GET_DICOM_INFO_LINE(DCM_XRayReceptorType                     ) //DcmTagKey(0x0018, 0x9420)
    GET_DICOM_INFO_LINE(DCM_AcquisitionProtocolName              ) //DcmTagKey(0x0018, 0x9423)
    GET_DICOM_INFO_LINE(DCM_AcquisitionProtocolDescription       ) //DcmTagKey(0x0018, 0x9424)
    GET_DICOM_INFO_LINE(DCM_ContrastBolusIngredientOpaque        ) //DcmTagKey(0x0018, 0x9425)
    GET_DICOM_INFO_LINE(DCM_DistanceReceptorPlaneToDetectorHousing ) //DcmTagKey(0x0018, 0x9426)
    GET_DICOM_INFO_LINE(DCM_IntensifierActiveShape               ) //DcmTagKey(0x0018, 0x9427)
    GET_DICOM_INFO_LINE(DCM_IntensifierActiveDimensions          ) //DcmTagKey(0x0018, 0x9428)
    GET_DICOM_INFO_LINE(DCM_PhysicalDetectorSize                 ) //DcmTagKey(0x0018, 0x9429)
    GET_DICOM_INFO_LINE(DCM_PositionOfIsocenterProjection        ) //DcmTagKey(0x0018, 0x9430)
    GET_DICOM_INFO_LINE(DCM_FieldOfViewSequence                  ) //DcmTagKey(0x0018, 0x9432)
    GET_DICOM_INFO_LINE(DCM_FieldOfViewDescription               ) //DcmTagKey(0x0018, 0x9433)
    GET_DICOM_INFO_LINE(DCM_ExposureControlSensingRegionsSequence ) //DcmTagKey(0x0018, 0x9434)
    GET_DICOM_INFO_LINE(DCM_ExposureControlSensingRegionShape    ) //DcmTagKey(0x0018, 0x9435)
    GET_DICOM_INFO_LINE(DCM_ExposureControlSensingRegionLeftVerticalEdge ) //DcmTagKey(0x0018, 0x9436)
    GET_DICOM_INFO_LINE(DCM_ExposureControlSensingRegionRightVerticalEdge ) //DcmTagKey(0x0018, 0x9437)
    GET_DICOM_INFO_LINE(DCM_ExposureControlSensingRegionUpperHorizontalEdge ) //DcmTagKey(0x0018, 0x9438)
    GET_DICOM_INFO_LINE(DCM_ExposureControlSensingRegionLowerHorizontalEdge ) //DcmTagKey(0x0018, 0x9439)
    GET_DICOM_INFO_LINE(DCM_CenterOfCircularExposureControlSensingRegion ) //DcmTagKey(0x0018, 0x9440)
    GET_DICOM_INFO_LINE(DCM_RadiusOfCircularExposureControlSensingRegion ) //DcmTagKey(0x0018, 0x9441)
    GET_DICOM_INFO_LINE(DCM_VerticesOfThePolygonalExposureControlSensingRegion ) //DcmTagKey(0x0018, 0x9442)
    GET_DICOM_INFO_LINE(DCM_ColumnAngulationPatient              ) //DcmTagKey(0x0018, 0x9447)
    GET_DICOM_INFO_LINE(DCM_BeamAngle                            ) //DcmTagKey(0x0018, 0x9449)
    GET_DICOM_INFO_LINE(DCM_FrameDetectorParametersSequence      ) //DcmTagKey(0x0018, 0x9451)
    GET_DICOM_INFO_LINE(DCM_CalculatedAnatomyThickness           ) //DcmTagKey(0x0018, 0x9452)
    GET_DICOM_INFO_LINE(DCM_CalibrationSequence                  ) //DcmTagKey(0x0018, 0x9455)
    GET_DICOM_INFO_LINE(DCM_ObjectThicknessSequence              ) //DcmTagKey(0x0018, 0x9456)
    GET_DICOM_INFO_LINE(DCM_PlaneIdentification                  ) //DcmTagKey(0x0018, 0x9457)
    GET_DICOM_INFO_LINE(DCM_FieldOfViewDimensionsInFloat         ) //DcmTagKey(0x0018, 0x9461)
    GET_DICOM_INFO_LINE(DCM_IsocenterReferenceSystemSequence     ) //DcmTagKey(0x0018, 0x9462)
    GET_DICOM_INFO_LINE(DCM_PositionerIsocenterPrimaryAngle      ) //DcmTagKey(0x0018, 0x9463)
    GET_DICOM_INFO_LINE(DCM_PositionerIsocenterSecondaryAngle    ) //DcmTagKey(0x0018, 0x9464)
    GET_DICOM_INFO_LINE(DCM_PositionerIsocenterDetectorRotationAngle ) //DcmTagKey(0x0018, 0x9465)
    GET_DICOM_INFO_LINE(DCM_TableXPositionToIsocenter            ) //DcmTagKey(0x0018, 0x9466)
    GET_DICOM_INFO_LINE(DCM_TableYPositionToIsocenter            ) //DcmTagKey(0x0018, 0x9467)
    GET_DICOM_INFO_LINE(DCM_TableZPositionToIsocenter            ) //DcmTagKey(0x0018, 0x9468)
    GET_DICOM_INFO_LINE(DCM_TableHorizontalRotationAngle         ) //DcmTagKey(0x0018, 0x9469)
    GET_DICOM_INFO_LINE(DCM_TableHeadTiltAngle                   ) //DcmTagKey(0x0018, 0x9470)
    GET_DICOM_INFO_LINE(DCM_TableCradleTiltAngle                 ) //DcmTagKey(0x0018, 0x9471)
    GET_DICOM_INFO_LINE(DCM_FrameDisplayShutterSequence          ) //DcmTagKey(0x0018, 0x9472)
    GET_DICOM_INFO_LINE(DCM_AcquiredImageAreaDoseProduct         ) //DcmTagKey(0x0018, 0x9473)
    GET_DICOM_INFO_LINE(DCM_CArmPositionerTabletopRelationship   ) //DcmTagKey(0x0018, 0x9474)
    GET_DICOM_INFO_LINE(DCM_XRayGeometrySequence                 ) //DcmTagKey(0x0018, 0x9476)
    GET_DICOM_INFO_LINE(DCM_IrradiationEventIdentificationSequence ) //DcmTagKey(0x0018, 0x9477)
    GET_DICOM_INFO_LINE(DCM_XRay3DFrameTypeSequence              ) //DcmTagKey(0x0018, 0x9504)
    GET_DICOM_INFO_LINE(DCM_ContributingSourcesSequence          ) //DcmTagKey(0x0018, 0x9506)
    GET_DICOM_INFO_LINE(DCM_XRay3DAcquisitionSequence            ) //DcmTagKey(0x0018, 0x9507)
    GET_DICOM_INFO_LINE(DCM_PrimaryPositionerScanArc             ) //DcmTagKey(0x0018, 0x9508)
    GET_DICOM_INFO_LINE(DCM_SecondaryPositionerScanArc           ) //DcmTagKey(0x0018, 0x9509)
    GET_DICOM_INFO_LINE(DCM_PrimaryPositionerScanStartAngle      ) //DcmTagKey(0x0018, 0x9510)
    GET_DICOM_INFO_LINE(DCM_SecondaryPositionerScanStartAngle    ) //DcmTagKey(0x0018, 0x9511)
    GET_DICOM_INFO_LINE(DCM_PrimaryPositionerIncrement           ) //DcmTagKey(0x0018, 0x9514)
    GET_DICOM_INFO_LINE(DCM_SecondaryPositionerIncrement         ) //DcmTagKey(0x0018, 0x9515)
    GET_DICOM_INFO_LINE(DCM_StartAcquisitionDateTime             ) //DcmTagKey(0x0018, 0x9516)
    GET_DICOM_INFO_LINE(DCM_EndAcquisitionDateTime               ) //DcmTagKey(0x0018, 0x9517)
    GET_DICOM_INFO_LINE(DCM_ApplicationName                      ) //DcmTagKey(0x0018, 0x9524)
    GET_DICOM_INFO_LINE(DCM_ApplicationVersion                   ) //DcmTagKey(0x0018, 0x9525)
    GET_DICOM_INFO_LINE(DCM_ApplicationManufacturer              ) //DcmTagKey(0x0018, 0x9526)
    GET_DICOM_INFO_LINE(DCM_AlgorithmType                        ) //DcmTagKey(0x0018, 0x9527)
    GET_DICOM_INFO_LINE(DCM_AlgorithmDescription                 ) //DcmTagKey(0x0018, 0x9528)
    GET_DICOM_INFO_LINE(DCM_XRay3DReconstructionSequence         ) //DcmTagKey(0x0018, 0x9530)
    GET_DICOM_INFO_LINE(DCM_ReconstructionDescription            ) //DcmTagKey(0x0018, 0x9531)
    GET_DICOM_INFO_LINE(DCM_PerProjectionAcquisitionSequence     ) //DcmTagKey(0x0018, 0x9538)
    GET_DICOM_INFO_LINE(DCM_DiffusionBMatrixSequence             ) //DcmTagKey(0x0018, 0x9601)
    GET_DICOM_INFO_LINE(DCM_DiffusionBValueXX                    ) //DcmTagKey(0x0018, 0x9602)
    GET_DICOM_INFO_LINE(DCM_DiffusionBValueXY                    ) //DcmTagKey(0x0018, 0x9603)
    GET_DICOM_INFO_LINE(DCM_DiffusionBValueXZ                    ) //DcmTagKey(0x0018, 0x9604)
    GET_DICOM_INFO_LINE(DCM_DiffusionBValueYY                    ) //DcmTagKey(0x0018, 0x9605)
    GET_DICOM_INFO_LINE(DCM_DiffusionBValueYZ                    ) //DcmTagKey(0x0018, 0x9606)
    GET_DICOM_INFO_LINE(DCM_DiffusionBValueZZ                    ) //DcmTagKey(0x0018, 0x9607)
    GET_DICOM_INFO_LINE(DCM_DecayCorrectionDateTime              ) //DcmTagKey(0x0018, 0x9701)
    GET_DICOM_INFO_LINE(DCM_StartDensityThreshold                ) //DcmTagKey(0x0018, 0x9715)
    GET_DICOM_INFO_LINE(DCM_StartRelativeDensityDifferenceThreshold ) //DcmTagKey(0x0018, 0x9716)
    GET_DICOM_INFO_LINE(DCM_StartCardiacTriggerCountThreshold    ) //DcmTagKey(0x0018, 0x9717)
    GET_DICOM_INFO_LINE(DCM_StartRespiratoryTriggerCountThreshold ) //DcmTagKey(0x0018, 0x9718)
    GET_DICOM_INFO_LINE(DCM_TerminationCountsThreshold           ) //DcmTagKey(0x0018, 0x9719)
    GET_DICOM_INFO_LINE(DCM_TerminationDensityThreshold          ) //DcmTagKey(0x0018, 0x9720)
    GET_DICOM_INFO_LINE(DCM_TerminationRelativeDensityThreshold  ) //DcmTagKey(0x0018, 0x9721)
    GET_DICOM_INFO_LINE(DCM_TerminationTimeThreshold             ) //DcmTagKey(0x0018, 0x9722)
    GET_DICOM_INFO_LINE(DCM_TerminationCardiacTriggerCountThreshold ) //DcmTagKey(0x0018, 0x9723)
    GET_DICOM_INFO_LINE(DCM_TerminationRespiratoryTriggerCountThreshold ) //DcmTagKey(0x0018, 0x9724)
    GET_DICOM_INFO_LINE(DCM_DetectorGeometry                     ) //DcmTagKey(0x0018, 0x9725)
    GET_DICOM_INFO_LINE(DCM_TransverseDetectorSeparation         ) //DcmTagKey(0x0018, 0x9726)
    GET_DICOM_INFO_LINE(DCM_AxialDetectorDimension               ) //DcmTagKey(0x0018, 0x9727)
    GET_DICOM_INFO_LINE(DCM_RadiopharmaceuticalAgentNumber       ) //DcmTagKey(0x0018, 0x9729)
    GET_DICOM_INFO_LINE(DCM_PETFrameAcquisitionSequence          ) //DcmTagKey(0x0018, 0x9732)
    GET_DICOM_INFO_LINE(DCM_PETDetectorMotionDetailsSequence     ) //DcmTagKey(0x0018, 0x9733)
    GET_DICOM_INFO_LINE(DCM_PETTableDynamicsSequence             ) //DcmTagKey(0x0018, 0x9734)
    GET_DICOM_INFO_LINE(DCM_PETPositionSequence                  ) //DcmTagKey(0x0018, 0x9735)
    GET_DICOM_INFO_LINE(DCM_PETFrameCorrectionFactorsSequence    ) //DcmTagKey(0x0018, 0x9736)
    GET_DICOM_INFO_LINE(DCM_RadiopharmaceuticalUsageSequence     ) //DcmTagKey(0x0018, 0x9737)
    GET_DICOM_INFO_LINE(DCM_AttenuationCorrectionSource          ) //DcmTagKey(0x0018, 0x9738)
    GET_DICOM_INFO_LINE(DCM_NumberOfIterations                   ) //DcmTagKey(0x0018, 0x9739)
    GET_DICOM_INFO_LINE(DCM_NumberOfSubsets                      ) //DcmTagKey(0x0018, 0x9740)
    GET_DICOM_INFO_LINE(DCM_PETReconstructionSequence            ) //DcmTagKey(0x0018, 0x9749)
    GET_DICOM_INFO_LINE(DCM_PETFrameTypeSequence                 ) //DcmTagKey(0x0018, 0x9751)
    GET_DICOM_INFO_LINE(DCM_TimeOfFlightInformationUsed          ) //DcmTagKey(0x0018, 0x9755)
    GET_DICOM_INFO_LINE(DCM_ReconstructionType                   ) //DcmTagKey(0x0018, 0x9756)
    GET_DICOM_INFO_LINE(DCM_DecayCorrected                       ) //DcmTagKey(0x0018, 0x9758)
    GET_DICOM_INFO_LINE(DCM_AttenuationCorrected                 ) //DcmTagKey(0x0018, 0x9759)
    GET_DICOM_INFO_LINE(DCM_ScatterCorrected                     ) //DcmTagKey(0x0018, 0x9760)
    GET_DICOM_INFO_LINE(DCM_DeadTimeCorrected                    ) //DcmTagKey(0x0018, 0x9761)
    GET_DICOM_INFO_LINE(DCM_GantryMotionCorrected                ) //DcmTagKey(0x0018, 0x9762)
    GET_DICOM_INFO_LINE(DCM_PatientMotionCorrected               ) //DcmTagKey(0x0018, 0x9763)
    GET_DICOM_INFO_LINE(DCM_CountLossNormalizationCorrected      ) //DcmTagKey(0x0018, 0x9764)
    GET_DICOM_INFO_LINE(DCM_RandomsCorrected                     ) //DcmTagKey(0x0018, 0x9765)
    GET_DICOM_INFO_LINE(DCM_NonUniformRadialSamplingCorrected    ) //DcmTagKey(0x0018, 0x9766)
    GET_DICOM_INFO_LINE(DCM_SensitivityCalibrated                ) //DcmTagKey(0x0018, 0x9767)
    GET_DICOM_INFO_LINE(DCM_DetectorNormalizationCorrection      ) //DcmTagKey(0x0018, 0x9768)
    GET_DICOM_INFO_LINE(DCM_IterativeReconstructionMethod        ) //DcmTagKey(0x0018, 0x9769)
    GET_DICOM_INFO_LINE(DCM_AttenuationCorrectionTemporalRelationship ) //DcmTagKey(0x0018, 0x9770)
    GET_DICOM_INFO_LINE(DCM_PatientPhysiologicalStateSequence    ) //DcmTagKey(0x0018, 0x9771)
    GET_DICOM_INFO_LINE(DCM_PatientPhysiologicalStateCodeSequence ) //DcmTagKey(0x0018, 0x9772)
    GET_DICOM_INFO_LINE(DCM_DepthsOfFocus                        ) //DcmTagKey(0x0018, 0x9801)
    GET_DICOM_INFO_LINE(DCM_ExcludedIntervalsSequence            ) //DcmTagKey(0x0018, 0x9803)
//    GET_DICOM_INFO_LINE(DCM_ExclusionStartDateTime               ) //DcmTagKey(0x0018, 0x9804)
    GET_DICOM_INFO_LINE(DCM_ExclusionDuration                    ) //DcmTagKey(0x0018, 0x9805)
    GET_DICOM_INFO_LINE(DCM_USImageDescriptionSequence           ) //DcmTagKey(0x0018, 0x9806)
    GET_DICOM_INFO_LINE(DCM_ImageDataTypeSequence                ) //DcmTagKey(0x0018, 0x9807)
    GET_DICOM_INFO_LINE(DCM_DataType                             ) //DcmTagKey(0x0018, 0x9808)
    GET_DICOM_INFO_LINE(DCM_TransducerScanPatternCodeSequence    ) //DcmTagKey(0x0018, 0x9809)
    GET_DICOM_INFO_LINE(DCM_AliasedDataType                      ) //DcmTagKey(0x0018, 0x980b)
    GET_DICOM_INFO_LINE(DCM_PositionMeasuringDeviceUsed          ) //DcmTagKey(0x0018, 0x980c)
    GET_DICOM_INFO_LINE(DCM_TransducerGeometryCodeSequence       ) //DcmTagKey(0x0018, 0x980d)
    GET_DICOM_INFO_LINE(DCM_TransducerBeamSteeringCodeSequence   ) //DcmTagKey(0x0018, 0x980e)
    GET_DICOM_INFO_LINE(DCM_TransducerApplicationCodeSequence    ) //DcmTagKey(0x0018, 0x980f)
    GET_DICOM_INFO_LINE(DCM_ContributingEquipmentSequence        ) //DcmTagKey(0x0018, 0xa001)
    GET_DICOM_INFO_LINE(DCM_ContributionDateTime                 ) //DcmTagKey(0x0018, 0xa002)
    GET_DICOM_INFO_LINE(DCM_ContributionDescription              ) //DcmTagKey(0x0018, 0xa003)
            info->append("</div>");
            info->append("<div style=\"color:#00007F\">");
    GET_DICOM_INFO_LINE(DCM_StudyInstanceUID                     ) //DcmTagKey(0x0020, 0x000d)
    GET_DICOM_INFO_LINE(DCM_SeriesInstanceUID                    ) //DcmTagKey(0x0020, 0x000e)
    GET_DICOM_INFO_LINE(DCM_StudyID                              ) //DcmTagKey(0x0020, 0x0010)
    GET_DICOM_INFO_LINE(DCM_SeriesNumber                         ) //DcmTagKey(0x0020, 0x0011)
    GET_DICOM_INFO_LINE(DCM_AcquisitionNumber                    ) //DcmTagKey(0x0020, 0x0012)
    GET_DICOM_INFO_LINE(DCM_InstanceNumber                       ) //DcmTagKey(0x0020, 0x0013)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_IsotopeNumber                ) //DcmTagKey(0x0020, 0x0014)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_PhaseNumber                  ) //DcmTagKey(0x0020, 0x0015)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_IntervalNumber               ) //DcmTagKey(0x0020, 0x0016)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_TimeSlotNumber               ) //DcmTagKey(0x0020, 0x0017)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_AngleNumber                  ) //DcmTagKey(0x0020, 0x0018)
    GET_DICOM_INFO_LINE(DCM_ItemNumber                           ) //DcmTagKey(0x0020, 0x0019)
    GET_DICOM_INFO_LINE(DCM_PatientOrientation                   ) //DcmTagKey(0x0020, 0x0020)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_OverlayNumber                ) //DcmTagKey(0x0020, 0x0022)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_CurveNumber                  ) //DcmTagKey(0x0020, 0x0024)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_LUTNumber                    ) //DcmTagKey(0x0020, 0x0026)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_ImagePosition               ) //DcmTagKey(0x0020, 0x0030)
    GET_DICOM_INFO_LINE(DCM_ImagePositionPatient                 ) //DcmTagKey(0x0020, 0x0032)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_ImageOrientation            ) //DcmTagKey(0x0020, 0x0035)
    GET_DICOM_INFO_LINE(DCM_ImageOrientationPatient              ) //DcmTagKey(0x0020, 0x0037)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_Location                    ) //DcmTagKey(0x0020, 0x0050)
    GET_DICOM_INFO_LINE(DCM_FrameOfReferenceUID                  ) //DcmTagKey(0x0020, 0x0052)
    GET_DICOM_INFO_LINE(DCM_Laterality                           ) //DcmTagKey(0x0020, 0x0060)
    GET_DICOM_INFO_LINE(DCM_ImageLaterality                      ) //DcmTagKey(0x0020, 0x0062)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_ImageGeometryType           ) //DcmTagKey(0x0020, 0x0070)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_MaskingImage                ) //DcmTagKey(0x0020, 0x0080)
    GET_DICOM_INFO_LINE(DCM_TemporalPositionIdentifier           ) //DcmTagKey(0x0020, 0x0100)
    GET_DICOM_INFO_LINE(DCM_NumberOfTemporalPositions            ) //DcmTagKey(0x0020, 0x0105)
    GET_DICOM_INFO_LINE(DCM_TemporalResolution                   ) //DcmTagKey(0x0020, 0x0110)
    GET_DICOM_INFO_LINE(DCM_SynchronizationFrameOfReferenceUID   ) //DcmTagKey(0x0020, 0x0200)
    GET_DICOM_INFO_LINE(DCM_SOPInstanceUIDOfConcatenationSource  ) //DcmTagKey(0x0020, 0x0242)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_SeriesInStudy                ) //DcmTagKey(0x0020, 0x1000)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_AcquisitionsInSeries        ) //DcmTagKey(0x0020, 0x1001)
    GET_DICOM_INFO_LINE(DCM_ImagesInAcquisition                  ) //DcmTagKey(0x0020, 0x1002)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_ImagesInSeries              ) //DcmTagKey(0x0020, 0x1003)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_AcquisitionsInStudy          ) //DcmTagKey(0x0020, 0x1004)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_ImagesInStudy               ) //DcmTagKey(0x0020, 0x1005)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_Reference                   ) //DcmTagKey(0x0020, 0x1020)
    GET_DICOM_INFO_LINE(DCM_PositionReferenceIndicator           ) //DcmTagKey(0x0020, 0x1040)
    GET_DICOM_INFO_LINE(DCM_SliceLocation                        ) //DcmTagKey(0x0020, 0x1041)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_OtherStudyNumbers            ) //DcmTagKey(0x0020, 0x1070)
    GET_DICOM_INFO_LINE(DCM_NumberOfPatientRelatedStudies        ) //DcmTagKey(0x0020, 0x1200)
    GET_DICOM_INFO_LINE(DCM_NumberOfPatientRelatedSeries         ) //DcmTagKey(0x0020, 0x1202)
    GET_DICOM_INFO_LINE(DCM_NumberOfPatientRelatedInstances      ) //DcmTagKey(0x0020, 0x1204)
    GET_DICOM_INFO_LINE(DCM_NumberOfStudyRelatedSeries           ) //DcmTagKey(0x0020, 0x1206)
    GET_DICOM_INFO_LINE(DCM_NumberOfStudyRelatedInstances        ) //DcmTagKey(0x0020, 0x1208)
    GET_DICOM_INFO_LINE(DCM_NumberOfSeriesRelatedInstances       ) //DcmTagKey(0x0020, 0x1209)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_ModifyingDeviceID           ) //DcmTagKey(0x0020, 0x3401)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_ModifiedImageID             ) //DcmTagKey(0x0020, 0x3402)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_ModifiedImageDate           ) //DcmTagKey(0x0020, 0x3403)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_ModifyingDeviceManufacturer ) //DcmTagKey(0x0020, 0x3404)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_ModifiedImageTime           ) //DcmTagKey(0x0020, 0x3405)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_ModifiedImageDescription    ) //DcmTagKey(0x0020, 0x3406)
    GET_DICOM_INFO_LINE(DCM_ImageComments                        ) //DcmTagKey(0x0020, 0x4000)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_OriginalImageIdentification ) //DcmTagKey(0x0020, 0x5000)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_OriginalImageIdentificationNomenclature ) //DcmTagKey(0x0020, 0x5002)
    GET_DICOM_INFO_LINE(DCM_StackID                              ) //DcmTagKey(0x0020, 0x9056)
    GET_DICOM_INFO_LINE(DCM_InStackPositionNumber                ) //DcmTagKey(0x0020, 0x9057)
    GET_DICOM_INFO_LINE(DCM_FrameAnatomySequence                 ) //DcmTagKey(0x0020, 0x9071)
    GET_DICOM_INFO_LINE(DCM_FrameLaterality                      ) //DcmTagKey(0x0020, 0x9072)
    GET_DICOM_INFO_LINE(DCM_FrameContentSequence                 ) //DcmTagKey(0x0020, 0x9111)
    GET_DICOM_INFO_LINE(DCM_PlanePositionSequence                ) //DcmTagKey(0x0020, 0x9113)
    GET_DICOM_INFO_LINE(DCM_PlaneOrientationSequence             ) //DcmTagKey(0x0020, 0x9116)
    GET_DICOM_INFO_LINE(DCM_TemporalPositionIndex                ) //DcmTagKey(0x0020, 0x9128)
    GET_DICOM_INFO_LINE(DCM_NominalCardiacTriggerDelayTime       ) //DcmTagKey(0x0020, 0x9153)
    GET_DICOM_INFO_LINE(DCM_FrameAcquisitionNumber               ) //DcmTagKey(0x0020, 0x9156)
    GET_DICOM_INFO_LINE(DCM_DimensionIndexValues                 ) //DcmTagKey(0x0020, 0x9157)
    GET_DICOM_INFO_LINE(DCM_FrameComments                        ) //DcmTagKey(0x0020, 0x9158)
    GET_DICOM_INFO_LINE(DCM_ConcatenationUID                     ) //DcmTagKey(0x0020, 0x9161)
    GET_DICOM_INFO_LINE(DCM_InConcatenationNumber                ) //DcmTagKey(0x0020, 0x9162)
    GET_DICOM_INFO_LINE(DCM_InConcatenationTotalNumber           ) //DcmTagKey(0x0020, 0x9163)
    GET_DICOM_INFO_LINE(DCM_DimensionOrganizationUID             ) //DcmTagKey(0x0020, 0x9164)
    GET_DICOM_INFO_LINE(DCM_DimensionIndexPointer                ) //DcmTagKey(0x0020, 0x9165)
    GET_DICOM_INFO_LINE(DCM_FunctionalGroupPointer               ) //DcmTagKey(0x0020, 0x9167)
    GET_DICOM_INFO_LINE(DCM_DimensionIndexPrivateCreator         ) //DcmTagKey(0x0020, 0x9213)
    GET_DICOM_INFO_LINE(DCM_DimensionOrganizationSequence        ) //DcmTagKey(0x0020, 0x9221)
    GET_DICOM_INFO_LINE(DCM_DimensionIndexSequence               ) //DcmTagKey(0x0020, 0x9222)
    GET_DICOM_INFO_LINE(DCM_ConcatenationFrameOffsetNumber       ) //DcmTagKey(0x0020, 0x9228)
    GET_DICOM_INFO_LINE(DCM_FunctionalGroupPrivateCreator        ) //DcmTagKey(0x0020, 0x9238)
    GET_DICOM_INFO_LINE(DCM_NominalPercentageOfCardiacPhase      ) //DcmTagKey(0x0020, 0x9241)
    GET_DICOM_INFO_LINE(DCM_NominalPercentageOfRespiratoryPhase  ) //DcmTagKey(0x0020, 0x9245)
    GET_DICOM_INFO_LINE(DCM_StartingRespiratoryAmplitude         ) //DcmTagKey(0x0020, 0x9246)
    GET_DICOM_INFO_LINE(DCM_StartingRespiratoryPhase             ) //DcmTagKey(0x0020, 0x9247)
    GET_DICOM_INFO_LINE(DCM_EndingRespiratoryAmplitude           ) //DcmTagKey(0x0020, 0x9248)
    GET_DICOM_INFO_LINE(DCM_EndingRespiratoryPhase               ) //DcmTagKey(0x0020, 0x9249)
    GET_DICOM_INFO_LINE(DCM_RespiratoryTriggerType               ) //DcmTagKey(0x0020, 0x9250)
    GET_DICOM_INFO_LINE(DCM_RRIntervalTimeNominal                ) //DcmTagKey(0x0020, 0x9251)
    GET_DICOM_INFO_LINE(DCM_ActualCardiacTriggerDelayTime        ) //DcmTagKey(0x0020, 0x9252)
    GET_DICOM_INFO_LINE(DCM_RespiratorySynchronizationSequence   ) //DcmTagKey(0x0020, 0x9253)
    GET_DICOM_INFO_LINE(DCM_RespiratoryIntervalTime              ) //DcmTagKey(0x0020, 0x9254)
    GET_DICOM_INFO_LINE(DCM_NominalRespiratoryTriggerDelayTime   ) //DcmTagKey(0x0020, 0x9255)
    GET_DICOM_INFO_LINE(DCM_RespiratoryTriggerDelayThreshold     ) //DcmTagKey(0x0020, 0x9256)
    GET_DICOM_INFO_LINE(DCM_ActualRespiratoryTriggerDelayTime    ) //DcmTagKey(0x0020, 0x9257)
    GET_DICOM_INFO_LINE(DCM_ImagePositionVolume                  ) //DcmTagKey(0x0020, 0x9301)
    GET_DICOM_INFO_LINE(DCM_ImageOrientationVolume               ) //DcmTagKey(0x0020, 0x9302)
    GET_DICOM_INFO_LINE(DCM_UltrasoundAcquisitionGeometry        ) //DcmTagKey(0x0020, 0x9307)
    GET_DICOM_INFO_LINE(DCM_ApexPosition                         ) //DcmTagKey(0x0020, 0x9308)
    GET_DICOM_INFO_LINE(DCM_VolumeToTransducerMappingMatrix      ) //DcmTagKey(0x0020, 0x9309)
    GET_DICOM_INFO_LINE(DCM_VolumeToTableMappingMatrix           ) //DcmTagKey(0x0020, 0x930a)
    GET_DICOM_INFO_LINE(DCM_PatientFrameOfReferenceSource        ) //DcmTagKey(0x0020, 0x930c)
    GET_DICOM_INFO_LINE(DCM_TemporalPositionTimeOffset           ) //DcmTagKey(0x0020, 0x930d)
    GET_DICOM_INFO_LINE(DCM_PlanePositionVolumeSequence          ) //DcmTagKey(0x0020, 0x930e)
    GET_DICOM_INFO_LINE(DCM_PlaneOrientationVolumeSequence       ) //DcmTagKey(0x0020, 0x930f)
    GET_DICOM_INFO_LINE(DCM_TemporalPositionSequence             ) //DcmTagKey(0x0020, 0x9310)
    GET_DICOM_INFO_LINE(DCM_DimensionOrganizationType            ) //DcmTagKey(0x0020, 0x9311)
    GET_DICOM_INFO_LINE(DCM_VolumeFrameOfReferenceUID            ) //DcmTagKey(0x0020, 0x9312)
    GET_DICOM_INFO_LINE(DCM_TableFrameOfReferenceUID             ) //DcmTagKey(0x0020, 0x9313)
    GET_DICOM_INFO_LINE(DCM_DimensionDescriptionLabel            ) //DcmTagKey(0x0020, 0x9421)
    GET_DICOM_INFO_LINE(DCM_PatientOrientationInFrameSequence    ) //DcmTagKey(0x0020, 0x9450)
    GET_DICOM_INFO_LINE(DCM_FrameLabel                           ) //DcmTagKey(0x0020, 0x9453)
    GET_DICOM_INFO_LINE(DCM_AcquisitionIndex                     ) //DcmTagKey(0x0020, 0x9518)
    GET_DICOM_INFO_LINE(DCM_ContributingSOPInstancesReferenceSequence ) //DcmTagKey(0x0020, 0x9529)
    GET_DICOM_INFO_LINE(DCM_ReconstructionIndex                  ) //DcmTagKey(0x0020, 0x9536)
            info->append("</div>");
            info->append("<div style=\"color:#00007F\">");
    GET_DICOM_INFO_LINE(DCM_LightPathFilterPassThroughWavelength ) //DcmTagKey(0x0022, 0x0001)
    GET_DICOM_INFO_LINE(DCM_LightPathFilterPassBand              ) //DcmTagKey(0x0022, 0x0002)
    GET_DICOM_INFO_LINE(DCM_ImagePathFilterPassThroughWavelength ) //DcmTagKey(0x0022, 0x0003)
    GET_DICOM_INFO_LINE(DCM_ImagePathFilterPassBand              ) //DcmTagKey(0x0022, 0x0004)
    GET_DICOM_INFO_LINE(DCM_PatientEyeMovementCommanded          ) //DcmTagKey(0x0022, 0x0005)
    GET_DICOM_INFO_LINE(DCM_PatientEyeMovementCommandCodeSequence ) //DcmTagKey(0x0022, 0x0006)
    GET_DICOM_INFO_LINE(DCM_SphericalLensPower                   ) //DcmTagKey(0x0022, 0x0007)
    GET_DICOM_INFO_LINE(DCM_CylinderLensPower                    ) //DcmTagKey(0x0022, 0x0008)
    GET_DICOM_INFO_LINE(DCM_CylinderAxis                         ) //DcmTagKey(0x0022, 0x0009)
    GET_DICOM_INFO_LINE(DCM_EmmetropicMagnification              ) //DcmTagKey(0x0022, 0x000a)
    GET_DICOM_INFO_LINE(DCM_IntraOcularPressure                  ) //DcmTagKey(0x0022, 0x000b)
    GET_DICOM_INFO_LINE(DCM_HorizontalFieldOfView                ) //DcmTagKey(0x0022, 0x000c)
    GET_DICOM_INFO_LINE(DCM_PupilDilated                         ) //DcmTagKey(0x0022, 0x000d)
    GET_DICOM_INFO_LINE(DCM_DegreeOfDilation                     ) //DcmTagKey(0x0022, 0x000e)
    GET_DICOM_INFO_LINE(DCM_StereoBaselineAngle                  ) //DcmTagKey(0x0022, 0x0010)
    GET_DICOM_INFO_LINE(DCM_StereoBaselineDisplacement           ) //DcmTagKey(0x0022, 0x0011)
    GET_DICOM_INFO_LINE(DCM_StereoHorizontalPixelOffset          ) //DcmTagKey(0x0022, 0x0012)
    GET_DICOM_INFO_LINE(DCM_StereoVerticalPixelOffset            ) //DcmTagKey(0x0022, 0x0013)
    GET_DICOM_INFO_LINE(DCM_StereoRotation                       ) //DcmTagKey(0x0022, 0x0014)
    GET_DICOM_INFO_LINE(DCM_AcquisitionDeviceTypeCodeSequence    ) //DcmTagKey(0x0022, 0x0015)
    GET_DICOM_INFO_LINE(DCM_IlluminationTypeCodeSequence         ) //DcmTagKey(0x0022, 0x0016)
    GET_DICOM_INFO_LINE(DCM_LightPathFilterTypeStackCodeSequence ) //DcmTagKey(0x0022, 0x0017)
    GET_DICOM_INFO_LINE(DCM_ImagePathFilterTypeStackCodeSequence ) //DcmTagKey(0x0022, 0x0018)
    GET_DICOM_INFO_LINE(DCM_LensesCodeSequence                   ) //DcmTagKey(0x0022, 0x0019)
    GET_DICOM_INFO_LINE(DCM_ChannelDescriptionCodeSequence       ) //DcmTagKey(0x0022, 0x001a)
    GET_DICOM_INFO_LINE(DCM_RefractiveStateSequence              ) //DcmTagKey(0x0022, 0x001b)
    GET_DICOM_INFO_LINE(DCM_MydriaticAgentCodeSequence           ) //DcmTagKey(0x0022, 0x001c)
    GET_DICOM_INFO_LINE(DCM_RelativeImagePositionCodeSequence    ) //DcmTagKey(0x0022, 0x001d)
    GET_DICOM_INFO_LINE(DCM_StereoPairsSequence                  ) //DcmTagKey(0x0022, 0x0020)
    GET_DICOM_INFO_LINE(DCM_LeftImageSequence                    ) //DcmTagKey(0x0022, 0x0021)
    GET_DICOM_INFO_LINE(DCM_RightImageSequence                   ) //DcmTagKey(0x0022, 0x0022)
    GET_DICOM_INFO_LINE(DCM_AxialLengthOfTheEye                  ) //DcmTagKey(0x0022, 0x0030)
    GET_DICOM_INFO_LINE(DCM_OphthalmicFrameLocationSequence      ) //DcmTagKey(0x0022, 0x0031)
    GET_DICOM_INFO_LINE(DCM_ReferenceCoordinates                 ) //DcmTagKey(0x0022, 0x0032)
    GET_DICOM_INFO_LINE(DCM_DepthSpatialResolution               ) //DcmTagKey(0x0022, 0x0035)
    GET_DICOM_INFO_LINE(DCM_MaximumDepthDistortion               ) //DcmTagKey(0x0022, 0x0036)
    GET_DICOM_INFO_LINE(DCM_AlongScanSpatialResolution           ) //DcmTagKey(0x0022, 0x0037)
    GET_DICOM_INFO_LINE(DCM_MaximumAlongScanDistortion           ) //DcmTagKey(0x0022, 0x0038)
    GET_DICOM_INFO_LINE(DCM_OphthalmicImageOrientation           ) //DcmTagKey(0x0022, 0x0039)
    GET_DICOM_INFO_LINE(DCM_DepthOfTransverseImage               ) //DcmTagKey(0x0022, 0x0041)
    GET_DICOM_INFO_LINE(DCM_MydriaticAgentConcentrationUnitsSequence ) //DcmTagKey(0x0022, 0x0042)
    GET_DICOM_INFO_LINE(DCM_AcrossScanSpatialResolution          ) //DcmTagKey(0x0022, 0x0048)
    GET_DICOM_INFO_LINE(DCM_MaximumAcrossScanDistortion          ) //DcmTagKey(0x0022, 0x0049)
    GET_DICOM_INFO_LINE(DCM_MydriaticAgentConcentration          ) //DcmTagKey(0x0022, 0x004e)
    GET_DICOM_INFO_LINE(DCM_IlluminationWaveLength               ) //DcmTagKey(0x0022, 0x0055)
    GET_DICOM_INFO_LINE(DCM_IlluminationPower                    ) //DcmTagKey(0x0022, 0x0056)
    GET_DICOM_INFO_LINE(DCM_IlluminationBandwidth                ) //DcmTagKey(0x0022, 0x0057)
    GET_DICOM_INFO_LINE(DCM_MydriaticAgentSequence               ) //DcmTagKey(0x0022, 0x0058)
    GET_DICOM_INFO_LINE(DCM_OphthalmicAxialMeasurementsRightEyeSequence ) //DcmTagKey(0x0022, 0x1007)
    GET_DICOM_INFO_LINE(DCM_OphthalmicAxialMeasurementsLeftEyeSequence ) //DcmTagKey(0x0022, 0x1008)
    GET_DICOM_INFO_LINE(DCM_OphthalmicAxialMeasurementsDeviceType ) //DcmTagKey(0x0022, 0x1009)
    GET_DICOM_INFO_LINE(DCM_OphthalmicAxialLengthMeasurementsType ) //DcmTagKey(0x0022, 0x1010)
    GET_DICOM_INFO_LINE(DCM_OphthalmicAxialLengthSequence        ) //DcmTagKey(0x0022, 0x1012)
    GET_DICOM_INFO_LINE(DCM_OphthalmicAxialLength                ) //DcmTagKey(0x0022, 0x1019)
    GET_DICOM_INFO_LINE(DCM_LensStatusCodeSequence               ) //DcmTagKey(0x0022, 0x1024)
    GET_DICOM_INFO_LINE(DCM_VitreousStatusCodeSequence           ) //DcmTagKey(0x0022, 0x1025)
    GET_DICOM_INFO_LINE(DCM_IOLFormulaCodeSequence               ) //DcmTagKey(0x0022, 0x1028)
    GET_DICOM_INFO_LINE(DCM_IOLFormulaDetail                     ) //DcmTagKey(0x0022, 0x1029)
    GET_DICOM_INFO_LINE(DCM_KeratometerIndex                     ) //DcmTagKey(0x0022, 0x1033)
//    GET_DICOM_INFO_LINE(DCM_SourceofOphthalmicAxialLengthCodeSequence ) //DcmTagKey(0x0022, 0x1035)
    GET_DICOM_INFO_LINE(DCM_TargetRefraction                     ) //DcmTagKey(0x0022, 0x1037)
    GET_DICOM_INFO_LINE(DCM_RefractiveProcedureOccurred          ) //DcmTagKey(0x0022, 0x1039)
    GET_DICOM_INFO_LINE(DCM_RefractiveSurgeryTypeCodeSequence    ) //DcmTagKey(0x0022, 0x1040)
    GET_DICOM_INFO_LINE(DCM_OphthalmicUltrasoundMethodCodeSequence ) //DcmTagKey(0x0022, 0x1044)
    GET_DICOM_INFO_LINE(DCM_OphthalmicAxialLengthMeasurementsSequence ) //DcmTagKey(0x0022, 0x1050)
    GET_DICOM_INFO_LINE(DCM_IOLPower                             ) //DcmTagKey(0x0022, 0x1053)
    GET_DICOM_INFO_LINE(DCM_PredictedRefractiveError             ) //DcmTagKey(0x0022, 0x1054)
    GET_DICOM_INFO_LINE(DCM_OphthalmicAxialLengthVelocity        ) //DcmTagKey(0x0022, 0x1059)
    GET_DICOM_INFO_LINE(DCM_LensStatusDescription                ) //DcmTagKey(0x0022, 0x1065)
    GET_DICOM_INFO_LINE(DCM_VitreousStatusDescription            ) //DcmTagKey(0x0022, 0x1066)
    GET_DICOM_INFO_LINE(DCM_IOLPowerSequence                     ) //DcmTagKey(0x0022, 0x1090)
    GET_DICOM_INFO_LINE(DCM_LensConstantSequence                 ) //DcmTagKey(0x0022, 0x1092)
    GET_DICOM_INFO_LINE(DCM_IOLManufacturer                      ) //DcmTagKey(0x0022, 0x1093)
    GET_DICOM_INFO_LINE(DCM_ImplantName                          ) //DcmTagKey(0x0022, 0x1095)
    GET_DICOM_INFO_LINE(DCM_KeratometryMeasurementTypeCodeSequence ) //DcmTagKey(0x0022, 0x1096)
    GET_DICOM_INFO_LINE(DCM_ImplantPartNumber                    ) //DcmTagKey(0x0022, 0x1097)
    GET_DICOM_INFO_LINE(DCM_ReferencedOphthalmicAxialMeasurementsSequence ) //DcmTagKey(0x0022, 0x1100)
    GET_DICOM_INFO_LINE(DCM_OphthalmicAxialLengthMeasurementsSegmentNameCodeSequence ) //DcmTagKey(0x0022, 0x1101)
    GET_DICOM_INFO_LINE(DCM_RefractiveErrorBeforeRefractiveSurgeryCodeSequence ) //DcmTagKey(0x0022, 0x1103)
    GET_DICOM_INFO_LINE(DCM_IOLPowerForExactEmmetropia           ) //DcmTagKey(0x0022, 0x1121)
    GET_DICOM_INFO_LINE(DCM_IOLPowerForExactTargetRefraction     ) //DcmTagKey(0x0022, 0x1122)
    GET_DICOM_INFO_LINE(DCM_AnteriorChamberDepthDefinitionCodeSequence ) //DcmTagKey(0x0022, 0x1125)
    GET_DICOM_INFO_LINE(DCM_LensThicknessSequence                ) //DcmTagKey(0x0022, 0x1127)
    GET_DICOM_INFO_LINE(DCM_AnteriorChamberDepthSequence         ) //DcmTagKey(0x0022, 0x1128)
    GET_DICOM_INFO_LINE(DCM_LensThickness                        ) //DcmTagKey(0x0022, 0x1130)
    GET_DICOM_INFO_LINE(DCM_AnteriorChamberDepth                 ) //DcmTagKey(0x0022, 0x1131)
//    GET_DICOM_INFO_LINE(DCM_SourceofLensThicknessDataCodeSequence ) //DcmTagKey(0x0022, 0x1132)
//    GET_DICOM_INFO_LINE(DCM_SourceofAnteriorChamberDepthDataCodeSequence ) //DcmTagKey(0x0022, 0x1133)
//    GET_DICOM_INFO_LINE(DCM_SourceofRefractiveMeasurementsSequence ) //DcmTagKey(0x0022, 0x1134)
//    GET_DICOM_INFO_LINE(DCM_SourceofRefractiveMeasurementsCodeSequence ) //DcmTagKey(0x0022, 0x1135)
    GET_DICOM_INFO_LINE(DCM_OphthalmicAxialLengthMeasurementModified ) //DcmTagKey(0x0022, 0x1140)
    GET_DICOM_INFO_LINE(DCM_OphthalmicAxialLengthDataSourceCodeSequence ) //DcmTagKey(0x0022, 0x1150)
//    GET_DICOM_INFO_LINE(DCM_SignaltoNoiseRatio                   ) //DcmTagKey(0x0022, 0x1155)
    GET_DICOM_INFO_LINE(DCM_OphthalmicAxialLengthDataSourceDescription ) //DcmTagKey(0x0022, 0x1159)
    GET_DICOM_INFO_LINE(DCM_OphthalmicAxialLengthMeasurementsTotalLengthSequence ) //DcmTagKey(0x0022, 0x1210)
    GET_DICOM_INFO_LINE(DCM_OphthalmicAxialLengthMeasurementsSegmentalLengthSequence ) //DcmTagKey(0x0022, 0x1211)
    GET_DICOM_INFO_LINE(DCM_OphthalmicAxialLengthMeasurementsLengthSummationSequence ) //DcmTagKey(0x0022, 0x1212)
    GET_DICOM_INFO_LINE(DCM_UltrasoundOphthalmicAxialLengthMeasurementsSequence ) //DcmTagKey(0x0022, 0x1220)
    GET_DICOM_INFO_LINE(DCM_OpticalOphthalmicAxialLengthMeasurementsSequence ) //DcmTagKey(0x0022, 0x1225)
    GET_DICOM_INFO_LINE(DCM_UltrasoundSelectedOphthalmicAxialLengthSequence ) //DcmTagKey(0x0022, 0x1230)
    GET_DICOM_INFO_LINE(DCM_OphthalmicAxialLengthSelectionMethodCodeSequence ) //DcmTagKey(0x0022, 0x1250)
    GET_DICOM_INFO_LINE(DCM_OpticalSelectedOphthalmicAxialLengthSequence ) //DcmTagKey(0x0022, 0x1255)
    GET_DICOM_INFO_LINE(DCM_SelectedSegmentalOphthalmicAxialLengthSequence ) //DcmTagKey(0x0022, 0x1257)
    GET_DICOM_INFO_LINE(DCM_SelectedTotalOphthalmicAxialLengthSequence ) //DcmTagKey(0x0022, 0x1260)
    GET_DICOM_INFO_LINE(DCM_OphthalmicAxialLengthQualityMetricSequence ) //DcmTagKey(0x0022, 0x1262)
//    GET_DICOM_INFO_LINE(DCM_OphthalmicAxialLengthQualityMetricTypeCodeSequence ) //DcmTagKey(0x0022, 0x1265)
    GET_DICOM_INFO_LINE(DCM_IntraocularLensCalculationsRightEyeSequence ) //DcmTagKey(0x0022, 0x1300)
    GET_DICOM_INFO_LINE(DCM_IntraocularLensCalculationsLeftEyeSequence ) //DcmTagKey(0x0022, 0x1310)
    GET_DICOM_INFO_LINE(DCM_ReferencedOphthalmicAxialLengthMeasurementQCImageSequence ) //DcmTagKey(0x0022, 0x1330)
            info->append("</div>");
            info->append("<div style=\"color:#00007F\">");
    GET_DICOM_INFO_LINE(DCM_VisualFieldHorizontalExtent          ) //DcmTagKey(0x0024, 0x0010)
    GET_DICOM_INFO_LINE(DCM_VisualFieldVerticalExtent            ) //DcmTagKey(0x0024, 0x0011)
    GET_DICOM_INFO_LINE(DCM_VisualFieldShape                     ) //DcmTagKey(0x0024, 0x0012)
    GET_DICOM_INFO_LINE(DCM_ScreeningTestModeCodeSequence        ) //DcmTagKey(0x0024, 0x0016)
    GET_DICOM_INFO_LINE(DCM_MaximumStimulusLuminance             ) //DcmTagKey(0x0024, 0x0018)
    GET_DICOM_INFO_LINE(DCM_BackgroundLuminance                  ) //DcmTagKey(0x0024, 0x0020)
    GET_DICOM_INFO_LINE(DCM_StimulusColorCodeSequence            ) //DcmTagKey(0x0024, 0x0021)
    GET_DICOM_INFO_LINE(DCM_BackgroundIlluminationColorCodeSequence ) //DcmTagKey(0x0024, 0x0024)
    GET_DICOM_INFO_LINE(DCM_StimulusArea                         ) //DcmTagKey(0x0024, 0x0025)
    GET_DICOM_INFO_LINE(DCM_StimulusPresentationTime             ) //DcmTagKey(0x0024, 0x0028)
    GET_DICOM_INFO_LINE(DCM_FixationSequence                     ) //DcmTagKey(0x0024, 0x0032)
    GET_DICOM_INFO_LINE(DCM_FixationMonitoringCodeSequence       ) //DcmTagKey(0x0024, 0x0033)
    GET_DICOM_INFO_LINE(DCM_VisualFieldCatchTrialSequence        ) //DcmTagKey(0x0024, 0x0034)
    GET_DICOM_INFO_LINE(DCM_FixationCheckedQuantity              ) //DcmTagKey(0x0024, 0x0035)
    GET_DICOM_INFO_LINE(DCM_PatientNotProperlyFixatedQuantity    ) //DcmTagKey(0x0024, 0x0036)
    GET_DICOM_INFO_LINE(DCM_PresentedVisualStimuliDataFlag       ) //DcmTagKey(0x0024, 0x0037)
    GET_DICOM_INFO_LINE(DCM_NumberOfVisualStimuli                ) //DcmTagKey(0x0024, 0x0038)
    GET_DICOM_INFO_LINE(DCM_ExcessiveFixationLossesDataFlag      ) //DcmTagKey(0x0024, 0x0039)
    GET_DICOM_INFO_LINE(DCM_ExcessiveFixationLosses              ) //DcmTagKey(0x0024, 0x0040)
    GET_DICOM_INFO_LINE(DCM_StimuliRetestingQuantity             ) //DcmTagKey(0x0024, 0x0042)
    GET_DICOM_INFO_LINE(DCM_CommentsOnPatientPerformanceOfVisualField ) //DcmTagKey(0x0024, 0x0044)
    GET_DICOM_INFO_LINE(DCM_FalseNegativesEstimateFlag           ) //DcmTagKey(0x0024, 0x0045)
    GET_DICOM_INFO_LINE(DCM_FalseNegativesEstimate               ) //DcmTagKey(0x0024, 0x0046)
    GET_DICOM_INFO_LINE(DCM_NegativeCatchTrialsQuantity          ) //DcmTagKey(0x0024, 0x0048)
    GET_DICOM_INFO_LINE(DCM_FalseNegativesQuantity               ) //DcmTagKey(0x0024, 0x0050)
    GET_DICOM_INFO_LINE(DCM_ExcessiveFalseNegativesDataFlag      ) //DcmTagKey(0x0024, 0x0051)
    GET_DICOM_INFO_LINE(DCM_ExcessiveFalseNegatives              ) //DcmTagKey(0x0024, 0x0052)
    GET_DICOM_INFO_LINE(DCM_FalsePositivesEstimateFlag           ) //DcmTagKey(0x0024, 0x0053)
    GET_DICOM_INFO_LINE(DCM_FalsePositivesEstimate               ) //DcmTagKey(0x0024, 0x0054)
    GET_DICOM_INFO_LINE(DCM_CatchTrialsDataFlag                  ) //DcmTagKey(0x0024, 0x0055)
    GET_DICOM_INFO_LINE(DCM_PositiveCatchTrialsQuantity          ) //DcmTagKey(0x0024, 0x0056)
    GET_DICOM_INFO_LINE(DCM_TestPointNormalsDataFlag             ) //DcmTagKey(0x0024, 0x0057)
    GET_DICOM_INFO_LINE(DCM_TestPointNormalsSequence             ) //DcmTagKey(0x0024, 0x0058)
    GET_DICOM_INFO_LINE(DCM_GlobalDeviationProbabilityNormalsFlag ) //DcmTagKey(0x0024, 0x0059)
    GET_DICOM_INFO_LINE(DCM_FalsePositivesQuantity               ) //DcmTagKey(0x0024, 0x0060)
    GET_DICOM_INFO_LINE(DCM_ExcessiveFalsePositivesDataFlag      ) //DcmTagKey(0x0024, 0x0061)
    GET_DICOM_INFO_LINE(DCM_ExcessiveFalsePositives              ) //DcmTagKey(0x0024, 0x0062)
    GET_DICOM_INFO_LINE(DCM_VisualFieldTestNormalsFlag           ) //DcmTagKey(0x0024, 0x0063)
    GET_DICOM_INFO_LINE(DCM_ResultsNormalsSequence               ) //DcmTagKey(0x0024, 0x0064)
    GET_DICOM_INFO_LINE(DCM_AgeCorrectedSensitivityDeviationAlgorithmSequence ) //DcmTagKey(0x0024, 0x0065)
    GET_DICOM_INFO_LINE(DCM_GlobalDeviationFromNormal            ) //DcmTagKey(0x0024, 0x0066)
    GET_DICOM_INFO_LINE(DCM_GeneralizedDefectSensitivityDeviationAlgorithmSequence ) //DcmTagKey(0x0024, 0x0067)
//    GET_DICOM_INFO_LINE(DCM_LocalizedDeviationfromNormal         ) //DcmTagKey(0x0024, 0x0068)
    GET_DICOM_INFO_LINE(DCM_PatientReliabilityIndicator          ) //DcmTagKey(0x0024, 0x0069)
    GET_DICOM_INFO_LINE(DCM_VisualFieldMeanSensitivity           ) //DcmTagKey(0x0024, 0x0070)
    GET_DICOM_INFO_LINE(DCM_GlobalDeviationProbability           ) //DcmTagKey(0x0024, 0x0071)
    GET_DICOM_INFO_LINE(DCM_LocalDeviationProbabilityNormalsFlag ) //DcmTagKey(0x0024, 0x0072)
    GET_DICOM_INFO_LINE(DCM_LocalizedDeviationProbability        ) //DcmTagKey(0x0024, 0x0073)
    GET_DICOM_INFO_LINE(DCM_ShortTermFluctuationCalculated       ) //DcmTagKey(0x0024, 0x0074)
    GET_DICOM_INFO_LINE(DCM_ShortTermFluctuation                 ) //DcmTagKey(0x0024, 0x0075)
    GET_DICOM_INFO_LINE(DCM_ShortTermFluctuationProbabilityCalculated ) //DcmTagKey(0x0024, 0x0076)
    GET_DICOM_INFO_LINE(DCM_ShortTermFluctuationProbability      ) //DcmTagKey(0x0024, 0x0077)
    GET_DICOM_INFO_LINE(DCM_CorrectedLocalizedDeviationFromNormalCalculated ) //DcmTagKey(0x0024, 0x0078)
    GET_DICOM_INFO_LINE(DCM_CorrectedLocalizedDeviationFromNormal ) //DcmTagKey(0x0024, 0x0079)
    GET_DICOM_INFO_LINE(DCM_CorrectedLocalizedDeviationFromNormalProbabilityCalculated ) //DcmTagKey(0x0024, 0x0080)
    GET_DICOM_INFO_LINE(DCM_CorrectedLocalizedDeviationFromNormalProbability ) //DcmTagKey(0x0024, 0x0081)
    GET_DICOM_INFO_LINE(DCM_GlobalDeviationProbabilitySequence   ) //DcmTagKey(0x0024, 0x0083)
    GET_DICOM_INFO_LINE(DCM_LocalizedDeviationProbabilitySequence ) //DcmTagKey(0x0024, 0x0085)
    GET_DICOM_INFO_LINE(DCM_FovealSensitivityMeasured            ) //DcmTagKey(0x0024, 0x0086)
    GET_DICOM_INFO_LINE(DCM_FovealSensitivity                    ) //DcmTagKey(0x0024, 0x0087)
    GET_DICOM_INFO_LINE(DCM_VisualFieldTestDuration              ) //DcmTagKey(0x0024, 0x0088)
    GET_DICOM_INFO_LINE(DCM_VisualFieldTestPointSequence         ) //DcmTagKey(0x0024, 0x0089)
    GET_DICOM_INFO_LINE(DCM_VisualFieldTestPointXCoordinate      ) //DcmTagKey(0x0024, 0x0090)
    GET_DICOM_INFO_LINE(DCM_VisualFieldTestPointYCoordinate      ) //DcmTagKey(0x0024, 0x0091)
    GET_DICOM_INFO_LINE(DCM_AgeCorrectedSensitivityDeviationValue ) //DcmTagKey(0x0024, 0x0092)
    GET_DICOM_INFO_LINE(DCM_StimulusResults                      ) //DcmTagKey(0x0024, 0x0093)
    GET_DICOM_INFO_LINE(DCM_SensitivityValue                     ) //DcmTagKey(0x0024, 0x0094)
    GET_DICOM_INFO_LINE(DCM_RetestStimulusSeen                   ) //DcmTagKey(0x0024, 0x0095)
    GET_DICOM_INFO_LINE(DCM_RetestSensitivityValue               ) //DcmTagKey(0x0024, 0x0096)
    GET_DICOM_INFO_LINE(DCM_VisualFieldTestPointNormalsSequence  ) //DcmTagKey(0x0024, 0x0097)
    GET_DICOM_INFO_LINE(DCM_QuantifiedDefect                     ) //DcmTagKey(0x0024, 0x0098)
    GET_DICOM_INFO_LINE(DCM_GeneralizedDefectCorrectedSensitivityDeviationFlag ) //DcmTagKey(0x0024, 0x0102)
    GET_DICOM_INFO_LINE(DCM_GeneralizedDefectCorrectedSensitivityDeviationValue ) //DcmTagKey(0x0024, 0x0103)
    GET_DICOM_INFO_LINE(DCM_GeneralizedDefectCorrectedSensitivityDeviationProbabilityValue ) //DcmTagKey(0x0024, 0x0104)
    GET_DICOM_INFO_LINE(DCM_MinimumSensitivityValue              ) //DcmTagKey(0x0024, 0x0105)
    GET_DICOM_INFO_LINE(DCM_BlindSpotLocalized                   ) //DcmTagKey(0x0024, 0x0106)
    GET_DICOM_INFO_LINE(DCM_BlindSpotXCoordinate                 ) //DcmTagKey(0x0024, 0x0107)
    GET_DICOM_INFO_LINE(DCM_BlindSpotYCoordinate                 ) //DcmTagKey(0x0024, 0x0108)
    GET_DICOM_INFO_LINE(DCM_VisualAcuityMeasurementSequence      ) //DcmTagKey(0x0024, 0x0110)
    GET_DICOM_INFO_LINE(DCM_RefractiveParametersUsedOnPatientSequence ) //DcmTagKey(0x0024, 0x0112)
    GET_DICOM_INFO_LINE(DCM_MeasurementLaterality                ) //DcmTagKey(0x0024, 0x0113)
    GET_DICOM_INFO_LINE(DCM_OphthalmicPatientClinicalInformationLeftEyeSequence ) //DcmTagKey(0x0024, 0x0114)
    GET_DICOM_INFO_LINE(DCM_OphthalmicPatientClinicalInformationRightEyeSequence ) //DcmTagKey(0x0024, 0x0115)
    GET_DICOM_INFO_LINE(DCM_FovealPointNormativeDataFlag         ) //DcmTagKey(0x0024, 0x0117)
    GET_DICOM_INFO_LINE(DCM_FovealPointProbabilityValue          ) //DcmTagKey(0x0024, 0x0118)
    GET_DICOM_INFO_LINE(DCM_ScreeningBaselineMeasured            ) //DcmTagKey(0x0024, 0x0120)
    GET_DICOM_INFO_LINE(DCM_ScreeningBaselineMeasuredSequence    ) //DcmTagKey(0x0024, 0x0122)
    GET_DICOM_INFO_LINE(DCM_ScreeningBaselineType                ) //DcmTagKey(0x0024, 0x0124)
    GET_DICOM_INFO_LINE(DCM_ScreeningBaselineValue               ) //DcmTagKey(0x0024, 0x0126)
    GET_DICOM_INFO_LINE(DCM_AlgorithmSource                      ) //DcmTagKey(0x0024, 0x0202)
    GET_DICOM_INFO_LINE(DCM_DataSetName                          ) //DcmTagKey(0x0024, 0x0306)
    GET_DICOM_INFO_LINE(DCM_DataSetVersion                       ) //DcmTagKey(0x0024, 0x0307)
    GET_DICOM_INFO_LINE(DCM_DataSetSource                        ) //DcmTagKey(0x0024, 0x0308)
    GET_DICOM_INFO_LINE(DCM_DataSetDescription                   ) //DcmTagKey(0x0024, 0x0309)
    GET_DICOM_INFO_LINE(DCM_VisualFieldTestReliabilityGlobalIndexSequence ) //DcmTagKey(0x0024, 0x0317)
    GET_DICOM_INFO_LINE(DCM_VisualFieldGlobalResultsIndexSequence ) //DcmTagKey(0x0024, 0x0320)
    GET_DICOM_INFO_LINE(DCM_DataObservationSequence              ) //DcmTagKey(0x0024, 0x0325)
    GET_DICOM_INFO_LINE(DCM_IndexNormalsFlag                     ) //DcmTagKey(0x0024, 0x0338)
    GET_DICOM_INFO_LINE(DCM_IndexProbability                     ) //DcmTagKey(0x0024, 0x0341)
    GET_DICOM_INFO_LINE(DCM_IndexProbabilitySequence             ) //DcmTagKey(0x0024, 0x0344)
            info->append("</div>");
            info->append("<div style=\"color:#0000FF\">");
    GET_DICOM_INFO_LINE(DCM_SamplesPerPixel                      ) //DcmTagKey(0x0028, 0x0002)
    GET_DICOM_INFO_LINE(DCM_SamplesPerPixelUsed                  ) //DcmTagKey(0x0028, 0x0003)
    GET_DICOM_INFO_LINE(DCM_PhotometricInterpretation            ) //DcmTagKey(0x0028, 0x0004)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_ImageDimensions             ) //DcmTagKey(0x0028, 0x0005)
    GET_DICOM_INFO_LINE(DCM_PlanarConfiguration                  ) //DcmTagKey(0x0028, 0x0006)
    GET_DICOM_INFO_LINE(DCM_NumberOfFrames                       ) //DcmTagKey(0x0028, 0x0008)
    GET_DICOM_INFO_LINE(DCM_FrameIncrementPointer                ) //DcmTagKey(0x0028, 0x0009)
    GET_DICOM_INFO_LINE(DCM_FrameDimensionPointer                ) //DcmTagKey(0x0028, 0x000a)
    GET_DICOM_INFO_LINE(DCM_Rows                                 ) //DcmTagKey(0x0028, 0x0010)
    GET_DICOM_INFO_LINE(DCM_Columns                              ) //DcmTagKey(0x0028, 0x0011)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_Planes                       ) //DcmTagKey(0x0028, 0x0012)
    GET_DICOM_INFO_LINE(DCM_UltrasoundColorDataPresent           ) //DcmTagKey(0x0028, 0x0014)
    GET_DICOM_INFO_LINE(DCM_PixelSpacing                         ) //DcmTagKey(0x0028, 0x0030)
    GET_DICOM_INFO_LINE(DCM_ZoomFactor                           ) //DcmTagKey(0x0028, 0x0031)
    GET_DICOM_INFO_LINE(DCM_ZoomCenter                           ) //DcmTagKey(0x0028, 0x0032)
    GET_DICOM_INFO_LINE(DCM_PixelAspectRatio                     ) //DcmTagKey(0x0028, 0x0034)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_ImageFormat                 ) //DcmTagKey(0x0028, 0x0040)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_ManipulatedImage            ) //DcmTagKey(0x0028, 0x0050)
    GET_DICOM_INFO_LINE(DCM_CorrectedImage                       ) //DcmTagKey(0x0028, 0x0051)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_CompressionRecognitionCode ) //DcmTagKey(0x0028, 0x005f)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_CompressionCode             ) //DcmTagKey(0x0028, 0x0060)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_CompressionOriginator    ) //DcmTagKey(0x0028, 0x0061)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_CompressionLabel         ) //DcmTagKey(0x0028, 0x0062)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_CompressionDescription   ) //DcmTagKey(0x0028, 0x0063)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_CompressionSequence      ) //DcmTagKey(0x0028, 0x0065)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_CompressionStepPointers  ) //DcmTagKey(0x0028, 0x0066)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_RepeatInterval           ) //DcmTagKey(0x0028, 0x0068)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_BitsGrouped              ) //DcmTagKey(0x0028, 0x0069)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_PerimeterTable           ) //DcmTagKey(0x0028, 0x0070)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_PerimeterValue           ) //DcmTagKey(0x0028, 0x0071)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_PredictorRows            ) //DcmTagKey(0x0028, 0x0080)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_PredictorColumns         ) //DcmTagKey(0x0028, 0x0081)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_PredictorConstants       ) //DcmTagKey(0x0028, 0x0082)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_BlockedPixels            ) //DcmTagKey(0x0028, 0x0090)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_BlockRows                ) //DcmTagKey(0x0028, 0x0091)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_BlockColumns             ) //DcmTagKey(0x0028, 0x0092)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_RowOverlap               ) //DcmTagKey(0x0028, 0x0093)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_ColumnOverlap            ) //DcmTagKey(0x0028, 0x0094)
    GET_DICOM_INFO_LINE(DCM_BitsAllocated                        ) //DcmTagKey(0x0028, 0x0100)
    GET_DICOM_INFO_LINE(DCM_BitsStored                           ) //DcmTagKey(0x0028, 0x0101)
    GET_DICOM_INFO_LINE(DCM_HighBit                              ) //DcmTagKey(0x0028, 0x0102)
    GET_DICOM_INFO_LINE(DCM_PixelRepresentation                  ) //DcmTagKey(0x0028, 0x0103)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_SmallestValidPixelValue     ) //DcmTagKey(0x0028, 0x0104)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_LargestValidPixelValue      ) //DcmTagKey(0x0028, 0x0105)
    GET_DICOM_INFO_LINE(DCM_SmallestImagePixelValue              ) //DcmTagKey(0x0028, 0x0106)
    GET_DICOM_INFO_LINE(DCM_LargestImagePixelValue               ) //DcmTagKey(0x0028, 0x0107)
    GET_DICOM_INFO_LINE(DCM_SmallestPixelValueInSeries           ) //DcmTagKey(0x0028, 0x0108)
    GET_DICOM_INFO_LINE(DCM_LargestPixelValueInSeries            ) //DcmTagKey(0x0028, 0x0109)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_SmallestImagePixelValueInPlane ) //DcmTagKey(0x0028, 0x0110)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_LargestImagePixelValueInPlane ) //DcmTagKey(0x0028, 0x0111)
    GET_DICOM_INFO_LINE(DCM_PixelPaddingValue                    ) //DcmTagKey(0x0028, 0x0120)
    GET_DICOM_INFO_LINE(DCM_PixelPaddingRangeLimit               ) //DcmTagKey(0x0028, 0x0121)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_ImageLocation               ) //DcmTagKey(0x0028, 0x0200)
    GET_DICOM_INFO_LINE(DCM_QualityControlImage                  ) //DcmTagKey(0x0028, 0x0300)
    GET_DICOM_INFO_LINE(DCM_BurnedInAnnotation                   ) //DcmTagKey(0x0028, 0x0301)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_TransformLabel           ) //DcmTagKey(0x0028, 0x0400)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_TransformVersionNumber   ) //DcmTagKey(0x0028, 0x0401)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_NumberOfTransformSteps   ) //DcmTagKey(0x0028, 0x0402)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_SequenceOfCompressedData ) //DcmTagKey(0x0028, 0x0403)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_DetailsOfCoefficients    ) //DcmTagKey(0x0028, 0x0404)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_RowsForNthOrderCoefficients ) //DcmTagKey(0x0028, 0x0410)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_ColumnsForNthOrderCoefficients ) //DcmTagKey(0x0028, 0x0411)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_CoefficientCoding        ) //DcmTagKey(0x0028, 0x0412)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_CoefficientCodingPointers ) //DcmTagKey(0x0028, 0x0413)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_DCTLabel                 ) //DcmTagKey(0x0028, 0x0700)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_DataBlockDescription     ) //DcmTagKey(0x0028, 0x0701)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_DataBlock                ) //DcmTagKey(0x0028, 0x0702)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_NormalizationFactorFormat ) //DcmTagKey(0x0028, 0x0710)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_ZonalMapNumberFormat     ) //DcmTagKey(0x0028, 0x0720)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_ZonalMapLocation         ) //DcmTagKey(0x0028, 0x0721)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_ZonalMapFormat           ) //DcmTagKey(0x0028, 0x0722)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_AdaptiveMapFormat        ) //DcmTagKey(0x0028, 0x0730)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_CodeNumberFormat         ) //DcmTagKey(0x0028, 0x0740)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_CodeLabel                ) //DcmTagKey(0x0028, 0x0800)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_NumberOfTables           ) //DcmTagKey(0x0028, 0x0802)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_CodeTableLocation        ) //DcmTagKey(0x0028, 0x0803)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_BitsForCodeWord          ) //DcmTagKey(0x0028, 0x0804)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_ImageDataLocation        ) //DcmTagKey(0x0028, 0x0808)
    GET_DICOM_INFO_LINE(DCM_PixelSpacingCalibrationType          ) //DcmTagKey(0x0028, 0x0a02)
    GET_DICOM_INFO_LINE(DCM_PixelSpacingCalibrationDescription   ) //DcmTagKey(0x0028, 0x0a04)
    GET_DICOM_INFO_LINE(DCM_PixelIntensityRelationship           ) //DcmTagKey(0x0028, 0x1040)
    GET_DICOM_INFO_LINE(DCM_PixelIntensityRelationshipSign       ) //DcmTagKey(0x0028, 0x1041)
    GET_DICOM_INFO_LINE(DCM_WindowCenter                         ) //DcmTagKey(0x0028, 0x1050)
    GET_DICOM_INFO_LINE(DCM_WindowWidth                          ) //DcmTagKey(0x0028, 0x1051)
    GET_DICOM_INFO_LINE(DCM_RescaleIntercept                     ) //DcmTagKey(0x0028, 0x1052)
    GET_DICOM_INFO_LINE(DCM_RescaleSlope                         ) //DcmTagKey(0x0028, 0x1053)
    GET_DICOM_INFO_LINE(DCM_RescaleType                          ) //DcmTagKey(0x0028, 0x1054)
    GET_DICOM_INFO_LINE(DCM_WindowCenterWidthExplanation         ) //DcmTagKey(0x0028, 0x1055)
    GET_DICOM_INFO_LINE(DCM_VOILUTFunction                       ) //DcmTagKey(0x0028, 0x1056)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_GrayScale                   ) //DcmTagKey(0x0028, 0x1080)
    GET_DICOM_INFO_LINE(DCM_RecommendedViewingMode               ) //DcmTagKey(0x0028, 0x1090)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_GrayLookupTableDescriptor   ) //DcmTagKey(0x0028, 0x1100)
    GET_DICOM_INFO_LINE(DCM_RedPaletteColorLookupTableDescriptor ) //DcmTagKey(0x0028, 0x1101)
    GET_DICOM_INFO_LINE(DCM_GreenPaletteColorLookupTableDescriptor ) //DcmTagKey(0x0028, 0x1102)
    GET_DICOM_INFO_LINE(DCM_BluePaletteColorLookupTableDescriptor ) //DcmTagKey(0x0028, 0x1103)
    GET_DICOM_INFO_LINE(DCM_AlphaPaletteColorLookupTableDescriptor ) //DcmTagKey(0x0028, 0x1104)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_LargeRedPaletteColorLookupTableDescriptor ) //DcmTagKey(0x0028, 0x1111)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_LargeGreenPaletteColorLookupTableDescriptor ) //DcmTagKey(0x0028, 0x1112)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_LargeBluePaletteColorLookupTableDescriptor ) //DcmTagKey(0x0028, 0x1113)
    GET_DICOM_INFO_LINE(DCM_PaletteColorLookupTableUID           ) //DcmTagKey(0x0028, 0x1199)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_GrayLookupTableData         ) //DcmTagKey(0x0028, 0x1200)
    GET_DICOM_INFO_LINE(DCM_RedPaletteColorLookupTableData       ) //DcmTagKey(0x0028, 0x1201)
    GET_DICOM_INFO_LINE(DCM_GreenPaletteColorLookupTableData     ) //DcmTagKey(0x0028, 0x1202)
    GET_DICOM_INFO_LINE(DCM_BluePaletteColorLookupTableData      ) //DcmTagKey(0x0028, 0x1203)
    GET_DICOM_INFO_LINE(DCM_AlphaPaletteColorLookupTableData     ) //DcmTagKey(0x0028, 0x1204)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_LargeRedPaletteColorLookupTableData ) //DcmTagKey(0x0028, 0x1211)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_LargeGreenPaletteColorLookupTableData ) //DcmTagKey(0x0028, 0x1212)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_LargeBluePaletteColorLookupTableData ) //DcmTagKey(0x0028, 0x1213)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_LargePaletteColorLookupTableUID ) //DcmTagKey(0x0028, 0x1214)
    GET_DICOM_INFO_LINE(DCM_SegmentedRedPaletteColorLookupTableData ) //DcmTagKey(0x0028, 0x1221)
    GET_DICOM_INFO_LINE(DCM_SegmentedGreenPaletteColorLookupTableData ) //DcmTagKey(0x0028, 0x1222)
    GET_DICOM_INFO_LINE(DCM_SegmentedBluePaletteColorLookupTableData ) //DcmTagKey(0x0028, 0x1223)
    GET_DICOM_INFO_LINE(DCM_BreastImplantPresent                 ) //DcmTagKey(0x0028, 0x1300)
    GET_DICOM_INFO_LINE(DCM_PartialView                          ) //DcmTagKey(0x0028, 0x1350)
    GET_DICOM_INFO_LINE(DCM_PartialViewDescription               ) //DcmTagKey(0x0028, 0x1351)
    GET_DICOM_INFO_LINE(DCM_PartialViewCodeSequence              ) //DcmTagKey(0x0028, 0x1352)
    GET_DICOM_INFO_LINE(DCM_SpatialLocationsPreserved            ) //DcmTagKey(0x0028, 0x135a)
    GET_DICOM_INFO_LINE(DCM_DataFrameAssignmentSequence          ) //DcmTagKey(0x0028, 0x1401)
    GET_DICOM_INFO_LINE(DCM_DataPathAssignment                   ) //DcmTagKey(0x0028, 0x1402)
    GET_DICOM_INFO_LINE(DCM_BitsMappedToColorLookupTable         ) //DcmTagKey(0x0028, 0x1403)
    GET_DICOM_INFO_LINE(DCM_BlendingLUT1Sequence                 ) //DcmTagKey(0x0028, 0x1404)
    GET_DICOM_INFO_LINE(DCM_BlendingLUT1TransferFunction         ) //DcmTagKey(0x0028, 0x1405)
    GET_DICOM_INFO_LINE(DCM_BlendingWeightConstant               ) //DcmTagKey(0x0028, 0x1406)
    GET_DICOM_INFO_LINE(DCM_BlendingLookupTableDescriptor        ) //DcmTagKey(0x0028, 0x1407)
    GET_DICOM_INFO_LINE(DCM_BlendingLookupTableData              ) //DcmTagKey(0x0028, 0x1408)
    GET_DICOM_INFO_LINE(DCM_EnhancedPaletteColorLookupTableSequence ) //DcmTagKey(0x0028, 0x140b)
    GET_DICOM_INFO_LINE(DCM_BlendingLUT2Sequence                 ) //DcmTagKey(0x0028, 0x140c)
    GET_DICOM_INFO_LINE(DCM_BlendingLUT2TransferFunction         ) //DcmTagKey(0x0028, 0x140d)
    GET_DICOM_INFO_LINE(DCM_DataPathID                           ) //DcmTagKey(0x0028, 0x140e)
    GET_DICOM_INFO_LINE(DCM_RGBLUTTransferFunction               ) //DcmTagKey(0x0028, 0x140f)
    GET_DICOM_INFO_LINE(DCM_AlphaLUTTransferFunction             ) //DcmTagKey(0x0028, 0x1410)
    GET_DICOM_INFO_LINE(DCM_ICCProfile                           ) //DcmTagKey(0x0028, 0x2000)
    GET_DICOM_INFO_LINE(DCM_LossyImageCompression                ) //DcmTagKey(0x0028, 0x2110)
    GET_DICOM_INFO_LINE(DCM_LossyImageCompressionRatio           ) //DcmTagKey(0x0028, 0x2112)
    GET_DICOM_INFO_LINE(DCM_LossyImageCompressionMethod          ) //DcmTagKey(0x0028, 0x2114)
    GET_DICOM_INFO_LINE(DCM_ModalityLUTSequence                  ) //DcmTagKey(0x0028, 0x3000)
    GET_DICOM_INFO_LINE(DCM_LUTDescriptor                        ) //DcmTagKey(0x0028, 0x3002)
    GET_DICOM_INFO_LINE(DCM_LUTExplanation                       ) //DcmTagKey(0x0028, 0x3003)
    GET_DICOM_INFO_LINE(DCM_ModalityLUTType                      ) //DcmTagKey(0x0028, 0x3004)
    GET_DICOM_INFO_LINE(DCM_LUTData                              ) //DcmTagKey(0x0028, 0x3006)
    GET_DICOM_INFO_LINE(DCM_VOILUTSequence                       ) //DcmTagKey(0x0028, 0x3010)
    GET_DICOM_INFO_LINE(DCM_SoftcopyVOILUTSequence               ) //DcmTagKey(0x0028, 0x3110)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_ImagePresentationComments   ) //DcmTagKey(0x0028, 0x4000)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_BiPlaneAcquisitionSequence   ) //DcmTagKey(0x0028, 0x5000)
    GET_DICOM_INFO_LINE(DCM_RepresentativeFrameNumber            ) //DcmTagKey(0x0028, 0x6010)
    GET_DICOM_INFO_LINE(DCM_FrameNumbersOfInterest               ) //DcmTagKey(0x0028, 0x6020)
    GET_DICOM_INFO_LINE(DCM_FrameOfInterestDescription           ) //DcmTagKey(0x0028, 0x6022)
    GET_DICOM_INFO_LINE(DCM_FrameOfInterestType                  ) //DcmTagKey(0x0028, 0x6023)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_MaskPointers                 ) //DcmTagKey(0x0028, 0x6030)
    GET_DICOM_INFO_LINE(DCM_RWavePointer                         ) //DcmTagKey(0x0028, 0x6040)
    GET_DICOM_INFO_LINE(DCM_MaskSubtractionSequence              ) //DcmTagKey(0x0028, 0x6100)
    GET_DICOM_INFO_LINE(DCM_MaskOperation                        ) //DcmTagKey(0x0028, 0x6101)
    GET_DICOM_INFO_LINE(DCM_ApplicableFrameRange                 ) //DcmTagKey(0x0028, 0x6102)
    GET_DICOM_INFO_LINE(DCM_MaskFrameNumbers                     ) //DcmTagKey(0x0028, 0x6110)
    GET_DICOM_INFO_LINE(DCM_ContrastFrameAveraging               ) //DcmTagKey(0x0028, 0x6112)
    GET_DICOM_INFO_LINE(DCM_MaskSubPixelShift                    ) //DcmTagKey(0x0028, 0x6114)
    GET_DICOM_INFO_LINE(DCM_TIDOffset                            ) //DcmTagKey(0x0028, 0x6120)
    GET_DICOM_INFO_LINE(DCM_MaskOperationExplanation             ) //DcmTagKey(0x0028, 0x6190)
    GET_DICOM_INFO_LINE(DCM_PixelDataProviderURL                 ) //DcmTagKey(0x0028, 0x7fe0)
    GET_DICOM_INFO_LINE(DCM_DataPointRows                        ) //DcmTagKey(0x0028, 0x9001)
    GET_DICOM_INFO_LINE(DCM_DataPointColumns                     ) //DcmTagKey(0x0028, 0x9002)
    GET_DICOM_INFO_LINE(DCM_SignalDomainColumns                  ) //DcmTagKey(0x0028, 0x9003)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_LargestMonochromePixelValue  ) //DcmTagKey(0x0028, 0x9099)
    GET_DICOM_INFO_LINE(DCM_DataRepresentation                   ) //DcmTagKey(0x0028, 0x9108)
    GET_DICOM_INFO_LINE(DCM_PixelMeasuresSequence                ) //DcmTagKey(0x0028, 0x9110)
    GET_DICOM_INFO_LINE(DCM_FrameVOILUTSequence                  ) //DcmTagKey(0x0028, 0x9132)
    GET_DICOM_INFO_LINE(DCM_PixelValueTransformationSequence     ) //DcmTagKey(0x0028, 0x9145)
    GET_DICOM_INFO_LINE(DCM_SignalDomainRows                     ) //DcmTagKey(0x0028, 0x9235)
    GET_DICOM_INFO_LINE(DCM_DisplayFilterPercentage              ) //DcmTagKey(0x0028, 0x9411)
    GET_DICOM_INFO_LINE(DCM_FramePixelShiftSequence              ) //DcmTagKey(0x0028, 0x9415)
    GET_DICOM_INFO_LINE(DCM_SubtractionItemID                    ) //DcmTagKey(0x0028, 0x9416)
    GET_DICOM_INFO_LINE(DCM_PixelIntensityRelationshipLUTSequence ) //DcmTagKey(0x0028, 0x9422)
    GET_DICOM_INFO_LINE(DCM_FramePixelDataPropertiesSequence     ) //DcmTagKey(0x0028, 0x9443)
    GET_DICOM_INFO_LINE(DCM_GeometricalProperties                ) //DcmTagKey(0x0028, 0x9444)
    GET_DICOM_INFO_LINE(DCM_GeometricMaximumDistortion           ) //DcmTagKey(0x0028, 0x9445)
    GET_DICOM_INFO_LINE(DCM_ImageProcessingApplied               ) //DcmTagKey(0x0028, 0x9446)
    GET_DICOM_INFO_LINE(DCM_MaskSelectionMode                    ) //DcmTagKey(0x0028, 0x9454)
    GET_DICOM_INFO_LINE(DCM_LUTFunction                          ) //DcmTagKey(0x0028, 0x9474)
    GET_DICOM_INFO_LINE(DCM_MaskVisibilityPercentage             ) //DcmTagKey(0x0028, 0x9478)
    GET_DICOM_INFO_LINE(DCM_PixelShiftSequence                   ) //DcmTagKey(0x0028, 0x9501)
    GET_DICOM_INFO_LINE(DCM_RegionPixelShiftSequence             ) //DcmTagKey(0x0028, 0x9502)
    GET_DICOM_INFO_LINE(DCM_VerticesOfTheRegion                  ) //DcmTagKey(0x0028, 0x9503)
    GET_DICOM_INFO_LINE(DCM_MultiFramePresentationSequence       ) //DcmTagKey(0x0028, 0x9505)
    GET_DICOM_INFO_LINE(DCM_PixelShiftFrameRange                 ) //DcmTagKey(0x0028, 0x9506)
    GET_DICOM_INFO_LINE(DCM_LUTFrameRange                        ) //DcmTagKey(0x0028, 0x9507)
    GET_DICOM_INFO_LINE(DCM_ImageToEquipmentMappingMatrix        ) //DcmTagKey(0x0028, 0x9520)
    GET_DICOM_INFO_LINE(DCM_EquipmentCoordinateSystemIdentification ) //DcmTagKey(0x0028, 0x9537)
            info->append("</div>");


/*
            info->append("<div style=\"color:#00007F\">");
//    GET_DICOM_INFO_LINE(DCM_RETIRED_StudyStatusID                ) //DcmTagKey(0x0032, 0x000a)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_StudyPriorityID              ) //DcmTagKey(0x0032, 0x000c)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_StudyIDIssuer                ) //DcmTagKey(0x0032, 0x0012)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_StudyVerifiedDate            ) //DcmTagKey(0x0032, 0x0032)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_StudyVerifiedTime            ) //DcmTagKey(0x0032, 0x0033)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_StudyReadDate                ) //DcmTagKey(0x0032, 0x0034)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_StudyReadTime                ) //DcmTagKey(0x0032, 0x0035)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ScheduledStudyStartDate      ) //DcmTagKey(0x0032, 0x1000)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ScheduledStudyStartTime      ) //DcmTagKey(0x0032, 0x1001)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ScheduledStudyStopDate       ) //DcmTagKey(0x0032, 0x1010)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ScheduledStudyStopTime       ) //DcmTagKey(0x0032, 0x1011)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ScheduledStudyLocation       ) //DcmTagKey(0x0032, 0x1020)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ScheduledStudyLocationAETitle ) //DcmTagKey(0x0032, 0x1021)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ReasonForStudy               ) //DcmTagKey(0x0032, 0x1030)
    GET_DICOM_INFO_LINE(DCM_RequestingPhysicianIdentificationSequence ) //DcmTagKey(0x0032, 0x1031)
    GET_DICOM_INFO_LINE(DCM_RequestingPhysician                  ) //DcmTagKey(0x0032, 0x1032)
    GET_DICOM_INFO_LINE(DCM_RequestingService                    ) //DcmTagKey(0x0032, 0x1033)
    GET_DICOM_INFO_LINE(DCM_RequestingServiceCodeSequence        ) //DcmTagKey(0x0032, 0x1034)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_StudyArrivalDate             ) //DcmTagKey(0x0032, 0x1040)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_StudyArrivalTime             ) //DcmTagKey(0x0032, 0x1041)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_StudyCompletionDate          ) //DcmTagKey(0x0032, 0x1050)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_StudyCompletionTime          ) //DcmTagKey(0x0032, 0x1051)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_StudyComponentStatusID       ) //DcmTagKey(0x0032, 0x1055)
    GET_DICOM_INFO_LINE(DCM_RequestedProcedureDescription        ) //DcmTagKey(0x0032, 0x1060)
    GET_DICOM_INFO_LINE(DCM_RequestedProcedureCodeSequence       ) //DcmTagKey(0x0032, 0x1064)
    GET_DICOM_INFO_LINE(DCM_RequestedContrastAgent               ) //DcmTagKey(0x0032, 0x1070)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_StudyComments                ) //DcmTagKey(0x0032, 0x4000)
            info->append("</div>");
            info->append("<div style=\"color:#00007F\">");
    GET_DICOM_INFO_LINE(DCM_ReferencedPatientAliasSequence       ) //DcmTagKey(0x0038, 0x0004)
    GET_DICOM_INFO_LINE(DCM_VisitStatusID                        ) //DcmTagKey(0x0038, 0x0008)
    GET_DICOM_INFO_LINE(DCM_AdmissionID                          ) //DcmTagKey(0x0038, 0x0010)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_IssuerOfAdmissionID          ) //DcmTagKey(0x0038, 0x0011)
    GET_DICOM_INFO_LINE(DCM_IssuerOfAdmissionIDSequence          ) //DcmTagKey(0x0038, 0x0014)
    GET_DICOM_INFO_LINE(DCM_RouteOfAdmissions                    ) //DcmTagKey(0x0038, 0x0016)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ScheduledAdmissionDate       ) //DcmTagKey(0x0038, 0x001a)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ScheduledAdmissionTime       ) //DcmTagKey(0x0038, 0x001b)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ScheduledDischargeDate       ) //DcmTagKey(0x0038, 0x001c)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ScheduledDischargeTime       ) //DcmTagKey(0x0038, 0x001d)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ScheduledPatientInstitutionResidence ) //DcmTagKey(0x0038, 0x001e)
    GET_DICOM_INFO_LINE(DCM_AdmittingDate                        ) //DcmTagKey(0x0038, 0x0020)
    GET_DICOM_INFO_LINE(DCM_AdmittingTime                        ) //DcmTagKey(0x0038, 0x0021)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_DischargeDate                ) //DcmTagKey(0x0038, 0x0030)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_DischargeTime                ) //DcmTagKey(0x0038, 0x0032)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_DischargeDiagnosisDescription ) //DcmTagKey(0x0038, 0x0040)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_DischargeDiagnosisCodeSequence ) //DcmTagKey(0x0038, 0x0044)
    GET_DICOM_INFO_LINE(DCM_SpecialNeeds                         ) //DcmTagKey(0x0038, 0x0050)
    GET_DICOM_INFO_LINE(DCM_ServiceEpisodeID                     ) //DcmTagKey(0x0038, 0x0060)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_IssuerOfServiceEpisodeID     ) //DcmTagKey(0x0038, 0x0061)
    GET_DICOM_INFO_LINE(DCM_ServiceEpisodeDescription            ) //DcmTagKey(0x0038, 0x0062)
    GET_DICOM_INFO_LINE(DCM_IssuerOfServiceEpisodeIDSequence     ) //DcmTagKey(0x0038, 0x0064)
    GET_DICOM_INFO_LINE(DCM_PertinentDocumentsSequence           ) //DcmTagKey(0x0038, 0x0100)
    GET_DICOM_INFO_LINE(DCM_CurrentPatientLocation               ) //DcmTagKey(0x0038, 0x0300)
    GET_DICOM_INFO_LINE(DCM_PatientInstitutionResidence          ) //DcmTagKey(0x0038, 0x0400)
    GET_DICOM_INFO_LINE(DCM_PatientState                         ) //DcmTagKey(0x0038, 0x0500)
    GET_DICOM_INFO_LINE(DCM_PatientClinicalTrialParticipationSequence ) //DcmTagKey(0x0038, 0x0502)
    GET_DICOM_INFO_LINE(DCM_VisitComments                        ) //DcmTagKey(0x0038, 0x4000)
            info->append("</div>");
            info->append("<div style=\"color:#00007F\">");
    GET_DICOM_INFO_LINE(DCM_WaveformOriginality                  ) //DcmTagKey(0x003a, 0x0004)
    GET_DICOM_INFO_LINE(DCM_NumberOfWaveformChannels             ) //DcmTagKey(0x003a, 0x0005)
    GET_DICOM_INFO_LINE(DCM_NumberOfWaveformSamples              ) //DcmTagKey(0x003a, 0x0010)
    GET_DICOM_INFO_LINE(DCM_SamplingFrequency                    ) //DcmTagKey(0x003a, 0x001a)
    GET_DICOM_INFO_LINE(DCM_MultiplexGroupLabel                  ) //DcmTagKey(0x003a, 0x0020)
    GET_DICOM_INFO_LINE(DCM_ChannelDefinitionSequence            ) //DcmTagKey(0x003a, 0x0200)
    GET_DICOM_INFO_LINE(DCM_WaveformChannelNumber                ) //DcmTagKey(0x003a, 0x0202)
    GET_DICOM_INFO_LINE(DCM_ChannelLabel                         ) //DcmTagKey(0x003a, 0x0203)
    GET_DICOM_INFO_LINE(DCM_ChannelStatus                        ) //DcmTagKey(0x003a, 0x0205)
    GET_DICOM_INFO_LINE(DCM_ChannelSourceSequence                ) //DcmTagKey(0x003a, 0x0208)
    GET_DICOM_INFO_LINE(DCM_ChannelSourceModifiersSequence       ) //DcmTagKey(0x003a, 0x0209)
    GET_DICOM_INFO_LINE(DCM_SourceWaveformSequence               ) //DcmTagKey(0x003a, 0x020a)
    GET_DICOM_INFO_LINE(DCM_ChannelDerivationDescription         ) //DcmTagKey(0x003a, 0x020c)
    GET_DICOM_INFO_LINE(DCM_ChannelSensitivity                   ) //DcmTagKey(0x003a, 0x0210)
    GET_DICOM_INFO_LINE(DCM_ChannelSensitivityUnitsSequence      ) //DcmTagKey(0x003a, 0x0211)
    GET_DICOM_INFO_LINE(DCM_ChannelSensitivityCorrectionFactor   ) //DcmTagKey(0x003a, 0x0212)
    GET_DICOM_INFO_LINE(DCM_ChannelBaseline                      ) //DcmTagKey(0x003a, 0x0213)
    GET_DICOM_INFO_LINE(DCM_ChannelTimeSkew                      ) //DcmTagKey(0x003a, 0x0214)
    GET_DICOM_INFO_LINE(DCM_ChannelSampleSkew                    ) //DcmTagKey(0x003a, 0x0215)
    GET_DICOM_INFO_LINE(DCM_ChannelOffset                        ) //DcmTagKey(0x003a, 0x0218)
    GET_DICOM_INFO_LINE(DCM_WaveformBitsStored                   ) //DcmTagKey(0x003a, 0x021a)
    GET_DICOM_INFO_LINE(DCM_FilterLowFrequency                   ) //DcmTagKey(0x003a, 0x0220)
    GET_DICOM_INFO_LINE(DCM_FilterHighFrequency                  ) //DcmTagKey(0x003a, 0x0221)
    GET_DICOM_INFO_LINE(DCM_NotchFilterFrequency                 ) //DcmTagKey(0x003a, 0x0222)
    GET_DICOM_INFO_LINE(DCM_NotchFilterBandwidth                 ) //DcmTagKey(0x003a, 0x0223)
    GET_DICOM_INFO_LINE(DCM_WaveformDataDisplayScale             ) //DcmTagKey(0x003a, 0x0230)
    GET_DICOM_INFO_LINE(DCM_WaveformDisplayBackgroundCIELabValue ) //DcmTagKey(0x003a, 0x0231)
    GET_DICOM_INFO_LINE(DCM_WaveformPresentationGroupSequence    ) //DcmTagKey(0x003a, 0x0240)
    GET_DICOM_INFO_LINE(DCM_PresentationGroupNumber              ) //DcmTagKey(0x003a, 0x0241)
    GET_DICOM_INFO_LINE(DCM_ChannelDisplaySequence               ) //DcmTagKey(0x003a, 0x0242)
    GET_DICOM_INFO_LINE(DCM_ChannelRecommendedDisplayCIELabValue ) //DcmTagKey(0x003a, 0x0244)
    GET_DICOM_INFO_LINE(DCM_ChannelPosition                      ) //DcmTagKey(0x003a, 0x0245)
    GET_DICOM_INFO_LINE(DCM_DisplayShadingFlag                   ) //DcmTagKey(0x003a, 0x0246)
    GET_DICOM_INFO_LINE(DCM_FractionalChannelDisplayScale        ) //DcmTagKey(0x003a, 0x0247)
    GET_DICOM_INFO_LINE(DCM_AbsoluteChannelDisplayScale          ) //DcmTagKey(0x003a, 0x0248)
    GET_DICOM_INFO_LINE(DCM_MultiplexedAudioChannelsDescriptionCodeSequence ) //DcmTagKey(0x003a, 0x0300)
    GET_DICOM_INFO_LINE(DCM_ChannelIdentificationCode            ) //DcmTagKey(0x003a, 0x0301)
    GET_DICOM_INFO_LINE(DCM_ChannelMode                          ) //DcmTagKey(0x003a, 0x0302)
            info->append("</div>");
            info->append("<div style=\"color:#00007F\">");
    GET_DICOM_INFO_LINE(DCM_ScheduledStationAETitle              ) //DcmTagKey(0x0040, 0x0001)
    GET_DICOM_INFO_LINE(DCM_ScheduledProcedureStepStartDate      ) //DcmTagKey(0x0040, 0x0002)
    GET_DICOM_INFO_LINE(DCM_ScheduledProcedureStepStartTime      ) //DcmTagKey(0x0040, 0x0003)
    GET_DICOM_INFO_LINE(DCM_ScheduledProcedureStepEndDate        ) //DcmTagKey(0x0040, 0x0004)
    GET_DICOM_INFO_LINE(DCM_ScheduledProcedureStepEndTime        ) //DcmTagKey(0x0040, 0x0005)
    GET_DICOM_INFO_LINE(DCM_ScheduledPerformingPhysicianName     ) //DcmTagKey(0x0040, 0x0006)
    GET_DICOM_INFO_LINE(DCM_ScheduledProcedureStepDescription    ) //DcmTagKey(0x0040, 0x0007)
    GET_DICOM_INFO_LINE(DCM_ScheduledProtocolCodeSequence        ) //DcmTagKey(0x0040, 0x0008)
    GET_DICOM_INFO_LINE(DCM_ScheduledProcedureStepID             ) //DcmTagKey(0x0040, 0x0009)
    GET_DICOM_INFO_LINE(DCM_StageCodeSequence                    ) //DcmTagKey(0x0040, 0x000a)
    GET_DICOM_INFO_LINE(DCM_ScheduledPerformingPhysicianIdentificationSequence ) //DcmTagKey(0x0040, 0x000b)
    GET_DICOM_INFO_LINE(DCM_ScheduledStationName                 ) //DcmTagKey(0x0040, 0x0010)
    GET_DICOM_INFO_LINE(DCM_ScheduledProcedureStepLocation       ) //DcmTagKey(0x0040, 0x0011)
    GET_DICOM_INFO_LINE(DCM_PreMedication                        ) //DcmTagKey(0x0040, 0x0012)
    GET_DICOM_INFO_LINE(DCM_ScheduledProcedureStepStatus         ) //DcmTagKey(0x0040, 0x0020)
    GET_DICOM_INFO_LINE(DCM_OrderPlacerIdentifierSequence        ) //DcmTagKey(0x0040, 0x0026)
    GET_DICOM_INFO_LINE(DCM_OrderFillerIdentifierSequence        ) //DcmTagKey(0x0040, 0x0027)
    GET_DICOM_INFO_LINE(DCM_LocalNamespaceEntityID               ) //DcmTagKey(0x0040, 0x0031)
    GET_DICOM_INFO_LINE(DCM_UniversalEntityID                    ) //DcmTagKey(0x0040, 0x0032)
    GET_DICOM_INFO_LINE(DCM_UniversalEntityIDType                ) //DcmTagKey(0x0040, 0x0033)
    GET_DICOM_INFO_LINE(DCM_IdentifierTypeCode                   ) //DcmTagKey(0x0040, 0x0035)
    GET_DICOM_INFO_LINE(DCM_AssigningFacilitySequence            ) //DcmTagKey(0x0040, 0x0036)
    GET_DICOM_INFO_LINE(DCM_AssigningJurisdictionCodeSequence    ) //DcmTagKey(0x0040, 0x0039)
    GET_DICOM_INFO_LINE(DCM_AssigningAgencyOrDepartmentCodeSequence ) //DcmTagKey(0x0040, 0x003a)
    GET_DICOM_INFO_LINE(DCM_ScheduledProcedureStepSequence       ) //DcmTagKey(0x0040, 0x0100)
    GET_DICOM_INFO_LINE(DCM_ReferencedNonImageCompositeSOPInstanceSequence ) //DcmTagKey(0x0040, 0x0220)
    GET_DICOM_INFO_LINE(DCM_PerformedStationAETitle              ) //DcmTagKey(0x0040, 0x0241)
    GET_DICOM_INFO_LINE(DCM_PerformedStationName                 ) //DcmTagKey(0x0040, 0x0242)
    GET_DICOM_INFO_LINE(DCM_PerformedLocation                    ) //DcmTagKey(0x0040, 0x0243)
    GET_DICOM_INFO_LINE(DCM_PerformedProcedureStepStartDate      ) //DcmTagKey(0x0040, 0x0244)
    GET_DICOM_INFO_LINE(DCM_PerformedProcedureStepStartTime      ) //DcmTagKey(0x0040, 0x0245)
    GET_DICOM_INFO_LINE(DCM_PerformedProcedureStepEndDate        ) //DcmTagKey(0x0040, 0x0250)
    GET_DICOM_INFO_LINE(DCM_PerformedProcedureStepEndTime        ) //DcmTagKey(0x0040, 0x0251)
    GET_DICOM_INFO_LINE(DCM_PerformedProcedureStepStatus         ) //DcmTagKey(0x0040, 0x0252)
    GET_DICOM_INFO_LINE(DCM_PerformedProcedureStepID             ) //DcmTagKey(0x0040, 0x0253)
    GET_DICOM_INFO_LINE(DCM_PerformedProcedureStepDescription    ) //DcmTagKey(0x0040, 0x0254)
    GET_DICOM_INFO_LINE(DCM_PerformedProcedureTypeDescription    ) //DcmTagKey(0x0040, 0x0255)
    GET_DICOM_INFO_LINE(DCM_PerformedProtocolCodeSequence        ) //DcmTagKey(0x0040, 0x0260)
    GET_DICOM_INFO_LINE(DCM_PerformedProtocolType                ) //DcmTagKey(0x0040, 0x0261)
    GET_DICOM_INFO_LINE(DCM_ScheduledStepAttributesSequence      ) //DcmTagKey(0x0040, 0x0270)
    GET_DICOM_INFO_LINE(DCM_RequestAttributesSequence            ) //DcmTagKey(0x0040, 0x0275)
    GET_DICOM_INFO_LINE(DCM_CommentsOnThePerformedProcedureStep  ) //DcmTagKey(0x0040, 0x0280)
    GET_DICOM_INFO_LINE(DCM_PerformedProcedureStepDiscontinuationReasonCodeSequence ) //DcmTagKey(0x0040, 0x0281)
    GET_DICOM_INFO_LINE(DCM_QuantitySequence                     ) //DcmTagKey(0x0040, 0x0293)
    GET_DICOM_INFO_LINE(DCM_Quantity                             ) //DcmTagKey(0x0040, 0x0294)
    GET_DICOM_INFO_LINE(DCM_MeasuringUnitsSequence               ) //DcmTagKey(0x0040, 0x0295)
    GET_DICOM_INFO_LINE(DCM_BillingItemSequence                  ) //DcmTagKey(0x0040, 0x0296)
    GET_DICOM_INFO_LINE(DCM_TotalTimeOfFluoroscopy               ) //DcmTagKey(0x0040, 0x0300)
    GET_DICOM_INFO_LINE(DCM_TotalNumberOfExposures               ) //DcmTagKey(0x0040, 0x0301)
    GET_DICOM_INFO_LINE(DCM_EntranceDose                         ) //DcmTagKey(0x0040, 0x0302)
    GET_DICOM_INFO_LINE(DCM_ExposedArea                          ) //DcmTagKey(0x0040, 0x0303)
    GET_DICOM_INFO_LINE(DCM_DistanceSourceToEntrance             ) //DcmTagKey(0x0040, 0x0306)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_DistanceSourceToSupport      ) //DcmTagKey(0x0040, 0x0307)
    GET_DICOM_INFO_LINE(DCM_ExposureDoseSequence                 ) //DcmTagKey(0x0040, 0x030e)
    GET_DICOM_INFO_LINE(DCM_CommentsOnRadiationDose              ) //DcmTagKey(0x0040, 0x0310)
    GET_DICOM_INFO_LINE(DCM_XRayOutput                           ) //DcmTagKey(0x0040, 0x0312)
    GET_DICOM_INFO_LINE(DCM_HalfValueLayer                       ) //DcmTagKey(0x0040, 0x0314)
    GET_DICOM_INFO_LINE(DCM_OrganDose                            ) //DcmTagKey(0x0040, 0x0316)
    GET_DICOM_INFO_LINE(DCM_OrganExposed                         ) //DcmTagKey(0x0040, 0x0318)
    GET_DICOM_INFO_LINE(DCM_BillingProcedureStepSequence         ) //DcmTagKey(0x0040, 0x0320)
    GET_DICOM_INFO_LINE(DCM_FilmConsumptionSequence              ) //DcmTagKey(0x0040, 0x0321)
    GET_DICOM_INFO_LINE(DCM_BillingSuppliesAndDevicesSequence    ) //DcmTagKey(0x0040, 0x0324)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ReferencedProcedureStepSequence ) //DcmTagKey(0x0040, 0x0330)
    GET_DICOM_INFO_LINE(DCM_PerformedSeriesSequence              ) //DcmTagKey(0x0040, 0x0340)
    GET_DICOM_INFO_LINE(DCM_CommentsOnTheScheduledProcedureStep  ) //DcmTagKey(0x0040, 0x0400)
    GET_DICOM_INFO_LINE(DCM_ProtocolContextSequence              ) //DcmTagKey(0x0040, 0x0440)
    GET_DICOM_INFO_LINE(DCM_ContentItemModifierSequence          ) //DcmTagKey(0x0040, 0x0441)
    GET_DICOM_INFO_LINE(DCM_ScheduledSpecimenSequence            ) //DcmTagKey(0x0040, 0x0500)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_SpecimenAccessionNumber      ) //DcmTagKey(0x0040, 0x050a)
    GET_DICOM_INFO_LINE(DCM_ContainerIdentifier                  ) //DcmTagKey(0x0040, 0x0512)
    GET_DICOM_INFO_LINE(DCM_IssuerOfTheContainerIdentifierSequence ) //DcmTagKey(0x0040, 0x0513)
    GET_DICOM_INFO_LINE(DCM_AlternateContainerIdentifierSequence ) //DcmTagKey(0x0040, 0x0515)
    GET_DICOM_INFO_LINE(DCM_ContainerTypeCodeSequence            ) //DcmTagKey(0x0040, 0x0518)
    GET_DICOM_INFO_LINE(DCM_ContainerDescription                 ) //DcmTagKey(0x0040, 0x051a)
    GET_DICOM_INFO_LINE(DCM_ContainerComponentSequence           ) //DcmTagKey(0x0040, 0x0520)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_SpecimenSequence             ) //DcmTagKey(0x0040, 0x0550)
    GET_DICOM_INFO_LINE(DCM_SpecimenIdentifier                   ) //DcmTagKey(0x0040, 0x0551)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_SpecimenDescriptionSequenceTrial ) //DcmTagKey(0x0040, 0x0552)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_SpecimenDescriptionTrial     ) //DcmTagKey(0x0040, 0x0553)
    GET_DICOM_INFO_LINE(DCM_SpecimenUID                          ) //DcmTagKey(0x0040, 0x0554)
    GET_DICOM_INFO_LINE(DCM_AcquisitionContextSequence           ) //DcmTagKey(0x0040, 0x0555)
    GET_DICOM_INFO_LINE(DCM_AcquisitionContextDescription        ) //DcmTagKey(0x0040, 0x0556)
    GET_DICOM_INFO_LINE(DCM_SpecimenDescriptionSequence          ) //DcmTagKey(0x0040, 0x0560)
    GET_DICOM_INFO_LINE(DCM_IssuerOfTheSpecimenIdentifierSequence ) //DcmTagKey(0x0040, 0x0562)
    GET_DICOM_INFO_LINE(DCM_SpecimenTypeCodeSequence             ) //DcmTagKey(0x0040, 0x059a)
    GET_DICOM_INFO_LINE(DCM_SpecimenShortDescription             ) //DcmTagKey(0x0040, 0x0600)
    GET_DICOM_INFO_LINE(DCM_SpecimenDetailedDescription          ) //DcmTagKey(0x0040, 0x0602)
    GET_DICOM_INFO_LINE(DCM_SpecimenPreparationSequence          ) //DcmTagKey(0x0040, 0x0610)
    GET_DICOM_INFO_LINE(DCM_SpecimenPreparationStepContentItemSequence ) //DcmTagKey(0x0040, 0x0612)
    GET_DICOM_INFO_LINE(DCM_SpecimenLocalizationContentItemSequence ) //DcmTagKey(0x0040, 0x0620)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_SlideIdentifier              ) //DcmTagKey(0x0040, 0x06fa)
    GET_DICOM_INFO_LINE(DCM_ImageCenterPointCoordinatesSequence  ) //DcmTagKey(0x0040, 0x071a)
    GET_DICOM_INFO_LINE(DCM_XOffsetInSlideCoordinateSystem       ) //DcmTagKey(0x0040, 0x072a)
    GET_DICOM_INFO_LINE(DCM_YOffsetInSlideCoordinateSystem       ) //DcmTagKey(0x0040, 0x073a)
    GET_DICOM_INFO_LINE(DCM_ZOffsetInSlideCoordinateSystem       ) //DcmTagKey(0x0040, 0x074a)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_PixelSpacingSequence         ) //DcmTagKey(0x0040, 0x08d8)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_CoordinateSystemAxisCodeSequence ) //DcmTagKey(0x0040, 0x08da)
    GET_DICOM_INFO_LINE(DCM_MeasurementUnitsCodeSequence         ) //DcmTagKey(0x0040, 0x08ea)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_VitalStainCodeSequenceTrial  ) //DcmTagKey(0x0040, 0x09f8)
    GET_DICOM_INFO_LINE(DCM_RequestedProcedureID                 ) //DcmTagKey(0x0040, 0x1001)
    GET_DICOM_INFO_LINE(DCM_ReasonForTheRequestedProcedure       ) //DcmTagKey(0x0040, 0x1002)
    GET_DICOM_INFO_LINE(DCM_RequestedProcedurePriority           ) //DcmTagKey(0x0040, 0x1003)
    GET_DICOM_INFO_LINE(DCM_PatientTransportArrangements         ) //DcmTagKey(0x0040, 0x1004)
    GET_DICOM_INFO_LINE(DCM_RequestedProcedureLocation           ) //DcmTagKey(0x0040, 0x1005)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_PlacerOrderNumberProcedure   ) //DcmTagKey(0x0040, 0x1006)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_FillerOrderNumberProcedure   ) //DcmTagKey(0x0040, 0x1007)
    GET_DICOM_INFO_LINE(DCM_ConfidentialityCode                  ) //DcmTagKey(0x0040, 0x1008)
    GET_DICOM_INFO_LINE(DCM_ReportingPriority                    ) //DcmTagKey(0x0040, 0x1009)
    GET_DICOM_INFO_LINE(DCM_ReasonForRequestedProcedureCodeSequence ) //DcmTagKey(0x0040, 0x100a)
    GET_DICOM_INFO_LINE(DCM_NamesOfIntendedRecipientsOfResults   ) //DcmTagKey(0x0040, 0x1010)
    GET_DICOM_INFO_LINE(DCM_IntendedRecipientsOfResultsIdentificationSequence ) //DcmTagKey(0x0040, 0x1011)
    GET_DICOM_INFO_LINE(DCM_ReasonForPerformedProcedureCodeSequence ) //DcmTagKey(0x0040, 0x1012)
    GET_DICOM_INFO_LINE(DCM_PersonIdentificationCodeSequence     ) //DcmTagKey(0x0040, 0x1101)
    GET_DICOM_INFO_LINE(DCM_PersonAddress                        ) //DcmTagKey(0x0040, 0x1102)
    GET_DICOM_INFO_LINE(DCM_PersonTelephoneNumbers               ) //DcmTagKey(0x0040, 0x1103)
    GET_DICOM_INFO_LINE(DCM_RequestedProcedureComments           ) //DcmTagKey(0x0040, 0x1400)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ReasonForTheImagingServiceRequest ) //DcmTagKey(0x0040, 0x2001)
    GET_DICOM_INFO_LINE(DCM_IssueDateOfImagingServiceRequest     ) //DcmTagKey(0x0040, 0x2004)
    GET_DICOM_INFO_LINE(DCM_IssueTimeOfImagingServiceRequest     ) //DcmTagKey(0x0040, 0x2005)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_PlacerOrderNumberImagingServiceRequestRetired ) //DcmTagKey(0x0040, 0x2006)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_FillerOrderNumberImagingServiceRequestRetired ) //DcmTagKey(0x0040, 0x2007)
    GET_DICOM_INFO_LINE(DCM_OrderEnteredBy                       ) //DcmTagKey(0x0040, 0x2008)
    GET_DICOM_INFO_LINE(DCM_OrderEntererLocation                 ) //DcmTagKey(0x0040, 0x2009)
    GET_DICOM_INFO_LINE(DCM_OrderCallbackPhoneNumber             ) //DcmTagKey(0x0040, 0x2010)
    GET_DICOM_INFO_LINE(DCM_PlacerOrderNumberImagingServiceRequest ) //DcmTagKey(0x0040, 0x2016)
    GET_DICOM_INFO_LINE(DCM_FillerOrderNumberImagingServiceRequest ) //DcmTagKey(0x0040, 0x2017)
    GET_DICOM_INFO_LINE(DCM_ImagingServiceRequestComments        ) //DcmTagKey(0x0040, 0x2400)
    GET_DICOM_INFO_LINE(DCM_ConfidentialityConstraintOnPatientDataDescription ) //DcmTagKey(0x0040, 0x3001)
    GET_DICOM_INFO_LINE(DCM_GeneralPurposeScheduledProcedureStepStatus ) //DcmTagKey(0x0040, 0x4001)
    GET_DICOM_INFO_LINE(DCM_GeneralPurposePerformedProcedureStepStatus ) //DcmTagKey(0x0040, 0x4002)
    GET_DICOM_INFO_LINE(DCM_GeneralPurposeScheduledProcedureStepPriority ) //DcmTagKey(0x0040, 0x4003)
    GET_DICOM_INFO_LINE(DCM_ScheduledProcessingApplicationsCodeSequence ) //DcmTagKey(0x0040, 0x4004)
    GET_DICOM_INFO_LINE(DCM_ScheduledProcedureStepStartDateTime  ) //DcmTagKey(0x0040, 0x4005)
    GET_DICOM_INFO_LINE(DCM_MultipleCopiesFlag                   ) //DcmTagKey(0x0040, 0x4006)
    GET_DICOM_INFO_LINE(DCM_PerformedProcessingApplicationsCodeSequence ) //DcmTagKey(0x0040, 0x4007)
    GET_DICOM_INFO_LINE(DCM_HumanPerformerCodeSequence           ) //DcmTagKey(0x0040, 0x4009)
    GET_DICOM_INFO_LINE(DCM_ScheduledProcedureStepModificationDateTime ) //DcmTagKey(0x0040, 0x4010)
    GET_DICOM_INFO_LINE(DCM_ExpectedCompletionDateTime           ) //DcmTagKey(0x0040, 0x4011)
    GET_DICOM_INFO_LINE(DCM_ResultingGeneralPurposePerformedProcedureStepsSequence ) //DcmTagKey(0x0040, 0x4015)
    GET_DICOM_INFO_LINE(DCM_ReferencedGeneralPurposeScheduledProcedureStepSequence ) //DcmTagKey(0x0040, 0x4016)
    GET_DICOM_INFO_LINE(DCM_ScheduledWorkitemCodeSequence        ) //DcmTagKey(0x0040, 0x4018)
    GET_DICOM_INFO_LINE(DCM_PerformedWorkitemCodeSequence        ) //DcmTagKey(0x0040, 0x4019)
    GET_DICOM_INFO_LINE(DCM_InputAvailabilityFlag                ) //DcmTagKey(0x0040, 0x4020)
    GET_DICOM_INFO_LINE(DCM_InputInformationSequence             ) //DcmTagKey(0x0040, 0x4021)
    GET_DICOM_INFO_LINE(DCM_RelevantInformationSequence          ) //DcmTagKey(0x0040, 0x4022)
    GET_DICOM_INFO_LINE(DCM_ReferencedGeneralPurposeScheduledProcedureStepTransactionUID ) //DcmTagKey(0x0040, 0x4023)
    GET_DICOM_INFO_LINE(DCM_ScheduledStationNameCodeSequence     ) //DcmTagKey(0x0040, 0x4025)
    GET_DICOM_INFO_LINE(DCM_ScheduledStationClassCodeSequence    ) //DcmTagKey(0x0040, 0x4026)
    GET_DICOM_INFO_LINE(DCM_ScheduledStationGeographicLocationCodeSequence ) //DcmTagKey(0x0040, 0x4027)
    GET_DICOM_INFO_LINE(DCM_PerformedStationNameCodeSequence     ) //DcmTagKey(0x0040, 0x4028)
    GET_DICOM_INFO_LINE(DCM_PerformedStationClassCodeSequence    ) //DcmTagKey(0x0040, 0x4029)
    GET_DICOM_INFO_LINE(DCM_PerformedStationGeographicLocationCodeSequence ) //DcmTagKey(0x0040, 0x4030)
    GET_DICOM_INFO_LINE(DCM_RequestedSubsequentWorkitemCodeSequence ) //DcmTagKey(0x0040, 0x4031)
    GET_DICOM_INFO_LINE(DCM_NonDICOMOutputCodeSequence           ) //DcmTagKey(0x0040, 0x4032)
    GET_DICOM_INFO_LINE(DCM_OutputInformationSequence            ) //DcmTagKey(0x0040, 0x4033)
    GET_DICOM_INFO_LINE(DCM_ScheduledHumanPerformersSequence     ) //DcmTagKey(0x0040, 0x4034)
    GET_DICOM_INFO_LINE(DCM_ActualHumanPerformersSequence        ) //DcmTagKey(0x0040, 0x4035)
    GET_DICOM_INFO_LINE(DCM_HumanPerformerOrganization           ) //DcmTagKey(0x0040, 0x4036)
    GET_DICOM_INFO_LINE(DCM_HumanPerformerName                   ) //DcmTagKey(0x0040, 0x4037)
    GET_DICOM_INFO_LINE(DCM_RawDataHandling                      ) //DcmTagKey(0x0040, 0x4040)
    GET_DICOM_INFO_LINE(DCM_EntranceDoseInmGy                    ) //DcmTagKey(0x0040, 0x8302)
    GET_DICOM_INFO_LINE(DCM_ReferencedImageRealWorldValueMappingSequence ) //DcmTagKey(0x0040, 0x9094)
    GET_DICOM_INFO_LINE(DCM_RealWorldValueMappingSequence        ) //DcmTagKey(0x0040, 0x9096)
    GET_DICOM_INFO_LINE(DCM_PixelValueMappingCodeSequence        ) //DcmTagKey(0x0040, 0x9098)
    GET_DICOM_INFO_LINE(DCM_LUTLabel                             ) //DcmTagKey(0x0040, 0x9210)
    GET_DICOM_INFO_LINE(DCM_RealWorldValueLastValueMapped        ) //DcmTagKey(0x0040, 0x9211)
    GET_DICOM_INFO_LINE(DCM_RealWorldValueLUTData                ) //DcmTagKey(0x0040, 0x9212)
    GET_DICOM_INFO_LINE(DCM_RealWorldValueFirstValueMapped       ) //DcmTagKey(0x0040, 0x9216)
    GET_DICOM_INFO_LINE(DCM_RealWorldValueIntercept              ) //DcmTagKey(0x0040, 0x9224)
    GET_DICOM_INFO_LINE(DCM_RealWorldValueSlope                  ) //DcmTagKey(0x0040, 0x9225)
    GET_DICOM_INFO_LINE(DCM_RelationshipType                     ) //DcmTagKey(0x0040, 0xa010)
    GET_DICOM_INFO_LINE(DCM_VerifyingOrganization                ) //DcmTagKey(0x0040, 0xa027)
    GET_DICOM_INFO_LINE(DCM_VerificationDateTime                 ) //DcmTagKey(0x0040, 0xa030)
    GET_DICOM_INFO_LINE(DCM_ObservationDateTime                  ) //DcmTagKey(0x0040, 0xa032)
    GET_DICOM_INFO_LINE(DCM_ValueType                            ) //DcmTagKey(0x0040, 0xa040)
    GET_DICOM_INFO_LINE(DCM_ConceptNameCodeSequence              ) //DcmTagKey(0x0040, 0xa043)
    GET_DICOM_INFO_LINE(DCM_ContinuityOfContent                  ) //DcmTagKey(0x0040, 0xa050)
    GET_DICOM_INFO_LINE(DCM_VerifyingObserverSequence            ) //DcmTagKey(0x0040, 0xa073)
    GET_DICOM_INFO_LINE(DCM_VerifyingObserverName                ) //DcmTagKey(0x0040, 0xa075)
    GET_DICOM_INFO_LINE(DCM_AuthorObserverSequence               ) //DcmTagKey(0x0040, 0xa078)
    GET_DICOM_INFO_LINE(DCM_ParticipantSequence                  ) //DcmTagKey(0x0040, 0xa07a)
    GET_DICOM_INFO_LINE(DCM_CustodialOrganizationSequence        ) //DcmTagKey(0x0040, 0xa07c)
    GET_DICOM_INFO_LINE(DCM_ParticipationType                    ) //DcmTagKey(0x0040, 0xa080)
    GET_DICOM_INFO_LINE(DCM_ParticipationDateTime                ) //DcmTagKey(0x0040, 0xa082)
    GET_DICOM_INFO_LINE(DCM_ObserverType                         ) //DcmTagKey(0x0040, 0xa084)
    GET_DICOM_INFO_LINE(DCM_VerifyingObserverIdentificationCodeSequence ) //DcmTagKey(0x0040, 0xa088)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_EquivalentCDADocumentSequence ) //DcmTagKey(0x0040, 0xa090)
    GET_DICOM_INFO_LINE(DCM_ReferencedWaveformChannels           ) //DcmTagKey(0x0040, 0xa0b0)
    GET_DICOM_INFO_LINE(DCM_DateTime                             ) //DcmTagKey(0x0040, 0xa120)
    GET_DICOM_INFO_LINE(DCM_Date                                 ) //DcmTagKey(0x0040, 0xa121)
    GET_DICOM_INFO_LINE(DCM_Time                                 ) //DcmTagKey(0x0040, 0xa122)
    GET_DICOM_INFO_LINE(DCM_PersonName                           ) //DcmTagKey(0x0040, 0xa123)
    GET_DICOM_INFO_LINE(DCM_UID                                  ) //DcmTagKey(0x0040, 0xa124)
    GET_DICOM_INFO_LINE(DCM_TemporalRangeType                    ) //DcmTagKey(0x0040, 0xa130)
    GET_DICOM_INFO_LINE(DCM_ReferencedSamplePositions            ) //DcmTagKey(0x0040, 0xa132)
    GET_DICOM_INFO_LINE(DCM_ReferencedFrameNumbers               ) //DcmTagKey(0x0040, 0xa136)
    GET_DICOM_INFO_LINE(DCM_ReferencedTimeOffsets                ) //DcmTagKey(0x0040, 0xa138)
    GET_DICOM_INFO_LINE(DCM_ReferencedDateTime                   ) //DcmTagKey(0x0040, 0xa13a)
    GET_DICOM_INFO_LINE(DCM_TextValue                            ) //DcmTagKey(0x0040, 0xa160)
    GET_DICOM_INFO_LINE(DCM_ConceptCodeSequence                  ) //DcmTagKey(0x0040, 0xa168)
    GET_DICOM_INFO_LINE(DCM_PurposeOfReferenceCodeSequence       ) //DcmTagKey(0x0040, 0xa170)
    GET_DICOM_INFO_LINE(DCM_AnnotationGroupNumber                ) //DcmTagKey(0x0040, 0xa180)
    GET_DICOM_INFO_LINE(DCM_ModifierCodeSequence                 ) //DcmTagKey(0x0040, 0xa195)
    GET_DICOM_INFO_LINE(DCM_MeasuredValueSequence                ) //DcmTagKey(0x0040, 0xa300)
    GET_DICOM_INFO_LINE(DCM_NumericValueQualifierCodeSequence    ) //DcmTagKey(0x0040, 0xa301)
    GET_DICOM_INFO_LINE(DCM_NumericValue                         ) //DcmTagKey(0x0040, 0xa30a)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_AddressTrial                 ) //DcmTagKey(0x0040, 0xa353)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_TelephoneNumberTrial         ) //DcmTagKey(0x0040, 0xa354)
    GET_DICOM_INFO_LINE(DCM_PredecessorDocumentsSequence         ) //DcmTagKey(0x0040, 0xa360)
    GET_DICOM_INFO_LINE(DCM_ReferencedRequestSequence            ) //DcmTagKey(0x0040, 0xa370)
    GET_DICOM_INFO_LINE(DCM_PerformedProcedureCodeSequence       ) //DcmTagKey(0x0040, 0xa372)
    GET_DICOM_INFO_LINE(DCM_CurrentRequestedProcedureEvidenceSequence ) //DcmTagKey(0x0040, 0xa375)
    GET_DICOM_INFO_LINE(DCM_PertinentOtherEvidenceSequence       ) //DcmTagKey(0x0040, 0xa385)
    GET_DICOM_INFO_LINE(DCM_HL7StructuredDocumentReferenceSequence ) //DcmTagKey(0x0040, 0xa390)
    GET_DICOM_INFO_LINE(DCM_CompletionFlag                       ) //DcmTagKey(0x0040, 0xa491)
    GET_DICOM_INFO_LINE(DCM_CompletionFlagDescription            ) //DcmTagKey(0x0040, 0xa492)
    GET_DICOM_INFO_LINE(DCM_VerificationFlag                     ) //DcmTagKey(0x0040, 0xa493)
    GET_DICOM_INFO_LINE(DCM_ArchiveRequested                     ) //DcmTagKey(0x0040, 0xa494)
    GET_DICOM_INFO_LINE(DCM_PreliminaryFlag                      ) //DcmTagKey(0x0040, 0xa496)
    GET_DICOM_INFO_LINE(DCM_ContentTemplateSequence              ) //DcmTagKey(0x0040, 0xa504)
    GET_DICOM_INFO_LINE(DCM_IdenticalDocumentsSequence           ) //DcmTagKey(0x0040, 0xa525)
    GET_DICOM_INFO_LINE(DCM_ContentSequence                      ) //DcmTagKey(0x0040, 0xa730)
    GET_DICOM_INFO_LINE(DCM_WaveformAnnotationSequence           ) //DcmTagKey(0x0040, 0xb020)
    GET_DICOM_INFO_LINE(DCM_TemplateIdentifier                   ) //DcmTagKey(0x0040, 0xdb00)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_TemplateVersion              ) //DcmTagKey(0x0040, 0xdb06)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_TemplateLocalVersion         ) //DcmTagKey(0x0040, 0xdb07)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_TemplateExtensionFlag        ) //DcmTagKey(0x0040, 0xdb0b)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_TemplateExtensionOrganizationUID ) //DcmTagKey(0x0040, 0xdb0c)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_TemplateExtensionCreatorUID  ) //DcmTagKey(0x0040, 0xdb0d)
    GET_DICOM_INFO_LINE(DCM_ReferencedContentItemIdentifier      ) //DcmTagKey(0x0040, 0xdb73)
    GET_DICOM_INFO_LINE(DCM_HL7InstanceIdentifier                ) //DcmTagKey(0x0040, 0xe001)
    GET_DICOM_INFO_LINE(DCM_HL7DocumentEffectiveTime             ) //DcmTagKey(0x0040, 0xe004)
    GET_DICOM_INFO_LINE(DCM_HL7DocumentTypeCodeSequence          ) //DcmTagKey(0x0040, 0xe006)
    GET_DICOM_INFO_LINE(DCM_RetrieveURI                          ) //DcmTagKey(0x0040, 0xe010)
    GET_DICOM_INFO_LINE(DCM_RetrieveLocationUID                  ) //DcmTagKey(0x0040, 0xe011)
            info->append("</div>");
            info->append("<div style=\"color:#00007F\">");
    GET_DICOM_INFO_LINE(DCM_DocumentTitle                        ) //DcmTagKey(0x0042, 0x0010)
    GET_DICOM_INFO_LINE(DCM_EncapsulatedDocument                 ) //DcmTagKey(0x0042, 0x0011)
    GET_DICOM_INFO_LINE(DCM_MIMETypeOfEncapsulatedDocument       ) //DcmTagKey(0x0042, 0x0012)
    GET_DICOM_INFO_LINE(DCM_SourceInstanceSequence               ) //DcmTagKey(0x0042, 0x0013)
    GET_DICOM_INFO_LINE(DCM_ListOfMIMETypes                      ) //DcmTagKey(0x0042, 0x0014)
            info->append("</div>");
            info->append("<div style=\"color:#00007F\">");
    GET_DICOM_INFO_LINE(DCM_ProductPackageIdentifier             ) //DcmTagKey(0x0044, 0x0001)
    GET_DICOM_INFO_LINE(DCM_SubstanceAdministrationApproval      ) //DcmTagKey(0x0044, 0x0002)
    GET_DICOM_INFO_LINE(DCM_ApprovalStatusFurtherDescription     ) //DcmTagKey(0x0044, 0x0003)
    GET_DICOM_INFO_LINE(DCM_ApprovalStatusDateTime               ) //DcmTagKey(0x0044, 0x0004)
    GET_DICOM_INFO_LINE(DCM_ProductTypeCodeSequence              ) //DcmTagKey(0x0044, 0x0007)
    GET_DICOM_INFO_LINE(DCM_ProductName                          ) //DcmTagKey(0x0044, 0x0008)
    GET_DICOM_INFO_LINE(DCM_ProductDescription                   ) //DcmTagKey(0x0044, 0x0009)
    GET_DICOM_INFO_LINE(DCM_ProductLotIdentifier                 ) //DcmTagKey(0x0044, 0x000a)
    GET_DICOM_INFO_LINE(DCM_ProductExpirationDateTime            ) //DcmTagKey(0x0044, 0x000b)
    GET_DICOM_INFO_LINE(DCM_SubstanceAdministrationDateTime      ) //DcmTagKey(0x0044, 0x0010)
    GET_DICOM_INFO_LINE(DCM_SubstanceAdministrationNotes         ) //DcmTagKey(0x0044, 0x0011)
    GET_DICOM_INFO_LINE(DCM_SubstanceAdministrationDeviceID      ) //DcmTagKey(0x0044, 0x0012)
    GET_DICOM_INFO_LINE(DCM_ProductParameterSequence             ) //DcmTagKey(0x0044, 0x0013)
    GET_DICOM_INFO_LINE(DCM_SubstanceAdministrationParameterSequence ) //DcmTagKey(0x0044, 0x0019)
            info->append("</div>");
            info->append("<div style=\"color:#00007F\">");
    GET_DICOM_INFO_LINE(DCM_LensDescription                      ) //DcmTagKey(0x0046, 0x0012)
    GET_DICOM_INFO_LINE(DCM_RightLensSequence                    ) //DcmTagKey(0x0046, 0x0014)
    GET_DICOM_INFO_LINE(DCM_LeftLensSequence                     ) //DcmTagKey(0x0046, 0x0015)
    GET_DICOM_INFO_LINE(DCM_UnspecifiedLateralityLensSequence    ) //DcmTagKey(0x0046, 0x0016)
    GET_DICOM_INFO_LINE(DCM_CylinderSequence                     ) //DcmTagKey(0x0046, 0x0018)
    GET_DICOM_INFO_LINE(DCM_PrismSequence                        ) //DcmTagKey(0x0046, 0x0028)
    GET_DICOM_INFO_LINE(DCM_HorizontalPrismPower                 ) //DcmTagKey(0x0046, 0x0030)
    GET_DICOM_INFO_LINE(DCM_HorizontalPrismBase                  ) //DcmTagKey(0x0046, 0x0032)
    GET_DICOM_INFO_LINE(DCM_VerticalPrismPower                   ) //DcmTagKey(0x0046, 0x0034)
    GET_DICOM_INFO_LINE(DCM_VerticalPrismBase                    ) //DcmTagKey(0x0046, 0x0036)
    GET_DICOM_INFO_LINE(DCM_LensSegmentType                      ) //DcmTagKey(0x0046, 0x0038)
    GET_DICOM_INFO_LINE(DCM_OpticalTransmittance                 ) //DcmTagKey(0x0046, 0x0040)
    GET_DICOM_INFO_LINE(DCM_ChannelWidth                         ) //DcmTagKey(0x0046, 0x0042)
    GET_DICOM_INFO_LINE(DCM_PupilSize                            ) //DcmTagKey(0x0046, 0x0044)
    GET_DICOM_INFO_LINE(DCM_CornealSize                          ) //DcmTagKey(0x0046, 0x0046)
    GET_DICOM_INFO_LINE(DCM_AutorefractionRightEyeSequence       ) //DcmTagKey(0x0046, 0x0050)
    GET_DICOM_INFO_LINE(DCM_AutorefractionLeftEyeSequence        ) //DcmTagKey(0x0046, 0x0052)
    GET_DICOM_INFO_LINE(DCM_DistancePupillaryDistance            ) //DcmTagKey(0x0046, 0x0060)
    GET_DICOM_INFO_LINE(DCM_NearPupillaryDistance                ) //DcmTagKey(0x0046, 0x0062)
    GET_DICOM_INFO_LINE(DCM_IntermediatePupillaryDistance        ) //DcmTagKey(0x0046, 0x0063)
    GET_DICOM_INFO_LINE(DCM_OtherPupillaryDistance               ) //DcmTagKey(0x0046, 0x0064)
    GET_DICOM_INFO_LINE(DCM_KeratometryRightEyeSequence          ) //DcmTagKey(0x0046, 0x0070)
    GET_DICOM_INFO_LINE(DCM_KeratometryLeftEyeSequence           ) //DcmTagKey(0x0046, 0x0071)
    GET_DICOM_INFO_LINE(DCM_SteepKeratometricAxisSequence        ) //DcmTagKey(0x0046, 0x0074)
    GET_DICOM_INFO_LINE(DCM_RadiusOfCurvature                    ) //DcmTagKey(0x0046, 0x0075)
    GET_DICOM_INFO_LINE(DCM_KeratometricPower                    ) //DcmTagKey(0x0046, 0x0076)
    GET_DICOM_INFO_LINE(DCM_KeratometricAxis                     ) //DcmTagKey(0x0046, 0x0077)
    GET_DICOM_INFO_LINE(DCM_FlatKeratometricAxisSequence         ) //DcmTagKey(0x0046, 0x0080)
    GET_DICOM_INFO_LINE(DCM_BackgroundColor                      ) //DcmTagKey(0x0046, 0x0092)
    GET_DICOM_INFO_LINE(DCM_Optotype                             ) //DcmTagKey(0x0046, 0x0094)
    GET_DICOM_INFO_LINE(DCM_OptotypePresentation                 ) //DcmTagKey(0x0046, 0x0095)
    GET_DICOM_INFO_LINE(DCM_SubjectiveRefractionRightEyeSequence ) //DcmTagKey(0x0046, 0x0097)
    GET_DICOM_INFO_LINE(DCM_SubjectiveRefractionLeftEyeSequence  ) //DcmTagKey(0x0046, 0x0098)
    GET_DICOM_INFO_LINE(DCM_AddNearSequence                      ) //DcmTagKey(0x0046, 0x0100)
    GET_DICOM_INFO_LINE(DCM_AddIntermediateSequence              ) //DcmTagKey(0x0046, 0x0101)
    GET_DICOM_INFO_LINE(DCM_AddOtherSequence                     ) //DcmTagKey(0x0046, 0x0102)
    GET_DICOM_INFO_LINE(DCM_AddPower                             ) //DcmTagKey(0x0046, 0x0104)
    GET_DICOM_INFO_LINE(DCM_ViewingDistance                      ) //DcmTagKey(0x0046, 0x0106)
    GET_DICOM_INFO_LINE(DCM_VisualAcuityTypeCodeSequence         ) //DcmTagKey(0x0046, 0x0121)
    GET_DICOM_INFO_LINE(DCM_VisualAcuityRightEyeSequence         ) //DcmTagKey(0x0046, 0x0122)
    GET_DICOM_INFO_LINE(DCM_VisualAcuityLeftEyeSequence          ) //DcmTagKey(0x0046, 0x0123)
    GET_DICOM_INFO_LINE(DCM_VisualAcuityBothEyesOpenSequence     ) //DcmTagKey(0x0046, 0x0124)
    GET_DICOM_INFO_LINE(DCM_ViewingDistanceType                  ) //DcmTagKey(0x0046, 0x0125)
    GET_DICOM_INFO_LINE(DCM_VisualAcuityModifiers                ) //DcmTagKey(0x0046, 0x0135)
    GET_DICOM_INFO_LINE(DCM_DecimalVisualAcuity                  ) //DcmTagKey(0x0046, 0x0137)
    GET_DICOM_INFO_LINE(DCM_OptotypeDetailedDefinition           ) //DcmTagKey(0x0046, 0x0139)
    GET_DICOM_INFO_LINE(DCM_ReferencedRefractiveMeasurementsSequence ) //DcmTagKey(0x0046, 0x0145)
    GET_DICOM_INFO_LINE(DCM_SpherePower                          ) //DcmTagKey(0x0046, 0x0146)
    GET_DICOM_INFO_LINE(DCM_CylinderPower                        ) //DcmTagKey(0x0046, 0x0147)
            info->append("</div>");
            info->append("<div style=\"color:#00007F\">");
    GET_DICOM_INFO_LINE(DCM_ImagedVolumeWidth                    ) //DcmTagKey(0x0048, 0x0001)
    GET_DICOM_INFO_LINE(DCM_ImagedVolumeHeight                   ) //DcmTagKey(0x0048, 0x0002)
    GET_DICOM_INFO_LINE(DCM_ImagedVolumeDepth                    ) //DcmTagKey(0x0048, 0x0003)
    GET_DICOM_INFO_LINE(DCM_TotalPixelMatrixColumns              ) //DcmTagKey(0x0048, 0x0006)
    GET_DICOM_INFO_LINE(DCM_TotalPixelMatrixRows                 ) //DcmTagKey(0x0048, 0x0007)
    GET_DICOM_INFO_LINE(DCM_TotalPixelMatrixOriginSequence       ) //DcmTagKey(0x0048, 0x0008)
    GET_DICOM_INFO_LINE(DCM_SpecimenLabelInImage                 ) //DcmTagKey(0x0048, 0x0010)
    GET_DICOM_INFO_LINE(DCM_FocusMethod                          ) //DcmTagKey(0x0048, 0x0011)
    GET_DICOM_INFO_LINE(DCM_ExtendedDepthOfField                 ) //DcmTagKey(0x0048, 0x0012)
    GET_DICOM_INFO_LINE(DCM_NumberOfFocalPlanes                  ) //DcmTagKey(0x0048, 0x0013)
    GET_DICOM_INFO_LINE(DCM_DistanceBetweenFocalPlanes           ) //DcmTagKey(0x0048, 0x0014)
    GET_DICOM_INFO_LINE(DCM_RecommendedAbsentPixelCIELabValue    ) //DcmTagKey(0x0048, 0x0015)
    GET_DICOM_INFO_LINE(DCM_IlluminatorTypeCodeSequence          ) //DcmTagKey(0x0048, 0x0100)
    GET_DICOM_INFO_LINE(DCM_ImageOrientationSlide                ) //DcmTagKey(0x0048, 0x0102)
    GET_DICOM_INFO_LINE(DCM_OpticalPathSequence                  ) //DcmTagKey(0x0048, 0x0105)
    GET_DICOM_INFO_LINE(DCM_OpticalPathIdentifier                ) //DcmTagKey(0x0048, 0x0106)
    GET_DICOM_INFO_LINE(DCM_OpticalPathDescription               ) //DcmTagKey(0x0048, 0x0107)
    GET_DICOM_INFO_LINE(DCM_IlluminationColorCodeSequence        ) //DcmTagKey(0x0048, 0x0108)
    GET_DICOM_INFO_LINE(DCM_SpecimenReferenceSequence            ) //DcmTagKey(0x0048, 0x0110)
    GET_DICOM_INFO_LINE(DCM_CondenserLensPower                   ) //DcmTagKey(0x0048, 0x0111)
    GET_DICOM_INFO_LINE(DCM_ObjectiveLensPower                   ) //DcmTagKey(0x0048, 0x0112)
    GET_DICOM_INFO_LINE(DCM_ObjectiveLensNumericalAperture       ) //DcmTagKey(0x0048, 0x0113)
    GET_DICOM_INFO_LINE(DCM_PaletteColorLookupTableSequence      ) //DcmTagKey(0x0048, 0x0120)
    GET_DICOM_INFO_LINE(DCM_ReferencedImageNavigationSequence    ) //DcmTagKey(0x0048, 0x0200)
    GET_DICOM_INFO_LINE(DCM_TopLeftHandCornerOfLocalizerArea     ) //DcmTagKey(0x0048, 0x0201)
    GET_DICOM_INFO_LINE(DCM_BottomRightHandCornerOfLocalizerArea ) //DcmTagKey(0x0048, 0x0202)
    GET_DICOM_INFO_LINE(DCM_OpticalPathIdentificationSequence    ) //DcmTagKey(0x0048, 0x0207)
    GET_DICOM_INFO_LINE(DCM_PlanePositionSlideSequence           ) //DcmTagKey(0x0048, 0x021a)
    GET_DICOM_INFO_LINE(DCM_ColumnPositionInTotalImagePixelMatrix ) //DcmTagKey(0x0048, 0x021e)
    GET_DICOM_INFO_LINE(DCM_RowPositionInTotalImagePixelMatrix   ) //DcmTagKey(0x0048, 0x021f)
    GET_DICOM_INFO_LINE(DCM_PixelOriginInterpretation            ) //DcmTagKey(0x0048, 0x0301)
            info->append("</div>");
            info->append("<div style=\"color:#00007F\">");
    GET_DICOM_INFO_LINE(DCM_CalibrationImage                     ) //DcmTagKey(0x0050, 0x0004)
    GET_DICOM_INFO_LINE(DCM_DeviceSequence                       ) //DcmTagKey(0x0050, 0x0010)
    GET_DICOM_INFO_LINE(DCM_ContainerComponentTypeCodeSequence   ) //DcmTagKey(0x0050, 0x0012)
    GET_DICOM_INFO_LINE(DCM_ContainerComponentThickness          ) //DcmTagKey(0x0050, 0x0013)
    GET_DICOM_INFO_LINE(DCM_DeviceLength                         ) //DcmTagKey(0x0050, 0x0014)
    GET_DICOM_INFO_LINE(DCM_ContainerComponentWidth              ) //DcmTagKey(0x0050, 0x0015)
    GET_DICOM_INFO_LINE(DCM_DeviceDiameter                       ) //DcmTagKey(0x0050, 0x0016)
    GET_DICOM_INFO_LINE(DCM_DeviceDiameterUnits                  ) //DcmTagKey(0x0050, 0x0017)
    GET_DICOM_INFO_LINE(DCM_DeviceVolume                         ) //DcmTagKey(0x0050, 0x0018)
    GET_DICOM_INFO_LINE(DCM_InterMarkerDistance                  ) //DcmTagKey(0x0050, 0x0019)
    GET_DICOM_INFO_LINE(DCM_ContainerComponentMaterial           ) //DcmTagKey(0x0050, 0x001a)
    GET_DICOM_INFO_LINE(DCM_ContainerComponentID                 ) //DcmTagKey(0x0050, 0x001b)
    GET_DICOM_INFO_LINE(DCM_ContainerComponentLength             ) //DcmTagKey(0x0050, 0x001c)
    GET_DICOM_INFO_LINE(DCM_ContainerComponentDiameter           ) //DcmTagKey(0x0050, 0x001d)
    GET_DICOM_INFO_LINE(DCM_ContainerComponentDescription        ) //DcmTagKey(0x0050, 0x001e)
    GET_DICOM_INFO_LINE(DCM_DeviceDescription                    ) //DcmTagKey(0x0050, 0x0020)
            info->append("</div>");
            info->append("<div style=\"color:#00007F\">");
    GET_DICOM_INFO_LINE(DCM_EnergyWindowVector                   ) //DcmTagKey(0x0054, 0x0010)
    GET_DICOM_INFO_LINE(DCM_NumberOfEnergyWindows                ) //DcmTagKey(0x0054, 0x0011)
    GET_DICOM_INFO_LINE(DCM_EnergyWindowInformationSequence      ) //DcmTagKey(0x0054, 0x0012)
    GET_DICOM_INFO_LINE(DCM_EnergyWindowRangeSequence            ) //DcmTagKey(0x0054, 0x0013)
    GET_DICOM_INFO_LINE(DCM_EnergyWindowLowerLimit               ) //DcmTagKey(0x0054, 0x0014)
    GET_DICOM_INFO_LINE(DCM_EnergyWindowUpperLimit               ) //DcmTagKey(0x0054, 0x0015)
    GET_DICOM_INFO_LINE(DCM_RadiopharmaceuticalInformationSequence ) //DcmTagKey(0x0054, 0x0016)
    GET_DICOM_INFO_LINE(DCM_ResidualSyringeCounts                ) //DcmTagKey(0x0054, 0x0017)
    GET_DICOM_INFO_LINE(DCM_EnergyWindowName                     ) //DcmTagKey(0x0054, 0x0018)
    GET_DICOM_INFO_LINE(DCM_DetectorVector                       ) //DcmTagKey(0x0054, 0x0020)
    GET_DICOM_INFO_LINE(DCM_NumberOfDetectors                    ) //DcmTagKey(0x0054, 0x0021)
    GET_DICOM_INFO_LINE(DCM_DetectorInformationSequence          ) //DcmTagKey(0x0054, 0x0022)
    GET_DICOM_INFO_LINE(DCM_PhaseVector                          ) //DcmTagKey(0x0054, 0x0030)
    GET_DICOM_INFO_LINE(DCM_NumberOfPhases                       ) //DcmTagKey(0x0054, 0x0031)
    GET_DICOM_INFO_LINE(DCM_PhaseInformationSequence             ) //DcmTagKey(0x0054, 0x0032)
    GET_DICOM_INFO_LINE(DCM_NumberOfFramesInPhase                ) //DcmTagKey(0x0054, 0x0033)
    GET_DICOM_INFO_LINE(DCM_PhaseDelay                           ) //DcmTagKey(0x0054, 0x0036)
    GET_DICOM_INFO_LINE(DCM_PauseBetweenFrames                   ) //DcmTagKey(0x0054, 0x0038)
    GET_DICOM_INFO_LINE(DCM_PhaseDescription                     ) //DcmTagKey(0x0054, 0x0039)
    GET_DICOM_INFO_LINE(DCM_RotationVector                       ) //DcmTagKey(0x0054, 0x0050)
    GET_DICOM_INFO_LINE(DCM_NumberOfRotations                    ) //DcmTagKey(0x0054, 0x0051)
    GET_DICOM_INFO_LINE(DCM_RotationInformationSequence          ) //DcmTagKey(0x0054, 0x0052)
    GET_DICOM_INFO_LINE(DCM_NumberOfFramesInRotation             ) //DcmTagKey(0x0054, 0x0053)
    GET_DICOM_INFO_LINE(DCM_RRIntervalVector                     ) //DcmTagKey(0x0054, 0x0060)
    GET_DICOM_INFO_LINE(DCM_NumberOfRRIntervals                  ) //DcmTagKey(0x0054, 0x0061)
    GET_DICOM_INFO_LINE(DCM_GatedInformationSequence             ) //DcmTagKey(0x0054, 0x0062)
    GET_DICOM_INFO_LINE(DCM_DataInformationSequence              ) //DcmTagKey(0x0054, 0x0063)
    GET_DICOM_INFO_LINE(DCM_TimeSlotVector                       ) //DcmTagKey(0x0054, 0x0070)
    GET_DICOM_INFO_LINE(DCM_NumberOfTimeSlots                    ) //DcmTagKey(0x0054, 0x0071)
    GET_DICOM_INFO_LINE(DCM_TimeSlotInformationSequence          ) //DcmTagKey(0x0054, 0x0072)
    GET_DICOM_INFO_LINE(DCM_TimeSlotTime                         ) //DcmTagKey(0x0054, 0x0073)
    GET_DICOM_INFO_LINE(DCM_SliceVector                          ) //DcmTagKey(0x0054, 0x0080)
    GET_DICOM_INFO_LINE(DCM_NumberOfSlices                       ) //DcmTagKey(0x0054, 0x0081)
    GET_DICOM_INFO_LINE(DCM_AngularViewVector                    ) //DcmTagKey(0x0054, 0x0090)
    GET_DICOM_INFO_LINE(DCM_TimeSliceVector                      ) //DcmTagKey(0x0054, 0x0100)
    GET_DICOM_INFO_LINE(DCM_NumberOfTimeSlices                   ) //DcmTagKey(0x0054, 0x0101)
    GET_DICOM_INFO_LINE(DCM_StartAngle                           ) //DcmTagKey(0x0054, 0x0200)
    GET_DICOM_INFO_LINE(DCM_TypeOfDetectorMotion                 ) //DcmTagKey(0x0054, 0x0202)
    GET_DICOM_INFO_LINE(DCM_TriggerVector                        ) //DcmTagKey(0x0054, 0x0210)
    GET_DICOM_INFO_LINE(DCM_NumberOfTriggersInPhase              ) //DcmTagKey(0x0054, 0x0211)
    GET_DICOM_INFO_LINE(DCM_ViewCodeSequence                     ) //DcmTagKey(0x0054, 0x0220)
    GET_DICOM_INFO_LINE(DCM_ViewModifierCodeSequence             ) //DcmTagKey(0x0054, 0x0222)
    GET_DICOM_INFO_LINE(DCM_RadionuclideCodeSequence             ) //DcmTagKey(0x0054, 0x0300)
    GET_DICOM_INFO_LINE(DCM_AdministrationRouteCodeSequence      ) //DcmTagKey(0x0054, 0x0302)
    GET_DICOM_INFO_LINE(DCM_RadiopharmaceuticalCodeSequence      ) //DcmTagKey(0x0054, 0x0304)
    GET_DICOM_INFO_LINE(DCM_CalibrationDataSequence              ) //DcmTagKey(0x0054, 0x0306)
    GET_DICOM_INFO_LINE(DCM_EnergyWindowNumber                   ) //DcmTagKey(0x0054, 0x0308)
    GET_DICOM_INFO_LINE(DCM_ImageID                              ) //DcmTagKey(0x0054, 0x0400)
    GET_DICOM_INFO_LINE(DCM_PatientOrientationCodeSequence       ) //DcmTagKey(0x0054, 0x0410)
    GET_DICOM_INFO_LINE(DCM_PatientOrientationModifierCodeSequence ) //DcmTagKey(0x0054, 0x0412)
    GET_DICOM_INFO_LINE(DCM_PatientGantryRelationshipCodeSequence ) //DcmTagKey(0x0054, 0x0414)
    GET_DICOM_INFO_LINE(DCM_SliceProgressionDirection            ) //DcmTagKey(0x0054, 0x0500)
    GET_DICOM_INFO_LINE(DCM_SeriesType                           ) //DcmTagKey(0x0054, 0x1000)
    GET_DICOM_INFO_LINE(DCM_Units                                ) //DcmTagKey(0x0054, 0x1001)
    GET_DICOM_INFO_LINE(DCM_CountsSource                         ) //DcmTagKey(0x0054, 0x1002)
    GET_DICOM_INFO_LINE(DCM_ReprojectionMethod                   ) //DcmTagKey(0x0054, 0x1004)
    GET_DICOM_INFO_LINE(DCM_RandomsCorrectionMethod              ) //DcmTagKey(0x0054, 0x1100)
    GET_DICOM_INFO_LINE(DCM_AttenuationCorrectionMethod          ) //DcmTagKey(0x0054, 0x1101)
    GET_DICOM_INFO_LINE(DCM_DecayCorrection                      ) //DcmTagKey(0x0054, 0x1102)
    GET_DICOM_INFO_LINE(DCM_ReconstructionMethod                 ) //DcmTagKey(0x0054, 0x1103)
    GET_DICOM_INFO_LINE(DCM_DetectorLinesOfResponseUsed          ) //DcmTagKey(0x0054, 0x1104)
    GET_DICOM_INFO_LINE(DCM_ScatterCorrectionMethod              ) //DcmTagKey(0x0054, 0x1105)
    GET_DICOM_INFO_LINE(DCM_AxialAcceptance                      ) //DcmTagKey(0x0054, 0x1200)
    GET_DICOM_INFO_LINE(DCM_AxialMash                            ) //DcmTagKey(0x0054, 0x1201)
    GET_DICOM_INFO_LINE(DCM_TransverseMash                       ) //DcmTagKey(0x0054, 0x1202)
    GET_DICOM_INFO_LINE(DCM_DetectorElementSize                  ) //DcmTagKey(0x0054, 0x1203)
    GET_DICOM_INFO_LINE(DCM_CoincidenceWindowWidth               ) //DcmTagKey(0x0054, 0x1210)
    GET_DICOM_INFO_LINE(DCM_SecondaryCountsType                  ) //DcmTagKey(0x0054, 0x1220)
    GET_DICOM_INFO_LINE(DCM_FrameReferenceTime                   ) //DcmTagKey(0x0054, 0x1300)
    GET_DICOM_INFO_LINE(DCM_PrimaryPromptsCountsAccumulated      ) //DcmTagKey(0x0054, 0x1310)
    GET_DICOM_INFO_LINE(DCM_SecondaryCountsAccumulated           ) //DcmTagKey(0x0054, 0x1311)
    GET_DICOM_INFO_LINE(DCM_SliceSensitivityFactor               ) //DcmTagKey(0x0054, 0x1320)
    GET_DICOM_INFO_LINE(DCM_DecayFactor                          ) //DcmTagKey(0x0054, 0x1321)
    GET_DICOM_INFO_LINE(DCM_DoseCalibrationFactor                ) //DcmTagKey(0x0054, 0x1322)
    GET_DICOM_INFO_LINE(DCM_ScatterFractionFactor                ) //DcmTagKey(0x0054, 0x1323)
    GET_DICOM_INFO_LINE(DCM_DeadTimeFactor                       ) //DcmTagKey(0x0054, 0x1324)
    GET_DICOM_INFO_LINE(DCM_ImageIndex                           ) //DcmTagKey(0x0054, 0x1330)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_CountsIncluded               ) //DcmTagKey(0x0054, 0x1400)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_DeadTimeCorrectionFlag       ) //DcmTagKey(0x0054, 0x1401)
            info->append("</div>");
            info->append("<div style=\"color:#00007F\">");
    GET_DICOM_INFO_LINE(DCM_HistogramSequence                    ) //DcmTagKey(0x0060, 0x3000)
    GET_DICOM_INFO_LINE(DCM_HistogramNumberOfBins                ) //DcmTagKey(0x0060, 0x3002)
    GET_DICOM_INFO_LINE(DCM_HistogramFirstBinValue               ) //DcmTagKey(0x0060, 0x3004)
    GET_DICOM_INFO_LINE(DCM_HistogramLastBinValue                ) //DcmTagKey(0x0060, 0x3006)
    GET_DICOM_INFO_LINE(DCM_HistogramBinWidth                    ) //DcmTagKey(0x0060, 0x3008)
    GET_DICOM_INFO_LINE(DCM_HistogramExplanation                 ) //DcmTagKey(0x0060, 0x3010)
    GET_DICOM_INFO_LINE(DCM_HistogramData                        ) //DcmTagKey(0x0060, 0x3020)
            info->append("</div>");
            info->append("<div style=\"color:#00007F\">");
    GET_DICOM_INFO_LINE(DCM_SegmentationType                     ) //DcmTagKey(0x0062, 0x0001)
    GET_DICOM_INFO_LINE(DCM_SegmentSequence                      ) //DcmTagKey(0x0062, 0x0002)
    GET_DICOM_INFO_LINE(DCM_SegmentedPropertyCategoryCodeSequence ) //DcmTagKey(0x0062, 0x0003)
    GET_DICOM_INFO_LINE(DCM_SegmentNumber                        ) //DcmTagKey(0x0062, 0x0004)
    GET_DICOM_INFO_LINE(DCM_SegmentLabel                         ) //DcmTagKey(0x0062, 0x0005)
    GET_DICOM_INFO_LINE(DCM_SegmentDescription                   ) //DcmTagKey(0x0062, 0x0006)
    GET_DICOM_INFO_LINE(DCM_SegmentAlgorithmType                 ) //DcmTagKey(0x0062, 0x0008)
    GET_DICOM_INFO_LINE(DCM_SegmentAlgorithmName                 ) //DcmTagKey(0x0062, 0x0009)
    GET_DICOM_INFO_LINE(DCM_SegmentIdentificationSequence        ) //DcmTagKey(0x0062, 0x000a)
    GET_DICOM_INFO_LINE(DCM_ReferencedSegmentNumber              ) //DcmTagKey(0x0062, 0x000b)
    GET_DICOM_INFO_LINE(DCM_RecommendedDisplayGrayscaleValue     ) //DcmTagKey(0x0062, 0x000c)
    GET_DICOM_INFO_LINE(DCM_RecommendedDisplayCIELabValue        ) //DcmTagKey(0x0062, 0x000d)
    GET_DICOM_INFO_LINE(DCM_MaximumFractionalValue               ) //DcmTagKey(0x0062, 0x000e)
    GET_DICOM_INFO_LINE(DCM_SegmentedPropertyTypeCodeSequence    ) //DcmTagKey(0x0062, 0x000f)
    GET_DICOM_INFO_LINE(DCM_SegmentationFractionalType           ) //DcmTagKey(0x0062, 0x0010)
            info->append("</div>");
            info->append("<div style=\"color:#00007F\">");
    GET_DICOM_INFO_LINE(DCM_DeformableRegistrationSequence       ) //DcmTagKey(0x0064, 0x0002)
    GET_DICOM_INFO_LINE(DCM_SourceFrameOfReferenceUID            ) //DcmTagKey(0x0064, 0x0003)
    GET_DICOM_INFO_LINE(DCM_DeformableRegistrationGridSequence   ) //DcmTagKey(0x0064, 0x0005)
    GET_DICOM_INFO_LINE(DCM_GridDimensions                       ) //DcmTagKey(0x0064, 0x0007)
    GET_DICOM_INFO_LINE(DCM_GridResolution                       ) //DcmTagKey(0x0064, 0x0008)
    GET_DICOM_INFO_LINE(DCM_VectorGridData                       ) //DcmTagKey(0x0064, 0x0009)
    GET_DICOM_INFO_LINE(DCM_PreDeformationMatrixRegistrationSequence ) //DcmTagKey(0x0064, 0x000f)
    GET_DICOM_INFO_LINE(DCM_PostDeformationMatrixRegistrationSequence ) //DcmTagKey(0x0064, 0x0010)
            info->append("</div>");
            info->append("<div style=\"color:#00007F\">");
    GET_DICOM_INFO_LINE(DCM_NumberOfSurfaces                     ) //DcmTagKey(0x0066, 0x0001)
    GET_DICOM_INFO_LINE(DCM_SurfaceSequence                      ) //DcmTagKey(0x0066, 0x0002)
    GET_DICOM_INFO_LINE(DCM_SurfaceNumber                        ) //DcmTagKey(0x0066, 0x0003)
    GET_DICOM_INFO_LINE(DCM_SurfaceComments                      ) //DcmTagKey(0x0066, 0x0004)
    GET_DICOM_INFO_LINE(DCM_SurfaceProcessing                    ) //DcmTagKey(0x0066, 0x0009)
    GET_DICOM_INFO_LINE(DCM_SurfaceProcessingRatio               ) //DcmTagKey(0x0066, 0x000a)
    GET_DICOM_INFO_LINE(DCM_SurfaceProcessingDescription         ) //DcmTagKey(0x0066, 0x000b)
    GET_DICOM_INFO_LINE(DCM_RecommendedPresentationOpacity       ) //DcmTagKey(0x0066, 0x000c)
    GET_DICOM_INFO_LINE(DCM_RecommendedPresentationType          ) //DcmTagKey(0x0066, 0x000d)
    GET_DICOM_INFO_LINE(DCM_FiniteVolume                         ) //DcmTagKey(0x0066, 0x000e)
    GET_DICOM_INFO_LINE(DCM_Manifold                             ) //DcmTagKey(0x0066, 0x0010)
    GET_DICOM_INFO_LINE(DCM_SurfacePointsSequence                ) //DcmTagKey(0x0066, 0x0011)
    GET_DICOM_INFO_LINE(DCM_SurfacePointsNormalsSequence         ) //DcmTagKey(0x0066, 0x0012)
    GET_DICOM_INFO_LINE(DCM_SurfaceMeshPrimitivesSequence        ) //DcmTagKey(0x0066, 0x0013)
    GET_DICOM_INFO_LINE(DCM_NumberOfSurfacePoints                ) //DcmTagKey(0x0066, 0x0015)
    GET_DICOM_INFO_LINE(DCM_PointCoordinatesData                 ) //DcmTagKey(0x0066, 0x0016)
    GET_DICOM_INFO_LINE(DCM_PointPositionAccuracy                ) //DcmTagKey(0x0066, 0x0017)
    GET_DICOM_INFO_LINE(DCM_MeanPointDistance                    ) //DcmTagKey(0x0066, 0x0018)
    GET_DICOM_INFO_LINE(DCM_MaximumPointDistance                 ) //DcmTagKey(0x0066, 0x0019)
    GET_DICOM_INFO_LINE(DCM_PointsBoundingBoxCoordinates         ) //DcmTagKey(0x0066, 0x001a)
    GET_DICOM_INFO_LINE(DCM_AxisOfRotation                       ) //DcmTagKey(0x0066, 0x001b)
    GET_DICOM_INFO_LINE(DCM_CenterOfRotation                     ) //DcmTagKey(0x0066, 0x001c)
    GET_DICOM_INFO_LINE(DCM_NumberOfVectors                      ) //DcmTagKey(0x0066, 0x001e)
    GET_DICOM_INFO_LINE(DCM_VectorDimensionality                 ) //DcmTagKey(0x0066, 0x001f)
    GET_DICOM_INFO_LINE(DCM_VectorAccuracy                       ) //DcmTagKey(0x0066, 0x0020)
    GET_DICOM_INFO_LINE(DCM_VectorCoordinateData                 ) //DcmTagKey(0x0066, 0x0021)
    GET_DICOM_INFO_LINE(DCM_TrianglePointIndexList               ) //DcmTagKey(0x0066, 0x0023)
    GET_DICOM_INFO_LINE(DCM_EdgePointIndexList                   ) //DcmTagKey(0x0066, 0x0024)
    GET_DICOM_INFO_LINE(DCM_VertexPointIndexList                 ) //DcmTagKey(0x0066, 0x0025)
    GET_DICOM_INFO_LINE(DCM_TriangleStripSequence                ) //DcmTagKey(0x0066, 0x0026)
    GET_DICOM_INFO_LINE(DCM_TriangleFanSequence                  ) //DcmTagKey(0x0066, 0x0027)
    GET_DICOM_INFO_LINE(DCM_LineSequence                         ) //DcmTagKey(0x0066, 0x0028)
    GET_DICOM_INFO_LINE(DCM_PrimitivePointIndexList              ) //DcmTagKey(0x0066, 0x0029)
    GET_DICOM_INFO_LINE(DCM_SurfaceCount                         ) //DcmTagKey(0x0066, 0x002a)
    GET_DICOM_INFO_LINE(DCM_ReferencedSurfaceSequence            ) //DcmTagKey(0x0066, 0x002b)
    GET_DICOM_INFO_LINE(DCM_ReferencedSurfaceNumber              ) //DcmTagKey(0x0066, 0x002c)
    GET_DICOM_INFO_LINE(DCM_SegmentSurfaceGenerationAlgorithmIdentificationSequence ) //DcmTagKey(0x0066, 0x002d)
    GET_DICOM_INFO_LINE(DCM_SegmentSurfaceSourceInstanceSequence ) //DcmTagKey(0x0066, 0x002e)
    GET_DICOM_INFO_LINE(DCM_AlgorithmFamilyCodeSequence          ) //DcmTagKey(0x0066, 0x002f)
    GET_DICOM_INFO_LINE(DCM_AlgorithmNameCodeSequence            ) //DcmTagKey(0x0066, 0x0030)
    GET_DICOM_INFO_LINE(DCM_AlgorithmVersion                     ) //DcmTagKey(0x0066, 0x0031)
    GET_DICOM_INFO_LINE(DCM_AlgorithmParameters                  ) //DcmTagKey(0x0066, 0x0032)
    GET_DICOM_INFO_LINE(DCM_FacetSequence                        ) //DcmTagKey(0x0066, 0x0034)
    GET_DICOM_INFO_LINE(DCM_SurfaceProcessingAlgorithmIdentificationSequence ) //DcmTagKey(0x0066, 0x0035)
    GET_DICOM_INFO_LINE(DCM_AlgorithmName                        ) //DcmTagKey(0x0066, 0x0036)
            info->append("</div>");
            info->append("<div style=\"color:#00007F\">");
    GET_DICOM_INFO_LINE(DCM_ImplantSize                          ) //DcmTagKey(0x0068, 0x6210)
    GET_DICOM_INFO_LINE(DCM_ImplantTemplateVersion               ) //DcmTagKey(0x0068, 0x6221)
    GET_DICOM_INFO_LINE(DCM_ReplacedImplantTemplateSequence      ) //DcmTagKey(0x0068, 0x6222)
    GET_DICOM_INFO_LINE(DCM_ImplantType                          ) //DcmTagKey(0x0068, 0x6223)
    GET_DICOM_INFO_LINE(DCM_DerivationImplantTemplateSequence    ) //DcmTagKey(0x0068, 0x6224)
    GET_DICOM_INFO_LINE(DCM_OriginalImplantTemplateSequence      ) //DcmTagKey(0x0068, 0x6225)
    GET_DICOM_INFO_LINE(DCM_EffectiveDateTime                    ) //DcmTagKey(0x0068, 0x6226)
    GET_DICOM_INFO_LINE(DCM_ImplantTargetAnatomySequence         ) //DcmTagKey(0x0068, 0x6230)
    GET_DICOM_INFO_LINE(DCM_InformationFromManufacturerSequence  ) //DcmTagKey(0x0068, 0x6260)
    GET_DICOM_INFO_LINE(DCM_NotificationFromManufacturerSequence ) //DcmTagKey(0x0068, 0x6265)
    GET_DICOM_INFO_LINE(DCM_InformationIssueDateTime             ) //DcmTagKey(0x0068, 0x6270)
    GET_DICOM_INFO_LINE(DCM_InformationSummary                   ) //DcmTagKey(0x0068, 0x6280)
    GET_DICOM_INFO_LINE(DCM_ImplantRegulatoryDisapprovalCodeSequence ) //DcmTagKey(0x0068, 0x62a0)
    GET_DICOM_INFO_LINE(DCM_OverallTemplateSpatialTolerance      ) //DcmTagKey(0x0068, 0x62a5)
    GET_DICOM_INFO_LINE(DCM_HPGLDocumentSequence                 ) //DcmTagKey(0x0068, 0x62c0)
    GET_DICOM_INFO_LINE(DCM_HPGLDocumentID                       ) //DcmTagKey(0x0068, 0x62d0)
    GET_DICOM_INFO_LINE(DCM_HPGLDocumentLabel                    ) //DcmTagKey(0x0068, 0x62d5)
    GET_DICOM_INFO_LINE(DCM_ViewOrientationCodeSequence          ) //DcmTagKey(0x0068, 0x62e0)
    GET_DICOM_INFO_LINE(DCM_ViewOrientationModifier              ) //DcmTagKey(0x0068, 0x62f0)
    GET_DICOM_INFO_LINE(DCM_HPGLDocumentScaling                  ) //DcmTagKey(0x0068, 0x62f2)
    GET_DICOM_INFO_LINE(DCM_HPGLDocument                         ) //DcmTagKey(0x0068, 0x6300)
    GET_DICOM_INFO_LINE(DCM_HPGLContourPenNumber                 ) //DcmTagKey(0x0068, 0x6310)
    GET_DICOM_INFO_LINE(DCM_HPGLPenSequence                      ) //DcmTagKey(0x0068, 0x6320)
    GET_DICOM_INFO_LINE(DCM_HPGLPenNumber                        ) //DcmTagKey(0x0068, 0x6330)
    GET_DICOM_INFO_LINE(DCM_HPGLPenLabel                         ) //DcmTagKey(0x0068, 0x6340)
    GET_DICOM_INFO_LINE(DCM_HPGLPenDescription                   ) //DcmTagKey(0x0068, 0x6345)
    GET_DICOM_INFO_LINE(DCM_RecommendedRotationPoint             ) //DcmTagKey(0x0068, 0x6346)
    GET_DICOM_INFO_LINE(DCM_BoundingRectangle                    ) //DcmTagKey(0x0068, 0x6347)
    GET_DICOM_INFO_LINE(DCM_ImplantTemplate3DModelSurfaceNumber  ) //DcmTagKey(0x0068, 0x6350)
    GET_DICOM_INFO_LINE(DCM_SurfaceModelDescriptionSequence      ) //DcmTagKey(0x0068, 0x6360)
    GET_DICOM_INFO_LINE(DCM_SurfaceModelLabel                    ) //DcmTagKey(0x0068, 0x6380)
    GET_DICOM_INFO_LINE(DCM_SurfaceModelScalingFactor            ) //DcmTagKey(0x0068, 0x6390)
    GET_DICOM_INFO_LINE(DCM_MaterialsCodeSequence                ) //DcmTagKey(0x0068, 0x63a0)
    GET_DICOM_INFO_LINE(DCM_CoatingMaterialsCodeSequence         ) //DcmTagKey(0x0068, 0x63a4)
    GET_DICOM_INFO_LINE(DCM_ImplantTypeCodeSequence              ) //DcmTagKey(0x0068, 0x63a8)
    GET_DICOM_INFO_LINE(DCM_FixationMethodCodeSequence           ) //DcmTagKey(0x0068, 0x63ac)
    GET_DICOM_INFO_LINE(DCM_MatingFeatureSetsSequence            ) //DcmTagKey(0x0068, 0x63b0)
    GET_DICOM_INFO_LINE(DCM_MatingFeatureSetID                   ) //DcmTagKey(0x0068, 0x63c0)
    GET_DICOM_INFO_LINE(DCM_MatingFeatureSetLabel                ) //DcmTagKey(0x0068, 0x63d0)
    GET_DICOM_INFO_LINE(DCM_MatingFeatureSequence                ) //DcmTagKey(0x0068, 0x63e0)
    GET_DICOM_INFO_LINE(DCM_MatingFeatureID                      ) //DcmTagKey(0x0068, 0x63f0)
    GET_DICOM_INFO_LINE(DCM_MatingFeatureDegreeOfFreedomSequence ) //DcmTagKey(0x0068, 0x6400)
    GET_DICOM_INFO_LINE(DCM_DegreeOfFreedomID                    ) //DcmTagKey(0x0068, 0x6410)
    GET_DICOM_INFO_LINE(DCM_DegreeOfFreedomType                  ) //DcmTagKey(0x0068, 0x6420)
    GET_DICOM_INFO_LINE(DCM_TwoDMatingFeatureCoordinatesSequence ) //DcmTagKey(0x0068, 0x6430)
    GET_DICOM_INFO_LINE(DCM_ReferencedHPGLDocumentID             ) //DcmTagKey(0x0068, 0x6440)
    GET_DICOM_INFO_LINE(DCM_TwoDMatingPoint                      ) //DcmTagKey(0x0068, 0x6450)
    GET_DICOM_INFO_LINE(DCM_TwoDMatingAxes                       ) //DcmTagKey(0x0068, 0x6460)
    GET_DICOM_INFO_LINE(DCM_TwoDDegreeOfFreedomSequence          ) //DcmTagKey(0x0068, 0x6470)
    GET_DICOM_INFO_LINE(DCM_ThreeDDegreeOfFreedomAxis            ) //DcmTagKey(0x0068, 0x6490)
    GET_DICOM_INFO_LINE(DCM_RangeOfFreedom                       ) //DcmTagKey(0x0068, 0x64a0)
    GET_DICOM_INFO_LINE(DCM_ThreeDMatingPoint                    ) //DcmTagKey(0x0068, 0x64c0)
    GET_DICOM_INFO_LINE(DCM_ThreeDMatingAxes                     ) //DcmTagKey(0x0068, 0x64d0)
    GET_DICOM_INFO_LINE(DCM_TwoDDegreeOfFreedomAxis              ) //DcmTagKey(0x0068, 0x64f0)
    GET_DICOM_INFO_LINE(DCM_PlanningLandmarkPointSequence        ) //DcmTagKey(0x0068, 0x6500)
    GET_DICOM_INFO_LINE(DCM_PlanningLandmarkLineSequence         ) //DcmTagKey(0x0068, 0x6510)
    GET_DICOM_INFO_LINE(DCM_PlanningLandmarkPlaneSequence        ) //DcmTagKey(0x0068, 0x6520)
    GET_DICOM_INFO_LINE(DCM_PlanningLandmarkID                   ) //DcmTagKey(0x0068, 0x6530)
    GET_DICOM_INFO_LINE(DCM_PlanningLandmarkDescription          ) //DcmTagKey(0x0068, 0x6540)
    GET_DICOM_INFO_LINE(DCM_PlanningLandmarkIdentificationCodeSequence ) //DcmTagKey(0x0068, 0x6545)
    GET_DICOM_INFO_LINE(DCM_TwoDPointCoordinatesSequence         ) //DcmTagKey(0x0068, 0x6550)
    GET_DICOM_INFO_LINE(DCM_TwoDPointCoordinates                 ) //DcmTagKey(0x0068, 0x6560)
    GET_DICOM_INFO_LINE(DCM_ThreeDPointCoordinates               ) //DcmTagKey(0x0068, 0x6590)
    GET_DICOM_INFO_LINE(DCM_TwoDLineCoordinatesSequence          ) //DcmTagKey(0x0068, 0x65a0)
    GET_DICOM_INFO_LINE(DCM_TwoDLineCoordinates                  ) //DcmTagKey(0x0068, 0x65b0)
    GET_DICOM_INFO_LINE(DCM_ThreeDLineCoordinates                ) //DcmTagKey(0x0068, 0x65d0)
    GET_DICOM_INFO_LINE(DCM_TwoDPlaneCoordinatesSequence         ) //DcmTagKey(0x0068, 0x65e0)
    GET_DICOM_INFO_LINE(DCM_TwoDPlaneIntersection                ) //DcmTagKey(0x0068, 0x65f0)
    GET_DICOM_INFO_LINE(DCM_ThreeDPlaneOrigin                    ) //DcmTagKey(0x0068, 0x6610)
    GET_DICOM_INFO_LINE(DCM_ThreeDPlaneNormal                    ) //DcmTagKey(0x0068, 0x6620)
            info->append("</div>");
            info->append("<div style=\"color:#00007F\">");
    GET_DICOM_INFO_LINE(DCM_GraphicAnnotationSequence            ) //DcmTagKey(0x0070, 0x0001)
    GET_DICOM_INFO_LINE(DCM_GraphicLayer                         ) //DcmTagKey(0x0070, 0x0002)
    GET_DICOM_INFO_LINE(DCM_BoundingBoxAnnotationUnits           ) //DcmTagKey(0x0070, 0x0003)
    GET_DICOM_INFO_LINE(DCM_AnchorPointAnnotationUnits           ) //DcmTagKey(0x0070, 0x0004)
    GET_DICOM_INFO_LINE(DCM_GraphicAnnotationUnits               ) //DcmTagKey(0x0070, 0x0005)
    GET_DICOM_INFO_LINE(DCM_UnformattedTextValue                 ) //DcmTagKey(0x0070, 0x0006)
    GET_DICOM_INFO_LINE(DCM_TextObjectSequence                   ) //DcmTagKey(0x0070, 0x0008)
    GET_DICOM_INFO_LINE(DCM_GraphicObjectSequence                ) //DcmTagKey(0x0070, 0x0009)
    GET_DICOM_INFO_LINE(DCM_BoundingBoxTopLeftHandCorner         ) //DcmTagKey(0x0070, 0x0010)
    GET_DICOM_INFO_LINE(DCM_BoundingBoxBottomRightHandCorner     ) //DcmTagKey(0x0070, 0x0011)
    GET_DICOM_INFO_LINE(DCM_BoundingBoxTextHorizontalJustification ) //DcmTagKey(0x0070, 0x0012)
    GET_DICOM_INFO_LINE(DCM_AnchorPoint                          ) //DcmTagKey(0x0070, 0x0014)
    GET_DICOM_INFO_LINE(DCM_AnchorPointVisibility                ) //DcmTagKey(0x0070, 0x0015)
    GET_DICOM_INFO_LINE(DCM_GraphicDimensions                    ) //DcmTagKey(0x0070, 0x0020)
    GET_DICOM_INFO_LINE(DCM_NumberOfGraphicPoints                ) //DcmTagKey(0x0070, 0x0021)
    GET_DICOM_INFO_LINE(DCM_GraphicData                          ) //DcmTagKey(0x0070, 0x0022)
    GET_DICOM_INFO_LINE(DCM_GraphicType                          ) //DcmTagKey(0x0070, 0x0023)
    GET_DICOM_INFO_LINE(DCM_GraphicFilled                        ) //DcmTagKey(0x0070, 0x0024)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ImageRotationRetired         ) //DcmTagKey(0x0070, 0x0040)
    GET_DICOM_INFO_LINE(DCM_ImageHorizontalFlip                  ) //DcmTagKey(0x0070, 0x0041)
    GET_DICOM_INFO_LINE(DCM_ImageRotation                        ) //DcmTagKey(0x0070, 0x0042)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_DisplayedAreaTopLeftHandCornerTrial ) //DcmTagKey(0x0070, 0x0050)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_DisplayedAreaBottomRightHandCornerTrial ) //DcmTagKey(0x0070, 0x0051)
    GET_DICOM_INFO_LINE(DCM_DisplayedAreaTopLeftHandCorner       ) //DcmTagKey(0x0070, 0x0052)
    GET_DICOM_INFO_LINE(DCM_DisplayedAreaBottomRightHandCorner   ) //DcmTagKey(0x0070, 0x0053)
    GET_DICOM_INFO_LINE(DCM_DisplayedAreaSelectionSequence       ) //DcmTagKey(0x0070, 0x005a)
    GET_DICOM_INFO_LINE(DCM_GraphicLayerSequence                 ) //DcmTagKey(0x0070, 0x0060)
    GET_DICOM_INFO_LINE(DCM_GraphicLayerOrder                    ) //DcmTagKey(0x0070, 0x0062)
    GET_DICOM_INFO_LINE(DCM_GraphicLayerRecommendedDisplayGrayscaleValue ) //DcmTagKey(0x0070, 0x0066)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_GraphicLayerRecommendedDisplayRGBValue ) //DcmTagKey(0x0070, 0x0067)
    GET_DICOM_INFO_LINE(DCM_GraphicLayerDescription              ) //DcmTagKey(0x0070, 0x0068)
    GET_DICOM_INFO_LINE(DCM_ContentLabel                         ) //DcmTagKey(0x0070, 0x0080)
    GET_DICOM_INFO_LINE(DCM_ContentDescription                   ) //DcmTagKey(0x0070, 0x0081)
    GET_DICOM_INFO_LINE(DCM_PresentationCreationDate             ) //DcmTagKey(0x0070, 0x0082)
    GET_DICOM_INFO_LINE(DCM_PresentationCreationTime             ) //DcmTagKey(0x0070, 0x0083)
    GET_DICOM_INFO_LINE(DCM_ContentCreatorName                   ) //DcmTagKey(0x0070, 0x0084)
    GET_DICOM_INFO_LINE(DCM_ContentCreatorIdentificationCodeSequence ) //DcmTagKey(0x0070, 0x0086)
    GET_DICOM_INFO_LINE(DCM_AlternateContentDescriptionSequence  ) //DcmTagKey(0x0070, 0x0087)
    GET_DICOM_INFO_LINE(DCM_PresentationSizeMode                 ) //DcmTagKey(0x0070, 0x0100)
    GET_DICOM_INFO_LINE(DCM_PresentationPixelSpacing             ) //DcmTagKey(0x0070, 0x0101)
    GET_DICOM_INFO_LINE(DCM_PresentationPixelAspectRatio         ) //DcmTagKey(0x0070, 0x0102)
    GET_DICOM_INFO_LINE(DCM_PresentationPixelMagnificationRatio  ) //DcmTagKey(0x0070, 0x0103)
    GET_DICOM_INFO_LINE(DCM_GraphicGroupLabel                    ) //DcmTagKey(0x0070, 0x0207)
    GET_DICOM_INFO_LINE(DCM_GraphicGroupDescription              ) //DcmTagKey(0x0070, 0x0208)
    GET_DICOM_INFO_LINE(DCM_CompoundGraphicSequence              ) //DcmTagKey(0x0070, 0x0209)
    GET_DICOM_INFO_LINE(DCM_CompoundGraphicInstanceID            ) //DcmTagKey(0x0070, 0x0226)
    GET_DICOM_INFO_LINE(DCM_FontName                             ) //DcmTagKey(0x0070, 0x0227)
    GET_DICOM_INFO_LINE(DCM_FontNameType                         ) //DcmTagKey(0x0070, 0x0228)
    GET_DICOM_INFO_LINE(DCM_CSSFontName                          ) //DcmTagKey(0x0070, 0x0229)
    GET_DICOM_INFO_LINE(DCM_RotationAngle                        ) //DcmTagKey(0x0070, 0x0230)
    GET_DICOM_INFO_LINE(DCM_TextStyleSequence                    ) //DcmTagKey(0x0070, 0x0231)
    GET_DICOM_INFO_LINE(DCM_LineStyleSequence                    ) //DcmTagKey(0x0070, 0x0232)
    GET_DICOM_INFO_LINE(DCM_FillStyleSequence                    ) //DcmTagKey(0x0070, 0x0233)
    GET_DICOM_INFO_LINE(DCM_GraphicGroupSequence                 ) //DcmTagKey(0x0070, 0x0234)
    GET_DICOM_INFO_LINE(DCM_TextColorCIELabValue                 ) //DcmTagKey(0x0070, 0x0241)
    GET_DICOM_INFO_LINE(DCM_HorizontalAlignment                  ) //DcmTagKey(0x0070, 0x0242)
    GET_DICOM_INFO_LINE(DCM_VerticalAlignment                    ) //DcmTagKey(0x0070, 0x0243)
    GET_DICOM_INFO_LINE(DCM_ShadowStyle                          ) //DcmTagKey(0x0070, 0x0244)
    GET_DICOM_INFO_LINE(DCM_ShadowOffsetX                        ) //DcmTagKey(0x0070, 0x0245)
    GET_DICOM_INFO_LINE(DCM_ShadowOffsetY                        ) //DcmTagKey(0x0070, 0x0246)
    GET_DICOM_INFO_LINE(DCM_ShadowColorCIELabValue               ) //DcmTagKey(0x0070, 0x0247)
    GET_DICOM_INFO_LINE(DCM_Underlined                           ) //DcmTagKey(0x0070, 0x0248)
    GET_DICOM_INFO_LINE(DCM_Bold                                 ) //DcmTagKey(0x0070, 0x0249)
    GET_DICOM_INFO_LINE(DCM_Italic                               ) //DcmTagKey(0x0070, 0x0250)
    GET_DICOM_INFO_LINE(DCM_PatternOnColorCIELabValue            ) //DcmTagKey(0x0070, 0x0251)
    GET_DICOM_INFO_LINE(DCM_PatternOffColorCIELabValue           ) //DcmTagKey(0x0070, 0x0252)
    GET_DICOM_INFO_LINE(DCM_LineThickness                        ) //DcmTagKey(0x0070, 0x0253)
    GET_DICOM_INFO_LINE(DCM_LineDashingStyle                     ) //DcmTagKey(0x0070, 0x0254)
    GET_DICOM_INFO_LINE(DCM_LinePattern                          ) //DcmTagKey(0x0070, 0x0255)
    GET_DICOM_INFO_LINE(DCM_FillPattern                          ) //DcmTagKey(0x0070, 0x0256)
    GET_DICOM_INFO_LINE(DCM_FillMode                             ) //DcmTagKey(0x0070, 0x0257)
    GET_DICOM_INFO_LINE(DCM_ShadowOpacity                        ) //DcmTagKey(0x0070, 0x0258)
    GET_DICOM_INFO_LINE(DCM_GapLength                            ) //DcmTagKey(0x0070, 0x0261)
    GET_DICOM_INFO_LINE(DCM_DiameterOfVisibility                 ) //DcmTagKey(0x0070, 0x0262)
    GET_DICOM_INFO_LINE(DCM_RotationPoint                        ) //DcmTagKey(0x0070, 0x0273)
    GET_DICOM_INFO_LINE(DCM_TickAlignment                        ) //DcmTagKey(0x0070, 0x0274)
    GET_DICOM_INFO_LINE(DCM_ShowTickLabel                        ) //DcmTagKey(0x0070, 0x0278)
    GET_DICOM_INFO_LINE(DCM_TickLabelAlignment                   ) //DcmTagKey(0x0070, 0x0279)
    GET_DICOM_INFO_LINE(DCM_CompoundGraphicUnits                 ) //DcmTagKey(0x0070, 0x0282)
    GET_DICOM_INFO_LINE(DCM_PatternOnOpacity                     ) //DcmTagKey(0x0070, 0x0284)
    GET_DICOM_INFO_LINE(DCM_PatternOffOpacity                    ) //DcmTagKey(0x0070, 0x0285)
    GET_DICOM_INFO_LINE(DCM_MajorTicksSequence                   ) //DcmTagKey(0x0070, 0x0287)
    GET_DICOM_INFO_LINE(DCM_TickPosition                         ) //DcmTagKey(0x0070, 0x0288)
    GET_DICOM_INFO_LINE(DCM_TickLabel                            ) //DcmTagKey(0x0070, 0x0289)
    GET_DICOM_INFO_LINE(DCM_CompoundGraphicType                  ) //DcmTagKey(0x0070, 0x0294)
    GET_DICOM_INFO_LINE(DCM_GraphicGroupID                       ) //DcmTagKey(0x0070, 0x0295)
    GET_DICOM_INFO_LINE(DCM_ShapeType                            ) //DcmTagKey(0x0070, 0x0306)
    GET_DICOM_INFO_LINE(DCM_RegistrationSequence                 ) //DcmTagKey(0x0070, 0x0308)
    GET_DICOM_INFO_LINE(DCM_MatrixRegistrationSequence           ) //DcmTagKey(0x0070, 0x0309)
    GET_DICOM_INFO_LINE(DCM_MatrixSequence                       ) //DcmTagKey(0x0070, 0x030a)
    GET_DICOM_INFO_LINE(DCM_FrameOfReferenceTransformationMatrixType ) //DcmTagKey(0x0070, 0x030c)
    GET_DICOM_INFO_LINE(DCM_RegistrationTypeCodeSequence         ) //DcmTagKey(0x0070, 0x030d)
    GET_DICOM_INFO_LINE(DCM_FiducialDescription                  ) //DcmTagKey(0x0070, 0x030f)
    GET_DICOM_INFO_LINE(DCM_FiducialIdentifier                   ) //DcmTagKey(0x0070, 0x0310)
    GET_DICOM_INFO_LINE(DCM_FiducialIdentifierCodeSequence       ) //DcmTagKey(0x0070, 0x0311)
    GET_DICOM_INFO_LINE(DCM_ContourUncertaintyRadius             ) //DcmTagKey(0x0070, 0x0312)
    GET_DICOM_INFO_LINE(DCM_UsedFiducialsSequence                ) //DcmTagKey(0x0070, 0x0314)
    GET_DICOM_INFO_LINE(DCM_GraphicCoordinatesDataSequence       ) //DcmTagKey(0x0070, 0x0318)
    GET_DICOM_INFO_LINE(DCM_FiducialUID                          ) //DcmTagKey(0x0070, 0x031a)
    GET_DICOM_INFO_LINE(DCM_FiducialSetSequence                  ) //DcmTagKey(0x0070, 0x031c)
    GET_DICOM_INFO_LINE(DCM_FiducialSequence                     ) //DcmTagKey(0x0070, 0x031e)
    GET_DICOM_INFO_LINE(DCM_GraphicLayerRecommendedDisplayCIELabValue ) //DcmTagKey(0x0070, 0x0401)
    GET_DICOM_INFO_LINE(DCM_BlendingSequence                     ) //DcmTagKey(0x0070, 0x0402)
    GET_DICOM_INFO_LINE(DCM_RelativeOpacity                      ) //DcmTagKey(0x0070, 0x0403)
    GET_DICOM_INFO_LINE(DCM_ReferencedSpatialRegistrationSequence ) //DcmTagKey(0x0070, 0x0404)
    GET_DICOM_INFO_LINE(DCM_BlendingPosition                     ) //DcmTagKey(0x0070, 0x0405)
            info->append("</div>");
            info->append("<div style=\"color:#00007F\">");
    GET_DICOM_INFO_LINE(DCM_HangingProtocolName                  ) //DcmTagKey(0x0072, 0x0002)
    GET_DICOM_INFO_LINE(DCM_HangingProtocolDescription           ) //DcmTagKey(0x0072, 0x0004)
    GET_DICOM_INFO_LINE(DCM_HangingProtocolLevel                 ) //DcmTagKey(0x0072, 0x0006)
    GET_DICOM_INFO_LINE(DCM_HangingProtocolCreator               ) //DcmTagKey(0x0072, 0x0008)
    GET_DICOM_INFO_LINE(DCM_HangingProtocolCreationDateTime      ) //DcmTagKey(0x0072, 0x000a)
    GET_DICOM_INFO_LINE(DCM_HangingProtocolDefinitionSequence    ) //DcmTagKey(0x0072, 0x000c)
    GET_DICOM_INFO_LINE(DCM_HangingProtocolUserIdentificationCodeSequence ) //DcmTagKey(0x0072, 0x000e)
    GET_DICOM_INFO_LINE(DCM_HangingProtocolUserGroupName         ) //DcmTagKey(0x0072, 0x0010)
    GET_DICOM_INFO_LINE(DCM_SourceHangingProtocolSequence        ) //DcmTagKey(0x0072, 0x0012)
    GET_DICOM_INFO_LINE(DCM_NumberOfPriorsReferenced             ) //DcmTagKey(0x0072, 0x0014)
    GET_DICOM_INFO_LINE(DCM_ImageSetsSequence                    ) //DcmTagKey(0x0072, 0x0020)
    GET_DICOM_INFO_LINE(DCM_ImageSetSelectorSequence             ) //DcmTagKey(0x0072, 0x0022)
    GET_DICOM_INFO_LINE(DCM_ImageSetSelectorUsageFlag            ) //DcmTagKey(0x0072, 0x0024)
    GET_DICOM_INFO_LINE(DCM_SelectorAttribute                    ) //DcmTagKey(0x0072, 0x0026)
    GET_DICOM_INFO_LINE(DCM_SelectorValueNumber                  ) //DcmTagKey(0x0072, 0x0028)
    GET_DICOM_INFO_LINE(DCM_TimeBasedImageSetsSequence           ) //DcmTagKey(0x0072, 0x0030)
    GET_DICOM_INFO_LINE(DCM_ImageSetNumber                       ) //DcmTagKey(0x0072, 0x0032)
    GET_DICOM_INFO_LINE(DCM_ImageSetSelectorCategory             ) //DcmTagKey(0x0072, 0x0034)
    GET_DICOM_INFO_LINE(DCM_RelativeTime                         ) //DcmTagKey(0x0072, 0x0038)
    GET_DICOM_INFO_LINE(DCM_RelativeTimeUnits                    ) //DcmTagKey(0x0072, 0x003a)
    GET_DICOM_INFO_LINE(DCM_AbstractPriorValue                   ) //DcmTagKey(0x0072, 0x003c)
    GET_DICOM_INFO_LINE(DCM_AbstractPriorCodeSequence            ) //DcmTagKey(0x0072, 0x003e)
    GET_DICOM_INFO_LINE(DCM_ImageSetLabel                        ) //DcmTagKey(0x0072, 0x0040)
    GET_DICOM_INFO_LINE(DCM_SelectorAttributeVR                  ) //DcmTagKey(0x0072, 0x0050)
    GET_DICOM_INFO_LINE(DCM_SelectorSequencePointer              ) //DcmTagKey(0x0072, 0x0052)
    GET_DICOM_INFO_LINE(DCM_SelectorSequencePointerPrivateCreator ) //DcmTagKey(0x0072, 0x0054)
    GET_DICOM_INFO_LINE(DCM_SelectorAttributePrivateCreator      ) //DcmTagKey(0x0072, 0x0056)
    GET_DICOM_INFO_LINE(DCM_SelectorATValue                      ) //DcmTagKey(0x0072, 0x0060)
    GET_DICOM_INFO_LINE(DCM_SelectorCSValue                      ) //DcmTagKey(0x0072, 0x0062)
    GET_DICOM_INFO_LINE(DCM_SelectorISValue                      ) //DcmTagKey(0x0072, 0x0064)
    GET_DICOM_INFO_LINE(DCM_SelectorLOValue                      ) //DcmTagKey(0x0072, 0x0066)
    GET_DICOM_INFO_LINE(DCM_SelectorLTValue                      ) //DcmTagKey(0x0072, 0x0068)
    GET_DICOM_INFO_LINE(DCM_SelectorPNValue                      ) //DcmTagKey(0x0072, 0x006a)
    GET_DICOM_INFO_LINE(DCM_SelectorSHValue                      ) //DcmTagKey(0x0072, 0x006c)
    GET_DICOM_INFO_LINE(DCM_SelectorSTValue                      ) //DcmTagKey(0x0072, 0x006e)
    GET_DICOM_INFO_LINE(DCM_SelectorUTValue                      ) //DcmTagKey(0x0072, 0x0070)
    GET_DICOM_INFO_LINE(DCM_SelectorDSValue                      ) //DcmTagKey(0x0072, 0x0072)
    GET_DICOM_INFO_LINE(DCM_SelectorFDValue                      ) //DcmTagKey(0x0072, 0x0074)
    GET_DICOM_INFO_LINE(DCM_SelectorFLValue                      ) //DcmTagKey(0x0072, 0x0076)
    GET_DICOM_INFO_LINE(DCM_SelectorULValue                      ) //DcmTagKey(0x0072, 0x0078)
    GET_DICOM_INFO_LINE(DCM_SelectorUSValue                      ) //DcmTagKey(0x0072, 0x007a)
    GET_DICOM_INFO_LINE(DCM_SelectorSLValue                      ) //DcmTagKey(0x0072, 0x007c)
    GET_DICOM_INFO_LINE(DCM_SelectorSSValue                      ) //DcmTagKey(0x0072, 0x007e)
    GET_DICOM_INFO_LINE(DCM_SelectorCodeSequenceValue            ) //DcmTagKey(0x0072, 0x0080)
    GET_DICOM_INFO_LINE(DCM_NumberOfScreens                      ) //DcmTagKey(0x0072, 0x0100)
    GET_DICOM_INFO_LINE(DCM_NominalScreenDefinitionSequence      ) //DcmTagKey(0x0072, 0x0102)
    GET_DICOM_INFO_LINE(DCM_NumberOfVerticalPixels               ) //DcmTagKey(0x0072, 0x0104)
    GET_DICOM_INFO_LINE(DCM_NumberOfHorizontalPixels             ) //DcmTagKey(0x0072, 0x0106)
    GET_DICOM_INFO_LINE(DCM_DisplayEnvironmentSpatialPosition    ) //DcmTagKey(0x0072, 0x0108)
    GET_DICOM_INFO_LINE(DCM_ScreenMinimumGrayscaleBitDepth       ) //DcmTagKey(0x0072, 0x010a)
    GET_DICOM_INFO_LINE(DCM_ScreenMinimumColorBitDepth           ) //DcmTagKey(0x0072, 0x010c)
    GET_DICOM_INFO_LINE(DCM_ApplicationMaximumRepaintTime        ) //DcmTagKey(0x0072, 0x010e)
    GET_DICOM_INFO_LINE(DCM_DisplaySetsSequence                  ) //DcmTagKey(0x0072, 0x0200)
    GET_DICOM_INFO_LINE(DCM_DisplaySetNumber                     ) //DcmTagKey(0x0072, 0x0202)
    GET_DICOM_INFO_LINE(DCM_DisplaySetLabel                      ) //DcmTagKey(0x0072, 0x0203)
    GET_DICOM_INFO_LINE(DCM_DisplaySetPresentationGroup          ) //DcmTagKey(0x0072, 0x0204)
    GET_DICOM_INFO_LINE(DCM_DisplaySetPresentationGroupDescription ) //DcmTagKey(0x0072, 0x0206)
    GET_DICOM_INFO_LINE(DCM_PartialDataDisplayHandling           ) //DcmTagKey(0x0072, 0x0208)
    GET_DICOM_INFO_LINE(DCM_SynchronizedScrollingSequence        ) //DcmTagKey(0x0072, 0x0210)
    GET_DICOM_INFO_LINE(DCM_DisplaySetScrollingGroup             ) //DcmTagKey(0x0072, 0x0212)
    GET_DICOM_INFO_LINE(DCM_NavigationIndicatorSequence          ) //DcmTagKey(0x0072, 0x0214)
    GET_DICOM_INFO_LINE(DCM_NavigationDisplaySet                 ) //DcmTagKey(0x0072, 0x0216)
    GET_DICOM_INFO_LINE(DCM_ReferenceDisplaySets                 ) //DcmTagKey(0x0072, 0x0218)
    GET_DICOM_INFO_LINE(DCM_ImageBoxesSequence                   ) //DcmTagKey(0x0072, 0x0300)
    GET_DICOM_INFO_LINE(DCM_ImageBoxNumber                       ) //DcmTagKey(0x0072, 0x0302)
    GET_DICOM_INFO_LINE(DCM_ImageBoxLayoutType                   ) //DcmTagKey(0x0072, 0x0304)
    GET_DICOM_INFO_LINE(DCM_ImageBoxTileHorizontalDimension      ) //DcmTagKey(0x0072, 0x0306)
    GET_DICOM_INFO_LINE(DCM_ImageBoxTileVerticalDimension        ) //DcmTagKey(0x0072, 0x0308)
    GET_DICOM_INFO_LINE(DCM_ImageBoxScrollDirection              ) //DcmTagKey(0x0072, 0x0310)
    GET_DICOM_INFO_LINE(DCM_ImageBoxSmallScrollType              ) //DcmTagKey(0x0072, 0x0312)
    GET_DICOM_INFO_LINE(DCM_ImageBoxSmallScrollAmount            ) //DcmTagKey(0x0072, 0x0314)
    GET_DICOM_INFO_LINE(DCM_ImageBoxLargeScrollType              ) //DcmTagKey(0x0072, 0x0316)
    GET_DICOM_INFO_LINE(DCM_ImageBoxLargeScrollAmount            ) //DcmTagKey(0x0072, 0x0318)
    GET_DICOM_INFO_LINE(DCM_ImageBoxOverlapPriority              ) //DcmTagKey(0x0072, 0x0320)
    GET_DICOM_INFO_LINE(DCM_CineRelativeToRealTime               ) //DcmTagKey(0x0072, 0x0330)
    GET_DICOM_INFO_LINE(DCM_FilterOperationsSequence             ) //DcmTagKey(0x0072, 0x0400)
    GET_DICOM_INFO_LINE(DCM_FilterByCategory                     ) //DcmTagKey(0x0072, 0x0402)
    GET_DICOM_INFO_LINE(DCM_FilterByAttributePresence            ) //DcmTagKey(0x0072, 0x0404)
    GET_DICOM_INFO_LINE(DCM_FilterByOperator                     ) //DcmTagKey(0x0072, 0x0406)
    GET_DICOM_INFO_LINE(DCM_StructuredDisplayBackgroundCIELabValue ) //DcmTagKey(0x0072, 0x0420)
    GET_DICOM_INFO_LINE(DCM_EmptyImageBoxCIELabValue             ) //DcmTagKey(0x0072, 0x0421)
    GET_DICOM_INFO_LINE(DCM_StructuredDisplayImageBoxSequence    ) //DcmTagKey(0x0072, 0x0422)
    GET_DICOM_INFO_LINE(DCM_StructuredDisplayTextBoxSequence     ) //DcmTagKey(0x0072, 0x0424)
    GET_DICOM_INFO_LINE(DCM_ReferencedFirstFrameSequence         ) //DcmTagKey(0x0072, 0x0427)
    GET_DICOM_INFO_LINE(DCM_ImageBoxSynchronizationSequence      ) //DcmTagKey(0x0072, 0x0430)
    GET_DICOM_INFO_LINE(DCM_SynchronizedImageBoxList             ) //DcmTagKey(0x0072, 0x0432)
    GET_DICOM_INFO_LINE(DCM_TypeOfSynchronization                ) //DcmTagKey(0x0072, 0x0434)
    GET_DICOM_INFO_LINE(DCM_BlendingOperationType                ) //DcmTagKey(0x0072, 0x0500)
    GET_DICOM_INFO_LINE(DCM_ReformattingOperationType            ) //DcmTagKey(0x0072, 0x0510)
    GET_DICOM_INFO_LINE(DCM_ReformattingThickness                ) //DcmTagKey(0x0072, 0x0512)
    GET_DICOM_INFO_LINE(DCM_ReformattingInterval                 ) //DcmTagKey(0x0072, 0x0514)
    GET_DICOM_INFO_LINE(DCM_ReformattingOperationInitialViewDirection ) //DcmTagKey(0x0072, 0x0516)
    GET_DICOM_INFO_LINE(DCM_ThreeDRenderingType                  ) //DcmTagKey(0x0072, 0x0520)
    GET_DICOM_INFO_LINE(DCM_SortingOperationsSequence            ) //DcmTagKey(0x0072, 0x0600)
    GET_DICOM_INFO_LINE(DCM_SortByCategory                       ) //DcmTagKey(0x0072, 0x0602)
    GET_DICOM_INFO_LINE(DCM_SortingDirection                     ) //DcmTagKey(0x0072, 0x0604)
    GET_DICOM_INFO_LINE(DCM_DisplaySetPatientOrientation         ) //DcmTagKey(0x0072, 0x0700)
    GET_DICOM_INFO_LINE(DCM_VOIType                              ) //DcmTagKey(0x0072, 0x0702)
    GET_DICOM_INFO_LINE(DCM_PseudoColorType                      ) //DcmTagKey(0x0072, 0x0704)
    GET_DICOM_INFO_LINE(DCM_ShowGrayscaleInverted                ) //DcmTagKey(0x0072, 0x0706)
    GET_DICOM_INFO_LINE(DCM_ShowImageTrueSizeFlag                ) //DcmTagKey(0x0072, 0x0710)
    GET_DICOM_INFO_LINE(DCM_ShowGraphicAnnotationFlag            ) //DcmTagKey(0x0072, 0x0712)
    GET_DICOM_INFO_LINE(DCM_ShowPatientDemographicsFlag          ) //DcmTagKey(0x0072, 0x0714)
    GET_DICOM_INFO_LINE(DCM_ShowAcquisitionTechniquesFlag        ) //DcmTagKey(0x0072, 0x0716)
    GET_DICOM_INFO_LINE(DCM_DisplaySetHorizontalJustification    ) //DcmTagKey(0x0072, 0x0717)
    GET_DICOM_INFO_LINE(DCM_DisplaySetVerticalJustification      ) //DcmTagKey(0x0072, 0x0718)
            info->append("</div>");
            info->append("<div style=\"color:#00007F\">");
    GET_DICOM_INFO_LINE(DCM_UnifiedProcedureStepState            ) //DcmTagKey(0x0074, 0x1000)
    GET_DICOM_INFO_LINE(DCM_UnifiedProcedureStepProgressInformationSequence ) //DcmTagKey(0x0074, 0x1002)
    GET_DICOM_INFO_LINE(DCM_UnifiedProcedureStepProgress         ) //DcmTagKey(0x0074, 0x1004)
    GET_DICOM_INFO_LINE(DCM_UnifiedProcedureStepProgressDescription ) //DcmTagKey(0x0074, 0x1006)
    GET_DICOM_INFO_LINE(DCM_UnifiedProcedureStepCommunicationsURISequence ) //DcmTagKey(0x0074, 0x1008)
    GET_DICOM_INFO_LINE(DCM_ContactURI                           ) //DcmTagKey(0x0074, 0x100a)
    GET_DICOM_INFO_LINE(DCM_ContactDisplayName                   ) //DcmTagKey(0x0074, 0x100c)
    GET_DICOM_INFO_LINE(DCM_UnifiedProcedureStepDiscontinuationReasonCodeSequence ) //DcmTagKey(0x0074, 0x100e)
    GET_DICOM_INFO_LINE(DCM_BeamTaskSequence                     ) //DcmTagKey(0x0074, 0x1020)
    GET_DICOM_INFO_LINE(DCM_BeamTaskType                         ) //DcmTagKey(0x0074, 0x1022)
    GET_DICOM_INFO_LINE(DCM_BeamOrderIndex                       ) //DcmTagKey(0x0074, 0x1024)
    GET_DICOM_INFO_LINE(DCM_DeliveryVerificationImageSequence    ) //DcmTagKey(0x0074, 0x1030)
    GET_DICOM_INFO_LINE(DCM_VerificationImageTiming              ) //DcmTagKey(0x0074, 0x1032)
    GET_DICOM_INFO_LINE(DCM_DoubleExposureFlag                   ) //DcmTagKey(0x0074, 0x1034)
    GET_DICOM_INFO_LINE(DCM_DoubleExposureOrdering               ) //DcmTagKey(0x0074, 0x1036)
    GET_DICOM_INFO_LINE(DCM_DoubleExposureMeterset               ) //DcmTagKey(0x0074, 0x1038)
    GET_DICOM_INFO_LINE(DCM_DoubleExposureFieldDelta             ) //DcmTagKey(0x0074, 0x103a)
    GET_DICOM_INFO_LINE(DCM_RelatedReferenceRTImageSequence      ) //DcmTagKey(0x0074, 0x1040)
    GET_DICOM_INFO_LINE(DCM_GeneralMachineVerificationSequence   ) //DcmTagKey(0x0074, 0x1042)
    GET_DICOM_INFO_LINE(DCM_ConventionalMachineVerificationSequence ) //DcmTagKey(0x0074, 0x1044)
    GET_DICOM_INFO_LINE(DCM_IonMachineVerificationSequence       ) //DcmTagKey(0x0074, 0x1046)
    GET_DICOM_INFO_LINE(DCM_FailedAttributesSequence             ) //DcmTagKey(0x0074, 0x1048)
    GET_DICOM_INFO_LINE(DCM_OverriddenAttributesSequence         ) //DcmTagKey(0x0074, 0x104a)
    GET_DICOM_INFO_LINE(DCM_ConventionalControlPointVerificationSequence ) //DcmTagKey(0x0074, 0x104c)
    GET_DICOM_INFO_LINE(DCM_IonControlPointVerificationSequence  ) //DcmTagKey(0x0074, 0x104e)
    GET_DICOM_INFO_LINE(DCM_AttributeOccurrenceSequence          ) //DcmTagKey(0x0074, 0x1050)
    GET_DICOM_INFO_LINE(DCM_AttributeOccurrencePointer           ) //DcmTagKey(0x0074, 0x1052)
    GET_DICOM_INFO_LINE(DCM_AttributeItemSelector                ) //DcmTagKey(0x0074, 0x1054)
    GET_DICOM_INFO_LINE(DCM_AttributeOccurrencePrivateCreator    ) //DcmTagKey(0x0074, 0x1056)
    GET_DICOM_INFO_LINE(DCM_ScheduledProcedureStepPriority       ) //DcmTagKey(0x0074, 0x1200)
    GET_DICOM_INFO_LINE(DCM_WorklistLabel                        ) //DcmTagKey(0x0074, 0x1202)
    GET_DICOM_INFO_LINE(DCM_ProcedureStepLabel                   ) //DcmTagKey(0x0074, 0x1204)
    GET_DICOM_INFO_LINE(DCM_ScheduledProcessingParametersSequence ) //DcmTagKey(0x0074, 0x1210)
    GET_DICOM_INFO_LINE(DCM_PerformedProcessingParametersSequence ) //DcmTagKey(0x0074, 0x1212)
    GET_DICOM_INFO_LINE(DCM_UnifiedProcedureStepPerformedProcedureSequence ) //DcmTagKey(0x0074, 0x1216)
    GET_DICOM_INFO_LINE(DCM_RelatedProcedureStepSequence         ) //DcmTagKey(0x0074, 0x1220)
    GET_DICOM_INFO_LINE(DCM_ProcedureStepRelationshipType        ) //DcmTagKey(0x0074, 0x1222)
    GET_DICOM_INFO_LINE(DCM_DeletionLock                         ) //DcmTagKey(0x0074, 0x1230)
    GET_DICOM_INFO_LINE(DCM_ReceivingAE                          ) //DcmTagKey(0x0074, 0x1234)
    GET_DICOM_INFO_LINE(DCM_RequestingAE                         ) //DcmTagKey(0x0074, 0x1236)
    GET_DICOM_INFO_LINE(DCM_ReasonForCancellation                ) //DcmTagKey(0x0074, 0x1238)
    GET_DICOM_INFO_LINE(DCM_SCPStatus                            ) //DcmTagKey(0x0074, 0x1242)
    GET_DICOM_INFO_LINE(DCM_SubscriptionListStatus               ) //DcmTagKey(0x0074, 0x1244)
    GET_DICOM_INFO_LINE(DCM_UnifiedProcedureStepListStatus       ) //DcmTagKey(0x0074, 0x1246)
            info->append("</div>");
            info->append("<div style=\"color:#00007F\">");
    GET_DICOM_INFO_LINE(DCM_ImplantAssemblyTemplateName          ) //DcmTagKey(0x0076, 0x0001)
    GET_DICOM_INFO_LINE(DCM_ImplantAssemblyTemplateIssuer        ) //DcmTagKey(0x0076, 0x0003)
    GET_DICOM_INFO_LINE(DCM_ImplantAssemblyTemplateVersion       ) //DcmTagKey(0x0076, 0x0006)
    GET_DICOM_INFO_LINE(DCM_ReplacedImplantAssemblyTemplateSequence ) //DcmTagKey(0x0076, 0x0008)
    GET_DICOM_INFO_LINE(DCM_ImplantAssemblyTemplateType          ) //DcmTagKey(0x0076, 0x000a)
    GET_DICOM_INFO_LINE(DCM_OriginalImplantAssemblyTemplateSequence ) //DcmTagKey(0x0076, 0x000c)
    GET_DICOM_INFO_LINE(DCM_DerivationImplantAssemblyTemplateSequence ) //DcmTagKey(0x0076, 0x000e)
    GET_DICOM_INFO_LINE(DCM_ImplantAssemblyTemplateTargetAnatomySequence ) //DcmTagKey(0x0076, 0x0010)
    GET_DICOM_INFO_LINE(DCM_ProcedureTypeCodeSequence            ) //DcmTagKey(0x0076, 0x0020)
    GET_DICOM_INFO_LINE(DCM_SurgicalTechnique                    ) //DcmTagKey(0x0076, 0x0030)
    GET_DICOM_INFO_LINE(DCM_ComponentTypesSequence               ) //DcmTagKey(0x0076, 0x0032)
    GET_DICOM_INFO_LINE(DCM_ComponentTypeCodeSequence            ) //DcmTagKey(0x0076, 0x0034)
    GET_DICOM_INFO_LINE(DCM_ExclusiveComponentType               ) //DcmTagKey(0x0076, 0x0036)
    GET_DICOM_INFO_LINE(DCM_MandatoryComponentType               ) //DcmTagKey(0x0076, 0x0038)
    GET_DICOM_INFO_LINE(DCM_ComponentSequence                    ) //DcmTagKey(0x0076, 0x0040)
    GET_DICOM_INFO_LINE(DCM_ComponentID                          ) //DcmTagKey(0x0076, 0x0055)
    GET_DICOM_INFO_LINE(DCM_ComponentAssemblySequence            ) //DcmTagKey(0x0076, 0x0060)
    GET_DICOM_INFO_LINE(DCM_Component1ReferencedID               ) //DcmTagKey(0x0076, 0x0070)
    GET_DICOM_INFO_LINE(DCM_Component1ReferencedMatingFeatureSetID ) //DcmTagKey(0x0076, 0x0080)
    GET_DICOM_INFO_LINE(DCM_Component1ReferencedMatingFeatureID  ) //DcmTagKey(0x0076, 0x0090)
    GET_DICOM_INFO_LINE(DCM_Component2ReferencedID               ) //DcmTagKey(0x0076, 0x00a0)
    GET_DICOM_INFO_LINE(DCM_Component2ReferencedMatingFeatureSetID ) //DcmTagKey(0x0076, 0x00b0)
    GET_DICOM_INFO_LINE(DCM_Component2ReferencedMatingFeatureID  ) //DcmTagKey(0x0076, 0x00c0)
            info->append("</div>");
            info->append("<div style=\"color:#00007F\">");
    GET_DICOM_INFO_LINE(DCM_ImplantTemplateGroupName             ) //DcmTagKey(0x0078, 0x0000)
    GET_DICOM_INFO_LINE(DCM_ImplantTemplateGroupDescription      ) //DcmTagKey(0x0078, 0x0010)
    GET_DICOM_INFO_LINE(DCM_ImplantTemplateGroupIssuer           ) //DcmTagKey(0x0078, 0x0020)
    GET_DICOM_INFO_LINE(DCM_ImplantTemplateGroupVersion          ) //DcmTagKey(0x0078, 0x0024)
    GET_DICOM_INFO_LINE(DCM_ReplacedImplantTemplateGroupSequence ) //DcmTagKey(0x0078, 0x0026)
    GET_DICOM_INFO_LINE(DCM_ImplantTemplateGroupTargetAnatomySequence ) //DcmTagKey(0x0078, 0x0028)
    GET_DICOM_INFO_LINE(DCM_ImplantTemplateGroupMembersSequence  ) //DcmTagKey(0x0078, 0x002a)
    GET_DICOM_INFO_LINE(DCM_ImplantTemplateGroupMemberID         ) //DcmTagKey(0x0078, 0x002e)
    GET_DICOM_INFO_LINE(DCM_ThreeDImplantTemplateGroupMemberMatchingPoint ) //DcmTagKey(0x0078, 0x0050)
    GET_DICOM_INFO_LINE(DCM_ThreeDImplantTemplateGroupMemberMatchingAxes ) //DcmTagKey(0x0078, 0x0060)
    GET_DICOM_INFO_LINE(DCM_ImplantTemplateGroupMemberMatching2DCoordinatesSequence ) //DcmTagKey(0x0078, 0x0070)
    GET_DICOM_INFO_LINE(DCM_TwoDImplantTemplateGroupMemberMatchingPoint ) //DcmTagKey(0x0078, 0x0090)
    GET_DICOM_INFO_LINE(DCM_TwoDImplantTemplateGroupMemberMatchingAxes ) //DcmTagKey(0x0078, 0x00a0)
    GET_DICOM_INFO_LINE(DCM_ImplantTemplateGroupVariationDimensionSequence ) //DcmTagKey(0x0078, 0x00b0)
    GET_DICOM_INFO_LINE(DCM_ImplantTemplateGroupVariationDimensionName ) //DcmTagKey(0x0078, 0x00b2)
    GET_DICOM_INFO_LINE(DCM_ImplantTemplateGroupVariationDimensionRankSequence ) //DcmTagKey(0x0078, 0x00b4)
    GET_DICOM_INFO_LINE(DCM_ReferencedImplantTemplateGroupMemberID ) //DcmTagKey(0x0078, 0x00b6)
    GET_DICOM_INFO_LINE(DCM_ImplantTemplateGroupVariationDimensionRank ) //DcmTagKey(0x0078, 0x00b8)
            info->append("</div>");
            info->append("<div style=\"color:#00007F\">");
    GET_DICOM_INFO_LINE(DCM_StorageMediaFileSetID                ) //DcmTagKey(0x0088, 0x0130)
    GET_DICOM_INFO_LINE(DCM_StorageMediaFileSetUID               ) //DcmTagKey(0x0088, 0x0140)
    GET_DICOM_INFO_LINE(DCM_IconImageSequence                    ) //DcmTagKey(0x0088, 0x0200)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_TopicTitle                   ) //DcmTagKey(0x0088, 0x0904)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_TopicSubject                 ) //DcmTagKey(0x0088, 0x0906)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_TopicAuthor                  ) //DcmTagKey(0x0088, 0x0910)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_TopicKeywords                ) //DcmTagKey(0x0088, 0x0912)
            info->append("</div>");
            info->append("<div style=\"color:#00007F\">");
    GET_DICOM_INFO_LINE(DCM_SOPInstanceStatus                    ) //DcmTagKey(0x0100, 0x0410)
    GET_DICOM_INFO_LINE(DCM_SOPAuthorizationDateTime             ) //DcmTagKey(0x0100, 0x0420)
    GET_DICOM_INFO_LINE(DCM_SOPAuthorizationComment              ) //DcmTagKey(0x0100, 0x0424)
    GET_DICOM_INFO_LINE(DCM_AuthorizationEquipmentCertificationNumber ) //DcmTagKey(0x0100, 0x0426)
            info->append("</div>");
            info->append("<div style=\"color:#00007F\">");
    GET_DICOM_INFO_LINE(DCM_MACIDNumber                          ) //DcmTagKey(0x0400, 0x0005)
    GET_DICOM_INFO_LINE(DCM_MACCalculationTransferSyntaxUID      ) //DcmTagKey(0x0400, 0x0010)
    GET_DICOM_INFO_LINE(DCM_MACAlgorithm                         ) //DcmTagKey(0x0400, 0x0015)
    GET_DICOM_INFO_LINE(DCM_DataElementsSigned                   ) //DcmTagKey(0x0400, 0x0020)
    GET_DICOM_INFO_LINE(DCM_DigitalSignatureUID                  ) //DcmTagKey(0x0400, 0x0100)
    GET_DICOM_INFO_LINE(DCM_DigitalSignatureDateTime             ) //DcmTagKey(0x0400, 0x0105)
    GET_DICOM_INFO_LINE(DCM_CertificateType                      ) //DcmTagKey(0x0400, 0x0110)
    GET_DICOM_INFO_LINE(DCM_CertificateOfSigner                  ) //DcmTagKey(0x0400, 0x0115)
    GET_DICOM_INFO_LINE(DCM_Signature                            ) //DcmTagKey(0x0400, 0x0120)
    GET_DICOM_INFO_LINE(DCM_CertifiedTimestampType               ) //DcmTagKey(0x0400, 0x0305)
    GET_DICOM_INFO_LINE(DCM_CertifiedTimestamp                   ) //DcmTagKey(0x0400, 0x0310)
    GET_DICOM_INFO_LINE(DCM_DigitalSignaturePurposeCodeSequence  ) //DcmTagKey(0x0400, 0x0401)
    GET_DICOM_INFO_LINE(DCM_ReferencedDigitalSignatureSequence   ) //DcmTagKey(0x0400, 0x0402)
    GET_DICOM_INFO_LINE(DCM_ReferencedSOPInstanceMACSequence     ) //DcmTagKey(0x0400, 0x0403)
    GET_DICOM_INFO_LINE(DCM_MAC                                  ) //DcmTagKey(0x0400, 0x0404)
    GET_DICOM_INFO_LINE(DCM_EncryptedAttributesSequence          ) //DcmTagKey(0x0400, 0x0500)
    GET_DICOM_INFO_LINE(DCM_EncryptedContentTransferSyntaxUID    ) //DcmTagKey(0x0400, 0x0510)
    GET_DICOM_INFO_LINE(DCM_EncryptedContent                     ) //DcmTagKey(0x0400, 0x0520)
    GET_DICOM_INFO_LINE(DCM_ModifiedAttributesSequence           ) //DcmTagKey(0x0400, 0x0550)
    GET_DICOM_INFO_LINE(DCM_OriginalAttributesSequence           ) //DcmTagKey(0x0400, 0x0561)
    GET_DICOM_INFO_LINE(DCM_AttributeModificationDateTime        ) //DcmTagKey(0x0400, 0x0562)
    GET_DICOM_INFO_LINE(DCM_ModifyingSystem                      ) //DcmTagKey(0x0400, 0x0563)
    GET_DICOM_INFO_LINE(DCM_SourceOfPreviousValues               ) //DcmTagKey(0x0400, 0x0564)
    GET_DICOM_INFO_LINE(DCM_ReasonForTheAttributeModification    ) //DcmTagKey(0x0400, 0x0565)
            info->append("</div>");
            info->append("<div style=\"color:#00007F\">");
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_CodeTableGroupLength     ) //DcmTagKey(0x1000, 0x0000)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_EscapeTriplet            ) //DcmTagKey(0x1000, 0x0010)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_RunLengthTriplet         ) //DcmTagKey(0x1000, 0x0011)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_HuffmanTableSize         ) //DcmTagKey(0x1000, 0x0012)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_HuffmanTableTriplet      ) //DcmTagKey(0x1000, 0x0013)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_ShiftTableSize           ) //DcmTagKey(0x1000, 0x0014)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_ShiftTableTriplet        ) //DcmTagKey(0x1000, 0x0015)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_ZonalMapGroupLength      ) //DcmTagKey(0x1010, 0x0000)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_ZonalMap                 ) //DcmTagKey(0x1010, 0x0004)
    GET_DICOM_INFO_LINE(DCM_NumberOfCopies                       ) //DcmTagKey(0x2000, 0x0010)
    GET_DICOM_INFO_LINE(DCM_PrinterConfigurationSequence         ) //DcmTagKey(0x2000, 0x001e)
    GET_DICOM_INFO_LINE(DCM_PrintPriority                        ) //DcmTagKey(0x2000, 0x0020)
    GET_DICOM_INFO_LINE(DCM_MediumType                           ) //DcmTagKey(0x2000, 0x0030)
    GET_DICOM_INFO_LINE(DCM_FilmDestination                      ) //DcmTagKey(0x2000, 0x0040)
    GET_DICOM_INFO_LINE(DCM_FilmSessionLabel                     ) //DcmTagKey(0x2000, 0x0050)
    GET_DICOM_INFO_LINE(DCM_MemoryAllocation                     ) //DcmTagKey(0x2000, 0x0060)
    GET_DICOM_INFO_LINE(DCM_MaximumMemoryAllocation              ) //DcmTagKey(0x2000, 0x0061)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ColorImagePrintingFlag       ) //DcmTagKey(0x2000, 0x0062)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_CollationFlag                ) //DcmTagKey(0x2000, 0x0063)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_AnnotationFlag               ) //DcmTagKey(0x2000, 0x0065)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ImageOverlayFlag             ) //DcmTagKey(0x2000, 0x0067)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_PresentationLUTFlag          ) //DcmTagKey(0x2000, 0x0069)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ImageBoxPresentationLUTFlag  ) //DcmTagKey(0x2000, 0x006a)
    GET_DICOM_INFO_LINE(DCM_MemoryBitDepth                       ) //DcmTagKey(0x2000, 0x00a0)
    GET_DICOM_INFO_LINE(DCM_PrintingBitDepth                     ) //DcmTagKey(0x2000, 0x00a1)
    GET_DICOM_INFO_LINE(DCM_MediaInstalledSequence               ) //DcmTagKey(0x2000, 0x00a2)
    GET_DICOM_INFO_LINE(DCM_OtherMediaAvailableSequence          ) //DcmTagKey(0x2000, 0x00a4)
    GET_DICOM_INFO_LINE(DCM_SupportedImageDisplayFormatsSequence ) //DcmTagKey(0x2000, 0x00a8)
    GET_DICOM_INFO_LINE(DCM_ReferencedFilmBoxSequence            ) //DcmTagKey(0x2000, 0x0500)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ReferencedStoredPrintSequence ) //DcmTagKey(0x2000, 0x0510)
    GET_DICOM_INFO_LINE(DCM_ImageDisplayFormat                   ) //DcmTagKey(0x2010, 0x0010)
    GET_DICOM_INFO_LINE(DCM_AnnotationDisplayFormatID            ) //DcmTagKey(0x2010, 0x0030)
    GET_DICOM_INFO_LINE(DCM_FilmOrientation                      ) //DcmTagKey(0x2010, 0x0040)
    GET_DICOM_INFO_LINE(DCM_FilmSizeID                           ) //DcmTagKey(0x2010, 0x0050)
    GET_DICOM_INFO_LINE(DCM_PrinterResolutionID                  ) //DcmTagKey(0x2010, 0x0052)
    GET_DICOM_INFO_LINE(DCM_DefaultPrinterResolutionID           ) //DcmTagKey(0x2010, 0x0054)
    GET_DICOM_INFO_LINE(DCM_MagnificationType                    ) //DcmTagKey(0x2010, 0x0060)
    GET_DICOM_INFO_LINE(DCM_SmoothingType                        ) //DcmTagKey(0x2010, 0x0080)
    GET_DICOM_INFO_LINE(DCM_DefaultMagnificationType             ) //DcmTagKey(0x2010, 0x00a6)
    GET_DICOM_INFO_LINE(DCM_OtherMagnificationTypesAvailable     ) //DcmTagKey(0x2010, 0x00a7)
    GET_DICOM_INFO_LINE(DCM_DefaultSmoothingType                 ) //DcmTagKey(0x2010, 0x00a8)
    GET_DICOM_INFO_LINE(DCM_OtherSmoothingTypesAvailable         ) //DcmTagKey(0x2010, 0x00a9)
    GET_DICOM_INFO_LINE(DCM_BorderDensity                        ) //DcmTagKey(0x2010, 0x0100)
    GET_DICOM_INFO_LINE(DCM_EmptyImageDensity                    ) //DcmTagKey(0x2010, 0x0110)
    GET_DICOM_INFO_LINE(DCM_MinDensity                           ) //DcmTagKey(0x2010, 0x0120)
    GET_DICOM_INFO_LINE(DCM_MaxDensity                           ) //DcmTagKey(0x2010, 0x0130)
    GET_DICOM_INFO_LINE(DCM_Trim                                 ) //DcmTagKey(0x2010, 0x0140)
    GET_DICOM_INFO_LINE(DCM_ConfigurationInformation             ) //DcmTagKey(0x2010, 0x0150)
    GET_DICOM_INFO_LINE(DCM_ConfigurationInformationDescription  ) //DcmTagKey(0x2010, 0x0152)
    GET_DICOM_INFO_LINE(DCM_MaximumCollatedFilms                 ) //DcmTagKey(0x2010, 0x0154)
    GET_DICOM_INFO_LINE(DCM_Illumination                         ) //DcmTagKey(0x2010, 0x015e)
    GET_DICOM_INFO_LINE(DCM_ReflectedAmbientLight                ) //DcmTagKey(0x2010, 0x0160)
    GET_DICOM_INFO_LINE(DCM_PrinterPixelSpacing                  ) //DcmTagKey(0x2010, 0x0376)
    GET_DICOM_INFO_LINE(DCM_ReferencedFilmSessionSequence        ) //DcmTagKey(0x2010, 0x0500)
    GET_DICOM_INFO_LINE(DCM_ReferencedImageBoxSequence           ) //DcmTagKey(0x2010, 0x0510)
    GET_DICOM_INFO_LINE(DCM_ReferencedBasicAnnotationBoxSequence ) //DcmTagKey(0x2010, 0x0520)
    GET_DICOM_INFO_LINE(DCM_ImageBoxPosition                     ) //DcmTagKey(0x2020, 0x0010)
    GET_DICOM_INFO_LINE(DCM_Polarity                             ) //DcmTagKey(0x2020, 0x0020)
    GET_DICOM_INFO_LINE(DCM_RequestedImageSize                   ) //DcmTagKey(0x2020, 0x0030)
    GET_DICOM_INFO_LINE(DCM_RequestedDecimateCropBehavior        ) //DcmTagKey(0x2020, 0x0040)
    GET_DICOM_INFO_LINE(DCM_RequestedResolutionID                ) //DcmTagKey(0x2020, 0x0050)
    GET_DICOM_INFO_LINE(DCM_RequestedImageSizeFlag               ) //DcmTagKey(0x2020, 0x00a0)
    GET_DICOM_INFO_LINE(DCM_DecimateCropResult                   ) //DcmTagKey(0x2020, 0x00a2)
    GET_DICOM_INFO_LINE(DCM_BasicGrayscaleImageSequence          ) //DcmTagKey(0x2020, 0x0110)
    GET_DICOM_INFO_LINE(DCM_BasicColorImageSequence              ) //DcmTagKey(0x2020, 0x0111)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ReferencedImageOverlayBoxSequence ) //DcmTagKey(0x2020, 0x0130)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ReferencedVOILUTBoxSequence  ) //DcmTagKey(0x2020, 0x0140)
    GET_DICOM_INFO_LINE(DCM_AnnotationPosition                   ) //DcmTagKey(0x2030, 0x0010)
    GET_DICOM_INFO_LINE(DCM_TextString                           ) //DcmTagKey(0x2030, 0x0020)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ReferencedOverlayPlaneSequence ) //DcmTagKey(0x2040, 0x0010)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ReferencedOverlayPlaneGroups ) //DcmTagKey(0x2040, 0x0011)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_OverlayPixelDataSequence     ) //DcmTagKey(0x2040, 0x0020)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_OverlayMagnificationType     ) //DcmTagKey(0x2040, 0x0060)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_OverlaySmoothingType         ) //DcmTagKey(0x2040, 0x0070)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_OverlayOrImageMagnification  ) //DcmTagKey(0x2040, 0x0072)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_MagnifyToNumberOfColumns     ) //DcmTagKey(0x2040, 0x0074)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_OverlayForegroundDensity     ) //DcmTagKey(0x2040, 0x0080)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_OverlayBackgroundDensity     ) //DcmTagKey(0x2040, 0x0082)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_OverlayMode                  ) //DcmTagKey(0x2040, 0x0090)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ThresholdDensity             ) //DcmTagKey(0x2040, 0x0100)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ReferencedImageBoxSequenceRetired ) //DcmTagKey(0x2040, 0x0500)
    GET_DICOM_INFO_LINE(DCM_PresentationLUTSequence              ) //DcmTagKey(0x2050, 0x0010)
    GET_DICOM_INFO_LINE(DCM_PresentationLUTShape                 ) //DcmTagKey(0x2050, 0x0020)
    GET_DICOM_INFO_LINE(DCM_ReferencedPresentationLUTSequence    ) //DcmTagKey(0x2050, 0x0500)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_PrintJobID                   ) //DcmTagKey(0x2100, 0x0010)
    GET_DICOM_INFO_LINE(DCM_ExecutionStatus                      ) //DcmTagKey(0x2100, 0x0020)
    GET_DICOM_INFO_LINE(DCM_ExecutionStatusInfo                  ) //DcmTagKey(0x2100, 0x0030)
    GET_DICOM_INFO_LINE(DCM_CreationDate                         ) //DcmTagKey(0x2100, 0x0040)
    GET_DICOM_INFO_LINE(DCM_CreationTime                         ) //DcmTagKey(0x2100, 0x0050)
    GET_DICOM_INFO_LINE(DCM_Originator                           ) //DcmTagKey(0x2100, 0x0070)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_DestinationAE                ) //DcmTagKey(0x2100, 0x0140)
    GET_DICOM_INFO_LINE(DCM_OwnerID                              ) //DcmTagKey(0x2100, 0x0160)
    GET_DICOM_INFO_LINE(DCM_NumberOfFilms                        ) //DcmTagKey(0x2100, 0x0170)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ReferencedPrintJobSequencePullStoredPrint ) //DcmTagKey(0x2100, 0x0500)
    GET_DICOM_INFO_LINE(DCM_PrinterStatus                        ) //DcmTagKey(0x2110, 0x0010)
    GET_DICOM_INFO_LINE(DCM_PrinterStatusInfo                    ) //DcmTagKey(0x2110, 0x0020)
    GET_DICOM_INFO_LINE(DCM_PrinterName                          ) //DcmTagKey(0x2110, 0x0030)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_PrintQueueID                 ) //DcmTagKey(0x2110, 0x0099)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_QueueStatus                  ) //DcmTagKey(0x2120, 0x0010)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_PrintJobDescriptionSequence  ) //DcmTagKey(0x2120, 0x0050)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ReferencedPrintJobSequence   ) //DcmTagKey(0x2120, 0x0070)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_PrintManagementCapabilitiesSequence ) //DcmTagKey(0x2130, 0x0010)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_PrinterCharacteristicsSequence ) //DcmTagKey(0x2130, 0x0015)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_FilmBoxContentSequence       ) //DcmTagKey(0x2130, 0x0030)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ImageBoxContentSequence      ) //DcmTagKey(0x2130, 0x0040)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_AnnotationContentSequence    ) //DcmTagKey(0x2130, 0x0050)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ImageOverlayBoxContentSequence ) //DcmTagKey(0x2130, 0x0060)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_PresentationLUTContentSequence ) //DcmTagKey(0x2130, 0x0080)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ProposedStudySequence        ) //DcmTagKey(0x2130, 0x00a0)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_OriginalImageSequence        ) //DcmTagKey(0x2130, 0x00c0)
    GET_DICOM_INFO_LINE(DCM_LabelUsingInformationExtractedFromInstances ) //DcmTagKey(0x2200, 0x0001)
    GET_DICOM_INFO_LINE(DCM_LabelText                            ) //DcmTagKey(0x2200, 0x0002)
    GET_DICOM_INFO_LINE(DCM_LabelStyleSelection                  ) //DcmTagKey(0x2200, 0x0003)
    GET_DICOM_INFO_LINE(DCM_MediaDisposition                     ) //DcmTagKey(0x2200, 0x0004)
    GET_DICOM_INFO_LINE(DCM_BarcodeValue                         ) //DcmTagKey(0x2200, 0x0005)
    GET_DICOM_INFO_LINE(DCM_BarcodeSymbology                     ) //DcmTagKey(0x2200, 0x0006)
    GET_DICOM_INFO_LINE(DCM_AllowMediaSplitting                  ) //DcmTagKey(0x2200, 0x0007)
    GET_DICOM_INFO_LINE(DCM_IncludeNonDICOMObjects               ) //DcmTagKey(0x2200, 0x0008)
    GET_DICOM_INFO_LINE(DCM_IncludeDisplayApplication            ) //DcmTagKey(0x2200, 0x0009)
    GET_DICOM_INFO_LINE(DCM_PreserveCompositeInstancesAfterMediaCreation ) //DcmTagKey(0x2200, 0x000a)
    GET_DICOM_INFO_LINE(DCM_TotalNumberOfPiecesOfMediaCreated    ) //DcmTagKey(0x2200, 0x000b)
    GET_DICOM_INFO_LINE(DCM_RequestedMediaApplicationProfile     ) //DcmTagKey(0x2200, 0x000c)
    GET_DICOM_INFO_LINE(DCM_ReferencedStorageMediaSequence       ) //DcmTagKey(0x2200, 0x000d)
    GET_DICOM_INFO_LINE(DCM_FailureAttributes                    ) //DcmTagKey(0x2200, 0x000e)
    GET_DICOM_INFO_LINE(DCM_AllowLossyCompression                ) //DcmTagKey(0x2200, 0x000f)
    GET_DICOM_INFO_LINE(DCM_RequestPriority                      ) //DcmTagKey(0x2200, 0x0020)
            info->append("</div>");
            info->append("<div style=\"color:#00007F\">");
    GET_DICOM_INFO_LINE(DCM_RTImageLabel                         ) //DcmTagKey(0x3002, 0x0002)
    GET_DICOM_INFO_LINE(DCM_RTImageName                          ) //DcmTagKey(0x3002, 0x0003)
    GET_DICOM_INFO_LINE(DCM_RTImageDescription                   ) //DcmTagKey(0x3002, 0x0004)
    GET_DICOM_INFO_LINE(DCM_ReportedValuesOrigin                 ) //DcmTagKey(0x3002, 0x000a)
    GET_DICOM_INFO_LINE(DCM_RTImagePlane                         ) //DcmTagKey(0x3002, 0x000c)
    GET_DICOM_INFO_LINE(DCM_XRayImageReceptorTranslation         ) //DcmTagKey(0x3002, 0x000d)
    GET_DICOM_INFO_LINE(DCM_XRayImageReceptorAngle               ) //DcmTagKey(0x3002, 0x000e)
    GET_DICOM_INFO_LINE(DCM_RTImageOrientation                   ) //DcmTagKey(0x3002, 0x0010)
    GET_DICOM_INFO_LINE(DCM_ImagePlanePixelSpacing               ) //DcmTagKey(0x3002, 0x0011)
    GET_DICOM_INFO_LINE(DCM_RTImagePosition                      ) //DcmTagKey(0x3002, 0x0012)
    GET_DICOM_INFO_LINE(DCM_RadiationMachineName                 ) //DcmTagKey(0x3002, 0x0020)
    GET_DICOM_INFO_LINE(DCM_RadiationMachineSAD                  ) //DcmTagKey(0x3002, 0x0022)
    GET_DICOM_INFO_LINE(DCM_RadiationMachineSSD                  ) //DcmTagKey(0x3002, 0x0024)
    GET_DICOM_INFO_LINE(DCM_RTImageSID                           ) //DcmTagKey(0x3002, 0x0026)
    GET_DICOM_INFO_LINE(DCM_SourceToReferenceObjectDistance      ) //DcmTagKey(0x3002, 0x0028)
    GET_DICOM_INFO_LINE(DCM_FractionNumber                       ) //DcmTagKey(0x3002, 0x0029)
    GET_DICOM_INFO_LINE(DCM_ExposureSequence                     ) //DcmTagKey(0x3002, 0x0030)
    GET_DICOM_INFO_LINE(DCM_MetersetExposure                     ) //DcmTagKey(0x3002, 0x0032)
    GET_DICOM_INFO_LINE(DCM_DiaphragmPosition                    ) //DcmTagKey(0x3002, 0x0034)
    GET_DICOM_INFO_LINE(DCM_FluenceMapSequence                   ) //DcmTagKey(0x3002, 0x0040)
    GET_DICOM_INFO_LINE(DCM_FluenceDataSource                    ) //DcmTagKey(0x3002, 0x0041)
    GET_DICOM_INFO_LINE(DCM_FluenceDataScale                     ) //DcmTagKey(0x3002, 0x0042)
    GET_DICOM_INFO_LINE(DCM_PrimaryFluenceModeSequence           ) //DcmTagKey(0x3002, 0x0050)
    GET_DICOM_INFO_LINE(DCM_FluenceMode                          ) //DcmTagKey(0x3002, 0x0051)
    GET_DICOM_INFO_LINE(DCM_FluenceModeID                        ) //DcmTagKey(0x3002, 0x0052)
    GET_DICOM_INFO_LINE(DCM_DVHType                              ) //DcmTagKey(0x3004, 0x0001)
    GET_DICOM_INFO_LINE(DCM_DoseUnits                            ) //DcmTagKey(0x3004, 0x0002)
    GET_DICOM_INFO_LINE(DCM_DoseType                             ) //DcmTagKey(0x3004, 0x0004)
    GET_DICOM_INFO_LINE(DCM_DoseComment                          ) //DcmTagKey(0x3004, 0x0006)
    GET_DICOM_INFO_LINE(DCM_NormalizationPoint                   ) //DcmTagKey(0x3004, 0x0008)
    GET_DICOM_INFO_LINE(DCM_DoseSummationType                    ) //DcmTagKey(0x3004, 0x000a)
    GET_DICOM_INFO_LINE(DCM_GridFrameOffsetVector                ) //DcmTagKey(0x3004, 0x000c)
    GET_DICOM_INFO_LINE(DCM_DoseGridScaling                      ) //DcmTagKey(0x3004, 0x000e)
    GET_DICOM_INFO_LINE(DCM_RTDoseROISequence                    ) //DcmTagKey(0x3004, 0x0010)
    GET_DICOM_INFO_LINE(DCM_DoseValue                            ) //DcmTagKey(0x3004, 0x0012)
    GET_DICOM_INFO_LINE(DCM_TissueHeterogeneityCorrection        ) //DcmTagKey(0x3004, 0x0014)
    GET_DICOM_INFO_LINE(DCM_DVHNormalizationPoint                ) //DcmTagKey(0x3004, 0x0040)
    GET_DICOM_INFO_LINE(DCM_DVHNormalizationDoseValue            ) //DcmTagKey(0x3004, 0x0042)
    GET_DICOM_INFO_LINE(DCM_DVHSequence                          ) //DcmTagKey(0x3004, 0x0050)
    GET_DICOM_INFO_LINE(DCM_DVHDoseScaling                       ) //DcmTagKey(0x3004, 0x0052)
    GET_DICOM_INFO_LINE(DCM_DVHVolumeUnits                       ) //DcmTagKey(0x3004, 0x0054)
    GET_DICOM_INFO_LINE(DCM_DVHNumberOfBins                      ) //DcmTagKey(0x3004, 0x0056)
    GET_DICOM_INFO_LINE(DCM_DVHData                              ) //DcmTagKey(0x3004, 0x0058)
    GET_DICOM_INFO_LINE(DCM_DVHReferencedROISequence             ) //DcmTagKey(0x3004, 0x0060)
    GET_DICOM_INFO_LINE(DCM_DVHROIContributionType               ) //DcmTagKey(0x3004, 0x0062)
    GET_DICOM_INFO_LINE(DCM_DVHMinimumDose                       ) //DcmTagKey(0x3004, 0x0070)
    GET_DICOM_INFO_LINE(DCM_DVHMaximumDose                       ) //DcmTagKey(0x3004, 0x0072)
    GET_DICOM_INFO_LINE(DCM_DVHMeanDose                          ) //DcmTagKey(0x3004, 0x0074)
    GET_DICOM_INFO_LINE(DCM_StructureSetLabel                    ) //DcmTagKey(0x3006, 0x0002)
    GET_DICOM_INFO_LINE(DCM_StructureSetName                     ) //DcmTagKey(0x3006, 0x0004)
    GET_DICOM_INFO_LINE(DCM_StructureSetDescription              ) //DcmTagKey(0x3006, 0x0006)
    GET_DICOM_INFO_LINE(DCM_StructureSetDate                     ) //DcmTagKey(0x3006, 0x0008)
    GET_DICOM_INFO_LINE(DCM_StructureSetTime                     ) //DcmTagKey(0x3006, 0x0009)
    GET_DICOM_INFO_LINE(DCM_ReferencedFrameOfReferenceSequence   ) //DcmTagKey(0x3006, 0x0010)
    GET_DICOM_INFO_LINE(DCM_RTReferencedStudySequence            ) //DcmTagKey(0x3006, 0x0012)
    GET_DICOM_INFO_LINE(DCM_RTReferencedSeriesSequence           ) //DcmTagKey(0x3006, 0x0014)
    GET_DICOM_INFO_LINE(DCM_ContourImageSequence                 ) //DcmTagKey(0x3006, 0x0016)
    GET_DICOM_INFO_LINE(DCM_StructureSetROISequence              ) //DcmTagKey(0x3006, 0x0020)
    GET_DICOM_INFO_LINE(DCM_ROINumber                            ) //DcmTagKey(0x3006, 0x0022)
    GET_DICOM_INFO_LINE(DCM_ReferencedFrameOfReferenceUID        ) //DcmTagKey(0x3006, 0x0024)
    GET_DICOM_INFO_LINE(DCM_ROIName                              ) //DcmTagKey(0x3006, 0x0026)
    GET_DICOM_INFO_LINE(DCM_ROIDescription                       ) //DcmTagKey(0x3006, 0x0028)
    GET_DICOM_INFO_LINE(DCM_ROIDisplayColor                      ) //DcmTagKey(0x3006, 0x002a)
    GET_DICOM_INFO_LINE(DCM_ROIVolume                            ) //DcmTagKey(0x3006, 0x002c)
    GET_DICOM_INFO_LINE(DCM_RTRelatedROISequence                 ) //DcmTagKey(0x3006, 0x0030)
    GET_DICOM_INFO_LINE(DCM_RTROIRelationship                    ) //DcmTagKey(0x3006, 0x0033)
    GET_DICOM_INFO_LINE(DCM_ROIGenerationAlgorithm               ) //DcmTagKey(0x3006, 0x0036)
    GET_DICOM_INFO_LINE(DCM_ROIGenerationDescription             ) //DcmTagKey(0x3006, 0x0038)
    GET_DICOM_INFO_LINE(DCM_ROIContourSequence                   ) //DcmTagKey(0x3006, 0x0039)
    GET_DICOM_INFO_LINE(DCM_ContourSequence                      ) //DcmTagKey(0x3006, 0x0040)
    GET_DICOM_INFO_LINE(DCM_ContourGeometricType                 ) //DcmTagKey(0x3006, 0x0042)
    GET_DICOM_INFO_LINE(DCM_ContourSlabThickness                 ) //DcmTagKey(0x3006, 0x0044)
    GET_DICOM_INFO_LINE(DCM_ContourOffsetVector                  ) //DcmTagKey(0x3006, 0x0045)
    GET_DICOM_INFO_LINE(DCM_NumberOfContourPoints                ) //DcmTagKey(0x3006, 0x0046)
    GET_DICOM_INFO_LINE(DCM_ContourNumber                        ) //DcmTagKey(0x3006, 0x0048)
    GET_DICOM_INFO_LINE(DCM_AttachedContours                     ) //DcmTagKey(0x3006, 0x0049)
    GET_DICOM_INFO_LINE(DCM_ContourData                          ) //DcmTagKey(0x3006, 0x0050)
    GET_DICOM_INFO_LINE(DCM_RTROIObservationsSequence            ) //DcmTagKey(0x3006, 0x0080)
    GET_DICOM_INFO_LINE(DCM_ObservationNumber                    ) //DcmTagKey(0x3006, 0x0082)
    GET_DICOM_INFO_LINE(DCM_ReferencedROINumber                  ) //DcmTagKey(0x3006, 0x0084)
    GET_DICOM_INFO_LINE(DCM_ROIObservationLabel                  ) //DcmTagKey(0x3006, 0x0085)
    GET_DICOM_INFO_LINE(DCM_RTROIIdentificationCodeSequence      ) //DcmTagKey(0x3006, 0x0086)
    GET_DICOM_INFO_LINE(DCM_ROIObservationDescription            ) //DcmTagKey(0x3006, 0x0088)
    GET_DICOM_INFO_LINE(DCM_RelatedRTROIObservationsSequence     ) //DcmTagKey(0x3006, 0x00a0)
    GET_DICOM_INFO_LINE(DCM_RTROIInterpretedType                 ) //DcmTagKey(0x3006, 0x00a4)
    GET_DICOM_INFO_LINE(DCM_ROIInterpreter                       ) //DcmTagKey(0x3006, 0x00a6)
    GET_DICOM_INFO_LINE(DCM_ROIPhysicalPropertiesSequence        ) //DcmTagKey(0x3006, 0x00b0)
    GET_DICOM_INFO_LINE(DCM_ROIPhysicalProperty                  ) //DcmTagKey(0x3006, 0x00b2)
    GET_DICOM_INFO_LINE(DCM_ROIPhysicalPropertyValue             ) //DcmTagKey(0x3006, 0x00b4)
    GET_DICOM_INFO_LINE(DCM_ROIElementalCompositionSequence      ) //DcmTagKey(0x3006, 0x00b6)
    GET_DICOM_INFO_LINE(DCM_ROIElementalCompositionAtomicNumber  ) //DcmTagKey(0x3006, 0x00b7)
    GET_DICOM_INFO_LINE(DCM_ROIElementalCompositionAtomicMassFraction ) //DcmTagKey(0x3006, 0x00b8)
    GET_DICOM_INFO_LINE(DCM_FrameOfReferenceRelationshipSequence ) //DcmTagKey(0x3006, 0x00c0)
    GET_DICOM_INFO_LINE(DCM_RelatedFrameOfReferenceUID           ) //DcmTagKey(0x3006, 0x00c2)
    GET_DICOM_INFO_LINE(DCM_FrameOfReferenceTransformationType   ) //DcmTagKey(0x3006, 0x00c4)
    GET_DICOM_INFO_LINE(DCM_FrameOfReferenceTransformationMatrix ) //DcmTagKey(0x3006, 0x00c6)
    GET_DICOM_INFO_LINE(DCM_FrameOfReferenceTransformationComment ) //DcmTagKey(0x3006, 0x00c8)
    GET_DICOM_INFO_LINE(DCM_MeasuredDoseReferenceSequence        ) //DcmTagKey(0x3008, 0x0010)
    GET_DICOM_INFO_LINE(DCM_MeasuredDoseDescription              ) //DcmTagKey(0x3008, 0x0012)
    GET_DICOM_INFO_LINE(DCM_MeasuredDoseType                     ) //DcmTagKey(0x3008, 0x0014)
    GET_DICOM_INFO_LINE(DCM_MeasuredDoseValue                    ) //DcmTagKey(0x3008, 0x0016)
    GET_DICOM_INFO_LINE(DCM_TreatmentSessionBeamSequence         ) //DcmTagKey(0x3008, 0x0020)
    GET_DICOM_INFO_LINE(DCM_TreatmentSessionIonBeamSequence      ) //DcmTagKey(0x3008, 0x0021)
    GET_DICOM_INFO_LINE(DCM_CurrentFractionNumber                ) //DcmTagKey(0x3008, 0x0022)
    GET_DICOM_INFO_LINE(DCM_TreatmentControlPointDate            ) //DcmTagKey(0x3008, 0x0024)
    GET_DICOM_INFO_LINE(DCM_TreatmentControlPointTime            ) //DcmTagKey(0x3008, 0x0025)
    GET_DICOM_INFO_LINE(DCM_TreatmentTerminationStatus           ) //DcmTagKey(0x3008, 0x002a)
    GET_DICOM_INFO_LINE(DCM_TreatmentTerminationCode             ) //DcmTagKey(0x3008, 0x002b)
    GET_DICOM_INFO_LINE(DCM_TreatmentVerificationStatus          ) //DcmTagKey(0x3008, 0x002c)
    GET_DICOM_INFO_LINE(DCM_ReferencedTreatmentRecordSequence    ) //DcmTagKey(0x3008, 0x0030)
    GET_DICOM_INFO_LINE(DCM_SpecifiedPrimaryMeterset             ) //DcmTagKey(0x3008, 0x0032)
    GET_DICOM_INFO_LINE(DCM_SpecifiedSecondaryMeterset           ) //DcmTagKey(0x3008, 0x0033)
    GET_DICOM_INFO_LINE(DCM_DeliveredPrimaryMeterset             ) //DcmTagKey(0x3008, 0x0036)
    GET_DICOM_INFO_LINE(DCM_DeliveredSecondaryMeterset           ) //DcmTagKey(0x3008, 0x0037)
    GET_DICOM_INFO_LINE(DCM_SpecifiedTreatmentTime               ) //DcmTagKey(0x3008, 0x003a)
    GET_DICOM_INFO_LINE(DCM_DeliveredTreatmentTime               ) //DcmTagKey(0x3008, 0x003b)
    GET_DICOM_INFO_LINE(DCM_ControlPointDeliverySequence         ) //DcmTagKey(0x3008, 0x0040)
    GET_DICOM_INFO_LINE(DCM_IonControlPointDeliverySequence      ) //DcmTagKey(0x3008, 0x0041)
    GET_DICOM_INFO_LINE(DCM_SpecifiedMeterset                    ) //DcmTagKey(0x3008, 0x0042)
    GET_DICOM_INFO_LINE(DCM_DeliveredMeterset                    ) //DcmTagKey(0x3008, 0x0044)
    GET_DICOM_INFO_LINE(DCM_MetersetRateSet                      ) //DcmTagKey(0x3008, 0x0045)
    GET_DICOM_INFO_LINE(DCM_MetersetRateDelivered                ) //DcmTagKey(0x3008, 0x0046)
    GET_DICOM_INFO_LINE(DCM_ScanSpotMetersetsDelivered           ) //DcmTagKey(0x3008, 0x0047)
    GET_DICOM_INFO_LINE(DCM_DoseRateDelivered                    ) //DcmTagKey(0x3008, 0x0048)
    GET_DICOM_INFO_LINE(DCM_TreatmentSummaryCalculatedDoseReferenceSequence ) //DcmTagKey(0x3008, 0x0050)
    GET_DICOM_INFO_LINE(DCM_CumulativeDoseToDoseReference        ) //DcmTagKey(0x3008, 0x0052)
    GET_DICOM_INFO_LINE(DCM_FirstTreatmentDate                   ) //DcmTagKey(0x3008, 0x0054)
    GET_DICOM_INFO_LINE(DCM_MostRecentTreatmentDate              ) //DcmTagKey(0x3008, 0x0056)
    GET_DICOM_INFO_LINE(DCM_NumberOfFractionsDelivered           ) //DcmTagKey(0x3008, 0x005a)
    GET_DICOM_INFO_LINE(DCM_OverrideSequence                     ) //DcmTagKey(0x3008, 0x0060)
    GET_DICOM_INFO_LINE(DCM_ParameterSequencePointer             ) //DcmTagKey(0x3008, 0x0061)
    GET_DICOM_INFO_LINE(DCM_OverrideParameterPointer             ) //DcmTagKey(0x3008, 0x0062)
    GET_DICOM_INFO_LINE(DCM_ParameterItemIndex                   ) //DcmTagKey(0x3008, 0x0063)
    GET_DICOM_INFO_LINE(DCM_MeasuredDoseReferenceNumber          ) //DcmTagKey(0x3008, 0x0064)
    GET_DICOM_INFO_LINE(DCM_ParameterPointer                     ) //DcmTagKey(0x3008, 0x0065)
    GET_DICOM_INFO_LINE(DCM_OverrideReason                       ) //DcmTagKey(0x3008, 0x0066)
    GET_DICOM_INFO_LINE(DCM_CorrectedParameterSequence           ) //DcmTagKey(0x3008, 0x0068)
    GET_DICOM_INFO_LINE(DCM_CorrectionValue                      ) //DcmTagKey(0x3008, 0x006a)
    GET_DICOM_INFO_LINE(DCM_CalculatedDoseReferenceSequence      ) //DcmTagKey(0x3008, 0x0070)
    GET_DICOM_INFO_LINE(DCM_CalculatedDoseReferenceNumber        ) //DcmTagKey(0x3008, 0x0072)
    GET_DICOM_INFO_LINE(DCM_CalculatedDoseReferenceDescription   ) //DcmTagKey(0x3008, 0x0074)
    GET_DICOM_INFO_LINE(DCM_CalculatedDoseReferenceDoseValue     ) //DcmTagKey(0x3008, 0x0076)
    GET_DICOM_INFO_LINE(DCM_StartMeterset                        ) //DcmTagKey(0x3008, 0x0078)
    GET_DICOM_INFO_LINE(DCM_EndMeterset                          ) //DcmTagKey(0x3008, 0x007a)
    GET_DICOM_INFO_LINE(DCM_ReferencedMeasuredDoseReferenceSequence ) //DcmTagKey(0x3008, 0x0080)
    GET_DICOM_INFO_LINE(DCM_ReferencedMeasuredDoseReferenceNumber ) //DcmTagKey(0x3008, 0x0082)
    GET_DICOM_INFO_LINE(DCM_ReferencedCalculatedDoseReferenceSequence ) //DcmTagKey(0x3008, 0x0090)
    GET_DICOM_INFO_LINE(DCM_ReferencedCalculatedDoseReferenceNumber ) //DcmTagKey(0x3008, 0x0092)
    GET_DICOM_INFO_LINE(DCM_BeamLimitingDeviceLeafPairsSequence  ) //DcmTagKey(0x3008, 0x00a0)
    GET_DICOM_INFO_LINE(DCM_RecordedWedgeSequence                ) //DcmTagKey(0x3008, 0x00b0)
    GET_DICOM_INFO_LINE(DCM_RecordedCompensatorSequence          ) //DcmTagKey(0x3008, 0x00c0)
    GET_DICOM_INFO_LINE(DCM_RecordedBlockSequence                ) //DcmTagKey(0x3008, 0x00d0)
    GET_DICOM_INFO_LINE(DCM_TreatmentSummaryMeasuredDoseReferenceSequence ) //DcmTagKey(0x3008, 0x00e0)
    GET_DICOM_INFO_LINE(DCM_RecordedSnoutSequence                ) //DcmTagKey(0x3008, 0x00f0)
    GET_DICOM_INFO_LINE(DCM_RecordedRangeShifterSequence         ) //DcmTagKey(0x3008, 0x00f2)
    GET_DICOM_INFO_LINE(DCM_RecordedLateralSpreadingDeviceSequence ) //DcmTagKey(0x3008, 0x00f4)
    GET_DICOM_INFO_LINE(DCM_RecordedRangeModulatorSequence       ) //DcmTagKey(0x3008, 0x00f6)
    GET_DICOM_INFO_LINE(DCM_RecordedSourceSequence               ) //DcmTagKey(0x3008, 0x0100)
    GET_DICOM_INFO_LINE(DCM_SourceSerialNumber                   ) //DcmTagKey(0x3008, 0x0105)
    GET_DICOM_INFO_LINE(DCM_TreatmentSessionApplicationSetupSequence ) //DcmTagKey(0x3008, 0x0110)
    GET_DICOM_INFO_LINE(DCM_ApplicationSetupCheck                ) //DcmTagKey(0x3008, 0x0116)
    GET_DICOM_INFO_LINE(DCM_RecordedBrachyAccessoryDeviceSequence ) //DcmTagKey(0x3008, 0x0120)
    GET_DICOM_INFO_LINE(DCM_ReferencedBrachyAccessoryDeviceNumber ) //DcmTagKey(0x3008, 0x0122)
    GET_DICOM_INFO_LINE(DCM_RecordedChannelSequence              ) //DcmTagKey(0x3008, 0x0130)
    GET_DICOM_INFO_LINE(DCM_SpecifiedChannelTotalTime            ) //DcmTagKey(0x3008, 0x0132)
    GET_DICOM_INFO_LINE(DCM_DeliveredChannelTotalTime            ) //DcmTagKey(0x3008, 0x0134)
    GET_DICOM_INFO_LINE(DCM_SpecifiedNumberOfPulses              ) //DcmTagKey(0x3008, 0x0136)
    GET_DICOM_INFO_LINE(DCM_DeliveredNumberOfPulses              ) //DcmTagKey(0x3008, 0x0138)
    GET_DICOM_INFO_LINE(DCM_SpecifiedPulseRepetitionInterval     ) //DcmTagKey(0x3008, 0x013a)
    GET_DICOM_INFO_LINE(DCM_DeliveredPulseRepetitionInterval     ) //DcmTagKey(0x3008, 0x013c)
    GET_DICOM_INFO_LINE(DCM_RecordedSourceApplicatorSequence     ) //DcmTagKey(0x3008, 0x0140)
    GET_DICOM_INFO_LINE(DCM_ReferencedSourceApplicatorNumber     ) //DcmTagKey(0x3008, 0x0142)
    GET_DICOM_INFO_LINE(DCM_RecordedChannelShieldSequence        ) //DcmTagKey(0x3008, 0x0150)
    GET_DICOM_INFO_LINE(DCM_ReferencedChannelShieldNumber        ) //DcmTagKey(0x3008, 0x0152)
    GET_DICOM_INFO_LINE(DCM_BrachyControlPointDeliveredSequence  ) //DcmTagKey(0x3008, 0x0160)
    GET_DICOM_INFO_LINE(DCM_SafePositionExitDate                 ) //DcmTagKey(0x3008, 0x0162)
    GET_DICOM_INFO_LINE(DCM_SafePositionExitTime                 ) //DcmTagKey(0x3008, 0x0164)
    GET_DICOM_INFO_LINE(DCM_SafePositionReturnDate               ) //DcmTagKey(0x3008, 0x0166)
    GET_DICOM_INFO_LINE(DCM_SafePositionReturnTime               ) //DcmTagKey(0x3008, 0x0168)
    GET_DICOM_INFO_LINE(DCM_CurrentTreatmentStatus               ) //DcmTagKey(0x3008, 0x0200)
    GET_DICOM_INFO_LINE(DCM_TreatmentStatusComment               ) //DcmTagKey(0x3008, 0x0202)
    GET_DICOM_INFO_LINE(DCM_FractionGroupSummarySequence         ) //DcmTagKey(0x3008, 0x0220)
    GET_DICOM_INFO_LINE(DCM_ReferencedFractionNumber             ) //DcmTagKey(0x3008, 0x0223)
    GET_DICOM_INFO_LINE(DCM_FractionGroupType                    ) //DcmTagKey(0x3008, 0x0224)
    GET_DICOM_INFO_LINE(DCM_BeamStopperPosition                  ) //DcmTagKey(0x3008, 0x0230)
    GET_DICOM_INFO_LINE(DCM_FractionStatusSummarySequence        ) //DcmTagKey(0x3008, 0x0240)
    GET_DICOM_INFO_LINE(DCM_TreatmentDate                        ) //DcmTagKey(0x3008, 0x0250)
    GET_DICOM_INFO_LINE(DCM_TreatmentTime                        ) //DcmTagKey(0x3008, 0x0251)
    GET_DICOM_INFO_LINE(DCM_RTPlanLabel                          ) //DcmTagKey(0x300a, 0x0002)
    GET_DICOM_INFO_LINE(DCM_RTPlanName                           ) //DcmTagKey(0x300a, 0x0003)
    GET_DICOM_INFO_LINE(DCM_RTPlanDescription                    ) //DcmTagKey(0x300a, 0x0004)
    GET_DICOM_INFO_LINE(DCM_RTPlanDate                           ) //DcmTagKey(0x300a, 0x0006)
    GET_DICOM_INFO_LINE(DCM_RTPlanTime                           ) //DcmTagKey(0x300a, 0x0007)
    GET_DICOM_INFO_LINE(DCM_TreatmentProtocols                   ) //DcmTagKey(0x300a, 0x0009)
    GET_DICOM_INFO_LINE(DCM_PlanIntent                           ) //DcmTagKey(0x300a, 0x000a)
    GET_DICOM_INFO_LINE(DCM_TreatmentSites                       ) //DcmTagKey(0x300a, 0x000b)
    GET_DICOM_INFO_LINE(DCM_RTPlanGeometry                       ) //DcmTagKey(0x300a, 0x000c)
    GET_DICOM_INFO_LINE(DCM_PrescriptionDescription              ) //DcmTagKey(0x300a, 0x000e)
    GET_DICOM_INFO_LINE(DCM_DoseReferenceSequence                ) //DcmTagKey(0x300a, 0x0010)
    GET_DICOM_INFO_LINE(DCM_DoseReferenceNumber                  ) //DcmTagKey(0x300a, 0x0012)
    GET_DICOM_INFO_LINE(DCM_DoseReferenceUID                     ) //DcmTagKey(0x300a, 0x0013)
    GET_DICOM_INFO_LINE(DCM_DoseReferenceStructureType           ) //DcmTagKey(0x300a, 0x0014)
    GET_DICOM_INFO_LINE(DCM_NominalBeamEnergyUnit                ) //DcmTagKey(0x300a, 0x0015)
    GET_DICOM_INFO_LINE(DCM_DoseReferenceDescription             ) //DcmTagKey(0x300a, 0x0016)
    GET_DICOM_INFO_LINE(DCM_DoseReferencePointCoordinates        ) //DcmTagKey(0x300a, 0x0018)
    GET_DICOM_INFO_LINE(DCM_NominalPriorDose                     ) //DcmTagKey(0x300a, 0x001a)
    GET_DICOM_INFO_LINE(DCM_DoseReferenceType                    ) //DcmTagKey(0x300a, 0x0020)
    GET_DICOM_INFO_LINE(DCM_ConstraintWeight                     ) //DcmTagKey(0x300a, 0x0021)
    GET_DICOM_INFO_LINE(DCM_DeliveryWarningDose                  ) //DcmTagKey(0x300a, 0x0022)
    GET_DICOM_INFO_LINE(DCM_DeliveryMaximumDose                  ) //DcmTagKey(0x300a, 0x0023)
    GET_DICOM_INFO_LINE(DCM_TargetMinimumDose                    ) //DcmTagKey(0x300a, 0x0025)
    GET_DICOM_INFO_LINE(DCM_TargetPrescriptionDose               ) //DcmTagKey(0x300a, 0x0026)
    GET_DICOM_INFO_LINE(DCM_TargetMaximumDose                    ) //DcmTagKey(0x300a, 0x0027)
    GET_DICOM_INFO_LINE(DCM_TargetUnderdoseVolumeFraction        ) //DcmTagKey(0x300a, 0x0028)
    GET_DICOM_INFO_LINE(DCM_OrganAtRiskFullVolumeDose            ) //DcmTagKey(0x300a, 0x002a)
    GET_DICOM_INFO_LINE(DCM_OrganAtRiskLimitDose                 ) //DcmTagKey(0x300a, 0x002b)
    GET_DICOM_INFO_LINE(DCM_OrganAtRiskMaximumDose               ) //DcmTagKey(0x300a, 0x002c)
    GET_DICOM_INFO_LINE(DCM_OrganAtRiskOverdoseVolumeFraction    ) //DcmTagKey(0x300a, 0x002d)
    GET_DICOM_INFO_LINE(DCM_ToleranceTableSequence               ) //DcmTagKey(0x300a, 0x0040)
    GET_DICOM_INFO_LINE(DCM_ToleranceTableNumber                 ) //DcmTagKey(0x300a, 0x0042)
    GET_DICOM_INFO_LINE(DCM_ToleranceTableLabel                  ) //DcmTagKey(0x300a, 0x0043)
    GET_DICOM_INFO_LINE(DCM_GantryAngleTolerance                 ) //DcmTagKey(0x300a, 0x0044)
    GET_DICOM_INFO_LINE(DCM_BeamLimitingDeviceAngleTolerance     ) //DcmTagKey(0x300a, 0x0046)
    GET_DICOM_INFO_LINE(DCM_BeamLimitingDeviceToleranceSequence  ) //DcmTagKey(0x300a, 0x0048)
    GET_DICOM_INFO_LINE(DCM_BeamLimitingDevicePositionTolerance  ) //DcmTagKey(0x300a, 0x004a)
    GET_DICOM_INFO_LINE(DCM_SnoutPositionTolerance               ) //DcmTagKey(0x300a, 0x004b)
    GET_DICOM_INFO_LINE(DCM_PatientSupportAngleTolerance         ) //DcmTagKey(0x300a, 0x004c)
    GET_DICOM_INFO_LINE(DCM_TableTopEccentricAngleTolerance      ) //DcmTagKey(0x300a, 0x004e)
    GET_DICOM_INFO_LINE(DCM_TableTopPitchAngleTolerance          ) //DcmTagKey(0x300a, 0x004f)
    GET_DICOM_INFO_LINE(DCM_TableTopRollAngleTolerance           ) //DcmTagKey(0x300a, 0x0050)
    GET_DICOM_INFO_LINE(DCM_TableTopVerticalPositionTolerance    ) //DcmTagKey(0x300a, 0x0051)
    GET_DICOM_INFO_LINE(DCM_TableTopLongitudinalPositionTolerance ) //DcmTagKey(0x300a, 0x0052)
    GET_DICOM_INFO_LINE(DCM_TableTopLateralPositionTolerance     ) //DcmTagKey(0x300a, 0x0053)
    GET_DICOM_INFO_LINE(DCM_RTPlanRelationship                   ) //DcmTagKey(0x300a, 0x0055)
    GET_DICOM_INFO_LINE(DCM_FractionGroupSequence                ) //DcmTagKey(0x300a, 0x0070)
    GET_DICOM_INFO_LINE(DCM_FractionGroupNumber                  ) //DcmTagKey(0x300a, 0x0071)
    GET_DICOM_INFO_LINE(DCM_FractionGroupDescription             ) //DcmTagKey(0x300a, 0x0072)
    GET_DICOM_INFO_LINE(DCM_NumberOfFractionsPlanned             ) //DcmTagKey(0x300a, 0x0078)
    GET_DICOM_INFO_LINE(DCM_NumberOfFractionPatternDigitsPerDay  ) //DcmTagKey(0x300a, 0x0079)
    GET_DICOM_INFO_LINE(DCM_RepeatFractionCycleLength            ) //DcmTagKey(0x300a, 0x007a)
    GET_DICOM_INFO_LINE(DCM_FractionPattern                      ) //DcmTagKey(0x300a, 0x007b)
    GET_DICOM_INFO_LINE(DCM_NumberOfBeams                        ) //DcmTagKey(0x300a, 0x0080)
    GET_DICOM_INFO_LINE(DCM_BeamDoseSpecificationPoint           ) //DcmTagKey(0x300a, 0x0082)
    GET_DICOM_INFO_LINE(DCM_BeamDose                             ) //DcmTagKey(0x300a, 0x0084)
    GET_DICOM_INFO_LINE(DCM_BeamMeterset                         ) //DcmTagKey(0x300a, 0x0086)
    GET_DICOM_INFO_LINE(DCM_BeamDosePointDepth                   ) //DcmTagKey(0x300a, 0x0088)
    GET_DICOM_INFO_LINE(DCM_BeamDosePointEquivalentDepth         ) //DcmTagKey(0x300a, 0x0089)
    GET_DICOM_INFO_LINE(DCM_BeamDosePointSSD                     ) //DcmTagKey(0x300a, 0x008a)
    GET_DICOM_INFO_LINE(DCM_NumberOfBrachyApplicationSetups      ) //DcmTagKey(0x300a, 0x00a0)
    GET_DICOM_INFO_LINE(DCM_BrachyApplicationSetupDoseSpecificationPoint ) //DcmTagKey(0x300a, 0x00a2)
    GET_DICOM_INFO_LINE(DCM_BrachyApplicationSetupDose           ) //DcmTagKey(0x300a, 0x00a4)
    GET_DICOM_INFO_LINE(DCM_BeamSequence                         ) //DcmTagKey(0x300a, 0x00b0)
    GET_DICOM_INFO_LINE(DCM_TreatmentMachineName                 ) //DcmTagKey(0x300a, 0x00b2)
    GET_DICOM_INFO_LINE(DCM_PrimaryDosimeterUnit                 ) //DcmTagKey(0x300a, 0x00b3)
    GET_DICOM_INFO_LINE(DCM_SourceAxisDistance                   ) //DcmTagKey(0x300a, 0x00b4)
    GET_DICOM_INFO_LINE(DCM_BeamLimitingDeviceSequence           ) //DcmTagKey(0x300a, 0x00b6)
    GET_DICOM_INFO_LINE(DCM_RTBeamLimitingDeviceType             ) //DcmTagKey(0x300a, 0x00b8)
    GET_DICOM_INFO_LINE(DCM_SourceToBeamLimitingDeviceDistance   ) //DcmTagKey(0x300a, 0x00ba)
    GET_DICOM_INFO_LINE(DCM_IsocenterToBeamLimitingDeviceDistance ) //DcmTagKey(0x300a, 0x00bb)
    GET_DICOM_INFO_LINE(DCM_NumberOfLeafJawPairs                 ) //DcmTagKey(0x300a, 0x00bc)
    GET_DICOM_INFO_LINE(DCM_LeafPositionBoundaries               ) //DcmTagKey(0x300a, 0x00be)
    GET_DICOM_INFO_LINE(DCM_BeamNumber                           ) //DcmTagKey(0x300a, 0x00c0)
    GET_DICOM_INFO_LINE(DCM_BeamName                             ) //DcmTagKey(0x300a, 0x00c2)
    GET_DICOM_INFO_LINE(DCM_BeamDescription                      ) //DcmTagKey(0x300a, 0x00c3)
    GET_DICOM_INFO_LINE(DCM_BeamType                             ) //DcmTagKey(0x300a, 0x00c4)
    GET_DICOM_INFO_LINE(DCM_RadiationType                        ) //DcmTagKey(0x300a, 0x00c6)
    GET_DICOM_INFO_LINE(DCM_HighDoseTechniqueType                ) //DcmTagKey(0x300a, 0x00c7)
    GET_DICOM_INFO_LINE(DCM_ReferenceImageNumber                 ) //DcmTagKey(0x300a, 0x00c8)
    GET_DICOM_INFO_LINE(DCM_PlannedVerificationImageSequence     ) //DcmTagKey(0x300a, 0x00ca)
    GET_DICOM_INFO_LINE(DCM_ImagingDeviceSpecificAcquisitionParameters ) //DcmTagKey(0x300a, 0x00cc)
    GET_DICOM_INFO_LINE(DCM_TreatmentDeliveryType                ) //DcmTagKey(0x300a, 0x00ce)
    GET_DICOM_INFO_LINE(DCM_NumberOfWedges                       ) //DcmTagKey(0x300a, 0x00d0)
    GET_DICOM_INFO_LINE(DCM_WedgeSequence                        ) //DcmTagKey(0x300a, 0x00d1)
    GET_DICOM_INFO_LINE(DCM_WedgeNumber                          ) //DcmTagKey(0x300a, 0x00d2)
    GET_DICOM_INFO_LINE(DCM_WedgeType                            ) //DcmTagKey(0x300a, 0x00d3)
    GET_DICOM_INFO_LINE(DCM_WedgeID                              ) //DcmTagKey(0x300a, 0x00d4)
    GET_DICOM_INFO_LINE(DCM_WedgeAngle                           ) //DcmTagKey(0x300a, 0x00d5)
    GET_DICOM_INFO_LINE(DCM_WedgeFactor                          ) //DcmTagKey(0x300a, 0x00d6)
    GET_DICOM_INFO_LINE(DCM_TotalWedgeTrayWaterEquivalentThickness ) //DcmTagKey(0x300a, 0x00d7)
    GET_DICOM_INFO_LINE(DCM_WedgeOrientation                     ) //DcmTagKey(0x300a, 0x00d8)
    GET_DICOM_INFO_LINE(DCM_IsocenterToWedgeTrayDistance         ) //DcmTagKey(0x300a, 0x00d9)
    GET_DICOM_INFO_LINE(DCM_SourceToWedgeTrayDistance            ) //DcmTagKey(0x300a, 0x00da)
    GET_DICOM_INFO_LINE(DCM_WedgeThinEdgePosition                ) //DcmTagKey(0x300a, 0x00db)
    GET_DICOM_INFO_LINE(DCM_BolusID                              ) //DcmTagKey(0x300a, 0x00dc)
    GET_DICOM_INFO_LINE(DCM_BolusDescription                     ) //DcmTagKey(0x300a, 0x00dd)
    GET_DICOM_INFO_LINE(DCM_NumberOfCompensators                 ) //DcmTagKey(0x300a, 0x00e0)
    GET_DICOM_INFO_LINE(DCM_MaterialID                           ) //DcmTagKey(0x300a, 0x00e1)
    GET_DICOM_INFO_LINE(DCM_TotalCompensatorTrayFactor           ) //DcmTagKey(0x300a, 0x00e2)
    GET_DICOM_INFO_LINE(DCM_CompensatorSequence                  ) //DcmTagKey(0x300a, 0x00e3)
    GET_DICOM_INFO_LINE(DCM_CompensatorNumber                    ) //DcmTagKey(0x300a, 0x00e4)
    GET_DICOM_INFO_LINE(DCM_CompensatorID                        ) //DcmTagKey(0x300a, 0x00e5)
    GET_DICOM_INFO_LINE(DCM_SourceToCompensatorTrayDistance      ) //DcmTagKey(0x300a, 0x00e6)
    GET_DICOM_INFO_LINE(DCM_CompensatorRows                      ) //DcmTagKey(0x300a, 0x00e7)
    GET_DICOM_INFO_LINE(DCM_CompensatorColumns                   ) //DcmTagKey(0x300a, 0x00e8)
    GET_DICOM_INFO_LINE(DCM_CompensatorPixelSpacing              ) //DcmTagKey(0x300a, 0x00e9)
    GET_DICOM_INFO_LINE(DCM_CompensatorPosition                  ) //DcmTagKey(0x300a, 0x00ea)
    GET_DICOM_INFO_LINE(DCM_CompensatorTransmissionData          ) //DcmTagKey(0x300a, 0x00eb)
    GET_DICOM_INFO_LINE(DCM_CompensatorThicknessData             ) //DcmTagKey(0x300a, 0x00ec)
    GET_DICOM_INFO_LINE(DCM_NumberOfBoli                         ) //DcmTagKey(0x300a, 0x00ed)
    GET_DICOM_INFO_LINE(DCM_CompensatorType                      ) //DcmTagKey(0x300a, 0x00ee)
    GET_DICOM_INFO_LINE(DCM_NumberOfBlocks                       ) //DcmTagKey(0x300a, 0x00f0)
    GET_DICOM_INFO_LINE(DCM_TotalBlockTrayFactor                 ) //DcmTagKey(0x300a, 0x00f2)
    GET_DICOM_INFO_LINE(DCM_TotalBlockTrayWaterEquivalentThickness ) //DcmTagKey(0x300a, 0x00f3)
    GET_DICOM_INFO_LINE(DCM_BlockSequence                        ) //DcmTagKey(0x300a, 0x00f4)
    GET_DICOM_INFO_LINE(DCM_BlockTrayID                          ) //DcmTagKey(0x300a, 0x00f5)
    GET_DICOM_INFO_LINE(DCM_SourceToBlockTrayDistance            ) //DcmTagKey(0x300a, 0x00f6)
    GET_DICOM_INFO_LINE(DCM_IsocenterToBlockTrayDistance         ) //DcmTagKey(0x300a, 0x00f7)
    GET_DICOM_INFO_LINE(DCM_BlockType                            ) //DcmTagKey(0x300a, 0x00f8)
    GET_DICOM_INFO_LINE(DCM_AccessoryCode                        ) //DcmTagKey(0x300a, 0x00f9)
    GET_DICOM_INFO_LINE(DCM_BlockDivergence                      ) //DcmTagKey(0x300a, 0x00fa)
    GET_DICOM_INFO_LINE(DCM_BlockMountingPosition                ) //DcmTagKey(0x300a, 0x00fb)
    GET_DICOM_INFO_LINE(DCM_BlockNumber                          ) //DcmTagKey(0x300a, 0x00fc)
    GET_DICOM_INFO_LINE(DCM_BlockName                            ) //DcmTagKey(0x300a, 0x00fe)
    GET_DICOM_INFO_LINE(DCM_BlockThickness                       ) //DcmTagKey(0x300a, 0x0100)
    GET_DICOM_INFO_LINE(DCM_BlockTransmission                    ) //DcmTagKey(0x300a, 0x0102)
    GET_DICOM_INFO_LINE(DCM_BlockNumberOfPoints                  ) //DcmTagKey(0x300a, 0x0104)
    GET_DICOM_INFO_LINE(DCM_BlockData                            ) //DcmTagKey(0x300a, 0x0106)
    GET_DICOM_INFO_LINE(DCM_ApplicatorSequence                   ) //DcmTagKey(0x300a, 0x0107)
    GET_DICOM_INFO_LINE(DCM_ApplicatorID                         ) //DcmTagKey(0x300a, 0x0108)
    GET_DICOM_INFO_LINE(DCM_ApplicatorType                       ) //DcmTagKey(0x300a, 0x0109)
    GET_DICOM_INFO_LINE(DCM_ApplicatorDescription                ) //DcmTagKey(0x300a, 0x010a)
    GET_DICOM_INFO_LINE(DCM_CumulativeDoseReferenceCoefficient   ) //DcmTagKey(0x300a, 0x010c)
    GET_DICOM_INFO_LINE(DCM_FinalCumulativeMetersetWeight        ) //DcmTagKey(0x300a, 0x010e)
    GET_DICOM_INFO_LINE(DCM_NumberOfControlPoints                ) //DcmTagKey(0x300a, 0x0110)
    GET_DICOM_INFO_LINE(DCM_ControlPointSequence                 ) //DcmTagKey(0x300a, 0x0111)
    GET_DICOM_INFO_LINE(DCM_ControlPointIndex                    ) //DcmTagKey(0x300a, 0x0112)
    GET_DICOM_INFO_LINE(DCM_NominalBeamEnergy                    ) //DcmTagKey(0x300a, 0x0114)
    GET_DICOM_INFO_LINE(DCM_DoseRateSet                          ) //DcmTagKey(0x300a, 0x0115)
    GET_DICOM_INFO_LINE(DCM_WedgePositionSequence                ) //DcmTagKey(0x300a, 0x0116)
    GET_DICOM_INFO_LINE(DCM_WedgePosition                        ) //DcmTagKey(0x300a, 0x0118)
    GET_DICOM_INFO_LINE(DCM_BeamLimitingDevicePositionSequence   ) //DcmTagKey(0x300a, 0x011a)
    GET_DICOM_INFO_LINE(DCM_LeafJawPositions                     ) //DcmTagKey(0x300a, 0x011c)
    GET_DICOM_INFO_LINE(DCM_GantryAngle                          ) //DcmTagKey(0x300a, 0x011e)
    GET_DICOM_INFO_LINE(DCM_GantryRotationDirection              ) //DcmTagKey(0x300a, 0x011f)
    GET_DICOM_INFO_LINE(DCM_BeamLimitingDeviceAngle              ) //DcmTagKey(0x300a, 0x0120)
    GET_DICOM_INFO_LINE(DCM_BeamLimitingDeviceRotationDirection  ) //DcmTagKey(0x300a, 0x0121)
    GET_DICOM_INFO_LINE(DCM_PatientSupportAngle                  ) //DcmTagKey(0x300a, 0x0122)
    GET_DICOM_INFO_LINE(DCM_PatientSupportRotationDirection      ) //DcmTagKey(0x300a, 0x0123)
    GET_DICOM_INFO_LINE(DCM_TableTopEccentricAxisDistance        ) //DcmTagKey(0x300a, 0x0124)
    GET_DICOM_INFO_LINE(DCM_TableTopEccentricAngle               ) //DcmTagKey(0x300a, 0x0125)
    GET_DICOM_INFO_LINE(DCM_TableTopEccentricRotationDirection   ) //DcmTagKey(0x300a, 0x0126)
    GET_DICOM_INFO_LINE(DCM_TableTopVerticalPosition             ) //DcmTagKey(0x300a, 0x0128)
    GET_DICOM_INFO_LINE(DCM_TableTopLongitudinalPosition         ) //DcmTagKey(0x300a, 0x0129)
    GET_DICOM_INFO_LINE(DCM_TableTopLateralPosition              ) //DcmTagKey(0x300a, 0x012a)
    GET_DICOM_INFO_LINE(DCM_IsocenterPosition                    ) //DcmTagKey(0x300a, 0x012c)
    GET_DICOM_INFO_LINE(DCM_SurfaceEntryPoint                    ) //DcmTagKey(0x300a, 0x012e)
    GET_DICOM_INFO_LINE(DCM_SourceToSurfaceDistance              ) //DcmTagKey(0x300a, 0x0130)
    GET_DICOM_INFO_LINE(DCM_CumulativeMetersetWeight             ) //DcmTagKey(0x300a, 0x0134)
    GET_DICOM_INFO_LINE(DCM_TableTopPitchAngle                   ) //DcmTagKey(0x300a, 0x0140)
    GET_DICOM_INFO_LINE(DCM_TableTopPitchRotationDirection       ) //DcmTagKey(0x300a, 0x0142)
    GET_DICOM_INFO_LINE(DCM_TableTopRollAngle                    ) //DcmTagKey(0x300a, 0x0144)
    GET_DICOM_INFO_LINE(DCM_TableTopRollRotationDirection        ) //DcmTagKey(0x300a, 0x0146)
    GET_DICOM_INFO_LINE(DCM_HeadFixationAngle                    ) //DcmTagKey(0x300a, 0x0148)
    GET_DICOM_INFO_LINE(DCM_GantryPitchAngle                     ) //DcmTagKey(0x300a, 0x014a)
    GET_DICOM_INFO_LINE(DCM_GantryPitchRotationDirection         ) //DcmTagKey(0x300a, 0x014c)
    GET_DICOM_INFO_LINE(DCM_GantryPitchAngleTolerance            ) //DcmTagKey(0x300a, 0x014e)
    GET_DICOM_INFO_LINE(DCM_PatientSetupSequence                 ) //DcmTagKey(0x300a, 0x0180)
    GET_DICOM_INFO_LINE(DCM_PatientSetupNumber                   ) //DcmTagKey(0x300a, 0x0182)
    GET_DICOM_INFO_LINE(DCM_PatientSetupLabel                    ) //DcmTagKey(0x300a, 0x0183)
    GET_DICOM_INFO_LINE(DCM_PatientAdditionalPosition            ) //DcmTagKey(0x300a, 0x0184)
    GET_DICOM_INFO_LINE(DCM_FixationDeviceSequence               ) //DcmTagKey(0x300a, 0x0190)
    GET_DICOM_INFO_LINE(DCM_FixationDeviceType                   ) //DcmTagKey(0x300a, 0x0192)
    GET_DICOM_INFO_LINE(DCM_FixationDeviceLabel                  ) //DcmTagKey(0x300a, 0x0194)
    GET_DICOM_INFO_LINE(DCM_FixationDeviceDescription            ) //DcmTagKey(0x300a, 0x0196)
    GET_DICOM_INFO_LINE(DCM_FixationDevicePosition               ) //DcmTagKey(0x300a, 0x0198)
    GET_DICOM_INFO_LINE(DCM_FixationDevicePitchAngle             ) //DcmTagKey(0x300a, 0x0199)
    GET_DICOM_INFO_LINE(DCM_FixationDeviceRollAngle              ) //DcmTagKey(0x300a, 0x019a)
    GET_DICOM_INFO_LINE(DCM_ShieldingDeviceSequence              ) //DcmTagKey(0x300a, 0x01a0)
    GET_DICOM_INFO_LINE(DCM_ShieldingDeviceType                  ) //DcmTagKey(0x300a, 0x01a2)
    GET_DICOM_INFO_LINE(DCM_ShieldingDeviceLabel                 ) //DcmTagKey(0x300a, 0x01a4)
    GET_DICOM_INFO_LINE(DCM_ShieldingDeviceDescription           ) //DcmTagKey(0x300a, 0x01a6)
    GET_DICOM_INFO_LINE(DCM_ShieldingDevicePosition              ) //DcmTagKey(0x300a, 0x01a8)
    GET_DICOM_INFO_LINE(DCM_SetupTechnique                       ) //DcmTagKey(0x300a, 0x01b0)
    GET_DICOM_INFO_LINE(DCM_SetupTechniqueDescription            ) //DcmTagKey(0x300a, 0x01b2)
    GET_DICOM_INFO_LINE(DCM_SetupDeviceSequence                  ) //DcmTagKey(0x300a, 0x01b4)
    GET_DICOM_INFO_LINE(DCM_SetupDeviceType                      ) //DcmTagKey(0x300a, 0x01b6)
    GET_DICOM_INFO_LINE(DCM_SetupDeviceLabel                     ) //DcmTagKey(0x300a, 0x01b8)
    GET_DICOM_INFO_LINE(DCM_SetupDeviceDescription               ) //DcmTagKey(0x300a, 0x01ba)
    GET_DICOM_INFO_LINE(DCM_SetupDeviceParameter                 ) //DcmTagKey(0x300a, 0x01bc)
    GET_DICOM_INFO_LINE(DCM_SetupReferenceDescription            ) //DcmTagKey(0x300a, 0x01d0)
    GET_DICOM_INFO_LINE(DCM_TableTopVerticalSetupDisplacement    ) //DcmTagKey(0x300a, 0x01d2)
    GET_DICOM_INFO_LINE(DCM_TableTopLongitudinalSetupDisplacement ) //DcmTagKey(0x300a, 0x01d4)
    GET_DICOM_INFO_LINE(DCM_TableTopLateralSetupDisplacement     ) //DcmTagKey(0x300a, 0x01d6)
    GET_DICOM_INFO_LINE(DCM_BrachyTreatmentTechnique             ) //DcmTagKey(0x300a, 0x0200)
    GET_DICOM_INFO_LINE(DCM_BrachyTreatmentType                  ) //DcmTagKey(0x300a, 0x0202)
    GET_DICOM_INFO_LINE(DCM_TreatmentMachineSequence             ) //DcmTagKey(0x300a, 0x0206)
    GET_DICOM_INFO_LINE(DCM_SourceSequence                       ) //DcmTagKey(0x300a, 0x0210)
    GET_DICOM_INFO_LINE(DCM_SourceNumber                         ) //DcmTagKey(0x300a, 0x0212)
    GET_DICOM_INFO_LINE(DCM_SourceType                           ) //DcmTagKey(0x300a, 0x0214)
    GET_DICOM_INFO_LINE(DCM_SourceManufacturer                   ) //DcmTagKey(0x300a, 0x0216)
    GET_DICOM_INFO_LINE(DCM_ActiveSourceDiameter                 ) //DcmTagKey(0x300a, 0x0218)
    GET_DICOM_INFO_LINE(DCM_ActiveSourceLength                   ) //DcmTagKey(0x300a, 0x021a)
    GET_DICOM_INFO_LINE(DCM_SourceEncapsulationNominalThickness  ) //DcmTagKey(0x300a, 0x0222)
    GET_DICOM_INFO_LINE(DCM_SourceEncapsulationNominalTransmission ) //DcmTagKey(0x300a, 0x0224)
    GET_DICOM_INFO_LINE(DCM_SourceIsotopeName                    ) //DcmTagKey(0x300a, 0x0226)
    GET_DICOM_INFO_LINE(DCM_SourceIsotopeHalfLife                ) //DcmTagKey(0x300a, 0x0228)
    GET_DICOM_INFO_LINE(DCM_SourceStrengthUnits                  ) //DcmTagKey(0x300a, 0x0229)
    GET_DICOM_INFO_LINE(DCM_ReferenceAirKermaRate                ) //DcmTagKey(0x300a, 0x022a)
    GET_DICOM_INFO_LINE(DCM_SourceStrength                       ) //DcmTagKey(0x300a, 0x022b)
    GET_DICOM_INFO_LINE(DCM_SourceStrengthReferenceDate          ) //DcmTagKey(0x300a, 0x022c)
    GET_DICOM_INFO_LINE(DCM_SourceStrengthReferenceTime          ) //DcmTagKey(0x300a, 0x022e)
    GET_DICOM_INFO_LINE(DCM_ApplicationSetupSequence             ) //DcmTagKey(0x300a, 0x0230)
    GET_DICOM_INFO_LINE(DCM_ApplicationSetupType                 ) //DcmTagKey(0x300a, 0x0232)
    GET_DICOM_INFO_LINE(DCM_ApplicationSetupNumber               ) //DcmTagKey(0x300a, 0x0234)
    GET_DICOM_INFO_LINE(DCM_ApplicationSetupName                 ) //DcmTagKey(0x300a, 0x0236)
    GET_DICOM_INFO_LINE(DCM_ApplicationSetupManufacturer         ) //DcmTagKey(0x300a, 0x0238)
    GET_DICOM_INFO_LINE(DCM_TemplateNumber                       ) //DcmTagKey(0x300a, 0x0240)
    GET_DICOM_INFO_LINE(DCM_TemplateType                         ) //DcmTagKey(0x300a, 0x0242)
    GET_DICOM_INFO_LINE(DCM_TemplateName                         ) //DcmTagKey(0x300a, 0x0244)
    GET_DICOM_INFO_LINE(DCM_TotalReferenceAirKerma               ) //DcmTagKey(0x300a, 0x0250)
    GET_DICOM_INFO_LINE(DCM_BrachyAccessoryDeviceSequence        ) //DcmTagKey(0x300a, 0x0260)
    GET_DICOM_INFO_LINE(DCM_BrachyAccessoryDeviceNumber          ) //DcmTagKey(0x300a, 0x0262)
    GET_DICOM_INFO_LINE(DCM_BrachyAccessoryDeviceID              ) //DcmTagKey(0x300a, 0x0263)
    GET_DICOM_INFO_LINE(DCM_BrachyAccessoryDeviceType            ) //DcmTagKey(0x300a, 0x0264)
    GET_DICOM_INFO_LINE(DCM_BrachyAccessoryDeviceName            ) //DcmTagKey(0x300a, 0x0266)
    GET_DICOM_INFO_LINE(DCM_BrachyAccessoryDeviceNominalThickness ) //DcmTagKey(0x300a, 0x026a)
    GET_DICOM_INFO_LINE(DCM_BrachyAccessoryDeviceNominalTransmission ) //DcmTagKey(0x300a, 0x026c)
    GET_DICOM_INFO_LINE(DCM_ChannelSequence                      ) //DcmTagKey(0x300a, 0x0280)
    GET_DICOM_INFO_LINE(DCM_ChannelNumber                        ) //DcmTagKey(0x300a, 0x0282)
    GET_DICOM_INFO_LINE(DCM_ChannelLength                        ) //DcmTagKey(0x300a, 0x0284)
    GET_DICOM_INFO_LINE(DCM_ChannelTotalTime                     ) //DcmTagKey(0x300a, 0x0286)
    GET_DICOM_INFO_LINE(DCM_SourceMovementType                   ) //DcmTagKey(0x300a, 0x0288)
    GET_DICOM_INFO_LINE(DCM_NumberOfPulses                       ) //DcmTagKey(0x300a, 0x028a)
    GET_DICOM_INFO_LINE(DCM_PulseRepetitionInterval              ) //DcmTagKey(0x300a, 0x028c)
    GET_DICOM_INFO_LINE(DCM_SourceApplicatorNumber               ) //DcmTagKey(0x300a, 0x0290)
    GET_DICOM_INFO_LINE(DCM_SourceApplicatorID                   ) //DcmTagKey(0x300a, 0x0291)
    GET_DICOM_INFO_LINE(DCM_SourceApplicatorType                 ) //DcmTagKey(0x300a, 0x0292)
    GET_DICOM_INFO_LINE(DCM_SourceApplicatorName                 ) //DcmTagKey(0x300a, 0x0294)
    GET_DICOM_INFO_LINE(DCM_SourceApplicatorLength               ) //DcmTagKey(0x300a, 0x0296)
    GET_DICOM_INFO_LINE(DCM_SourceApplicatorManufacturer         ) //DcmTagKey(0x300a, 0x0298)
    GET_DICOM_INFO_LINE(DCM_SourceApplicatorWallNominalThickness ) //DcmTagKey(0x300a, 0x029c)
    GET_DICOM_INFO_LINE(DCM_SourceApplicatorWallNominalTransmission ) //DcmTagKey(0x300a, 0x029e)
    GET_DICOM_INFO_LINE(DCM_SourceApplicatorStepSize             ) //DcmTagKey(0x300a, 0x02a0)
    GET_DICOM_INFO_LINE(DCM_TransferTubeNumber                   ) //DcmTagKey(0x300a, 0x02a2)
    GET_DICOM_INFO_LINE(DCM_TransferTubeLength                   ) //DcmTagKey(0x300a, 0x02a4)
    GET_DICOM_INFO_LINE(DCM_ChannelShieldSequence                ) //DcmTagKey(0x300a, 0x02b0)
    GET_DICOM_INFO_LINE(DCM_ChannelShieldNumber                  ) //DcmTagKey(0x300a, 0x02b2)
    GET_DICOM_INFO_LINE(DCM_ChannelShieldID                      ) //DcmTagKey(0x300a, 0x02b3)
    GET_DICOM_INFO_LINE(DCM_ChannelShieldName                    ) //DcmTagKey(0x300a, 0x02b4)
    GET_DICOM_INFO_LINE(DCM_ChannelShieldNominalThickness        ) //DcmTagKey(0x300a, 0x02b8)
    GET_DICOM_INFO_LINE(DCM_ChannelShieldNominalTransmission     ) //DcmTagKey(0x300a, 0x02ba)
    GET_DICOM_INFO_LINE(DCM_FinalCumulativeTimeWeight            ) //DcmTagKey(0x300a, 0x02c8)
    GET_DICOM_INFO_LINE(DCM_BrachyControlPointSequence           ) //DcmTagKey(0x300a, 0x02d0)
    GET_DICOM_INFO_LINE(DCM_ControlPointRelativePosition         ) //DcmTagKey(0x300a, 0x02d2)
    GET_DICOM_INFO_LINE(DCM_ControlPoint3DPosition               ) //DcmTagKey(0x300a, 0x02d4)
    GET_DICOM_INFO_LINE(DCM_CumulativeTimeWeight                 ) //DcmTagKey(0x300a, 0x02d6)
    GET_DICOM_INFO_LINE(DCM_CompensatorDivergence                ) //DcmTagKey(0x300a, 0x02e0)
    GET_DICOM_INFO_LINE(DCM_CompensatorMountingPosition          ) //DcmTagKey(0x300a, 0x02e1)
    GET_DICOM_INFO_LINE(DCM_SourceToCompensatorDistance          ) //DcmTagKey(0x300a, 0x02e2)
    GET_DICOM_INFO_LINE(DCM_TotalCompensatorTrayWaterEquivalentThickness ) //DcmTagKey(0x300a, 0x02e3)
    GET_DICOM_INFO_LINE(DCM_IsocenterToCompensatorTrayDistance   ) //DcmTagKey(0x300a, 0x02e4)
    GET_DICOM_INFO_LINE(DCM_CompensatorColumnOffset              ) //DcmTagKey(0x300a, 0x02e5)
    GET_DICOM_INFO_LINE(DCM_IsocenterToCompensatorDistances      ) //DcmTagKey(0x300a, 0x02e6)
    GET_DICOM_INFO_LINE(DCM_CompensatorRelativeStoppingPowerRatio ) //DcmTagKey(0x300a, 0x02e7)
    GET_DICOM_INFO_LINE(DCM_CompensatorMillingToolDiameter       ) //DcmTagKey(0x300a, 0x02e8)
    GET_DICOM_INFO_LINE(DCM_IonRangeCompensatorSequence          ) //DcmTagKey(0x300a, 0x02ea)
    GET_DICOM_INFO_LINE(DCM_CompensatorDescription               ) //DcmTagKey(0x300a, 0x02eb)
    GET_DICOM_INFO_LINE(DCM_RadiationMassNumber                  ) //DcmTagKey(0x300a, 0x0302)
    GET_DICOM_INFO_LINE(DCM_RadiationAtomicNumber                ) //DcmTagKey(0x300a, 0x0304)
    GET_DICOM_INFO_LINE(DCM_RadiationChargeState                 ) //DcmTagKey(0x300a, 0x0306)
    GET_DICOM_INFO_LINE(DCM_ScanMode                             ) //DcmTagKey(0x300a, 0x0308)
    GET_DICOM_INFO_LINE(DCM_VirtualSourceAxisDistances           ) //DcmTagKey(0x300a, 0x030a)
    GET_DICOM_INFO_LINE(DCM_SnoutSequence                        ) //DcmTagKey(0x300a, 0x030c)
    GET_DICOM_INFO_LINE(DCM_SnoutPosition                        ) //DcmTagKey(0x300a, 0x030d)
    GET_DICOM_INFO_LINE(DCM_SnoutID                              ) //DcmTagKey(0x300a, 0x030f)
    GET_DICOM_INFO_LINE(DCM_NumberOfRangeShifters                ) //DcmTagKey(0x300a, 0x0312)
    GET_DICOM_INFO_LINE(DCM_RangeShifterSequence                 ) //DcmTagKey(0x300a, 0x0314)
    GET_DICOM_INFO_LINE(DCM_RangeShifterNumber                   ) //DcmTagKey(0x300a, 0x0316)
    GET_DICOM_INFO_LINE(DCM_RangeShifterID                       ) //DcmTagKey(0x300a, 0x0318)
    GET_DICOM_INFO_LINE(DCM_RangeShifterType                     ) //DcmTagKey(0x300a, 0x0320)
    GET_DICOM_INFO_LINE(DCM_RangeShifterDescription              ) //DcmTagKey(0x300a, 0x0322)
    GET_DICOM_INFO_LINE(DCM_NumberOfLateralSpreadingDevices      ) //DcmTagKey(0x300a, 0x0330)
    GET_DICOM_INFO_LINE(DCM_LateralSpreadingDeviceSequence       ) //DcmTagKey(0x300a, 0x0332)
    GET_DICOM_INFO_LINE(DCM_LateralSpreadingDeviceNumber         ) //DcmTagKey(0x300a, 0x0334)
    GET_DICOM_INFO_LINE(DCM_LateralSpreadingDeviceID             ) //DcmTagKey(0x300a, 0x0336)
    GET_DICOM_INFO_LINE(DCM_LateralSpreadingDeviceType           ) //DcmTagKey(0x300a, 0x0338)
    GET_DICOM_INFO_LINE(DCM_LateralSpreadingDeviceDescription    ) //DcmTagKey(0x300a, 0x033a)
    GET_DICOM_INFO_LINE(DCM_LateralSpreadingDeviceWaterEquivalentThickness ) //DcmTagKey(0x300a, 0x033c)
    GET_DICOM_INFO_LINE(DCM_NumberOfRangeModulators              ) //DcmTagKey(0x300a, 0x0340)
    GET_DICOM_INFO_LINE(DCM_RangeModulatorSequence               ) //DcmTagKey(0x300a, 0x0342)
    GET_DICOM_INFO_LINE(DCM_RangeModulatorNumber                 ) //DcmTagKey(0x300a, 0x0344)
    GET_DICOM_INFO_LINE(DCM_RangeModulatorID                     ) //DcmTagKey(0x300a, 0x0346)
    GET_DICOM_INFO_LINE(DCM_RangeModulatorType                   ) //DcmTagKey(0x300a, 0x0348)
    GET_DICOM_INFO_LINE(DCM_RangeModulatorDescription            ) //DcmTagKey(0x300a, 0x034a)
    GET_DICOM_INFO_LINE(DCM_BeamCurrentModulationID              ) //DcmTagKey(0x300a, 0x034c)
    GET_DICOM_INFO_LINE(DCM_PatientSupportType                   ) //DcmTagKey(0x300a, 0x0350)
    GET_DICOM_INFO_LINE(DCM_PatientSupportID                     ) //DcmTagKey(0x300a, 0x0352)
    GET_DICOM_INFO_LINE(DCM_PatientSupportAccessoryCode          ) //DcmTagKey(0x300a, 0x0354)
    GET_DICOM_INFO_LINE(DCM_FixationLightAzimuthalAngle          ) //DcmTagKey(0x300a, 0x0356)
    GET_DICOM_INFO_LINE(DCM_FixationLightPolarAngle              ) //DcmTagKey(0x300a, 0x0358)
    GET_DICOM_INFO_LINE(DCM_MetersetRate                         ) //DcmTagKey(0x300a, 0x035a)
    GET_DICOM_INFO_LINE(DCM_RangeShifterSettingsSequence         ) //DcmTagKey(0x300a, 0x0360)
    GET_DICOM_INFO_LINE(DCM_RangeShifterSetting                  ) //DcmTagKey(0x300a, 0x0362)
    GET_DICOM_INFO_LINE(DCM_IsocenterToRangeShifterDistance      ) //DcmTagKey(0x300a, 0x0364)
    GET_DICOM_INFO_LINE(DCM_RangeShifterWaterEquivalentThickness ) //DcmTagKey(0x300a, 0x0366)
    GET_DICOM_INFO_LINE(DCM_LateralSpreadingDeviceSettingsSequence ) //DcmTagKey(0x300a, 0x0370)
    GET_DICOM_INFO_LINE(DCM_LateralSpreadingDeviceSetting        ) //DcmTagKey(0x300a, 0x0372)
    GET_DICOM_INFO_LINE(DCM_IsocenterToLateralSpreadingDeviceDistance ) //DcmTagKey(0x300a, 0x0374)
    GET_DICOM_INFO_LINE(DCM_RangeModulatorSettingsSequence       ) //DcmTagKey(0x300a, 0x0380)
    GET_DICOM_INFO_LINE(DCM_RangeModulatorGatingStartValue       ) //DcmTagKey(0x300a, 0x0382)
    GET_DICOM_INFO_LINE(DCM_RangeModulatorGatingStopValue        ) //DcmTagKey(0x300a, 0x0384)
    GET_DICOM_INFO_LINE(DCM_RangeModulatorGatingStartWaterEquivalentThickness ) //DcmTagKey(0x300a, 0x0386)
    GET_DICOM_INFO_LINE(DCM_RangeModulatorGatingStopWaterEquivalentThickness ) //DcmTagKey(0x300a, 0x0388)
    GET_DICOM_INFO_LINE(DCM_IsocenterToRangeModulatorDistance    ) //DcmTagKey(0x300a, 0x038a)
    GET_DICOM_INFO_LINE(DCM_ScanSpotTuneID                       ) //DcmTagKey(0x300a, 0x0390)
    GET_DICOM_INFO_LINE(DCM_NumberOfScanSpotPositions            ) //DcmTagKey(0x300a, 0x0392)
    GET_DICOM_INFO_LINE(DCM_ScanSpotPositionMap                  ) //DcmTagKey(0x300a, 0x0394)
    GET_DICOM_INFO_LINE(DCM_ScanSpotMetersetWeights              ) //DcmTagKey(0x300a, 0x0396)
    GET_DICOM_INFO_LINE(DCM_ScanningSpotSize                     ) //DcmTagKey(0x300a, 0x0398)
    GET_DICOM_INFO_LINE(DCM_NumberOfPaintings                    ) //DcmTagKey(0x300a, 0x039a)
    GET_DICOM_INFO_LINE(DCM_IonToleranceTableSequence            ) //DcmTagKey(0x300a, 0x03a0)
    GET_DICOM_INFO_LINE(DCM_IonBeamSequence                      ) //DcmTagKey(0x300a, 0x03a2)
    GET_DICOM_INFO_LINE(DCM_IonBeamLimitingDeviceSequence        ) //DcmTagKey(0x300a, 0x03a4)
    GET_DICOM_INFO_LINE(DCM_IonBlockSequence                     ) //DcmTagKey(0x300a, 0x03a6)
    GET_DICOM_INFO_LINE(DCM_IonControlPointSequence              ) //DcmTagKey(0x300a, 0x03a8)
    GET_DICOM_INFO_LINE(DCM_IonWedgeSequence                     ) //DcmTagKey(0x300a, 0x03aa)
    GET_DICOM_INFO_LINE(DCM_IonWedgePositionSequence             ) //DcmTagKey(0x300a, 0x03ac)
    GET_DICOM_INFO_LINE(DCM_ReferencedSetupImageSequence         ) //DcmTagKey(0x300a, 0x0401)
    GET_DICOM_INFO_LINE(DCM_SetupImageComment                    ) //DcmTagKey(0x300a, 0x0402)
    GET_DICOM_INFO_LINE(DCM_MotionSynchronizationSequence        ) //DcmTagKey(0x300a, 0x0410)
    GET_DICOM_INFO_LINE(DCM_ControlPointOrientation              ) //DcmTagKey(0x300a, 0x0412)
    GET_DICOM_INFO_LINE(DCM_GeneralAccessorySequence             ) //DcmTagKey(0x300a, 0x0420)
    GET_DICOM_INFO_LINE(DCM_GeneralAccessoryID                   ) //DcmTagKey(0x300a, 0x0421)
    GET_DICOM_INFO_LINE(DCM_GeneralAccessoryDescription          ) //DcmTagKey(0x300a, 0x0422)
    GET_DICOM_INFO_LINE(DCM_GeneralAccessoryType                 ) //DcmTagKey(0x300a, 0x0423)
    GET_DICOM_INFO_LINE(DCM_GeneralAccessoryNumber               ) //DcmTagKey(0x300a, 0x0424)
    GET_DICOM_INFO_LINE(DCM_ApplicatorGeometrySequence           ) //DcmTagKey(0x300a, 0x0431)
    GET_DICOM_INFO_LINE(DCM_ApplicatorApertureShape              ) //DcmTagKey(0x300a, 0x0432)
    GET_DICOM_INFO_LINE(DCM_ApplicatorOpening                    ) //DcmTagKey(0x300a, 0x0433)
    GET_DICOM_INFO_LINE(DCM_ApplicatorOpeningX                   ) //DcmTagKey(0x300a, 0x0434)
    GET_DICOM_INFO_LINE(DCM_ApplicatorOpeningY                   ) //DcmTagKey(0x300a, 0x0435)
    GET_DICOM_INFO_LINE(DCM_SourceToApplicatorMountingPositionDistance ) //DcmTagKey(0x300a, 0x0436)
    GET_DICOM_INFO_LINE(DCM_ReferencedRTPlanSequence             ) //DcmTagKey(0x300c, 0x0002)
    GET_DICOM_INFO_LINE(DCM_ReferencedBeamSequence               ) //DcmTagKey(0x300c, 0x0004)
    GET_DICOM_INFO_LINE(DCM_ReferencedBeamNumber                 ) //DcmTagKey(0x300c, 0x0006)
    GET_DICOM_INFO_LINE(DCM_ReferencedReferenceImageNumber       ) //DcmTagKey(0x300c, 0x0007)
    GET_DICOM_INFO_LINE(DCM_StartCumulativeMetersetWeight        ) //DcmTagKey(0x300c, 0x0008)
    GET_DICOM_INFO_LINE(DCM_EndCumulativeMetersetWeight          ) //DcmTagKey(0x300c, 0x0009)
    GET_DICOM_INFO_LINE(DCM_ReferencedBrachyApplicationSetupSequence ) //DcmTagKey(0x300c, 0x000a)
    GET_DICOM_INFO_LINE(DCM_ReferencedBrachyApplicationSetupNumber ) //DcmTagKey(0x300c, 0x000c)
    GET_DICOM_INFO_LINE(DCM_ReferencedSourceNumber               ) //DcmTagKey(0x300c, 0x000e)
    GET_DICOM_INFO_LINE(DCM_ReferencedFractionGroupSequence      ) //DcmTagKey(0x300c, 0x0020)
    GET_DICOM_INFO_LINE(DCM_ReferencedFractionGroupNumber        ) //DcmTagKey(0x300c, 0x0022)
    GET_DICOM_INFO_LINE(DCM_ReferencedVerificationImageSequence  ) //DcmTagKey(0x300c, 0x0040)
    GET_DICOM_INFO_LINE(DCM_ReferencedReferenceImageSequence     ) //DcmTagKey(0x300c, 0x0042)
    GET_DICOM_INFO_LINE(DCM_ReferencedDoseReferenceSequence      ) //DcmTagKey(0x300c, 0x0050)
    GET_DICOM_INFO_LINE(DCM_ReferencedDoseReferenceNumber        ) //DcmTagKey(0x300c, 0x0051)
    GET_DICOM_INFO_LINE(DCM_BrachyReferencedDoseReferenceSequence ) //DcmTagKey(0x300c, 0x0055)
    GET_DICOM_INFO_LINE(DCM_ReferencedStructureSetSequence       ) //DcmTagKey(0x300c, 0x0060)
    GET_DICOM_INFO_LINE(DCM_ReferencedPatientSetupNumber         ) //DcmTagKey(0x300c, 0x006a)
    GET_DICOM_INFO_LINE(DCM_ReferencedDoseSequence               ) //DcmTagKey(0x300c, 0x0080)
    GET_DICOM_INFO_LINE(DCM_ReferencedToleranceTableNumber       ) //DcmTagKey(0x300c, 0x00a0)
    GET_DICOM_INFO_LINE(DCM_ReferencedBolusSequence              ) //DcmTagKey(0x300c, 0x00b0)
    GET_DICOM_INFO_LINE(DCM_ReferencedWedgeNumber                ) //DcmTagKey(0x300c, 0x00c0)
    GET_DICOM_INFO_LINE(DCM_ReferencedCompensatorNumber          ) //DcmTagKey(0x300c, 0x00d0)
    GET_DICOM_INFO_LINE(DCM_ReferencedBlockNumber                ) //DcmTagKey(0x300c, 0x00e0)
    GET_DICOM_INFO_LINE(DCM_ReferencedControlPointIndex          ) //DcmTagKey(0x300c, 0x00f0)
    GET_DICOM_INFO_LINE(DCM_ReferencedControlPointSequence       ) //DcmTagKey(0x300c, 0x00f2)
    GET_DICOM_INFO_LINE(DCM_ReferencedStartControlPointIndex     ) //DcmTagKey(0x300c, 0x00f4)
    GET_DICOM_INFO_LINE(DCM_ReferencedStopControlPointIndex      ) //DcmTagKey(0x300c, 0x00f6)
    GET_DICOM_INFO_LINE(DCM_ReferencedRangeShifterNumber         ) //DcmTagKey(0x300c, 0x0100)
    GET_DICOM_INFO_LINE(DCM_ReferencedLateralSpreadingDeviceNumber ) //DcmTagKey(0x300c, 0x0102)
    GET_DICOM_INFO_LINE(DCM_ReferencedRangeModulatorNumber       ) //DcmTagKey(0x300c, 0x0104)
    GET_DICOM_INFO_LINE(DCM_ApprovalStatus                       ) //DcmTagKey(0x300e, 0x0002)
    GET_DICOM_INFO_LINE(DCM_ReviewDate                           ) //DcmTagKey(0x300e, 0x0004)
    GET_DICOM_INFO_LINE(DCM_ReviewTime                           ) //DcmTagKey(0x300e, 0x0005)
    GET_DICOM_INFO_LINE(DCM_ReviewerName                         ) //DcmTagKey(0x300e, 0x0008)
            info->append("</div>");
            info->append("<div style=\"color:#00007F\">");
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_TextGroupLength             ) //DcmTagKey(0x4000, 0x0000)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_TextArbitrary               ) //DcmTagKey(0x4000, 0x0010)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_TextComments                ) //DcmTagKey(0x4000, 0x4000)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ResultsID                    ) //DcmTagKey(0x4008, 0x0040)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ResultsIDIssuer              ) //DcmTagKey(0x4008, 0x0042)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ReferencedInterpretationSequence ) //DcmTagKey(0x4008, 0x0050)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_InterpretationRecordedDate   ) //DcmTagKey(0x4008, 0x0100)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_InterpretationRecordedTime   ) //DcmTagKey(0x4008, 0x0101)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_InterpretationRecorder       ) //DcmTagKey(0x4008, 0x0102)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ReferenceToRecordedSound     ) //DcmTagKey(0x4008, 0x0103)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_InterpretationTranscriptionDate ) //DcmTagKey(0x4008, 0x0108)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_InterpretationTranscriptionTime ) //DcmTagKey(0x4008, 0x0109)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_InterpretationTranscriber    ) //DcmTagKey(0x4008, 0x010a)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_InterpretationText           ) //DcmTagKey(0x4008, 0x010b)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_InterpretationAuthor         ) //DcmTagKey(0x4008, 0x010c)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_InterpretationApproverSequence ) //DcmTagKey(0x4008, 0x0111)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_InterpretationApprovalDate   ) //DcmTagKey(0x4008, 0x0112)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_InterpretationApprovalTime   ) //DcmTagKey(0x4008, 0x0113)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_PhysicianApprovingInterpretation ) //DcmTagKey(0x4008, 0x0114)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_InterpretationDiagnosisDescription ) //DcmTagKey(0x4008, 0x0115)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_InterpretationDiagnosisCodeSequence ) //DcmTagKey(0x4008, 0x0117)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ResultsDistributionListSequence ) //DcmTagKey(0x4008, 0x0118)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_DistributionName             ) //DcmTagKey(0x4008, 0x0119)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_DistributionAddress          ) //DcmTagKey(0x4008, 0x011a)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_InterpretationID             ) //DcmTagKey(0x4008, 0x0200)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_InterpretationIDIssuer       ) //DcmTagKey(0x4008, 0x0202)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_InterpretationTypeID         ) //DcmTagKey(0x4008, 0x0210)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_InterpretationStatusID       ) //DcmTagKey(0x4008, 0x0212)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_Impressions                  ) //DcmTagKey(0x4008, 0x0300)
//    GET_DICOM_INFO_LINE(DCM_RETIRED_ResultsComments              ) //DcmTagKey(0x4008, 0x4000)
    GET_DICOM_INFO_LINE(DCM_MACParametersSequence                ) //DcmTagKey(0x4ffe, 0x0001)
            info->append("</div>");
            info->append("<div style=\"color:#00007F\">");
    GET_DICOM_INFO_LINE(DCM_SharedFunctionalGroupsSequence       ) //DcmTagKey(0x5200, 0x9229)
    GET_DICOM_INFO_LINE(DCM_PerFrameFunctionalGroupsSequence     ) //DcmTagKey(0x5200, 0x9230)
    GET_DICOM_INFO_LINE(DCM_WaveformSequence                     ) //DcmTagKey(0x5400, 0x0100)
    GET_DICOM_INFO_LINE(DCM_ChannelMinimumValue                  ) //DcmTagKey(0x5400, 0x0110)
    GET_DICOM_INFO_LINE(DCM_ChannelMaximumValue                  ) //DcmTagKey(0x5400, 0x0112)
    GET_DICOM_INFO_LINE(DCM_WaveformBitsAllocated                ) //DcmTagKey(0x5400, 0x1004)
    GET_DICOM_INFO_LINE(DCM_WaveformSampleInterpretation         ) //DcmTagKey(0x5400, 0x1006)
    GET_DICOM_INFO_LINE(DCM_WaveformPaddingValue                 ) //DcmTagKey(0x5400, 0x100a)
    GET_DICOM_INFO_LINE(DCM_WaveformData                         ) //DcmTagKey(0x5400, 0x1010)
    GET_DICOM_INFO_LINE(DCM_FirstOrderPhaseCorrectionAngle       ) //DcmTagKey(0x5600, 0x0010)
    GET_DICOM_INFO_LINE(DCM_SpectroscopyData                     ) //DcmTagKey(0x5600, 0x0020)
*/




/*
    GET_DICOM_INFO_LINE(DCM_PixelData                            ) //DcmTagKey(0x7fe0, 0x0010)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_CoefficientsSDVN         ) //DcmTagKey(0x7fe0, 0x0020)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_CoefficientsSDHN         ) //DcmTagKey(0x7fe0, 0x0030)
//    GET_DICOM_INFO_LINE(DCM_ACR_NEMA_2C_CoefficientsSDDN         ) //DcmTagKey(0x7fe0, 0x0040)
    GET_DICOM_INFO_LINE(DCM_DigitalSignaturesSequence            ) //DcmTagKey(0xfffa, 0xfffa)
    GET_DICOM_INFO_LINE(DCM_DataSetTrailingPadding               ) //DcmTagKey(0xfffc, 0xfffc)
    GET_DICOM_INFO_LINE(DCM_Item                                 ) //DcmTagKey(0xfffe, 0xe000)
    GET_DICOM_INFO_LINE(DCM_ItemDelimitationItem                 ) //DcmTagKey(0xfffe, 0xe00d)
    GET_DICOM_INFO_LINE(DCM_SequenceDelimitationItem             ) //DcmTagKey(0xfffe, 0xe0dd)

            info->append("</div>");
*/
}
#endif
